
(** This file contains the concretisation relation between separation-logic
  formulae and the concrete constructs of the example language. **)

Set Implicit Arguments.

Require Export FrameSep.

Section General.

(** [L] represents singleton locations, [K] set locations and [KE]
  environments, which behave like set locations. **)
Variable L K KE : Type. (** Typically, set of integers **)
Hypothesis L_comparable : Comparable L.
Hypothesis K_comparable : Comparable K.
Hypothesis KE_comparable : Comparable KE.

(** These sets have to be infinite. **)
Hypothesis L_infinite : forall ls,
  exists (l : L), ~ mem l ls.
Hypothesis K_infinite : forall ks,
  exists (k : K), ~ mem k ks.
Hypothesis KE_infinite : forall ks,
  exists (k : KE), ~ mem k ks.
Hypothesis L_infinite_pick : forall ls,
  Pickable (fun l : L => ~ mem l ls).
Hypothesis K_infinite_pick : forall ks,
  Pickable (fun k : K => ~ mem k ks).
Hypothesis KE_infinite_pick : forall ks,
  Pickable (fun k : KE => ~ mem k ks).

Variable ABasic : Type. (** Abstract basic values **)
Hypothesis ABasic_comparable : Comparable ABasic.
Hypothesis ABasicLat : LatticeDec.t ABasic.
Hypothesis ABasicTop : TopDec.t ABasic.
Variable ABasic_gamma : ABasic → ℘ basic_val.
Hypothesis ABasic_Gamma : Gamma.t _ _ ABasic_gamma.
Hypothesis ABasic_gamma_bot : forall b,
  ~ ABasic_gamma (⊥♯) b.
Hypothesis ABasic_gamma_top : forall b,
  ABasic_gamma (⊤♯) b.

Let ALoc := ALoc L K.
Let d_singleton_loc := singleton_loc _ : L → ALoc.
Let d_set_loc := set_loc _ (K := K) : K → ALoc.
Coercion d_singleton_loc : L >-> ALoc.
Coercion d_set_loc : K >-> ALoc.

Let ALoc_Equiv : EquivDec.t ALoc :=
  ALoc_Equiv _ _.

Let ALocS := ALocS L K.

Local Instance Local_Inhab_ALoc : Inhab ALoc.
  apply~ ALoc_inhab.
Defined.

Local Instance Local_Comparable_ALoc : Comparable ALoc.
  apply~ ALoc_comparable.
Defined.

Let AClosure := AClosure KE.
Let AValue := AValue L K KE ABasic.
Let ASValue := ASValue L K KE ABasic.
Let DObject := DObject L K KE ABasic.
Let AEnv := AEnv L K KE ABasic.

Variable Ext : Type.
Variable Ext_substs : Ext → list ALoc → ALoc → Ext.
Variable Ext_Esubsts : Ext → list KE → KE → Ext.
Variable ExtNames : Ext → list ALoc.
Variable ExtENames : Ext → list KE.

Let IFormula := IFormula L K KE ABasic.
Let EFormula := EFormula L K KE ABasic Ext.
Let IMembrane := IMembrane L K KE.
Let Membrane := Membrane L K KE.
Let DFormula := DFormula _ _ _ ABasic ExtNames ExtENames.

Notation "'emp'" := (IFormula_emp _ _ _ _) : sep_scope.
Notation "phi1 '*' phi2" := (IFormula_star phi1 phi2) : sep_scope.
Notation "h '-->' o" := (IFormula_loc h o)
  (at level 39, no associativity) : sep_scope.
Notation "ke '~~>' E" := (IFormula_env ke E)
  (at level 39, no associativity) : sep_scope.

Notation "h '--->' hs" := (Membrane_rewriting _ h hs)
  (at level 56) : sep_scope.
Notation "'ν' h" := (Membrane_new _ h)
  (at level 56) : sep_scope.
Notation "ke '~~~>' kes" := (Membrane_Erewriting _ _ ke kes)
  (at level 56) : sep_scope.
Notation "'ν~' ke" := (Membrane_Enew _ _ ke)
  (at level 56) : sep_scope.

Open Scope sep_scope.

Instance IFormula_Inhab : Inhab IFormula.
  applys prove_Inhab (emp : IFormula).
Defined.

Let Mread := Mread _ _ : Membrane → ALoc → option (list ALoc).
Let Mread_env := Mread_env _ : Membrane → KE → option (list KE).

Let MInnerInterface_correct :=
  MInnerInterface_correct _ _ _ : forall (h : ALoc) (M : Membrane), _.
Let MInInnerInterfaceDecidable :=
  MInInnerInterfaceDecidable _ _ _ : forall (h : ALoc) (M : Membrane), _.
Let MEInnerInterface_correct :=
  MEInnerInterface_correct _ _ _ : forall (ke : KE) (M : Membrane), _.
Let MEInInnerInterfaceDecidable :=
  MEInInnerInterfaceDecidable _ _ _ : forall (ke : KE) (M : Membrane), _.

Let EFormula_correct :=
  EFormula_correct _ _ _ ExtNames ExtENames : EFormula → Prop.

Instance Local_AClosure_LatDec : LatticeDec.t AClosure.
  apply~ AClosure_LatDec.
Defined.

Local Instance Local_AEnv_Inhab : Inhab AEnv.
  applys prove_Inhab. constructors.
Defined.

Local Instance Local_DObject_Inhab : Inhab DObject.
  apply Local_DObject_Inhab.
Defined.

Local Instance Local_DObject_Poset : PosetDec.t DObject.
  apply~ DObject_Poset. typeclass.
Defined.

Let read_DObject : DObject → property → option ASValue :=
  read_DObject _ _ _ _.
Let read_AEnv : AEnv → var → option AValue :=
  read_AEnv L_comparable K_comparable KE_comparable ABasicLat.

Local Instance Local_locInDom_decidable : forall (phi : IFormula) (h : ALoc) f,
    Decidable (locInDom phi h f).
  apply~ locInDom_decidable.
Defined.

Local Instance Local_envInDom_decidable : forall (phi : IFormula) ke,
    Decidable (envInDom phi ke).
  apply envInDom_decidable; typeclass.
Defined.

Let passThroughMembrane : IFormula → Membrane → option IFormula :=
  passThroughMembrane _ _ _ _ _.

Let MembraneCompose : Membrane → Membrane → Membrane :=
  MembraneCompose _ _ _.
Let MembraneCheckCompose : Membrane → Membrane → option Membrane :=
  MembraneCheckCompose _ _ _.


(** ** Concretisation of Formulae **)

(** *** Valuation **)

(** Valuation of abstract locations to concrete ones.  Note that
  concrete locations can be both [loc] (object locations) and [floc]
  (closure locations).  See [ConcreteOWhile] for more details. **)
Record valuation := make_valuation {
    valuation_singleton : L → option loc ;
    valuation_set : K → option (℘ loc) ;
    valuation_env : KE → option (℘ floc)
  }.

(** Relating abstract locations to concrete using a valuation. **)
Definition evalrho (rho : valuation) (al : ALoc) cl :=
  match al with
  | singleton_loc l => valuation_singleton rho l = Some cl
  | set_loc k => exists s, valuation_set rho k = Some s ∧ s cl
  end.

(** Similar to [evalrho], but for closure locations. **)
Definition fevalrho (rho : valuation) ke cl :=
  exists s, valuation_env rho ke = Some s ∧ s cl.

(** Abstract locations in the domain of a valuation.  Note that a [k]
  associated to an empty set is nevertheless in the domain, despite no
  concrete location correspond to it. **)
Inductive inDomRho : valuation → ALoc → Prop :=
  | inDomRho_L : forall rho l cl,
      valuation_singleton rho l = Some cl →
      inDomRho rho l
  | inDomRho_K : forall rho k s,
      valuation_set rho k = Some s →
      inDomRho rho k
  .

Inductive inDomRho_env : valuation → KE → Prop :=
  | inDomRho_KE : forall rho ke s,
      valuation_env rho ke = Some s →
      inDomRho_env rho ke
  .

(** Valuations are supposed to be injective. **)
Record valid_valuation (rho : valuation) : Prop := {
  valuation_injective : forall (h1 h2 : ALoc) cl,
    evalrho rho h1 cl →
    evalrho rho h2 cl →
    h1 = h2 ;
  valuation_injective_env : forall ke1 ke2 fl,
    fevalrho rho ke1 fl →
    fevalrho rho ke2 fl →
    ke1 = ke2
  }.

(** An empty valuation.  We sometimes need a default valuation. **)
Definition empty_valuation : valuation :=
  make_valuation (fun _ => None) (fun _ => None) (fun _ => None).

Lemma empty_valuation_valid : valid_valuation empty_valuation.
Proof.
  constructors.
  (** injective **)
  introv E1 E2. destruct h1; repeat inverts E1 as E1.
  (** injective (env) **)
  introv E1 E2. repeat inverts E1 as E1.
Qed.


(** *** Values **)

Definition ALocS_gamma (rho : valuation) : ALocS → ℘ loc :=
  immerse_gamma_prop (FinitePowerSet.gamma ALoc_Equiv) (evalrho rho).

Lemma ALocS_Gamma : forall rho, valid_valuation rho → Gamma.t _ _ (ALocS_gamma rho).
Proof.
  introv V. apply~ immerseGammaUnique.
   typeclass.
   apply~ valuation_injective.
Qed.

Lemma ALocS_gamma_bot : forall rho ah,
  ~ ALocS_gamma rho (⊥♯) ah.
Proof. introv (cl&A&R). exact (FinitePowerSet.gamma_bot A). Qed.


Definition AClosure_gamma (rho : valuation) : AClosure → ℘ (var * prog * floc) :=
  immerse_gamma_prop (FinitePowerSet.gamma (AClosure_basic_Equiv _)) (fun cloa cloc : _ * _ =>
    let '(x, p, ke) := cloa in
    let '(x', p', cl) := cloc in
    x = x' ∧ p = p' ∧ fevalrho rho ke cl).

Lemma AClosure_Gamma : forall rho,
  valid_valuation rho →
  Gamma.t _ _ (AClosure_gamma rho).
Proof.
  introv V. apply~ immerseGammaUnique.
   typeclass.
   introv. repeat match goal with c : _ * _ |- _ => destruct c as [? ?] end.
    introv (?&?&?) (?&?&?). fequals~. apply* valuation_injective_env.
Qed.

Lemma AClosure_gamma_bot : forall rho ah,
  ~ AClosure_gamma rho (⊥♯) ah.
Proof. introv (cl&A&R). exact (FinitePowerSet.gamma_bot A). Qed.


Local Transparent AValue_Lat.
Local Transparent ASValue_Lat.

Definition AValue_gamma (rho : valuation) : AValue → ℘ val :=
  immerse_gamma
    (productSumGamma
      (productSumGamma ABasic_gamma (ALocS_gamma rho))
      (AClosure_gamma rho))
    (fun v =>
      match v with
      | val_basic v => inl (inl v)
      | val_loc l => inl (inr l)
      | val_closure x p le => inr (x, p, le)
      end).

Lemma AValue_Gamma : forall rho, valid_valuation rho → Gamma.t _ _ (AValue_gamma rho).
Proof.
  introv V. apply immerseGamma. repeat apply ProductSumGamma.
   apply~ ABasic_Gamma.
   apply~ ALocS_Gamma.
   apply~ AClosure_Gamma.
Qed.

Lemma AValue_gamma_bot : forall rho av,
  ~ AValue_gamma rho (⊥♯) av.
Proof. introv A. destruct av; repeat inverts A as A. apply* ABasic_gamma_bot. Qed.


Definition ASValue_gamma (rho : valuation) : ASValue → ℘ (option val) :=
  immerse_gamma
    (productSumGamma (AValue_gamma rho) (LiftBot.gamma UnitLattice.gamma))
    (fun ov =>
      match ov with
      | Some av => inl av
      | None => inr tt
      end).

Lemma ASValue_Gamma : forall rho, valid_valuation rho → Gamma.t _ _ (ASValue_gamma rho).
Proof.
  introv V. apply immerseGamma. apply ProductSumGamma.
   apply~ AValue_Gamma.
   apply LiftBot.Gamma. typeclass.
Qed.

Lemma ASValue_gamma_bot : forall rho av,
  ~ ASValue_gamma rho (⊥♯) av.
Proof.
  introv A. destruct av; inverts A.
   apply* AValue_gamma_bot.
   typeclass.
Qed.


Lemma ASValue_undef_spec : forall rho,
  ASValue_gamma rho (ASValue_undef _ _ _ _) None.
Proof. introv. repeat constructors~. Qed.

Variable ExtC : Type.
Variable Ext_gamma : valuation → Ext → ℘ ExtC.
(*Hypothesis Ext_Gamma : forall rho,
  valid_valuation rho →
  Gamma.t _ _ (Ext_gamma rho).
Hypothesis Ext_gamma_bot : forall rho ext,
  valid_valuation rho →
  ~ Ext_gamma rho (⊥♯) ext.*)


(** *** Objects and Environments **)

(** As abstract objects are only partial, their concretisation is also
  partial. It thus links an abstract object [ao] to both a concrete
  object [o] and the domain [D]. **)
Definition DObject_gamma (rho : valuation) (ao : DObject) (o : obj) (D : property → Prop) :=
  forall f,
    match read_DObject ao f with
    | Some av => D f ∧ ASValue_gamma rho av (read_option o f)
    | None => ~ D f
    end.

Lemma DObject_gamma_monotone : forall rho,
  valid_valuation rho →
  forall ao1 ao2, ao1 ⊑♯ ao2 → DObject_gamma rho ao1 ⊑ DObject_gamma rho ao2.
Proof.
  introv V O G. unfolds. introv. forwards O': O f. forwards G': G f. fold read_DObject in O'.
  destruct (read_DObject ao1 f) eqn: E1; destruct (read_DObject ao2 f) eqn: E2; tryfalse.
   splits*. applys* @Gamma.monotone ASValue_Gamma O'.
   autos*.
Qed.


Definition AEnv_gamma (rho : valuation) (AE : AEnv) (E : Env) :=
  forall x,
    match read_AEnv AE x, read_option E x with
    | Some av, Some v => AValue_gamma rho av v
    | None, None => True
    | _, _ => False
    end.

Lemma AEnv_gamma_monotone : forall rho,
  valid_valuation rho →
  forall E1 E2, E1 ⊑♯ E2 → AEnv_gamma rho E1 ⊑ AEnv_gamma rho E2.
Proof.
  introv V O G. unfolds. intro xv. forwards O': O xv. forwards G': G xv.
  unfolds read_AEnv. destruct (DefinitionsSep.read_AEnv _ _ _ _ E1 xv) eqn: Eq1;
    destruct (DefinitionsSep.read_AEnv _ _ _ _ E2 xv) eqn: Eq2; tryfalse;
    destruct read_option; tryfalse.
   applys~ @Gamma.monotone AValue_Gamma O'.
   autos*.
Qed.


(** *** Entailment **)

Section Rho.

Variable rho : valuation.

Definition stateDomain :=
  ℘ (loc * property).

Definition stateDomainEmpty : stateDomain := fun _ => False.
Record stateDomainMerge (D1 D2 D : stateDomain) := {
    stateDomainMergeDisjunct : forall d, ~ D1 d ∨ ~ D2 d ;
    stateDomainMergeLeft : forall d, D1 d → D d ;
    stateDomainMergeRight : forall d, D2 d → D d ;
    stateDomainMergeInv : forall d, D d → D1 d ∨ D2 d
  }.
Definition stateDomainFrom L d : stateDomain :=
  fun lf => let (l, f) := lf in L l ∧ in_AObject_domain f d.

(** Recognising an empty heap. **)
Definition is_empty (K V : Type) (CK : Comparable K) (h : heap K V) :=
  forall k, ~ indom h k.
Arguments is_empty [K] [V] [CK] h.

(** Recognising a heap whose properties are not defined, expect maybe
  those satisfying [P], on which we say nothing. **)
Definition is_empty_but (K V : Type) (CK : Comparable K) P (h : heap K V) :=
  forall k, ~ P k → ~ indom h k.
Arguments is_empty_but [K] [V] [CK] P h.

(** Recognising a heap which is the disjoinct union of two others. **)
Record is_merge (K V : Type) (CK : Comparable K) (h1 h2 h : heap K V) := {
    is_merge_disjunct : forall k, indom h1 k → indom h2 k → False ;
    is_merge_left : forall k v, Heap.binds h1 k v → Heap.binds h k v ;
    is_merge_right : forall k v, Heap.binds h2 k v → Heap.binds h k v
  }.
Arguments is_merge [K] [V] [CK] h1 h2 h.

Inductive IFormulaEntails : State → stateDomain → IFormula → Prop :=

  | IFormulaEntails_emp : forall St,
    is_empty (state_heap St) →
    is_empty (state_closures St) →
    IFormulaEntails St stateDomainEmpty emp

  | IFormulaEntails_star : forall St1 St2 St D1 D2 D phi1 phi2,
    IFormulaEntails St1 D1 phi1 →
    IFormulaEntails St2 D2 phi2 →
    is_merge (state_heap St1) (state_heap St2) (state_heap St) →
    is_merge (state_closures St1) (state_closures St2) (state_closures St) →
    stateDomainMerge D1 D2 D →
    IFormulaEntails St D (phi1 * phi2)

  | IFormulaEntails_loc : forall St h ao,
    is_empty_but (evalrho rho h) (state_heap St) →
    (forall cl,
      evalrho rho h cl →
      exists o,
        DObject_gamma rho ao o (fun f => in_AObject_domain f (DObject_dom ao))
        ∧ Heap.binds (state_heap St) cl o) →
    is_empty (state_closures St) →
    IFormulaEntails St (stateDomainFrom (evalrho rho h) (DObject_dom ao)) (h --> ao)

  | IFormulaEntails_env : forall St ke AE,
    is_empty (state_heap St) →
    is_empty_but (fevalrho rho ke) (state_closures St) →
    (forall cl,
      fevalrho rho ke cl →
      exists E, AEnv_gamma rho AE E ∧ Heap.binds (state_closures St) cl E) →
    IFormulaEntails St stateDomainEmpty (ke ~~> AE)
  .

End Rho.


(** *** Valuations and Membranes **)

(** Membranes defines two environments: the inside and the outside.
  Each of these environments needs a valuation.  They have to be
  related with respect to this property. **)
Record valuationGoThroughMembrane (M : Membrane) (rho_out rho_in : valuation) := {
    (** The interfaces correspond. **)
    valuationGoThroughMembrane_dom_out : forall h,
      inDomRho rho_out h ↔ MInOuterInterface _ _ _ h M ;
    valuationGoThroughMembrane_dom_out_env : forall ke,
      inDomRho_env rho_out ke ↔ MEInOuterInterface _ _ _ ke M ;
    valuationGoThroughMembrane_dom_in : forall h,
      inDomRho rho_in h ↔ MInInnerInterface _ _ _ h M ;
    valuationGoThroughMembrane_dom_in_env : forall ke,
      inDomRho_env rho_in ke ↔ MEInInnerInterface _ _ _ ke M ;
    (** In a rewriting [h ---> hs], we have [γ h ⊑ ∪ γ hs]. **)
    valuationGoThroughMembrane_rewriting : forall h hs cl,
      mem (h ---> hs) M →
      evalrho rho_out h cl →
      exists h', mem h' hs ∧ evalrho rho_in h' cl ;
    (** In a rewriting [ke ~~~> kes], we have [γ ke ⊑ ∪ γ kes]. **)
    valuationGoThroughMembrane_rewriting_env : forall ke kes cl,
      mem (ke ~~~> kes) M →
      fevalrho rho_out ke cl →
      exists ke', mem ke' kes ∧ fevalrho rho_in ke' cl ;
    (** We can trace back concrete locations to either [ν h]
      constructions or to the outside valuation. **)
    valuationGoThroughMembrane_conservation : forall cl h,
      evalrho rho_in h cl →
      mem (ν h) M ∨ exists h', evalrho rho_out h' cl ;
    valuationGoThroughMembrane_conservation_env : forall cl ke,
      fevalrho rho_in ke cl →
      mem (ν~ ke) M ∨ exists ke', fevalrho rho_out ke' cl
  }.


(** *** Wrapping Up to Extended Formulae **)

Definition EFormulaGamma Phi (Stv : _ * _) :=
  let (St, v) := Stv in
  let phi := EFormulaFormula Phi in
  let M := EFormulaMembrane Phi in
  let ext := EFormulaExt Phi in
  exists D rho_out rho_in,
    valid_valuation rho_out ∧ valid_valuation rho_in
    ∧ valuationGoThroughMembrane M rho_out rho_in
    ∧ IFormulaEntails rho_in St D phi
    ∧ Ext_gamma rho_in ext v.

Definition DFormulaGamma (Phi : DFormula) Stv :=
  EFormulaGamma (proj1_sig Phi) Stv.


(** ** Order over formulae **)

(** *** Contexts **)

Section Param.

(** This section is parameterised to face the different version of
  each type: [Membrane] can be simple membranes, but also
  [DMembrane]. In this file, only [EFormulaExtends] is used, but File
  [AbstractSep.v] shows a use-case with other types. **)

Variable membrane : Type.
Variable simple_formula : Type.
Variable formula : Type.
Variable frameMembrane : membrane → formula → option formula.
Variable frameFormula : simple_formula → formula → option formula.

Inductive formulaContext :=
  | formulaContextMembrane : membrane → formulaContext
  | formulaContextFormula : simple_formula → formulaContext
  .

(** Extends a formula with a context expressed as a list. **)
Inductive formulaExtends : formula → list formulaContext → formula → Prop :=
  | formulaExtendsNone : forall Phi,
    formulaExtends Phi nil Phi

  | formulaExtendsMembrane : forall Phi1 Phi2 Phi3 M C,
    frameMembrane M Phi1 = Some Phi2 →
    formulaExtends Phi2 C Phi3 →
    formulaExtends Phi1 (formulaContextMembrane M :: C) Phi3

  | formulaExtendsFormula : forall Phi1 Phi2 Phi3 phi C,
    frameFormula phi Phi1 = Some Phi2 →
    formulaExtends Phi2 C Phi3 →
    formulaExtends Phi1 (formulaContextFormula phi :: C) Phi3
  .

Lemma formulaExtends_trans : forall Phi1 Phi2 Phi3 Ca Cb,
  formulaExtends Phi1 Ca Phi2 →
  formulaExtends Phi2 Cb Phi3 →
  formulaExtends Phi1 (Ca ++ Cb) Phi3.
Proof. introv E1. induction E1; introv E2; autos~; constructors*. Qed.

Lemma formulaExtends_func : forall Phi Phi1 Phi2 C,
  formulaExtends Phi C Phi1 →
  formulaExtends Phi C Phi2 →
  Phi1 = Phi2.
Proof.
  introv E1. induction E1; introv E2; inverts E2 as Ea Eb.
   autos*.
   rewrite Ea in *. inverts~ H.
   rewrite Ea in *. inverts~ H.
Qed.

End Param.

Definition EFormulaExtends : EFormula → list (formulaContext Membrane IFormula)
    → EFormula → Prop :=
  formulaExtends (frameMembrane _ _ _ (Ext := _))
    (frameFormula _ _ _ L_infinite K_infinite KE_infinite
      L_infinite_pick K_infinite_pick KE_infinite_pick
      _ _ Ext_substs Ext_Esubsts ExtNames ExtENames).


(** *** Orders **)

(** There are two possible formula ordering. The one we need for
  proving the correctness of the abstract semantics is the inclusion
  of their concretisations (given by [EFormulaOrder_concr]); but it
  can be tricky to prove, so instead we would prefer to use
  [EFormulaOrder_dom], which does not introduce an arbitrary context. **)

(** Formulae are considered equals if they have the same
  concretisation in every context. **)
Definition EFormulaOrder_concr (Phi1 Phi2 : EFormula) : Prop :=
  forall Phi1' Phi2' C,
    EFormulaExtends Phi1 C Phi1' →
    EFormulaExtends Phi2 C Phi2' →
    forall StExt,
      EFormulaGamma Phi1' StExt → EFormulaGamma Phi2' StExt.

(** Another ordering which does not require to consider every contexts
  of the formulae [Phi]. **)
Record EFormulaOrder_dom (Phi1 Phi2 : EFormula) := {
  EFormulaOrder_dom_same_interface :
    double_incl (EFormulaInterface Phi1) (EFormulaInterface Phi2) ;
  EFormulaOrder_dom_same_interface_env :
    double_incl (EFormulaEInterface Phi1) (EFormulaEInterface Phi2) ;
  EFormulaOrder_dom_same_domain :
    forall h, EFormulaInDomain _ _ Phi1 h ↔ EFormulaInDomain _ _ Phi2 h ;
  EFormulaOrder_dom_same_domain_env :
    forall ke, EFormulaInEDomain _ Phi1 ke ↔ EFormulaInEDomain _ Phi2 ke ;
  EFormulaOrder_dom_gamma :
    forall StExt,
      EFormulaGamma Phi1 StExt → EFormulaGamma Phi2 StExt
  }.

Lemma EFormulaOrder_dom_refl : forall Phi,
  EFormulaOrder_dom Phi Phi.
Proof. introv. constructors*; apply double_incl_refl. Qed.

Lemma EFormulaOrder_dom_trans : forall Phi1 Phi2 Phi3,
  EFormulaOrder_dom Phi1 Phi2 →
  EFormulaOrder_dom Phi2 Phi3 →
  EFormulaOrder_dom Phi1 Phi3.
Proof.
  introv O1 O2. constructors.
   eapply double_incl_trans.
    apply O1.
    apply O2.
   eapply double_incl_trans.
    apply O1.
    apply O2.
   introv. rewrite EFormulaOrder_dom_same_domain.
    apply O2.
    apply O1.
   introv. rewrite EFormulaOrder_dom_same_domain_env.
    apply O2.
    apply O1.
   introv G. applys EFormulaOrder_dom_gamma O2. apply* O1.
Qed.

Global Instance EFormula_Equiv : Equiv.t EFormula.
  applys Equiv.Make (fun Phi1 Phi2 =>
      EFormulaOrder_dom Phi1 Phi2 ∧ EFormulaOrder_dom Phi2 Phi1).
   introv. splits; apply EFormulaOrder_dom_refl.
   introv (O1&O2). autos*.
   introv (O1&O2) (O3&O4). splits; apply* EFormulaOrder_dom_trans.
Defined.

Global Instance EFormula_Poset : Poset.t EFormula.
  applys Poset.Make EFormulaOrder_dom.
   introv (?&?). autos*.
   introv ? ?. split*.
   apply~ EFormulaOrder_dom_trans.
Defined.

Global Instance DFormula_Equiv : Equiv.t DFormula.
  applys Equiv.Make (fun Phi1 Phi2 : DFormula =>
    Equiv.eq (proj1_sig Phi1) (proj1_sig Phi2)).
  introv. apply~ Equiv.refl.
  introv. apply~ Equiv.sym.
  introv. apply~ Equiv.trans.
Defined.

Global Instance DFormula_Poset : Poset.t DFormula.
  applys Poset.Make (fun Phi1 Phi2 : DFormula =>
     proj1_sig Phi1 ⊑ proj1_sig Phi2).
   introv E. apply~ Poset.refl.
   introv E. apply~ Poset.antisym.
   introv E. apply~ Poset.trans.
Defined.


(** We do not have [EFormulaOrder_concr] exactly equivalent to
  [EFormulaOrder_dom].  For instance, [EFormulaOrder_concr] consider
  equivalent any formula with an empty concretisation, while
  [EFormulaOrder_dom] does only if they have the same domains.  We
  however have the converse. **)
Lemma EFormulaOrder_dom_concr : forall Phi1 Phi2,
  Phi1 ⊑ Phi2 → EFormulaOrder_concr Phi1 Phi2.
Proof.
  introv O Ex1 Ex2 G. gen StExt Phi1 Phi2 Phi1' Phi2'.
  induction C as [|C]; introv O Ex1 G Ex2.
   inverts Ex1. inverts Ex2. apply* O.
   destruct C as [M|phi]; inverts Ex1 as E1 Ex1; inverts Ex2 as E2 Ex2.
    forwards~: IHC Ex1 G Ex2. constructors.
     eapply double_incl_trans.
      apply double_incl_sym. applys~ frameMembraneInterface E1.
      applys~ frameMembraneInterface E2.
     eapply double_incl_trans.
      apply double_incl_sym. applys~ frameMembraneEInterface E1.
      applys~ frameMembraneEInterface E2.
     introv. forwards Equ: frameMembraneInDomain E1. rewrite <- Equ.
      forwards Equ': frameMembraneInDomain E2. rewrite <- Equ'.
      apply* O.
     introv. forwards Equ: frameMembraneInEDomain E1. rewrite <- Equ.
      forwards Equ': frameMembraneInEDomain E2. rewrite <- Equ'.
      apply* O.
     introv G'. skip. (* Admitted *)
    forwards~: IHC Ex1 G Ex2. constructors.
     skip. (* Admitted *)
     skip. (* Admitted *)
     introv. forwards Equ: frameFormulaInDomain; [ apply E1 |]. rewrite <- Equ.
      forwards Equ': frameFormulaInDomain; [ apply E2 |]. rewrite <- Equ'.
      apply* O.
     introv. forwards Equ: frameFormulaInEDomain; [ apply E1 |]. rewrite <- Equ.
      forwards Equ': frameFormulaInEDomain; [ apply E2 |]. rewrite <- Equ'.
      apply* O.
     introv G'. skip. (* Admitted *)
Qed.

Lemma DFormulaOrder_dom_concr : forall Phi1 Phi2 : DFormula,
  Phi1 ⊑ Phi2 → EFormulaOrder_concr (proj1_sig Phi1) (proj1_sig Phi2).
Proof. introv O. applys~ EFormulaOrder_dom_concr. Qed.

Lemma EFormulaOrder_dom_extends : forall Phi1 Phi2 Phi1' Phi2' C,
  Phi1 ⊑ Phi2 →
  EFormulaExtends Phi1 C Phi1' →
  EFormulaExtends Phi2 C Phi2' →
  Phi1' ⊑ Phi2'.
Proof.
  introv O Ex1 Ex2. constructors.
   forwards SI: EFormulaOrder_dom_same_interface O. clear - Ex1 Ex2 SI.
    gen Phi1 Phi1' Phi2 Phi2'. induction C; introv Ex1 O Ex2;
      inverts Ex1 as E1 Ex1; inverts Ex2 as E2 Ex2.
     autos*.
     applys* IHC. eapply double_incl_trans.
      apply double_incl_sym. apply* frameMembraneInterface.
      apply* frameMembraneInterface.
     applys* IHC. skip. (* Admitted *)
   skip. (* Admitted *)
   skip. (* Admitted *)
   skip. (* Admitted *)
   introv G. applys* EFormulaOrder_dom_concr O G.
Qed.


(** *** Simpler Orderings **)

Hypothesis ExtPoset : PosetDec.t Ext.

(** The two orders of the previous section are difficult to compute in
  practise. This section thus presents some orders which can be
  computed in practise. Furthermore, these orders imply the orders of
  the previous section: they can be a great way of actually computing
  (an approximation of) [EFormulaOrder_concr] and [EFormulaOrder_dom]. **)

Lemma EFormula_eq_EFormulaOrder_dom : forall Phi1 Phi2 : EFormula,
  EFormula_eq _ _ _ _ Phi1 Phi2 →
  Phi1 ⊑ Phi2.
Proof.
  introv E. induction E.
   inverts H as E.
    skip. (* Admitted *)
    skip. (* Admitted *)
    skip. (* Admitted *)
   apply~ EFormulaOrder_dom_refl.
   skip. (* Admitted *)
   applys~ EFormulaOrder_dom_trans IHE1 IHE2.
Qed.

Lemma DFormula_eq_DFormulaOrder_dom : forall Phi1 Phi2 : DFormula,
  DFormula_eq _ Phi1 Phi2 →
  Phi1 ⊑ Phi2.
Proof. introv. apply EFormula_eq_EFormulaOrder_dom. Qed.


End General.

