
(** This file contains the definitions of the separation-logic objects
  used in the abstract semantics of the example language. **)

(** The definitions and implementations are not meant to be efficient
  in any way, but to be very close to the paper version of this logic.
  In particular, the use of binary search trees for variables and
  abstract locations would be much more efficient than the direct
  representation of formulae as… formulae. The idea here is to be able
  to extend this work with new costructions easily, not to be
  efficient. **)

Set Implicit Arguments.
Require Export LibHeap.
Require Export CombinationLats.
Require Export ConcreteOWhile.
Require Export TLC.LibList.

Notation " [ ] " := nil : list_scope.
Notation " [ x ] " := (cons x nil) : list_scope.
Notation " [ x ; .. ; y ] " := (cons x .. (cons y nil) ..) : list_scope.

(* [logics] doesn’t work as expected, so I’m removing it from this tactic. *)
Ltac case_if_post ::=
  tryfalse.


Section General.

Section LocationScope.

(** * Locations **)

(** Type for abstract locations. **)

(** [L] represents singleton locations, [K] set locations and [KE]
  environments, which behave like set locations. **)
Variable L K KE : Type. (** Typically, set of integers **)
Hypothesis L_comparable : Comparable L.
Hypothesis K_comparable : Comparable K.
Hypothesis KE_comparable : Comparable KE.

(** These sets have to be infinite. **)
Hypothesis L_infinite : forall ls,
  exists (l : L), ~ mem l ls.
Hypothesis K_infinite : forall ks,
  exists (k : K), ~ mem k ks.
Hypothesis KE_infinite : forall ks,
  exists (k : KE), ~ mem k ks.
Hypothesis L_infinite_pick : forall ls,
  Pickable (fun l : L => ~ mem l ls).
Hypothesis K_infinite_pick : forall ks,
  Pickable (fun k : K => ~ mem k ks).
Hypothesis KE_infinite_pick : forall ks,
  Pickable (fun k : KE => ~ mem k ks).


Global Instance L_Inhab : Inhab L :=
  prove_Inhab (@pick _ _ (L_infinite_pick nil)).
Global Instance K_Inhab : Inhab K :=
  prove_Inhab (@pick _ _ (K_infinite_pick nil)).
Global Instance KE_Inhab : Inhab KE :=
  prove_Inhab (@pick _ _ (KE_infinite_pick nil)).

Inductive ALoc :=
  | singleton_loc : L → ALoc
  | set_loc : K → ALoc
  .
Coercion singleton_loc : L >-> ALoc.
Coercion set_loc : K >-> ALoc.

Global Instance ALoc_comparable : Comparable ALoc.
  prove_comparable_simple_inductive. Defined.

Global Instance ALoc_Equiv : EquivDec.t ALoc :=
  Comparable_EquivDec ALoc_comparable.

Definition ALoc_is_singleton h :=
  exists l, h = singleton_loc l.

Global Instance ALoc_is_singleton_decidable : forall h,
    Decidable (ALoc_is_singleton h).
  introv. destruct h as [l|k].
   applys~ decidable_make true. fold_bool. rew_refl. exists* l.
   applys~ decidable_make false. fold_bool. rew_refl. introv (l&A). inverts A.
Defined.

Definition ALoc_is_set h :=
  exists k, h = set_loc k.

Global Instance ALoc_is_set_decidable : forall h,
    Decidable (ALoc_is_set h).
  introv. destruct h as [l|k].
   applys~ decidable_make false. fold_bool. rew_refl. introv (k&A). inverts A.
   applys~ decidable_make true. fold_bool. rew_refl. exists* k.
Defined.

Definition ALoc_same_type h h' :=
  match h, h' with
  | singleton_loc l, singleton_loc l' => True
  | set_loc l, set_loc l' => True
  | _, _ => False
  end.

Lemma ALoc_same_type_refl : forall h,
  ALoc_same_type h h.
Proof. introv. destruct h; simpls~. Qed.

Global Instance ALoc_same_type_decidable : forall h h',
    Decidable (ALoc_same_type h h').
  introv. destruct h as [l|k].
   applys~ Decidable_equiv (ALoc_is_singleton h').
    destruct h'; simpl.
     iff~; eexists; autos*.
     iff I; inverts I; false*.
    typeclass.
   applys~ Decidable_equiv (ALoc_is_set h').
    destruct h'; simpl.
     iff I; inverts I; false*.
     iff~; eexists; autos*.
    typeclass.
Defined.

Global Instance ALoc_inhab : Inhab ALoc :=
  prove_Inhab ((arbitrary : K) : ALoc).

(** Getting the singleton locations from a list of locations. **)
Definition filter_l : list ALoc → list L := fun hs =>
  concat (map (fun h =>
    match h with
    | singleton_loc l => [l]
    | set_loc k => []
    end) hs).

Lemma filter_l_correct : forall hs l,
  mem l (filter_l hs) = mem (l : ALoc) hs.
Proof.
  extens. unfolds filter_l. rewrite concat_mem. iff I.
   lets (ls&Mls&Ml): (rm I). rewrite map_mem in Mls. lets (h&Mh&E): (rm Mls).
    destruct h; substs; tryfalse. simpl in Ml. rew_refl in Ml. inverts* Ml.
   exists [l]. splits; [| simpl; rew_refl~ ]. rewrite map_mem. exists* l.
Qed.

Lemma filter_l_No_duplicates : forall hs,
  No_duplicates hs →
  No_duplicates (filter_l hs).
Proof.
  introv ND. induction hs.
   constructors~.
   unfolds filter_l. destruct a as [l|k]; rew_list.
    constructors~.
     introv I. rewrite Mem_mem with (H := _) in I. rewrite concat_mem in I. lets (Ll&IL&Il): (rm I).
      rewrite map_mem in IL. lets (h&Ih&Eh): (rm IL). destruct h; substs; tryfalse.
      simpl in Il. rew_refl in Il. inverts~ Il. inverts ND as ND1 ND2.
      false ND1. erewrite~ Mem_mem.
     apply~ IHhs. inverts~ ND.
    apply~ IHhs. inverts~ ND.
Qed.

(** Getting the set locations from a list of locations. **)
Definition filter_k : list ALoc → list K := fun hs =>
  concat (map (fun h =>
    match h with
    | singleton_loc l => []
    | set_loc k => [k]
    end) hs).

Lemma filter_k_correct : forall hs k,
  mem k (filter_k hs) = mem (k : ALoc) hs.
Proof.
  extens. unfolds filter_k. rewrite concat_mem. iff I.
   lets (ls&Mls&Ml): (rm I). rewrite map_mem in Mls. lets (h&Mh&E): (rm Mls).
    destruct h; substs; tryfalse. simpl in Ml. rew_refl in Ml. inverts* Ml.
   exists [k]. splits; [| simpl; rew_refl~ ]. rewrite map_mem. exists* k.
Qed.

Lemma filter_k_No_duplicates : forall hs,
  No_duplicates hs →
  No_duplicates (filter_k hs).
Proof.
  introv ND. induction hs.
   constructors~.
   unfolds filter_k. destruct a as [l|k]; rew_list.
    apply~ IHhs. inverts~ ND.
    constructors~.
     introv I. rewrite Mem_mem with (H := _) in I. rewrite concat_mem in I. lets (Ll&IL&Il): (rm I).
      rewrite map_mem in IL. lets (h&Ih&Eh): (rm IL). destruct h; substs; tryfalse.
      simpl in Il. rew_refl in Il. inverts~ Il. inverts ND as ND1 ND2.
      false ND1. erewrite* Mem_mem.
     apply~ IHhs. inverts~ ND.
Qed.


Section ValueScope.

(** * Values **)

(** ** Abstract values **)

(** *** Basic Values **)

Variable ABasic : Type. (** Abstract basic values **)
Hypothesis ABasic_comparable : Comparable ABasic.
Hypothesis ABasicLat : LatticeDec.t ABasic.
Hypothesis ABasicTop : TopDec.t ABasic.
Variable ABasic_gamma : ABasic → ℘ basic_val.
Hypothesis ABasic_Gamma : Gamma.t _ _ ABasic_gamma.
Hypothesis ABasic_gamma_bot : forall b,
  ~ ABasic_gamma (⊥♯) b.
Hypothesis ABasic_gamma_top : forall b,
  ABasic_gamma (⊤♯) b.

(** *** Abstract Values **)

(** Domain for storing locations.  The abstract locations are considered
  representing disjuncts sets of concrete locations. **)
Definition ALocS := FinitePowerSet.t ALoc.

(*
(** Domain to account for [nil]. **)
Definition Anil := LiftBot.t UnitLattice.t.

Instance Anil_Lat : Lattice.t Anil.
  eapply LiftBot.Lat. typeclass.
Defined.

Instance Anil_LatDec : LatticeDec.t Anil.
  eapply LiftBot.LatDec; typeclass.
Defined.

Instance Anil_TopDec : TopDec.t Anil.
  apply~ LiftBot.TopDec.
Defined.
*)

Definition AClosure := FinitePowerSet.t (var * prog * KE).

Definition AClosure_basic_Equiv : EquivDec.t (var * prog * KE).
  apply Comparable_EquivDec. typeclass.
Defined.

Instance AClosure_LatDec : LatticeDec.t AClosure.
  applys~ FinitePowerSet.Lat AClosure_basic_Equiv. typeclass.
Defined.

Definition AValue := (ABasic * ALocS * AClosure)%type.

Global Instance AValue_comparable : Comparable AValue.
  repeat apply prod_comparable; typeclass.
Defined.

Global Instance AValue_Lat : LatticeDec.t AValue.
  typeclass.
Defined.

Global Instance AValue_Top : TopDec.t AValue.
  repeat apply ProductTopDec; typeclass.
Defined.

Definition ALoc_AValue (h : ALoc) : AValue :=
  (⊥♯, FinitePowerSet.const h, ⊥♯).
Coercion ALoc_AValue : ALoc >-> AValue.

Definition AClosure_singleton p x ke : AClosure :=
  FinitePowerSet.const (p, x, ke).

Definition AClosure_AValue ac : AValue :=
  (⊥♯, ⊥♯, ac).
Coercion AClosure_AValue : AClosure >-> AValue.

Definition AValue_closure p x ke : AValue :=
  AClosure_singleton p x ke.


(** *** Substitutions of locations in abstract values. **)

(** Replacing a location by a list of location may not look like a
  substitution. But in the more precise definition of [AValue] as a
  list of abstract values, it substitutes one part of the list to
  another list. As the implementation of [AValue] is not given here
  (and left parametric), the “substitution” can perform additionnal
  approximations and it yields a greater result in the lattice. **)

Definition ALocS_substs : ALocS → ALoc → ALocS → ALocS :=
  FinitePowerSet.substs _ _.

Definition AValue_substs (av : AValue) hs h : AValue :=
  let '(ab, al, ac) := av in
  let al' := ALocS_substs al h (FinitePowerSet.from_list _ _ hs) in
  (ab, al', ac).

Notation "av '[[' hs '//' h ']]'" := (AValue_substs av hs h)
  (at level 25, no associativity).

Lemma AValue_substs_basic : forall hs (h h0 : ALoc),
  mem h0 hs →
  (h0 : AValue) ⊑♯ h [[ hs // h ]].
Proof.
  introv I. repeat split; try (apply PosetDec.refl; apply EquivDec.refl).
  unfolds AValue_substs, ALoc_AValue, ALocS_substs.
  eapply PosetDec.trans;
    [| apply PosetDec.refl; apply EquivDec.sym; apply FinitePowerSet.substs_basic ].
  forwards (S1&S2): (FinitePowerSet.from_list_spec (T := ALoc)). apply* S2.
Qed.

Lemma AValue_substs_nosubsts : forall hs (h : ALoc) av,
  ~ (h : AValue) ⊑♯ av →
  av =♯ av [[ hs // h ]].
Proof.
  introv NO. destruct av as [p ac]. destruct p as [ab al].
  do 2 (split; try apply EquivDec.refl). unfolds AValue_substs, ALoc_AValue, ALocS_substs.
  rewrite~ FinitePowerSet.substs_nosubsts. introv O. apply NO.
  repeat split; try apply BotDec.prop. apply~ O.
Qed.

Lemma AValue_substs_congruence : forall hs h av1 av2,
  av1 ⊑♯ av2 →
  av1 [[ hs // h ]] ⊑♯ av2 [[ hs // h ]].
Proof.
  introv O. destruct av1 as [p ac1]. destruct p as [ab1 al1].
  destruct av2 as [p ac2]. destruct p as [ab2 al2].
  repeat split; try apply O. apply~ FinitePowerSet.substs_congruence. apply O.
Qed.


(** Substitution of environment locations. **)

Definition AClosure_Esubsts (c : AClosure) kes ke : AClosure :=
  match FinitePowerSet.to_list c with
  | Some cs =>
    FinitePowerSet.from_list _ AClosure_basic_Equiv (concat (map (fun c : _ * _ * _ =>
      let '(p, x, ke') := c in
      ifb ke = ke' then
        map (fun ke => (p, x, ke)) kes
      else [(p, x, ke')]) cs))
  | None => (⊤♯)
  end.

Definition AValue_Esubsts (av : AValue) kes ke : AValue :=
  let '(ab, al, ac) := av in
  let ac' := AClosure_Esubsts ac kes ke in
  (ab, al, ac').

Notation "av '|[[' kes '//' ke ']]'" := (AValue_Esubsts av kes ke)
  (at level 25, no associativity).

Lemma AClosure_Esubsts_basic : forall kes p x ke ke0,
  mem ke0 kes →
  AClosure_singleton p x ke0 ⊑♯ AClosure_Esubsts (AClosure_singleton p x ke) kes ke.
Proof.
  introv I. evar (ac : AClosure).
  asserts Eq: (AClosure_Esubsts (AClosure_singleton p x ke) kes ke = ac).
   simpl. rew_list. cases_if as D.
    unfolds* ac.
    fold_bool. rew_refl in D. false*.
  rewrite Eq. unfolds ac. clear Eq ac. eapply PosetDec.trans;
    [| apply PosetDec.refl; apply EquivDec.sym; apply FinitePowerSet.from_list_simpl ].
  apply FinitePowerSet.list_order. introv M. simpl in M. rew_refl in M. inverts M; tryfalse.
  eexists. splits*. rewrite* map_mem.
Qed.

Lemma AClosure_Esubsts_nosubsts : forall kes ke ac,
  (forall p x, ~ AClosure_singleton p x ke ⊑♯ ac) →
  ac =♯ AClosure_Esubsts ac kes ke.
Proof.
  introv NO. unfolds AClosure_Esubsts. destruct FinitePowerSet.to_list eqn: E.
   eapply EquivDec.trans; [| apply EquivDec.sym; apply FinitePowerSet.from_list_simpl ].
    repeat unfolds in E. substs. apply eq_EquivDec. fequals. induction* l.
     rew_list. destruct a as [px ke']. destruct px as [p x]. cases_if as Eq.
      fold_bool. rew_refl in Eq. substs. false NO.
       apply FinitePowerSet.list_order. introv M. simpl in M. rew_refl in M.
       inverts M; tryfalse. eexists. splits*. simpl. rew_refl*.
      rew_list. fequals. apply IHl.
       introv A. false NO. apply FinitePowerSet.list_order.
       introv M. simpl in M. rew_refl in M. inverts M; tryfalse.
       forwards (x'&Ex'&Mx'): FinitePowerSet.order_list A.
        simpl. rew_refl*.
       eexists. splits; [ exact Ex' |]. simpl. rew_refl*.
   apply eq_EquivDec. apply~ FinitePowerSet.to_list_none.
Qed.

Lemma AClosure_Esubsts_congruence : forall kes ke ac1 ac2,
  ac1 ⊑♯ ac2 →
  AClosure_Esubsts ac1 kes ke ⊑♯ AClosure_Esubsts ac2 kes ke.
Proof.
  introv O. unfolds AClosure_Esubsts.
  destruct FinitePowerSet.to_list as [l1|] eqn: Eq1 at 2; try apply TopDec.prop.
  eapply PosetDec.trans;
   [| apply PosetDec.refl; apply EquivDec.sym; apply FinitePowerSet.from_list_simpl ].
  destruct FinitePowerSet.to_list as [l2|] eqn: Eq2 at 1.
   eapply PosetDec.trans.
    apply PosetDec.refl. apply FinitePowerSet.from_list_simpl.
    apply FinitePowerSet.list_order. introv M.
     eexists. splits*. rewrite concat_mem in *. lets (Lx&ML&Mx): (rm M).
     rewrite map_mem in ML.  lets (y&M2&EL): (rm ML).
     destruct y as [px' ke']. destruct px' as [p x'].
     asserts O': (FinitePowerSet.const (p, x', ke') ⊑♯ ac2).
      applys @PosetDec.trans O.
      applys FinitePowerSet.to_list_some AClosure_basic_Equiv Eq2. eexists. splits*.
     forwards EO: FinitePowerSet.to_list_some AClosure_basic_Equiv Eq1.
     rewrite EO in O'. lets (e'&Me'&Ee'): (rm O'). inverts Ee'.
     cases_if as I; fold_bool; rew_refl in I.
      substs. eexists. splits.
       rewrite map_mem. eexists. splits*.
       cases_if as I; autos~. fold_bool. rew_refl* in I.
      substs. eexists. splits.
       rewrite map_mem. eexists. splits*.
       cases_if as I'; autos~. fold_bool. rew_refl* in I'.
   forwards Eac1: FinitePowerSet.to_list_none AClosure_basic_Equiv Eq2.
    asserts (x&N): (exists x, ~ mem x l1).
     forwards (ke'&Nke): KE_infinite (map snd l1).
     exists (arbitrary : string, arbitrary : prog, ke'). introv A. apply Nke.
     rewrite map_mem. eexists. splits*.
    false N. forwards (I1&I2): FinitePowerSet.to_list_some AClosure_basic_Equiv Eq1.
    lets (x'&Mx'&Eq): (rm I1).
     forwards E': sup_top O.
      applys eq_EquivDec Eac1.
      eapply PosetDec.trans.
       apply TopDec.prop.
       apply PosetDec.refl. apply~ @EquivDec.sym.
     simpl in Eq. substs*.
Qed.

Lemma AValue_Esubsts_basic : forall kes p x ke ke0,
  mem ke0 kes →
  AValue_closure p x ke0 ⊑♯ AValue_closure p x ke |[[ kes // ke ]].
Proof.
  introv I. repeat split; try (apply PosetDec.refl; apply EquivDec.refl).
  apply~ AClosure_Esubsts_basic.
Qed.

Lemma AValue_Esubsts_nosubsts : forall kes ke av,
  (forall p x, ~ AValue_closure p x ke ⊑♯ av) →
  av =♯ av |[[ kes // ke ]].
Proof.
  introv NO. destruct av as [p ac]. destruct p as [ab al].
  split; try apply EquivDec.refl. apply~ AClosure_Esubsts_nosubsts.
  introv A. false NO. repeat split; try apply BotDec.prop. apply A.
Qed.

Lemma AValue_Esubsts_congruence : forall kes ke av1 av2,
  av1 ⊑♯ av2 →
  av1 |[[ kes // ke ]] ⊑♯ av2 |[[ kes // ke ]].
Proof.
  introv O. destruct av1 as [p ac1]. destruct p as [ab1 al1].
  destruct av2 as [p ac2]. destruct p as [ab2 al2].
  repeat split; try apply O. apply AClosure_Esubsts_congruence. inverts~ O.
Qed.


(** *** All the location identifiers found in an abstract value. **)

Definition AVNames (av : AValue) : list ALoc :=
  let '(ab, al, ac) := av in
  match FinitePowerSet.to_list al with
  | Some hs => hs
  | None => nil
  end.

Definition ACENames (ac : AClosure) : list KE :=
  match FinitePowerSet.to_list ac with
  | Some cs => map (fun c => snd c) cs
  | None => nil
  end.

Definition AVENames (av : AValue) : list KE :=
  let '(ab, al, ac) := av in
  ACENames ac.

Lemma AVNames_substs : forall av h hs h0,
  mem h0 (AVNames (av [[ hs // h ]]))
    = mem h (AVNames av) &&& mem h0 hs
    ||| mem h0 (AVNames av) &&& decide (h ≠ h0).
Proof.
  introv. extens. rew_refl.
  destruct av as [p ac]. destruct p as [ab al].
  simpl. destruct (FinitePowerSet.to_list al) as [hs'|] eqn: E.
   match goal with |- context [ FinitePowerSet.to_list ?av ] =>
     destruct (FinitePowerSet.to_list av) eqn: E' end.
    forwards~ El: FinitePowerSet.from_list_to_list E'.
     forwards~ Eal: FinitePowerSet.from_list_to_list E. destruct al as [al|]; inverts E.
     asserts (h1&N1&N2): (exists h1, ~ mem h1 hs ∧ ~ mem h1 hs').
      lets (l1&N): L_infinite (filter_l (hs ++ hs')). exists l1.
      rewrite~ filter_l_correct in N. rewrite mem_app in N. rew_refl in N. rew_logic~ in N.
     iff I.
      forwards [(O1&[N|T'])|(O1&O2)]: (FinitePowerSet.substs_inv (T := ALoc)).
       applys @EquivDec.sym El.
       apply* FinitePowerSet.from_list_spec.
       right. splits~. forwards (x&Ex&Mx): FinitePowerSet.order_list O1.
        simpl. rew_refl*.
        simpl in Ex. substs~.
       false N2. forwards G: FinitePowerSet.gamma_top h1.
        apply EquivDec.sym in T'. apply PosetDec.refl in T'.
        lets G': (Gamma.monotone _ _ T' h0 G). simpl in G'. rewrite <- Exists_exists_st in G'.
        rewrite Exists_iff_exists_mem with (CA := _) in G'. lets (a&Ma&Ea): (rm G').
        cases_if~. false N2. asserts O2: (FinitePowerSet.const h1 ⊑♯ (⊤♯)).
         apply TopDec.prop.
         lets O3: (PosetDec.trans _ _ _ O2 T').
          forwards* O4: FinitePowerSet.from_list_to_list hs'. apply EquivDec.sym in O4.
          apply PosetDec.refl in O4. lets O5: (PosetDec.trans _ _ _ O3 O4).
          rewrite FinitePowerSet.from_list_spec in O5. lets (?&?&E): (rm O5).
          simpl in E. substs~.
       left. splits~.
        forwards (h'&Eh'&Mh'): FinitePowerSet.order_list O2.
         simpl. rew_refl*.
         simpl in Eh'. substs~.
        apply FinitePowerSet.from_list_spec in O1. lets (h'&Mh'&Eh'): O1.
         simpl in Eh'. substs~.
      asserts O: (FinitePowerSet.const h0 ⊑♯ FinitePowerSet.from_list _ _ l).
        inverts I as (M1&M2).
         eapply FinitePowerSet.substs_mem_in.
          applys~ @EquivDec.sym El.
          eapply PosetDec.trans; [| applys @PosetDec.refl Eal ].
           apply* FinitePowerSet.from_list_spec.
          apply* FinitePowerSet.from_list_spec.
         eapply FinitePowerSet.substs_mem_out.
          applys~ @EquivDec.sym El.
          eapply PosetDec.trans; [| applys @PosetDec.refl Eal ].
           apply* FinitePowerSet.from_list_spec.
          introv A. rewrite~ <- FinitePowerSet.const_same in A.
       rewrite FinitePowerSet.from_list_spec in O. lets (e'&Me'&Ee'): (rm O).
        simpl in Ee'. substs~.
    forwards T: FinitePowerSet.to_list_none E'. iff I; tryfalse.
     destruct al as [ahs|]; tryfalse.
      asserts (h1&N1&N2): (exists h1, ~ mem h1 ahs ∧ ~ mem h1 hs).
       lets (l&N): L_infinite (filter_l (ahs ++ hs)). exists l.
       rewrite~ filter_l_correct in N. rewrite mem_app in N. rew_refl in N. rew_logic~ in N.
     forwards [(O1&[N|T'])|(O1&O2)]: FinitePowerSet.substs_inv ALoc_Equiv h h1.
       applys eq_EquivDec T.
       apply TopDec.prop.
      forwards (h1'&Eh1'&Mh1'): FinitePowerSet.order_list O1.
       simpl. rew_refl*.
       simpl in Eh1'. substs. false* N1.
      false N1. forwards G: FinitePowerSet.gamma_top h1.
       apply EquivDec.sym in T'. apply PosetDec.refl in T'.
       lets G': (Gamma.monotone _ _ T' h0 G). simpl in G'. rewrite <- Exists_exists_st in G'.
       rewrite Exists_iff_exists_mem in G'. lets (a&Ma&Ea): (rm G'). clear G'.
       cases_if~. inverts E. false N1. asserts O2: (FinitePowerSet.const h1 ⊑♯ (⊤♯)).
        apply TopDec.prop.
        lets O3: (PosetDec.trans _ _ _ O2 T').
         forwards* O4: FinitePowerSet.from_list_to_list hs'. apply EquivDec.sym in O4.
         apply PosetDec.refl in O4. lets O5: (PosetDec.trans _ _ _ O3 O4).
         rewrite FinitePowerSet.from_list_spec in O5. lets (?&?&Eh1): (rm O5).
         simpl in Eh1. substs~.
      false N2. rewrite FinitePowerSet.from_list_spec in O1. lets (?&M&Eq): (rm O1).
       simpl in Eq. substs~.
   iff I.
    forwards: FinitePowerSet.to_list_none (rm E). substs.
     destruct FinitePowerSet.to_list eqn: E'; false*.
    inverts I as (M1&M2); false*.
Qed.

Lemma ACENames_Esubsts : forall c ke kes ke0,
  mem ke0 (ACENames (AClosure_Esubsts c kes ke))
    = mem ke (ACENames c) &&& mem ke0 kes
    ||| mem ke0 (ACENames c) &&& decide (ke ≠ ke0).
Proof.
  introv. extens. rew_refl. unfolds ACENames.
  destruct (FinitePowerSet.to_list c) as [c'|] eqn: E.
   match goal with |- context [ FinitePowerSet.to_list ?av ] =>
     destruct (FinitePowerSet.to_list av) eqn: E' end.
    forwards~ El: FinitePowerSet.from_list_to_list AClosure_basic_Equiv E'.
     forwards~ Eal: FinitePowerSet.from_list_to_list AClosure_basic_Equiv E.
     unfolds AClosure_Esubsts. rewrite E in *.
     forwards El1: @EquivDec.trans El FinitePowerSet.from_list_simpl.
     apply EquivDec.sym in El1.
     forwards El2: @EquivDec.trans El1 FinitePowerSet.from_list_simpl. iff I.
      rewrite map_mem in I. lets (y&My&Ey): (rm I). apply EquivDec.sym in El2.
       forwards (x'&Ex'&Mx'): FinitePowerSet.order_list AClosure_basic_Equiv My.
        applys @PosetDec.refl El2.
       simpl in Ex'. substs. rewrite concat_mem in Mx'. lets (Lx&ML&MxL): (rm Mx').
        rewrite map_mem in ML. lets (y&My'&Ey): (rm ML).
        destruct y as [px ke'], px as [p x]. substs. cases_if as I'.
         fold_bool. rew_refl in I. substs. left. splits.
          rewrite map_mem. eexists. splits*.
           rewrite map_mem with (CA := _) in MxL. lets (y&My2&Ey): (rm MxL). rew_refl* in I'.
          rewrite map_mem in MxL. lets (y&My2&Ey): (rm MxL). rew_refl in I'. substs~.
         simpl in MxL. rew_refl in MxL. inverts MxL; tryfalse. right. splits.
          rewrite map_mem. eexists. splits*.
          introv A. substs. fold_bool. rew_refl* in I'.
      rewrite map_mem. inverts I as (I1&I2); rewrite map_mem in I1; lets (y&My&Ey): (rm I1).
       destruct y as [xp ke']. destruct xp as [x p].
        forwards (x'&Ex'&Mx'): FinitePowerSet.order_list AClosure_basic_Equiv.
         applys @PosetDec.refl El2.
         rewrite concat_mem. eexists. splits.
          rewrite map_mem. eexists. splits*.
          cases_if as I.
           rewrite map_mem. eexists. splits*.
           fold_bool. rew_refl in I. substs*.
         simpl in Ex'. substs. eexists. splits*.
       destruct y as [xp ke']. destruct xp as [x p].
        forwards (x'&Ex'&Mx'): FinitePowerSet.order_list AClosure_basic_Equiv.
         applys @PosetDec.refl El2.
         rewrite concat_mem. eexists. splits.
          rewrite map_mem. eexists. splits*.
          cases_if as I.
           fold_bool. rew_refl in I. substs*.
           simpl. rew_refl*.
         simpl in Ex'. substs. eexists. splits*.
    iff I; tryfalse. destruct c as [c|]; tryfalse. simpl in E'.
     forwards T: FinitePowerSet.to_list_none AClosure_basic_Equiv E'.
     asserts (keA&N1&N2): (exists keA, ~ mem keA (map snd c) ∧ ~ mem keA kes).
      forwards (keA&KeAP): KE_infinite (map snd c ++ kes). rewrite mem_app in KeAP.
      rew_refl in KeAP. rew_logic in KeAP. exists* keA.
     forwards O: @TopDec.prop (FinitePowerSet.const (arbitrary : var, arbitrary : prog, keA)).
     rewrite <- T in O. forwards (R1&R2): FinitePowerSet.from_list_spec AClosure_basic_Equiv.
     forwards (e'&Me'&Ee'): R1 O. rewrite concat_mem in Me'. lets (Le&ML&MeL): (rm Me').
     rewrite map_mem in ML. lets (y&My&Ey): (rm ML).
     destruct y as [xp ke']. destruct xp as [x p]. substs. cases_if as I'.
      rewrite map_mem in MeL. lets (y'&My'&Ey'): (rm MeL). inverts* Ey'.
      simpl in MeL. rew_refl in MeL. inverts MeL as MeL; tryfalse. inverts MeL.
       false N1. rewrite* map_mem.
   iff I.
    forwards: FinitePowerSet.to_list_none AClosure_basic_Equiv (rm E). substs.
     destruct FinitePowerSet.to_list eqn: E'; false*.
    inverts I as (M1&M2); false*.
Qed.

Lemma AVENames_Esubsts : forall av ke kes ke0,
  mem ke0 (AVENames (av |[[ kes // ke ]]))
    = mem ke (AVENames av) &&& mem ke0 kes
    ||| mem ke0 (AVENames av) &&& decide (ke ≠ ke0).
Proof.
  introv. destruct av as [p ac]. destruct p as [ab al].
  apply ACENames_Esubsts.
Qed.

Lemma AVNames_Esubsts : forall av ke kes h,
  mem h (AVNames (av |[[ kes // ke ]]))
  = mem h (AVNames av).
Proof. introv. destruct av as [p ac]. destruct p as [ab al]. reflexivity. Qed.

Lemma AVENames_substs : forall av h hs ke,
  mem ke (AVENames (av [[ hs // h ]]))
  = mem ke (AVENames av).
Proof. introv. destruct av as [p ac]. destruct p as [ab al]. reflexivity. Qed.

Local Opaque AValue_Lat.
Local Opaque AValue_substs.


(** ** Abstract store values **)
(** These also contain a special [undefined] abstract value. **)

Definition Aundef := LiftBot.t UnitLattice.t.

Instance Aundef_Lat : LatticeDec.t Aundef.
  eapply LiftBot.LatDec; typeclass.
Defined.

Instance Aundef_Top : TopDec.t Aundef.
  apply~ LiftBot.TopDec.
Defined.

Definition ASValue := (AValue * Aundef)%type.

Global Instance ASValue_comparable : Comparable ASValue.
  typeclass.
Defined.

Global Instance ASValue_Lat : LatticeDec.t ASValue.
  repeat apply ProductLatticeDec; typeclass.
Defined.

Global Instance ASValue_Top : TopDec.t ASValue.
  repeat apply ProductTopDec; typeclass.
Defined.

Definition AValue_ASValue (av : AValue) : ASValue := (av, ⊥♯).
Coercion AValue_ASValue : AValue >-> ASValue.

Definition ASValue_undef : ASValue := (⊥♯, ⊤♯).

(** *** Substitutions of locations in abstract store values. **)

Definition ASValue_substs (asv : ASValue) (hs : list ALoc) (h : ALoc) : ASValue :=
  let (av, au) := asv in
  (av [[ hs // h ]], au).

Notation "av '~[[' hs '//' h ']]'" := (ASValue_substs av hs h)
  (at level 25, no associativity).

Definition ASValue_Esubsts (asv : ASValue) (kes : list KE) (ke : KE) : ASValue :=
  let (av, au) := asv in
  (av |[[ kes // ke ]], au).

Notation "av '~|[[' kes '//' ke ']]'" := (ASValue_Esubsts av kes ke)
  (at level 25, no associativity).

Lemma ASValue_substs_basic : forall hs (h h0 : ALoc),
  mem h0 hs →
  (h0 : ASValue) ⊑♯ h ~[[ hs // h ]].
Proof.
  introv M. split.
   apply~ AValue_substs_basic.
   autos*.
Qed.

Lemma ASValue_substs_nosubsts : forall hs (h : ALoc) av,
  ~ (h : ASValue) ⊑♯ av →
  av =♯ av ~[[ hs // h ]].
Proof.
  introv O. destruct av as [av au]. split.
   apply~ AValue_substs_nosubsts. introv O'. false O. split~. apply BotDec.prop.
   autos*.
Qed.

Lemma ASValue_substs_congruence : forall hs h av1 av2,
  av1 ⊑♯ av2 →
  av1 ~[[ hs // h ]] ⊑♯ av2 ~[[ hs // h ]].
Proof.
  introv O. destruct av1 as [av1 au1], av2 as [av2 au2]. split.
   apply~ AValue_substs_congruence. inverts~ O.
   inverts~ O.
Qed.

Lemma ASValue_Esubsts_basic : forall kes p x (ke ke0 : KE),
  mem ke0 kes →
  (AValue_closure p x ke0 : ASValue) ⊑♯ AValue_closure p x ke ~|[[ kes // ke ]].
Proof.
  introv M. split.
   apply~ AValue_Esubsts_basic.
   autos*.
Qed.

Lemma ASValue_Esubsts_nosubsts : forall kes ke av,
  (forall p x, ~ (AValue_closure p x ke : ASValue) ⊑♯ av) →
  av =♯ av ~|[[ kes // ke ]].
Proof.
  introv O. destruct av as [av au]. split.
   apply~ AValue_Esubsts_nosubsts. introv O'. false O. split.
    apply O'.
    apply BotDec.prop.
   autos*.
Qed.

Lemma ASValue_Esubsts_congruence : forall kes ke av1 av2,
  av1 ⊑♯ av2 →
  av1 ~|[[ kes // ke ]] ⊑♯ av2 ~|[[ kes // ke ]].
Proof.
  introv O. destruct av1 as [av1 au1], av2 as [av2 au2]. split.
   apply~ AValue_Esubsts_congruence. inverts~ O.
   inverts~ O.
Qed.


(** *** All the location identifiers found in the abstract value. **)

Definition ASVNames (asv : ASValue) : list ALoc :=
  let (av, au) := asv in
  AVNames av.

Lemma ASVNames_substs : forall av h hs h0,
  mem h0 (ASVNames (av ~[[ hs // h ]]))
    = mem h (ASVNames av) &&& mem h0 hs
    ||| mem h0 (ASVNames av) &&& decide (h ≠ h0).
Proof. introv. destruct av as [av au]. apply AVNames_substs. Qed.

Definition ASVENames (asv : ASValue) : list KE :=
  let (av, au) := asv in
  AVENames av.

Lemma ASVENames_Esubsts : forall av ke kes ke0,
  mem ke0 (ASVENames (av ~|[[ kes // ke ]]))
    = mem ke (ASVENames av) &&& mem ke0 kes
    ||| mem ke0 (ASVENames av) &&& decide (ke ≠ ke0).
Proof. introv. destruct av as [av au]. apply AVENames_Esubsts. Qed.

Lemma ASVNames_Esubsts : forall av ke kes h,
  mem h (ASVNames (av ~|[[ kes // ke ]]))
  = mem h (ASVNames av).
Proof. introv. destruct av as [av au]. simpl. rewrite~ AVNames_Esubsts. Qed.

Lemma ASVENames_substs : forall av h hs ke,
  mem ke (ASVENames (av ~[[ hs // h ]]))
  = mem ke (ASVENames av).
Proof. introv. destruct av as [av au]. simpl. rewrite~ AVENames_substs. Qed.

Local Opaque ASValue_Lat.


(** * Objects **)

(** ** Definition **)

(** (Partial) objects can be represented in two ways. In either way, a
  description of a finite member of properties is given. One way stops
  here. The other describes a value representing every other
  properties, except for a finite number: it describes a cofinite
  number of fields. Lemma [AObject_cofinite_star_decompose] can help
  to understand the intuition behind of [AObject_cofinite]. **)

Inductive AObject : Type :=
  | AObject_finite : list (property * ASValue) → AObject
  | AObject_cofinite : list (property * ASValue) → list property → ASValue → AObject
  .

(** A given field can not be described twice. **)

Inductive AObject_correct : AObject → Prop :=
  | AObject_correct_finite : forall fvs,
    (** Each field only appears once **)
    (forall f v1 v2 n1 n2,
      Nth n1 fvs (f, v1) →
      Nth n2 fvs (f, v2) →
      n1 = n2) →
    AObject_correct (AObject_finite fvs)
  | AObject_correct_cofinite : forall fvs fs v,
    (** Each field only appears once in each list **)
    (forall f v1 v2 n1 n2,
      Nth n1 fvs (f, v1) →
      Nth n2 fvs (f, v2) →
      n1 = n2) →
    (forall f n1 n2,
      Nth n1 fs f →
      Nth n2 fs f →
      n1 = n2) →
    (** The finite and cofinite lists represent disjunct properties **)
    (forall f,
      mem_assoc f fvs →
      mem f fs) →
    AObject_correct (AObject_cofinite fvs fs v)
  .

Definition DObject := { o : AObject | AObject_correct o }.

(** An object whose all fields are undefined. **)
Definition AObject_undef : AObject :=
  AObject_cofinite nil nil ASValue_undef.

Definition DObject_undef : DObject.
  refine (exist _ AObject_undef _). constructors.
   introv A. inverts A.
   introv A. inverts A.
   introv A. inverts A.
Defined.

(** An empty partial object. **)
Definition AObject_empty : AObject :=
  AObject_finite nil.

Definition DObject_empty : DObject.
  refine (exist _ AObject_empty _). constructors.
   introv A. inverts A.
Defined.

(** A bottom object. **)
Definition AObject_bottom : AObject :=
  AObject_cofinite nil nil (⊥♯).

Definition DObject_bottom : DObject.
  refine (exist _ AObject_bottom _). constructors.
   introv A. inverts A.
   introv A. inverts A.
   introv A. inverts A.
Defined.

Global Instance AObject_comparable : Comparable AObject.
  prove_comparable_simple_inductive.
Defined.

Global Instance DObject_comparable : Comparable DObject.
  constructors. intros (o1&OK1) (o2&OK2). applys Decidable_equiv (o1 = o2).
   iff I; inverts~ I. fequals.
   typeclass.
Defined.


(** ** Reading an Writing **)

(** Reading a field of an object. **)
Definition read_AObject (o : AObject) (f : property) :=
  match o with
  | AObject_finite fvs =>
    if mem_assoc f fvs then Some (assoc f fvs)
    else None
  | AObject_cofinite fvs fs v =>
    if mem_assoc f fvs then Some (assoc f fvs)
    else if mem f fs then None
    else Some v
  end.

Definition read_DObject (o : DObject) :=
  read_AObject (proj1_sig o).

(** Writing a field of an object. **)
Definition write_AObject (o : AObject) (f : property) v : AObject :=
  match o with
  | AObject_finite fvs =>
    AObject_finite ((f, v) :: remove_assoc f fvs)
  | AObject_cofinite fvs fs vr =>
    ifb mem f fs then
      AObject_cofinite ((f, v) :: remove_assoc f fvs) fs vr
    else AObject_cofinite ((f, v) :: fvs) (f :: fs) vr
  end.

Definition write_DObject (o : DObject) (f : property) (v : ASValue) : DObject.
  refine (exist _ (write_AObject (proj1_sig o) f v) _).
  destruct o as [o OK]. destruct o as [fvs|fvs fs vr]; simpl; inverts OK as O1 O2 O3.
   constructors.
   (** Unicity of fields **)
   introv N1 N2. inverts N1 as N1; inverts N2 as N2.
    autos*.
    apply Nth_mem with (CA := _) in N2. apply mem_mem_assoc in N2.
     rewrite mem_assoc_remove_assoc_eq in N2. false*.
    eapply Nth_mem in N1. apply mem_mem_assoc in N1.
     rewrite mem_assoc_remove_assoc_eq in N1. false*.
    fequals. applys* remove_assoc_Nth_same N1 N2. typeclass.
   case_if as M; constructors.
    (** Unicity of fields in the finite list **)
    introv N1 N2. inverts N1 as N1; inverts N2 as N2.
     autos*.
     apply Nth_mem with (CA := _) in N2. apply mem_mem_assoc in N2.
      rewrite mem_assoc_remove_assoc_eq in N2. false*.
     apply Nth_mem with (CA := _) in N1. apply mem_mem_assoc in N1.
      rewrite mem_assoc_remove_assoc_eq in N1. false*.
     fequals. applys* remove_assoc_Nth_same N1 N2. typeclass.
    (** Unicity of fields in the cofinite list **)
    introv N1 N2. applys* O2 N1 N2.
    (** Finite and cofinite lists represent disjunct properties **)
    introv M. rewrite mem_assoc_cons in M. rew_refl in M. inverts M as M; autos~.
    tests~: (f = f0). rewrite~ mem_assoc_remove_assoc_neq in M.
    (** Unicity of fields in the finite list **)
    introv N1 N2. inverts N1 as N1; inverts N2 as N2.
     autos*.
     apply Nth_mem with (CA := _) in N2. apply mem_mem_assoc in N2. false* O3.
     apply Nth_mem with (CA := _) in N1. apply mem_mem_assoc in N1. false* O3.
     fequals. applys* O1 N1 N2.
    (** Unicity of fields in the cofinite list **)
    introv N1 N2. inverts N1 as N1; inverts N2 as N2; autos*.
     apply Nth_mem with (CA := _) in N2. false*.
     apply Nth_mem with (CA := _) in N1. false*.
    (** Finite and cofinite lists disjunct disjunct properties **)
    introv M. rewrite mem_assoc_cons in M. rew_refl in M. simpl. rew_refl.
    inverts M as M; autos*.
Defined.

Lemma read_AObject_undef : forall f,
  read_AObject AObject_undef f = Some ASValue_undef.
Proof. reflexivity. Qed.

Lemma read_AObject_empty : forall f,
  read_AObject AObject_empty f = None.
Proof. reflexivity. Qed.

Lemma read_AObject_bottom : forall f,
  read_AObject AObject_bottom f = Some (⊥♯).
Proof. reflexivity. Qed.

Lemma read_DObject_undef : forall f,
  read_DObject DObject_undef f = Some ASValue_undef.
Proof. reflexivity. Qed.

Lemma read_DObject_empty : forall f,
  read_DObject DObject_empty f = None.
Proof. reflexivity. Qed.

Lemma read_DObject_bottom : forall f,
  read_DObject DObject_bottom f = Some (⊥♯).
Proof. reflexivity. Qed.


(** Names present in an object. **)
Definition ONames_param N (ANames : ASValue → list N) (o : AObject) :=
  match o with
  | AObject_finite fvs =>
    concat (map (fun fv => ANames (snd fv)) fvs)
  | AObject_cofinite fvs fs v =>
    ANames v ++ concat (map (fun fv => ANames (snd fv)) fvs)
  end.

Definition ONames := ONames_param ASVNames.
Definition OENames := ONames_param ASVENames.

Definition DNames (o : DObject) :=
  ONames (proj1_sig o).

Definition DENames (o : DObject) :=
  OENames (proj1_sig o).


(** Renaming the identifiers of an object. **)
Definition AObject_map (substs : ASValue → ASValue)
    (o : AObject) : AObject :=
  match o with
  | AObject_finite fvs =>
    AObject_finite (LibList.map (fun fv =>
      let (f, v) := fv : _ * _ in
      (f, substs v)) fvs)
  | AObject_cofinite fvs fs v =>
    AObject_cofinite (LibList.map (fun fv =>
      let (f, v) := fv : _ * _ in
      (f, substs v)) fvs) fs (substs v)
  end.

Definition AObject_map_partial (substs : ASValue → option ASValue)
    (o : AObject) : option AObject :=
  match o with
  | AObject_finite fvs =>
    LibOption.map AObject_finite (LibList.map_partial (fun fv =>
      let (f, v) := fv : _ * _ in
      LibOption.map_on (substs v) (fun v' => (f, v'))) fvs)
  | AObject_cofinite fvs fs v =>
    LibOption.apply_on (substs v) (fun v' =>
      LibOption.map (fun fvs' => AObject_cofinite fvs' fs v')
        (LibList.map_partial (fun fv =>
          let (f, v) := fv : _ * _ in
          LibOption.map_on (substs v) (fun v' => (f, v'))) fvs))
  end.

Definition AObject_substs o hs h : AObject :=
  AObject_map (fun v => v ~[[ hs // h]]) o.

Definition AObject_Esubsts o hs h : AObject :=
  AObject_map (fun v => v ~|[[ hs // h]]) o.

Lemma AObject_map_correct : forall substs o,
  AObject_correct o →
  AObject_correct (AObject_map substs o).
Proof.
  introv C. destruct o as [fvs|fvs fs v]; inverts C as O1 O2 O3; constructors~.
   introv N1 N2. lets ((f1&v1')&Efv1&Nfv1): Nth_map_inv (rm N1).
    lets ((f2&v2')&Efv2&Nfv2): Nth_map_inv (rm N2). inverts Efv1. inverts* Efv2.
   introv N1 N2. lets ((f1&v1')&Efv1&Nfv1): Nth_map_inv (rm N1).
    lets ((f2&v2')&Efv2&Nfv2): Nth_map_inv (rm N2). inverts Efv1. inverts* Efv2.
   introv M. rewrite mem_assoc_map_fst in M. rewrite list_map_compose in M.
    apply O3. rewrite mem_assoc_map_fst.
    match goal with _ : context [ compose fst ?f ] |- _ =>
      asserts R: (compose fst f = fst) end.
    extens. intros (?&?). reflexivity.
    rewrite~ R in M.
Qed.

Lemma AObject_map_partial_correct : forall substs o o',
  AObject_correct o →
  AObject_map_partial substs o = Some o' →
  AObject_correct o'.
Proof.
  introv C E. destruct o as [fvs|fvs fs v]; inverts C as O1 O2 O3; simpl in E.
   lets (fvs1&E1&Efvs1): map_inv (rm E). apply map_partial_inv in E1.
    lets F1: Forall2_forall_Nth E1. apply Forall2_swap in E1. substs. constructors.
    introv N1 N2. forwards~ N1': Forall2_Nth_nth_def (f, v2) E1 N1.
    forwards~ N2': Forall2_Nth_nth_def (f, v1) E1 N2.
    forwards Eq1: Forall2_forall_Nth E1 N1 N1'. forwards Eq2: Forall2_forall_Nth E1 N2 N2'.
    do 2 destruct nth_def. lets (v1'&Ev1'&Eq1'): map_on_inv (rm Eq1).
    lets (v2'&Ev2'&Eq2'): map_on_inv (rm Eq2). inverts Eq1'. inverts Eq2'. apply* O1.
   lets (v1&E1&Ev1): apply_on_inv (rm E). lets (fvs1&E2&Efvs1): map_inv (rm Ev1).
    apply map_partial_inv in E2. lets F1: Forall2_forall_Nth E2. apply Forall2_swap in E2.
    substs. constructors~.
     introv N1 N2. forwards~ N1': Forall2_Nth_nth_def (f, v2) E2 N1.
      forwards~ N2': Forall2_Nth_nth_def (f, v1) E2 N2.
      forwards Eq1: Forall2_forall_Nth E2 N1 N1'. forwards Eq2: Forall2_forall_Nth E2 N2 N2'.
      do 2 destruct nth_def. lets (v1'&Ev1'&Eq1'): map_on_inv (rm Eq1).
      lets (v2'&Ev2'&Eq2'): map_on_inv (rm Eq2). inverts Eq1'. inverts Eq2'. apply* O1.
     introv M. forwards M': assoc_mem_assoc M. lets (n&N): mem_Nth M'.
      forwards~ N': Forall2_Nth_nth_def (f, ⊥♯ : ASValue) E2 N.
      forwards Eq: Forall2_forall_Nth E2 N N'. destruct nth_def.
      lets (v'&Ev'&Eq'): map_on_inv (rm Eq). inverts Eq'. apply O3.
      eapply mem_mem_assoc. apply* Nth_mem.
Qed.

Definition DObject_map substs (o : DObject) : DObject :=
  exist _ _ (AObject_map_correct substs (proj2_sig o)).

Definition DObject_map_partial substs (o : DObject) : option DObject :=
  let r := AObject_map_partial substs (proj1_sig o) in
  match r as r0 return r = r0 → _ with
  | Some r => fun E =>
    Some (exist _ r (AObject_map_partial_correct substs (proj2_sig o) E))
  | None => fun _ => None
  end eq_refl.

Definition DObject_substs (o : DObject) hs h : DObject :=
  exist _ (AObject_substs (proj1_sig o) hs h) (AObject_map_correct _ (proj2_sig o)).

Definition DObject_Esubsts (o : DObject) hs h : DObject :=
  exist _ (AObject_Esubsts (proj1_sig o) hs h) (AObject_map_correct _ (proj2_sig o)).


(** ** Properties **)

Lemma read_write_AObject_same : forall o f v,
  read_AObject (write_AObject o f v) f = Some v.
Proof.
  introv. destruct o as [fvs|fvs fs vr]; simpl.
   rewrite mem_assoc_cons. cases_if as I; fold_bool; rew_refl in I.
    case_if~. false*.
    rew_logic in I. false*.
   case_if as I; simpl; rewrite mem_assoc_cons; case_if as I'; fold_bool; rew_refl in I'.
    case_if~. false*.
    rew_logic in I'. false*.
    case_if~. false*.
    rew_logic in I'. false*.
Qed.

Lemma read_write_AObject_propagates : forall o f1 f2 v,
  f1 ≠ f2 →
  read_AObject (write_AObject o f1 v) f2 = read_AObject o f2.
Proof.
  introv D. destruct o as [fvs|fvs fs vr]; simpl.
   rewrite mem_assoc_cons. cases_if as I; fold_bool; rew_refl in I.
    inverts I as I; tryfalse~. rewrite~ mem_assoc_remove_assoc_neq in I. do 2 case_if*.
     rewrite~ assoc_remove_assoc.
    rew_logic in I. rewrite~ mem_assoc_remove_assoc_neq in I. case_if~. false*.
   case_if as I; simpl; rewrite mem_assoc_cons; cases_if as I'; fold_bool; rew_refl in I'.
    case_if~. false*.
    inverts I' as I; tryfalse~. rewrite~ mem_assoc_remove_assoc_neq in I. case_if~.
     rewrite~ assoc_remove_assoc.
    rew_logic in I'. rewrite~ mem_assoc_remove_assoc_neq in I'. do 2 case_if~; tryfalse*.
    inverts I' as I; tryfalse~. do 2 case_if~; false~.
    rew_logic in I'. repeat case_if~; fold_bool; rew_refl in *; tryfalse*.
Qed.

Lemma ONames_param_spec : forall N `{Comparable N} (ANames : ASValue → list N) o ns n f v,
  AObject_correct o →
  ONames_param ANames o = ns →
  read_AObject o f = Some v →
  mem n (ANames v) →
  mem n ns.
Proof.
  introv C E1 E2 M. substs. destruct o as [fvs|fvs fs vr]; simpl.
   rewrite concat_mem. simpl in E2. cases_if as I. fold_bool. rewrite mem_assoc_map_fst in I.
    rewrite map_mem in I. lets ((f'&v')&Mfv&Efv): (rm I). substs. inverts E2.
    eexists. splits*. rewrite map_mem. eexists. splits*.
    erewrite~ unique_assoc_assoc_mem. intros. inverts* C.
   rewrite mem_app. rew_refl. simpl in E2. cases_if as I; fold_bool; rew_refl in I.
    right. inverts E2. rewrite mem_assoc_map_fst in I. rewrite map_mem in I.
     lets ((f'&v')&Mfv&Efv): (rm I). substs. rewrite concat_mem.
      erewrite unique_assoc_assoc_mem in M; autos*.
       eexists. splits*. rewrite* map_mem.
       intros. inverts* C.
    left. cases_if. inverts* E2.
Qed.

Lemma ONames_param_inv : forall N `{Comparable N} (ANames : ASValue → list N) o ns n,
  AObject_correct o →
  ONames_param ANames o = ns →
  mem n ns →
  exists f v, read_AObject o f = Some v ∧ mem n (ANames v).
Proof.
  introv C E M. substs. destruct o as [fvs|fvs fs vr]; simpls.
   rewrite concat_mem in M. lets (ns&Mns&nns): (rm M). rewrite map_mem in Mns.
    lets ((f&v)&Mfv&Efv): (rm Mns). substs. exists f v. cases_if as I; fold_bool.
     rewrite mem_assoc_map_fst in I. rewrite map_mem in I.
      lets ((f'&v')&Mfv'&Efv'): (rm I). substs. splits~.
      erewrite~ unique_assoc_assoc_mem. intros. inverts* C.
     rew_refl in I. false I. rewrite mem_assoc_map_fst. rewrite* map_mem.
   rewrite mem_app in M. rew_refl in M. inverts M as M.
    forwards (f&NM): string_infinite (map fst fvs ++ fs). exists f vr.
     cases_if as I; fold_bool; rew_refl in I.
      false NM. rewrite mem_app. rew_refl. left. rewrite* mem_assoc_map_fst in I.
      cases_if as I'; fold_bool; rew_refl in I'.
       false NM. rewrite mem_app. rew_refl. right*.
       splits*.
    rewrite concat_mem in M. lets (ns&Mns&nns): (rm M). rewrite map_mem in Mns.
     lets ((f&v)&Mfv&Efv): (rm Mns). substs. exists f v. cases_if as I; fold_bool.
      rewrite mem_assoc_map_fst in I. rewrite map_mem in I.
       lets ((f'&v')&Mfv'&Efv'): (rm I). substs. splits~.
       erewrite~ unique_assoc_assoc_mem. intros. inverts* C.
      rew_refl in I. false I. rewrite mem_assoc_map_fst. rewrite* map_mem.
Qed.

Lemma ONames_substs : forall o h hs h0,
  mem h0 (ONames (AObject_substs o hs h))
  = mem h (ONames o) &&& mem h0 hs
  ||| mem h0 (ONames o) &&& decide (h ≠ h0).
Proof.
  introv. destruct o as [fvs|fvs fs v].
   simpl. extens. rew_refl. repeat rewrite concat_mem. iff I.
    lets (hs0&Mhs0&Mh0): (rm I). rewrite map_mem in Mhs0. lets (fv&Mfv&Efv): (rm Mhs0).
     rewrite map_mem in Mfv. lets (fv'&Mfv'&Efv'): (rm Mfv). destruct fv' as [f' v']. substs.
     simpl in Mh0. rewrite ASVNames_substs in Mh0. rew_refl in Mh0. inverts Mh0 as (M1&M2).
      left. splits~. eexists. rewrite* map_mem.
      right. splits~. eexists. rewrite* map_mem.
    inverts I as ((hs0&Mhs&Mh0)&I2); rewrite map_mem with (CA := _) in Mhs; lets ((f&v)&Mfv&Efv): (rm Mhs).
     eexists. splits.
      rewrite map_mem with (CA := _). eexists. splits*. rewrite* map_mem.
      substs. simpl. rewrite ASVNames_substs. rew_refl*.
     eexists. splits.
      rewrite map_mem with (CA := _). eexists. splits*. rewrite* map_mem.
      substs. simpl. rewrite ASVNames_substs. rew_refl*.
   simpl. extens. repeat rewrite mem_app. rew_refl. repeat rewrite concat_mem. iff I.
    lets [M|(hs0&Mhs0&Mh0)]: (rm I).
     rewrite ASVNames_substs in M. rew_refl in M. inverts* M.
     rewrite map_mem in Mhs0. lets (fv&Mfv&Efv): (rm Mhs0).
      rewrite map_mem in Mfv. lets (fv'&Mfv'&Efv'): (rm Mfv). destruct fv' as [f' v']. substs.
      simpl in Mh0. rewrite ASVNames_substs in Mh0. rew_refl in Mh0. inverts Mh0 as (M1&M2).
       left. splits~. right. eexists. rewrite* map_mem.
       right. splits~. right. eexists. rewrite* map_mem.
    inverts I as ([MA|(hs0&Mhs&Mh0)]&I2);
      try (rewrite map_mem in Mhs; lets ((f&v)&Mfv&Efv): (rm Mhs)).
     rewrite ASVNames_substs. rew_refl*.
     rewrite map_mem in Mhs. lets ((f'&v')&Mfv&Efv): (rm Mhs). substs.
      right. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits*. rewrite* map_mem.
       substs. simpl. rewrite ASVNames_substs. rew_refl*.
     rewrite ASVNames_substs. rew_refl*.
     rewrite map_mem in Mhs. lets ((f'&v')&Mfv&Efv): (rm Mhs). substs.
      right. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits*. rewrite* map_mem.
       substs. simpl. rewrite ASVNames_substs. rew_refl*.
Qed.

Lemma OENames_Esubsts : forall o ke kes ke0,
  mem ke0 (OENames (AObject_Esubsts o kes ke))
  = mem ke (OENames o) &&& mem ke0 kes
  ||| mem ke0 (OENames o) &&& decide (ke ≠ ke0).
Proof.
  introv. destruct o as [fvs|fvs fs v].
   simpl. extens. rew_refl. repeat rewrite concat_mem. iff I.
    lets (hs0&Mhs0&Mh0): (rm I). rewrite map_mem in Mhs0. lets (fv&Mfv&Efv): (rm Mhs0).
     rewrite map_mem in Mfv. lets (fv'&Mfv'&Efv'): (rm Mfv). destruct fv' as [f' v']. substs.
     simpl in Mh0. rewrite ASVENames_Esubsts in Mh0. rew_refl in Mh0. inverts Mh0 as (M1&M2).
      left. splits~. eexists. rewrite* map_mem.
      right. splits~. eexists. rewrite* map_mem.
    inverts I as ((hs0&Mhs&Mh0)&I2); rewrite map_mem in Mhs; lets ((f&v)&Mfv&Efv): (rm Mhs).
     eexists. splits.
      rewrite map_mem with (CA := _). eexists. splits*. rewrite* map_mem.
      substs. simpl. rewrite ASVENames_Esubsts. rew_refl*.
     eexists. splits.
      rewrite map_mem with (CA := _). eexists. splits*. rewrite* map_mem.
      substs. simpl. rewrite ASVENames_Esubsts. rew_refl*.
   simpl. extens. repeat rewrite mem_app. rew_refl. repeat rewrite concat_mem. iff I.
    lets [M|(hs0&Mhs0&Mh0)]: (rm I).
     rewrite ASVENames_Esubsts in M. rew_refl in M. inverts* M.
     rewrite map_mem in Mhs0. lets (fv&Mfv&Efv): (rm Mhs0).
      rewrite map_mem in Mfv. lets (fv'&Mfv'&Efv'): (rm Mfv). destruct fv' as [f' v']. substs.
      simpl in Mh0. rewrite ASVENames_Esubsts in Mh0. rew_refl in Mh0. inverts Mh0 as (M1&M2).
       left. splits~. right. eexists. rewrite* map_mem.
       right. splits~. right. eexists. rewrite* map_mem.
    inverts I as ([MA|(hs0&Mhs&Mh0)]&I2);
      try (rewrite map_mem in Mhs; lets ((f&v)&Mfv&Efv): (rm Mhs)).
     rewrite ASVENames_Esubsts. rew_refl*.
     rewrite map_mem in Mhs. lets ((f'&v')&Mfv&Efv): (rm Mhs). substs.
      right. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits*. rewrite* map_mem.
       substs. simpl. rewrite ASVENames_Esubsts. rew_refl*.
     rewrite ASVENames_Esubsts. rew_refl*.
     rewrite map_mem in Mhs. lets ((f'&v')&Mfv&Efv): (rm Mhs). substs.
      right. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits*. rewrite* map_mem.
       substs. simpl. rewrite ASVENames_Esubsts. rew_refl*.
Qed.

Lemma ONames_Esubsts : forall o ke kes h,
  mem h (ONames (AObject_Esubsts o kes ke))
  = mem h (ONames o).
Proof.
  introv. destruct o as [fvs|fvs fs v].
   simpl. extens. repeat rewrite concat_mem. iff (hs&Mhs&Ehs); rewrite map_mem in Mhs;
     lets ((f&v)&Mfv&Efv): (rm Mhs); substs.
    rewrite map_mem in Mfv. lets ((f'&v')&Mfv0&Efv): (rm Mfv). inverts Efv.
     simpl in Ehs. rewrite ASVNames_Esubsts in Ehs. eexists. splits*. rewrite* map_mem.
    eexists. splits.
     rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
     simpl. rewrite* ASVNames_Esubsts.
   simpl. extens. repeat rewrite mem_app. rew_refl. repeat rewrite concat_mem.
    iff [M|(hs&Mhs&Ehs)];
      try (rewrite map_mem in Mhs; lets ((f&v)&Mfv&Efv): (rm Mhs); substs);
      repeat rewrite ASVNames_Esubsts in *; autos*.
    rewrite map_mem in Mhs. lets ((f'&v')&Mfv0&Efv): (rm Mhs). substs.
     rewrite map_mem in Mfv0. lets ((f&v'')&Mfv&Efv'): (rm Mfv0). inverts Efv'.
     simpl in Ehs. rewrite ASVNames_Esubsts in Ehs.
     right. eexists. splits*. rewrite* map_mem.
    rewrite map_mem in Mhs. lets ((f'&v')&Mfv0&Efv): (rm Mhs). substs.
     right. eexists. splits.
      rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
      simpl. rewrite* ASVNames_Esubsts.
Qed.

Lemma OENames_substs : forall o h hs ke,
  mem ke (OENames (AObject_substs o hs h))
  = mem ke (OENames o).
Proof.
  introv. destruct o as [fvs|fvs fs v].
   simpl. extens. repeat rewrite concat_mem. iff (kes&Mkes&Ekes); rewrite map_mem in Mkes;
     lets ((f&v)&Mfv&Efv): (rm Mkes); substs.
    rewrite map_mem in Mfv. lets ((f'&v')&Mfv0&Efv): (rm Mfv). inverts Efv.
     simpl in Ekes. rewrite ASVENames_substs in Ekes. eexists. splits*. rewrite* map_mem.
    eexists. splits.
     rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
     simpl. rewrite* ASVENames_substs.
   simpl. extens. repeat rewrite mem_app. rew_refl. repeat rewrite concat_mem.
    iff [M|(kes&Mkes&Ekes)];
      try (rewrite map_mem in Mkes; lets ((f&v)&Mfv&Efv): (rm Mkes); substs);
      repeat rewrite ASVENames_substs in *; autos*.
    rewrite map_mem in Mkes. lets ((f'&v')&Mfv0&Efv): (rm Mkes). substs.
     rewrite map_mem in Mfv0. lets ((f&v'')&Mfv&Efv'): (rm Mfv0). inverts Efv'.
     simpl in Ekes. rewrite ASVENames_substs in Ekes.
     right. eexists. splits*. rewrite* map_mem.
    rewrite map_mem in Mkes. lets ((f'&v')&Mfv0&Efv): (rm Mkes). substs.
     right. eexists. splits.
      rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
      simpl. rewrite* ASVENames_substs.
Qed.

Lemma DNames_substs : forall o h hs h0,
  mem h0 (DNames (DObject_substs o hs h))
  = mem h (DNames o) &&& mem h0 hs
  ||| mem h0 (DNames o) &&& decide (h ≠ h0).
Proof. introv. unfolds DNames, DObject_substs. simpl. rewrite~ ONames_substs. Qed.

Lemma DENames_Esubsts : forall o ke kes ke0,
  mem ke0 (DENames (DObject_Esubsts o kes ke))
  = mem ke (DENames o) &&& mem ke0 kes
  ||| mem ke0 (DENames o) &&& decide (ke ≠ ke0).
Proof. introv. unfolds DENames, DObject_substs. simpl. rewrite~ OENames_Esubsts. Qed.

Lemma DNames_Esubsts : forall o ke kes h,
  mem h (DNames (DObject_Esubsts o kes ke))
  = mem h (DNames o).
Proof. introv. unfolds DNames, DObject_substs. simpl. rewrite~ ONames_Esubsts. Qed.

Lemma DENames_substs : forall o h hs ke,
  mem ke (DENames (DObject_substs o hs h))
  = mem ke (DENames o).
Proof. introv. unfolds DENames, DObject_substs. simpl. rewrite~ OENames_substs. Qed.


(** Falsehood of an object. **)
Definition DObject_incoherent o :=
  exists f v, read_DObject o f = Some v ∧ v =♯ ⊥♯.

Global Instance DObject_incoherent_decidable : forall o,
    Decidable (DObject_incoherent o).
  introv. destruct o as [o OK]. destruct o as [fvs|fvs fs v] eqn: E.
   applys Decidable_equiv (Exists (fun fv =>
     match read_AObject o (fst fv) with
     | Some v => v =♯ ⊥♯
     | None => False
     end) fvs).
    rewrite Exists_iff_exists_mem. iff I.
     lets ((f&vf)&Mf&Rf): (rm I).
      destruct read_AObject eqn: E'; tryfalse. exists f. substs*.
     lets (f&v&E'&B): (rm I). substs. unfolds in E'. simpl in E'. unfolds read_AObject.
      case_if as I; tryfalse.
      exists (f, v). simpl. rewrite I. inverts E'. splits~. apply~ assoc_mem_assoc.
    apply Exists_decidable. introv. destruct read_AObject; typeclass.
   applys Decidable_equiv ((v =♯ ⊥♯) ∨ Exists (fun fv =>
     match read_AObject o (fst fv) with
     | Some v => v =♯ ⊥♯
     | None => False
     end) fvs).
    rewrite Exists_iff_exists_mem. iff I.
     lets [E'|((f&vf)&Mf&Rf)]: (rm I).
      forwards (f&NM): string_infinite (map fst fvs ++ fs). exists f.
       unfolds read_DObject, read_AObject. simpl. cases_if as I'.
        fold_bool. rewrite mem_assoc_map_fst in I'. rewrite mem_app in NM. rew_refl* in NM.
        case_if*. fold_bool. rewrite mem_app in NM. rew_refl* in NM.
      destruct read_AObject eqn: E'; tryfalse. exists f. substs*.
     lets (f&v'&E'&B): (rm I). unfolds in E'. simpl in E'. unfolds read_AObject.
      case_if as I.
       right. rewrite E. exists (f, v'). simpl. rewrite I. inverts E'. splits~.
        apply~ assoc_mem_assoc.
       case_if as I'; tryfalse. left. inverts~ E'.
    apply or_decidable.
     typeclass.
     apply Exists_decidable. introv. destruct read_AObject; typeclass.
Defined.


(** Mapping objects **)

Lemma DObject_map_partial_AObject : forall substs o o',
  DObject_map_partial substs o = Some o' →
  AObject_map_partial substs (proj1_sig o) = Some (proj1_sig o').
Proof.
  introv E. unfolds in E. asserts A: (forall A B (o0 : option A) o1 f,
    match o0 as o return o0 = o → option B with
    | Some o' => fun E => Some (f o' E)
    | None => fun _ => None
    end eq_refl = Some o1 →
    exists o' E, o0 = Some o' ∧ o1 = f o' E).
   clear. introv E. destruct o0; inverts* E.
  lets (o''&E'&E1&E2): (rm A) (rm E). substs*.
Qed.

Lemma AObject_map_partial_DObject : forall substs o o',
  AObject_map_partial substs (proj1_sig o) = Some (proj1_sig o') →
  DObject_map_partial substs o = Some o'.
Proof.
  introv E. destruct o as [o OK], o' as [o' OK']. unfolds.
  asserts A: (forall A B (o0 : option A) o' f,
    o0 = Some o' →
    exists E', match o0 as o return o0 = o → option B with
    | Some o' => fun E => Some (f o' E)
    | None => fun _ => None
    end eq_refl = Some (f o' E')).
   clear. introv Eq. substs. eexists. fequals.
  forwards (E'&Eq): (rm A) E. rewrite Eq. repeat fequals.
Qed.

Lemma read_AObject_AObject_map : forall substs f o,
  AObject_correct o →
  read_AObject (AObject_map substs o) f = LibOption.map substs (read_AObject o f).
Proof.
  introv C. destruct o as [fvs|fvs fs v]; simpl; inverts C as O1 O2 O3;
    repeat (let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I);
    rewrite mem_assoc_map_fst in *; autos~;
    repeat match goal with I : istrue (mem _ _) |- _ =>
      let n := fresh "n" in let N := fresh "N" in forwards (n&N): mem_Nth (rm I) end;
    repeat match goal with N : Nth _ (map _ _) _ |- _ =>
      let x := fresh "x" in let E := fresh "E" in let N' := fresh "N" in
      forwards (x&E&N'): Nth_map_inv (rm N);
      match type of x with (_ * _)%type => destruct x; try inverts E as E
      | _ => idtac end end; simpls; substs.
   erewrite~ unique_assoc_assoc_mem.
    clear - O1. introv N1 N2. forwards ((f1&v1)&E1&N1'): Nth_map_inv (rm N1).
     forwards ((f2&v2)&E2&N2'): Nth_map_inv (rm N2). inverts E1. inverts* E2.
    rewrite map_mem with (CA := _). eexists. splits; [ apply* Nth_mem |]. simpl. fequals.
     erewrite* unique_assoc_assoc_mem. apply* Nth_mem.
   false I0. rewrite map_mem with (CA := _). eexists. splits; [ apply* Nth_mem |]. autos*.
   false I. rewrite map_mem with (CA := _). eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits; [ apply* Nth_mem |]. autos*.
    autos*.
   erewrite~ unique_assoc_assoc_mem.
    clear - O1. introv N1 N2. forwards ((f1&v1)&E1&N1'): Nth_map_inv (rm N1).
     forwards ((f2&v2)&E2&N2'): Nth_map_inv (rm N2). inverts E1. inverts* E2.
    rewrite map_mem with (CA := _). eexists. splits; [ apply* Nth_mem |]. simpl. fequals.
     erewrite* unique_assoc_assoc_mem. apply* Nth_mem.
   false I0. rewrite map_mem with (CA := _). eexists. splits; [ apply* Nth_mem |]. autos*.
   false I0. rewrite map_mem with (CA := _). eexists. splits; [ apply* Nth_mem |]. autos*.
   false I. rewrite map_mem with (CA := _). eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits; [ apply* Nth_mem |]. autos*.
    autos*.
   false I. rewrite map_mem with (CA := _). eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits; [ apply* Nth_mem |]. autos*.
    autos*.
Qed.

Lemma read_DObject_DObject_map : forall substs f o,
  read_DObject (DObject_map substs o) f = LibOption.map substs (read_DObject o f).
Proof. introv. destruct o as [o OK]. apply* read_AObject_AObject_map. Qed.

Lemma read_AObject_AObject_map_partial : forall substs f o o' v,
  AObject_correct o →
  AObject_map_partial substs o = Some o' →
  read_AObject o f = Some v →
  exists v', substs v = Some v' ∧ read_AObject o' f = Some v'.
Proof.
  introv C E1 E2. destruct o as [fvs|fvs fs vr]; simpls; inverts C as O1 O2 O3.
   cases_if as I; inverts E2. fold_bool. forwards (o1&E2&E3): map_inv (rm E1). substs.
    forwards F: map_partial_inv (rm E2). rewrite mem_assoc_map_fst in I.
    rewrite map_mem in I. lets ((f'&v')&Mfv&Efv): (rm I). substs.
    lets (n&N): mem_Nth (rm Mfv). forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
    forwards~ E1: Forall2_forall_Nth F N N'. lets (v&E2&E3): map_on_inv (rm E1).
    rewrite E3 in N'. simpl. eexists. splits.
     erewrite unique_assoc_assoc_mem; try eapply Nth_mem; autos*.
     cases_if as I'.
      erewrite* unique_assoc_assoc_mem.
       introv N1 N2. apply Forall2_swap in F.
        forwards N1': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N1.
        forwards N2': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N2.
        forwards~ E1': Forall2_forall_Nth F N1 N1'. destruct (nth_def (f', ⊥♯) n1 fvs).
        forwards~ E2': Forall2_forall_Nth F N2 N2'. destruct (nth_def (f', ⊥♯) n2 fvs).
        lets (v1&E1a&E1b): map_on_inv (rm E1'). lets (v2&E2a&E2b): map_on_inv (rm E2').
        inverts E1b. inverts* E2b.
       apply* Nth_mem.
      fold_bool. rew_refl in I'. false I'. rewrite mem_assoc_map_fst.
       rewrite map_mem with (CA := _). eexists. splits; [ eapply Nth_mem |]; autos*.
   cases_if as I; fold_bool.
    inverts E2. forwards (v1&E2&E3): apply_on_inv (rm E1).
     forwards (o1&E4&E5): map_inv (rm E3). substs.
     forwards F: map_partial_inv (rm E4). rewrite mem_assoc_map_fst in I.
     rewrite map_mem in I. lets ((f'&v')&Mfv&Efv): (rm I). substs.
     lets (n&N): mem_Nth (rm Mfv). forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
     forwards~ E1: Forall2_forall_Nth F N N'. lets (v&E3&E4): map_on_inv (rm E1).
     rewrite E4 in N'. simpl. eexists. splits.
      erewrite unique_assoc_assoc_mem; try eapply Nth_mem; autos*.
      cases_if as I'.
       erewrite* unique_assoc_assoc_mem.
        introv N1 N2. apply Forall2_swap in F.
         forwards N1': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N1.
         forwards N2': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N2.
         forwards~ E1': Forall2_forall_Nth F N1 N1'. destruct (nth_def (f', ⊥♯) n1 fvs).
         forwards~ E2': Forall2_forall_Nth F N2 N2'. destruct (nth_def (f', ⊥♯) n2 fvs).
         lets (v1'&E1a&E1b): map_on_inv (rm E1'). lets (v2'&E2a&E2b): map_on_inv (rm E2').
         inverts E1b. inverts* E2b.
        apply* Nth_mem.
       fold_bool. rew_refl in I'. false I'. rewrite mem_assoc_map_fst.
        rewrite map_mem with (CA := _). eexists. splits; [ eapply Nth_mem |]; autos*.
    rew_refl in I. cases_if as I'; inverts E2. fold_bool. rew_refl in I'.
     forwards (v1&E2&E3): apply_on_inv (rm E1).
     forwards (o1&E4&E5): map_inv (rm E3). substs.
     forwards F: map_partial_inv (rm E4). eexists. splits*. unfolds.
     cases_if as I2; fold_bool.
      false I. rewrite mem_assoc_map_fst in I2. rewrite map_mem in I2.
       lets ((f'&v')&Mfv&Efv): (rm I2). substs.
       lets (n&N): mem_Nth (rm Mfv). apply Forall2_swap in F.
       forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
       forwards~ E1: Forall2_forall_Nth F N N'. destruct (nth_def (f', ⊥♯) n fvs).
       lets (v1'&E1a&E1b): map_on_inv (rm E1). inverts E1b. rewrite mem_assoc_map_fst.
       rewrite map_mem with (CA := _). eexists. splits; [ eapply Nth_mem |]; autos*.
     rew_refl in I2. cases_if~. false*.
Qed.

Lemma read_AObject_AObject_map_partial_None : forall substs f o o',
  AObject_correct o →
  AObject_map_partial substs o = Some o' →
  read_AObject o f = None →
  read_AObject o' f = None.
Proof.
  introv C E1 E2. destruct o as [fvs|fvs fs vr]; simpls; inverts C as O1 O2 O3.
   cases_if as I; inverts E2. fold_bool. rew_refl in I. forwards (o1&E2&E3): map_inv (rm E1).
    substs. forwards F: map_partial_inv (rm E2). rewrite mem_assoc_map_fst in I.
    unfolds. cases_if as I'; fold_bool.
     rewrite mem_assoc_map_fst in I'. rewrite map_mem in I'. lets ((f'&v')&Mfv&Efv): (rm I').
      substs. lets (n&N): mem_Nth (rm Mfv). apply Forall2_swap in F.
      forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
      forwards~ E1: Forall2_forall_Nth F N N'. destruct (nth_def (f', ⊥♯) n fvs).
      lets (v1'&E1a&E1b): map_on_inv (rm E1). inverts E1b. false I.
      rewrite map_mem with (CA := _). eexists. splits; [ eapply Nth_mem |]; autos*.
     autos*.
   cases_if as I; fold_bool. cases_if as I'; fold_bool. rew_refl in I. clear E2.
    forwards (v1&E2&E3): apply_on_inv (rm E1). forwards (o1&E4&E5): map_inv (rm E3). substs.
    forwards F: map_partial_inv (rm E4). rewrite mem_assoc_map_fst in I.
    unfolds. cases_if as I2; fold_bool.
     rewrite mem_assoc_map_fst in I2. rewrite map_mem in I2. lets ((f'&v')&Mfv&Efv): (rm I2).
      substs. lets (n&N): mem_Nth (rm Mfv). apply Forall2_swap in F.
      forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
      forwards~ E1: Forall2_forall_Nth F N N'. destruct (nth_def (f', ⊥♯) n fvs).
      lets (v1'&E1a&E1b): map_on_inv (rm E1). inverts E1b. false I.
      rewrite map_mem with (CA := _). eexists. splits; [ eapply Nth_mem |]; autos*.
     cases_if~.
Qed.

Lemma read_AObject_AObject_map_partial_None_inv : forall substs f o o',
  AObject_correct o' →
  AObject_map_partial substs o = Some o' →
  read_AObject o' f = None →
  read_AObject o f = None.
Proof.
  introv C E1 E2. destruct o as [fvs|fvs fs vr]; simpls.
   cases_if as I; autos~. fold_bool. rew_refl in I. forwards (o1&E3&E4): map_inv (rm E1).
    substs. forwards F: map_partial_inv (rm E3). rewrite mem_assoc_map_fst in I.
    unfolds in E2. cases_if as I'; fold_bool. rewrite mem_assoc_map_fst in I'.
    rew_refl in I'. rewrite map_mem in I'. false (rm I'). rewrite map_mem in I.
    lets ((f'&v')&Mfv&Efv): (rm I). substs. lets (n&N): mem_Nth (rm Mfv).
    forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
    forwards~ E1: Forall2_forall_Nth F N N'. destruct (nth_def (f', ⊥♯) n o1).
    lets (v1'&E1a&E1b): map_on_inv (rm E1). inverts E1b. exists (f', v1'). splits~.
    applys~ Nth_mem N'.
   forwards (v1&E3&E4): apply_on_inv (rm E1). forwards (o1&E5&E6): map_inv (rm E4). substs.
    forwards F: map_partial_inv (rm E5). cases_if as I; fold_bool.
     rewrite mem_assoc_map_fst in I. unfolds in E2. cases_if as I'; fold_bool.
      rewrite mem_assoc_map_fst in I'. rew_refl in I'. rewrite map_mem in I'. false (rm I').
      rewrite map_mem in I. lets ((f'&v')&Mfv&Efv): (rm I). substs.
      lets (n&N): mem_Nth (rm Mfv). forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
      forwards~ E1: Forall2_forall_Nth F N N'. destruct (nth_def (f', ⊥♯) n o1).
      lets (v1'&E1a&E1b): map_on_inv (rm E1). inverts E1b. exists (f', v1'). splits~.
      applys~ Nth_mem N'.
     cases_if as I'; autos~. simpl in E2. cases_if as I2; tryfalse. cases_if as I3; tryfalse.
Qed.

Lemma read_DObject_DObject_map_partial : forall substs f o o' v,
  DObject_map_partial substs o = Some o' →
  read_DObject o f = Some v →
  exists v', substs v = Some v' ∧ read_DObject o' f = Some v'.
Proof.
  introv E1 E2. destruct o as [o OK], o' as [o' OK'].
  forwards E1': DObject_map_partial_AObject E1.
  applys~ read_AObject_AObject_map_partial OK E1' E2.
Qed.

Lemma read_DObject_DObject_map_partial_None : forall substs f o o',
  DObject_map_partial substs o = Some o' →
  read_DObject o f = None →
  read_DObject o' f = None.
Proof.
  introv E1 E2. destruct o as [o OK], o' as [o' OK'].
  forwards E1': DObject_map_partial_AObject E1.
  applys~ read_AObject_AObject_map_partial_None OK E1' E2.
Qed.

Lemma AObject_map_partial_None : forall substs o,
  AObject_correct o →
  AObject_map_partial substs o = None →
  exists f v, read_AObject o f = Some v ∧ substs v = None.
Proof.
  introv C E. destruct o as [fvs|fvs fs vr]; simpl in E.
   forwards E': option_map_none_inv (rm E).
    forwards E: map_partial_inv_none (rm E').
    rewrite Exists_iff_exists_mem in E. lets ((f&v)&Mfv&Efv): (rm E). clear E.
    destruct (substs v) eqn: E; tryfalse. exists f v. splits~.
    simpl. cases_if as I; fold_bool.
     erewrite* unique_assoc_assoc_mem. intros. inverts* C.
     rew_refl in I. false I. rewrite mem_assoc_map_fst. rewrite* map_mem.
   forwards E': apply_on_inv_none (rm E). inverts E' as E'.
    forwards (f&NM): string_infinite (map fst fvs ++ fs). exists f vr. splits~.
     simpl. cases_if as I; fold_bool.
      false NM. rewrite mem_app. rewrite mem_assoc_map_fst in I. rew_refl*.
      cases_if~; fold_bool. false NM. rewrite mem_app. rew_refl*.
    lets (v&Es&E): (rm E'). forwards E': option_map_none_inv (rm E).
     apply map_partial_inv_none in E'. rewrite Exists_iff_exists_mem in E'.
     lets ((f&v')&Mfv&Efv): (rm E'). destruct (substs v') eqn: E; tryfalse.
     exists f v'. splits~. simpl. cases_if as I; fold_bool.
      erewrite* unique_assoc_assoc_mem. intros. inverts* C.
      rew_refl in I. false I. rewrite mem_assoc_map_fst. rewrite* map_mem.
Qed.

Lemma DObject_map_partial_None : forall substs o,
  DObject_map_partial substs o = None →
  exists f v, read_DObject o f = Some v ∧ substs v = None.
Proof.
  introv E. destruct o as [o OK]. destruct (AObject_map_partial substs o) eqn: E'.
   rewrite AObject_map_partial_DObject with
     (o' := exist _ _ (AObject_map_partial_correct _ OK E')) in E; tryfalse. autos*.
  apply* AObject_map_partial_None.
Qed.


(** ** Domain **)

(** The domain of an object is either finite or cofinite. **)
Inductive AObject_domain :=
  (** Finite domain **)
  | AObject_domain_has : list property → AObject_domain
  (** Cofinite domain **)
  | AObject_domain_misses : list property → AObject_domain
  .

Definition AObject_domain_empty := AObject_domain_has nil.
Definition AObject_domain_full := AObject_domain_misses nil.

Global Instance AObject_domain_inhab : Inhab AObject_domain.
  applys prove_Inhab AObject_domain_empty.
Defined.

Global Instance AObject_domain_comparable : Comparable (AObject_domain).
  prove_comparable_simple_inductive.
Defined.


(** Describes whether a given field is in a given object domain. **)
Definition in_AObject_domain f d :=
  match d with
  | AObject_domain_has fs => mem f fs
  | AObject_domain_misses fs => ! mem f fs
  end.

Lemma in_AObject_domain_empty : forall f,
  in_AObject_domain f AObject_domain_empty = false.
Proof. introv. reflexivity. Qed.

Lemma in_AObject_domain_full : forall f,
  in_AObject_domain f AObject_domain_full = true.
Proof. introv. reflexivity. Qed.

Definition AObject_dom (o : AObject) :=
  match o with
  | AObject_finite fvs =>
    AObject_domain_has (map fst fvs)
  | AObject_cofinite fvs fs v =>
    AObject_domain_misses (removes (map fst fvs) fs)
  end.

Lemma AObject_dom_empty : AObject_dom AObject_empty = AObject_domain_empty.
Proof. reflexivity. Qed.

Lemma AObject_dom_undef : AObject_dom AObject_undef = AObject_domain_full.
Proof. reflexivity. Qed.

Definition DObject_dom (o : DObject) :=
  AObject_dom (proj1_sig o).

Lemma DObject_dom_empty : DObject_dom DObject_empty = AObject_domain_empty.
Proof. reflexivity. Qed.

Lemma DObject_dom_undef : DObject_dom DObject_undef = AObject_domain_full.
Proof. reflexivity. Qed.

Lemma AObject_dom_correct : forall o f,
  in_AObject_domain f (AObject_dom o) ↔ read_AObject o f ≠ None.
Proof.
  introv. unfolds in_AObject_domain. destruct o as [fvs|fvs fs v]; simpl; case_if as M;
    fold_bool; rewrite~ mem_assoc_map_fst in M; try rewrite removes_mem; rew_refl in *;
    iff~ I; try case_if~; try discriminate; tryfalse~; rew_logic*.
Qed.

Lemma DObject_dom_correct : forall o f,
  in_AObject_domain f (DObject_dom o) ↔ read_DObject o f ≠ None.
Proof. introv. apply AObject_dom_correct. Qed.

Lemma AObject_dom_non_empty : forall d,
  d ≠ AObject_domain_empty →
  exists f, in_AObject_domain f d.
Proof.
  introv D. destruct d as [l|l].
   destruct l as [|f ?].
    false* D.
    exists f. simpl. rew_refl*.
   lets (f&NI): string_infinite l. exists f. simpl. rew_refl*.
Qed.


Definition AObject_domain_compare d1 d2 :=
  forall f, in_AObject_domain f d1 = in_AObject_domain f d2.

Global Instance AObject_domain_compare_decidable : forall d1 d2,
    Decidable (AObject_domain_compare d1 d2).
  introv. applys Decidable_equiv
    match d1, d2 with
    | AObject_domain_has fs1, AObject_domain_has fs2 => double_incl fs1 fs2
    | AObject_domain_misses fs1, AObject_domain_misses fs2 => double_incl fs1 fs2
    | _, _ => False
    end.
   destruct d1, d2; iff I; tryfalse;
     (intro f; simpl; rewrite* (double_incl_mem f I))
     || ((eapply mem_double_incl; intro f; extens)
         || (lets (f&M): (string_infinite (l0 ++ l)); rewrite mem_app in M;
             rew_refl in M; rew_logic in M);
         lets I': (rm I) f; try apply eq_bool_iff in I';
         repeat rewrite <- AObject_dom_correct in I';
         simpl in I'; rew_refl* in I').
    rewrite~ <- iff_not_cancel.
   destruct d1, d2; simpl; typeclass.
Defined.

Lemma AObject_domain_compare_correct : forall o1 o2,
  AObject_domain_compare (AObject_dom o1) (AObject_dom o2) ↔
  (forall f, read_AObject o1 f ≠ None ↔ read_AObject o2 f ≠ None).
Proof.
  introv. iff I.
   introv. repeat rewrite <- AObject_dom_correct. applys eq_bool_iff I.
   introv. extens. repeat rewrite AObject_dom_correct. apply~ I.
Qed.

Lemma DObject_domain_compare_correct : forall o1 o2,
  AObject_domain_compare (DObject_dom o1) (DObject_dom o2) ↔
  (forall f, read_DObject o1 f ≠ None ↔ read_DObject o2 f ≠ None).
Proof. introv. apply~ AObject_domain_compare_correct. Qed.

Lemma AObject_domain_compare_refl : forall o,
  AObject_domain_compare o o.
Proof. intros o f. reflexivity. Qed.

Lemma AObject_domain_compare_sym : forall o1 o2,
  AObject_domain_compare o1 o2 →
  AObject_domain_compare o2 o1.
Proof. introv C. introv. rewrite~ C. Qed.

Lemma AObject_domain_compare_trans : forall o1 o2 o3,
  AObject_domain_compare o1 o2 →
  AObject_domain_compare o2 o3 →
  AObject_domain_compare o1 o3.
Proof. introv C12 C23. introv. rewrite~ C12. Qed.


Lemma AObject_map_partial_dom : forall m o1 o2,
  AObject_correct o1 →
  AObject_correct o2 →
  AObject_map_partial m o1 = Some o2 →
  AObject_domain_compare (AObject_dom o1) (AObject_dom o2).
Proof.
  introv C1 C2 E. intro f. extens. do 2 rewrite AObject_dom_correct.
  rewrite iff_not_cancel. iff I.
   applys~ read_AObject_AObject_map_partial_None E I.
   applys~ read_AObject_AObject_map_partial_None_inv E I.
Qed.

Lemma DObject_map_partial_dom : forall m o1 o2,
  DObject_map_partial m o1 = Some o2 →
  AObject_domain_compare (DObject_dom o1) (DObject_dom o2).
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2]. applys~ AObject_map_partial_dom m.
  applys~ DObject_map_partial_AObject E.
Qed.


Definition AObject_domain_intersect (d1 d2 : AObject_domain) :=
  exists f, in_AObject_domain f d1 ∧ in_AObject_domain f d2.

Lemma AObject_domain_intersect_sym : forall d1 d2,
  AObject_domain_intersect d1 d2 →
  AObject_domain_intersect d2 d1.
Proof. introv (f&I1&I2). exists* f. Qed.

Lemma AObject_domain_intersect_empty : forall d,
  ~ AObject_domain_intersect d AObject_domain_empty.
Proof. introv (f&_&I). rewrite* in_AObject_domain_empty in I. Qed.

Lemma AObject_domain_intersect_full : forall d,
  d ≠ AObject_domain_empty →
  AObject_domain_intersect d AObject_domain_full.
Proof.
  introv D. destruct d as [fv|fv].
   destruct fv; tryfalse~. exists p. splits~. simpl. rew_refl*.
   lets (f&N): string_infinite fv. exists f. splits~. simpl. rew_refl*.
Qed.

Lemma AObject_domain_intersect_misses : forall f1 f2,
  AObject_domain_intersect (AObject_domain_misses f1)
    (AObject_domain_misses f2).
Proof.
  (* By default, some Hypothesis are used by some tactics, but shouldn’t be. *)
  repeat match goal with H : _ |- _ => clear H end.
  introv. lets (f&N): (string_infinite (f1 ++ f2)). rewrite mem_app in N.
  rew_refl in N. rew_logic in N. exists f. simpl. splits; rew_refl*.
Qed.

Global Instance AObject_domain_intersect_decidable : forall d1 d2,
    Decidable (AObject_domain_intersect d1 d2).
  (* By default, some Hypothesis are used by some tactics, but shouldn’t be. *)
  repeat match goal with H : _ |- _ => clear H end.
  introv. destruct d1 as [f1|f1], d2 as [f2|f2].
   applys* Decidable_equiv (exists f, mem f f1 ∧ mem f f2).
    applys decidable_make (decide (Exists (fun f => mem f f2) f1)).
    extens. rew_refl. rewrite* Exists_iff_exists_mem.
   applys Decidable_equiv (exists f, mem f f1 ∧ ~ mem f f2).
    iff (f&I1&I2); eexists; splits~; simpls; rew_refl* in *; rew_refl~.
    applys decidable_make (decide (Exists (fun f => ~ mem f f2) f1)).
     extens. rew_refl. rewrite* Exists_iff_exists_mem.
   applys Decidable_equiv (exists f, ~ mem f f1 ∧ mem f f2).
    iff (f&I1&I2); eexists; splits~; simpls; rew_refl* in *; rew_refl~.
    applys decidable_make (decide (Exists (fun f => ~ mem f f1) f2)).
     extens. rew_refl. rewrite Exists_iff_exists_mem. iff* (f&I1&I2).
   applys Decidable_equiv True; [| typeclass ]. iff~ I.
    apply~ AObject_domain_intersect_misses.
Defined.

Record AObject_domain_merge d1 d2 d3 := {
  AObject_domain_merge_total : forall f,
    in_AObject_domain f d1 || in_AObject_domain f d2 = in_AObject_domain f d3 ;
  AObject_domain_merge_disjunct : forall f,
    in_AObject_domain f d1 → in_AObject_domain f d2 → False }.

Definition AObject_domain_simple_merge d1 d2 : { d3 | forall f,
    in_AObject_domain f d1 || in_AObject_domain f d2 = in_AObject_domain f d3 }.
   destruct d1 as [f1|f1], d2 as [f2|f2].
    apply (exist _ (AObject_domain_has (f1 ++ f2))). introv. simpl. rewrite~ mem_app.
    apply (exist _ (AObject_domain_misses (removes f1 f2))). introv. simpl.
     rewrite removes_mem. extens. rew_refl. rew_logic*.
    apply (exist _ (AObject_domain_misses (removes f2 f1))). introv. simpl.
     rewrite removes_mem. extens. rew_refl. rew_logic*.
    apply (exist _ (AObject_domain_misses
      (Remove_duplicates (filter (fun f => mem f f2) f1)))).
     introv. simpl. rewrite Remove_duplicates_mem. rewrite filter_mem_eq.
     extens. rew_refl. rew_logic*.
Defined.

Global Instance AObject_domain_merge_pickable_option : forall d1 d2,
    Pickable_option (AObject_domain_merge d1 d2).
  (* By default, some Hypothesis are used by some tactics, but shouldn’t be. *)
  repeat match goal with H : _ |- _ => clear H end.
  introv. destruct (decide (AObject_domain_intersect d1 d2)) eqn: E; fold_bool; rew_refl in E.
   applys pickable_option_make (@None AObject_domain).
    introv A. inverts A.
    introv (d3&M). lets (f&I1&I2): (rm E). false* AObject_domain_merge_disjunct.
   applys pickable_option_make (Some (proj1_sig (AObject_domain_simple_merge d1 d2))).
    introv Eq. inverts Eq. constructors.
     apply~ proj2_sig.
     introv I1 I2. false* E. exists f. splits~.
    introv _. eexists. autos*.
Defined.

Lemma AObject_domain_merge_congruence : forall d1 d1' d2 d2' d3 d3',
  AObject_domain_compare d1 d1' →
  AObject_domain_compare d2 d2' →
  AObject_domain_merge d1 d2 d3 →
  AObject_domain_merge d1' d2' d3' →
  AObject_domain_compare d3 d3'.
Proof.
  introv C1 C2 M1 M2. intro f. rewrite <- (AObject_domain_merge_total M1).
  rewrite <- (AObject_domain_merge_total M2). rewrite (C1 f). rewrite~ (C2 f).
Qed.

Lemma AObject_domain_merge_compare : forall d1 d2 d3 d4,
  AObject_domain_merge d1 d2 d3 →
  AObject_domain_merge d1 d2 d4 →
  AObject_domain_compare d3 d4.
Proof.
  introv M1 M2. applys~ AObject_domain_merge_congruence M1 M2;
    apply AObject_domain_compare_refl.
Qed.

Lemma AObject_domain_merge_defined_compare : forall d1 d1' d2 d2' d3,
  AObject_domain_compare d1 d1' →
  AObject_domain_compare d2 d2' →
  AObject_domain_merge d1 d2 d3 →
  AObject_domain_merge d1' d2' d3.
Proof.
  introv C1 C2 M. constructors.
   introv. rewrite <- (C1 f). rewrite <- (C2 f). apply~ M.
   introv I1 I2. rewrite <- (C1 f) in I1. rewrite <- (C2 f) in I2. apply* M.
Qed.

Lemma AObject_domain_merge_sym : forall d1 d2 d3,
  AObject_domain_merge d1 d2 d3 →
  AObject_domain_merge d2 d1 d3.
Proof.
  introv M. constructors.
   introv. rewrite <- (AObject_domain_merge_total M). extens. rew_refl*.
   introv I1 I2. apply* AObject_domain_merge_disjunct.
Qed.

Lemma AObject_domain_merge_empty : forall d,
  AObject_domain_merge d AObject_domain_empty d.
Proof.
  introv. constructors.
   introv. rewrite in_AObject_domain_empty. rew_bool~.
   introv. rewrite~ in_AObject_domain_empty.
Qed.

Lemma AObject_domain_merge_not_intersect : forall d1 d2 d3,
  AObject_domain_merge d1 d2 d3 →
  ~ AObject_domain_intersect d1 d2.
Proof. introv M (f&I1&I2). apply* M. Qed.

Lemma AObject_domain_not_intersect_merge : forall d1 d2,
  ~ AObject_domain_intersect d1 d2 →
  exists d3, AObject_domain_merge d1 d2 d3.
Proof.
  introv N. eexists (proj1_sig (AObject_domain_simple_merge d1 d2)). constructors.
   apply~ proj2_sig.
   introv I1 I2. false* N. exists f. splits~.
Qed.

Lemma DObject_incoherent_in_AObject_domain : forall o,
  DObject_incoherent o →
  exists f, in_AObject_domain f (DObject_dom o).
Proof. introv (f&v&R&E). exists f. rewrite DObject_dom_correct. rewrite R. discriminate. Qed.


(** ** Poset Structure **)

Definition DObject_order o1 o2 :=
  forall f,
    match read_DObject o1 f, read_DObject o2 f with
    | Some v1, Some v2 => v1 ⊑♯ v2
    | None, None => True
    | _, _ => False
    end.

Definition AObject_order o1 o2 :=
  forall f,
    match read_AObject o1 f, read_AObject o2 f with
    | Some v1, Some v2 => v1 ⊑♯ v2
    | None, None => True
    | _, _ => False
    end.

Local Instance AObject_order_decidable : forall o1 o2,
    Decidable (DObject_order o1 o2).
  intros (o1&OK1) (o2&OK2). unfolds DObject_order, read_DObject. simpl.
  destruct o1 as [fvs1|fvs1 fs1 vr1] eqn: E1; destruct o2 as [fvs2|fvs2 fs2 vr2] eqn: E2.
   applys Decidable_equiv (Forall (fun f =>
       match read_AObject o1 f, read_AObject o2 f with
       | Some v1, Some v2 => v1 ⊑♯ v2
       | None, None => True
       | _, _ => False
       end)
     (map fst (fvs1 ++ fvs2))).
    rewrite E1, E2. rewrite map_app. iff I.
     lets (O1&O2): Forall_app_inv (rm I). introv.
      tests I: (mem f (map fst (fvs1 ++ fvs2)));
        rewrite map_app in I; rewrite mem_app in I; rew_refl in I.
      inverts I as I.
       rewrite Forall_iff_forall_mem in O1. applys~ O1 I.
       rewrite Forall_iff_forall_mem in O2. applys~ O2 I.
      unfolds read_AObject. cases_if as I1; fold_bool.
       false I. left. rewrite~ <- mem_assoc_map_fst.
       cases_if~. cases_if as I2; fold_bool.
        false I. right. rewrite~ <- mem_assoc_map_fst.
     apply Forall_app.
      rewrite Forall_iff_forall_mem with (CA := _). introv M. apply I.
      rewrite Forall_iff_forall_mem with (CA := _). introv M. apply I.
    apply Forall_decidable. introv. do 2 destruct read_AObject; typeclass.
   applys Decidable_equiv False; [| typeclass ]. iff I; tryfalse.
    lets (f&M): (string_infinite (fs2 ++ map fst (fvs1 ++ fvs2))).
    rewrite map_app in M. repeat rewrite mem_app in M. rew_refl in M.
    forwards O: I f. cases_if as I1; fold_bool.
     false M. right. left. rewrite~ <- mem_assoc_map_fst.
     cases_if. cases_if as I1'; fold_bool.
      cases_if as I2; fold_bool.
       false* M.
   applys Decidable_equiv False; [| typeclass ]. iff I; tryfalse.
    lets (f&M): (string_infinite (fs1 ++ map fst (fvs1 ++ fvs2))).
    rewrite map_app in M. repeat rewrite mem_app in M. rew_refl in M.
    forwards O: I f. cases_if as I1; fold_bool.
     false M. right. left. rewrite~ <- mem_assoc_map_fst.
     cases_if as I1'; fold_bool.
      false* M.
      cases_if as I2; fold_bool.
       false M. right. right. rewrite~ <- mem_assoc_map_fst.
   applys Decidable_equiv (vr1 ⊑♯ vr2
       ∧ AObject_domain_compare (AObject_dom o1) (AObject_dom o2)
       ∧ Forall (fun f =>
           match read_AObject o1 f, read_AObject o2 f with
           | Some v1, Some v2 => v1 ⊑♯ v2
           | None, None => True
           | _, _ => False
           end)
     (map fst (fvs1 ++ fvs2))).
    rewrite E1, E2. rewrite map_app. iff I.
     lets (O0&II&I'): (rm I). simpl in II. lets (O1&O2): Forall_app_inv (rm I'). introv.
      tests I: (mem f (map fst (fvs1 ++ fvs2)));
        rewrite map_app in I; rewrite mem_app in I; rew_refl in I.
      inverts I as I.
       rewrite Forall_iff_forall_mem in O1. applys~ O1 I.
       rewrite Forall_iff_forall_mem in O2. applys~ O2 I.
      unfolds read_AObject. cases_if as I1; fold_bool.
       false I. left. rewrite~ <- mem_assoc_map_fst.
       cases_if as I1'; fold_bool.
        cases_if~. cases_if as I2; fold_bool.
         false I. right. rewrite~ <- mem_assoc_map_fst.
         cases_if as I2'; fold_bool.
          forwards (D1&D2): eq_bool_iff (II f). simpl in D2. repeat rewrite removes_mem in D2.
           rew_refl in D2. rew_logic in D2. false I. left. forwards* [I3|I3]: D2.
            rew_refl* in I2'.
        cases_if as I2; fold_bool.
         false I. right. rewrite~ <- mem_assoc_map_fst.
         cases_if as I2'; fold_bool.
          forwards (D1&D2): eq_bool_iff (II f). simpl in D1. repeat rewrite removes_mem in D1.
           rew_refl in D1. rew_logic in D1. forwards* [I3|I3]: D1.
            rew_refl* in I1'.
          apply~ O0.
     splits.
      lets (f&M): (string_infinite ((fs1 ++ fs2) ++ map fst (fvs1 ++ fvs2))).
       rewrite map_app in M. repeat rewrite mem_app in M. rew_refl in M.
       forwards O: I f. cases_if as I1; fold_bool.
        false M. right. left. rewrite~ <- mem_assoc_map_fst.
        cases_if as I1'; fold_bool.
         false* M.
         cases_if as I2; fold_bool.
          false M. right. right. rewrite~ <- mem_assoc_map_fst.
          cases_if~.
      apply DObject_domain_compare_correct with (o1 := exist _ _ OK1) (o2 := exist _ _ OK2).
       introv. lets I': I f. unfolds read_DObject, proj1_sig.
        do 2 destruct read_AObject; tryfalse; iff~; try discriminate.
      apply Forall_app.
       rewrite Forall_iff_forall_mem with (CA := _). introv M. apply I.
       rewrite Forall_iff_forall_mem with (CA := _). introv M. apply I.
    repeat apply and_decidable; try typeclass.
     apply Forall_decidable. introv. do 2 destruct read_AObject; typeclass.
Defined.

Lemma DObject_order_refl : forall o, DObject_order o o.
Proof. intros o f. destruct~ read_DObject. Qed.

Lemma DObject_order_trans : forall o1 o2 o3,
  DObject_order o1 o2 →
  DObject_order o2 o3 →
  DObject_order o1 o3.
Proof.
  introv O1 O2; intro f. lets O1': O1 f. lets O2': O2 f.
  do 3 destruct~ read_DObject; tryfalse. applys @PosetDec.trans O1' O2'.
Qed.

Definition DObject_equiv o1 o2 :=
  forall f,
    match read_DObject o1 f, read_DObject o2 f with
    | Some v1, Some v2 => v1 =♯ v2
    | None, None => True
    | _, _ => False
    end.

Definition AObject_equiv o1 o2 :=
  forall f,
    match read_AObject o1 f, read_AObject o2 f with
    | Some v1, Some v2 => v1 =♯ v2
    | None, None => True
    | _, _ => False
    end.

Lemma DObject_order_antisym : forall o1 o2,
  DObject_order o1 o2 →
  DObject_order o2 o1 →
  DObject_equiv o1 o2.
Proof.
  introv O1 O2; intro f. lets O1': O1 f. lets O2': O2 f.
  do 2 destruct~ read_DObject; tryfalse.
Qed.

Global Instance DObject_Equiv : EquivDec.t DObject.
  apply EquivDec.Make with DObject_equiv.
   introv; intro f. destruct~ read_DObject.
   introv E; intro f. lets E': E f. do 2 destruct~ read_DObject.
   introv E1 E2; intro f. lets E1': E1 f. lets E2': E2 f. do 3 destruct* read_DObject.
  introv. apply decidable_sumbool.
   applys Decidable_equiv (DObject_order x y ∧ DObject_order y x).
    iff O.
     apply* DObject_order_antisym.
     splits; intro f; lets O': O f; do 2 destruct~ read_DObject.
   typeclass.
Defined.

Global Instance DObject_Poset : PosetDec.t DObject.
  apply PosetDec.Make with DObject_Equiv DObject_order.
   introv E. intro f. lets E': E f. do 2 destruct~ read_DObject.
   apply* DObject_order_antisym.
   apply* DObject_order_trans.
   introv. apply decidable_sumbool. typeclass.
Defined.


(** ** Star on Objects **)

(** This operation takes two objects of different domains and merge them. **)

Definition AObject_star (o1 o2 : AObject) : option AObject :=
  'let merge := fun fvs1 fvs2 fs2 v2 =>
    ifb ~ Exists (fun fv1 => Exists (fun fv2 => fst fv1 = fst fv2) fvs2) fvs1 then
      ifb incl (map fst fvs1) fs2 then
        Some (AObject_cofinite (fvs1 ++ fvs2) fs2 v2)
      else None
    else None in
  match o1, o2 with
  | AObject_finite fvs1, AObject_finite fvs2 =>
    ifb ~ Exists (fun fv1 => Exists (fun fv2 => fst fv1 = fst fv2) fvs2) fvs1 then
      Some (AObject_finite (fvs1 ++ fvs2))
    else None
  | AObject_finite fvs1, AObject_cofinite fvs2 fs2 v2 =>
    merge fvs1 fvs2 fs2 v2
  | AObject_cofinite fvs1 fs1 v1, AObject_finite fvs2 =>
    merge fvs2 fvs1 fs1 v1
  | AObject_cofinite fvs1 fs1 v1, AObject_cofinite fvs2 fs2 v2 => None
  end.

Lemma AObject_star_correct : forall o1 o2 o3,
  AObject_star o1 o2 = Some o3 →
  AObject_correct o1 →
  AObject_correct o2 →
  AObject_correct o3.
Proof.
  introv E C1 C2. destruct o1 as [fvs1|fvs1 fs1 v1]; destruct o2 as [fvs2|fvs2 fs2 v2];
    inverts C1 as NDfv1 NDf1 MA1; inverts C2 as NDfv2 NDf2 MA2; simpl in E;
    (repeat let I := fresh "I" in cases_if as I); inverts E; constructors;
    try solve [
      (** Unicity of fields in the finite list **)
      introv N1 N2; lets [N1'|(n1'&En1'&N1')]: Nth_app_inv (rm N1);
        lets [N2'|(n2'&En2'&N2')]: Nth_app_inv (rm N2);
        try applys~ NDfv1 N1' N2';
        try (forwards: NDfv2 N1' N2'; substs*);
        try (forwards: NDfv1 N1' N2'; substs*);
        fold_bool; rew_refl in I; false I;
        do 2 (rewrite Exists_iff_exists_mem with (CA := _);
        eexists; splits; [ apply* Nth_mem |]); autos*
    | (** Unicity of fields in the cofinite list **)
      introv N1 N2; autos*
    | (** Finite and cofinite lists disjunct disjunct properties **)
      introv M; rewrite app_mem_assoc in M; rew_refl in M; forwards~ [M'|M']: (rm M);
      fold_bool; rew_refl in I0; rewrite incl_Forall in I0;
      rewrite Forall_iff_forall_mem in I0; rewrite~ mem_assoc_map_fst in M' ].
Qed.

Definition DObject_star (o1 o2 : DObject) : option DObject.
  destruct (AObject_star (proj1_sig o1) (proj1_sig o2)) as [o3|] eqn: E.
   refine (Some (exist _ o3 _)).
    destruct o1. destruct o2. apply* AObject_star_correct.
   exact None.
Defined.

Lemma DObject_star_AObject_star : forall o1 o2 o3,
  DObject_star o1 o2 = Some o3 →
  AObject_star (proj1_sig o1) (proj1_sig o2) = Some (proj1_sig o3).
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3].
  unfolds in E. simpls. asserts A: (forall A B (o0 : option A) o1 f,
    match o0 as o return o0 = o → option B with
    | Some o' => fun E => Some (f o' E)
    | None => fun _ => None
    end eq_refl = Some o1 →
    exists o' E, o0 = Some o' ∧ o1 = f o' E).
   clear. introv Eq. destruct o0; inverts* Eq.
  forwards (o'&E'&E1&E2): (rm A) (rm E). inverts~ E2.
Qed.

Lemma AObject_star_DObject_star : forall o1 o2 o3,
  AObject_star (proj1_sig o1) (proj1_sig o2) = Some (proj1_sig o3) →
  DObject_star o1 o2 = Some o3.
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3].
  unfolds. simpls. asserts A: (forall A B (o0 : option A) o' f,
    o0 = Some o' →
    exists E', match o0 as o return o0 = o → option B with
    | Some o' => fun E => Some (f o' E)
    | None => fun _ => None
    end eq_refl = Some (f o' E')).
   clear. introv Eq. substs. eexists. fequals.
  forwards (E'&Eq): (rm A) E. rewrite Eq. repeat fequals.
Qed.

Lemma DObject_star_AObject_star_None : forall o1 o2,
  DObject_star o1 o2 = None →
  AObject_star (proj1_sig o1) (proj1_sig o2) = None.
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2].
  unfolds in E. simpls. asserts A: (forall A B (o0 : option A) f,
    match o0 as o return o0 = o → option B with
    | Some o' => fun E => Some (f o' E)
    | None => fun _ => None
    end eq_refl = None →
    o0 = None).
   clear. introv Eq. destruct o0; inverts* Eq.
  applys A E.
Qed.

Lemma AObject_star_DObject_star_None : forall o1 o2,
  AObject_star (proj1_sig o1) (proj1_sig o2) = None →
  DObject_star o1 o2 = None.
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2].
  unfolds. simpls. asserts A: (forall A B (o0 : option A) f,
    o0 = None →
    match o0 as o return o0 = o → option B with
    | Some o' => fun E => Some (f o' E)
    | None => fun _ => None
    end eq_refl = None).
   clear. introv Eq. substs~.
  applys A E.
Qed.

Lemma AObject_star_commut : forall o1 o2 o3,
  AObject_correct o1 →
  AObject_correct o2 →
  AObject_star o1 o2 = Some o3 →
  exists o3', AObject_star o2 o1 = Some o3' ∧ AObject_equiv o3 o3'.
Proof.
  introv C1 C2 E. unfolds AObject_star. do 2 let_name. substs.
  destruct o1 as [fvs1|fvs1 fs1 vr1], o2 as [fvs2|fvs2 fs2 vr2]; simpl in E;
    (repeat let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I); inverts E.
   eexists. splits*. intro f. simpl.
    repeat (let I := fresh "I" in case_if as I; fold_bool; rew_refl in I).
     rewrite app_mem_assoc in I1. rew_refl in I1. inverts I1 as I1.
      rewrite~ app_assoc_l. rewrite~ app_assoc_r. introv M. false I.
       rewrite Exists_iff_exists_mem. eexists. splits.
        apply* mem_assoc_assoc.
        rewrite Exists_iff_exists_mem. eexists. splits.
         apply* mem_assoc_assoc.
         autos*.
      rewrite~ app_assoc_r. rewrite~ app_assoc_l. introv M. false I.
       rewrite Exists_iff_exists_mem. eexists. splits.
        apply* mem_assoc_assoc.
        rewrite Exists_iff_exists_mem. eexists. splits.
         apply* mem_assoc_assoc.
         autos*.
     false I2. rewrite app_mem_assoc in I1. rew_refl in I1. rewrite app_mem_assoc. rew_refl*.
     false I1. rewrite app_mem_assoc in I3. rew_refl in I3. rewrite app_mem_assoc. rew_refl*.
     autos*.
   apply not_not_elim in I. false I0. repeat rewrite Exists_iff_exists_mem with (CA := _) in *.
    lets (a&Ma&Ea): (rm I). rewrite Exists_iff_exists_mem in Ea. lets (a'&Ma'&Ea'): (rm Ea).
    eexists. splits*. rewrite* Exists_iff_exists_mem.
   eexists. splits*. intro f. simpl.
    repeat (let I := fresh "I" in case_if as I; fold_bool; rew_refl in I); autos*.
   eexists. splits*. intro f. simpl.
    repeat (let I := fresh "I" in case_if as I; fold_bool; rew_refl in I); autos*.
Qed.

Lemma DObject_star_commut : forall o1 o2 o3,
  DObject_star o1 o2 = Some o3 →
  exists o3', DObject_star o2 o1 = Some o3' ∧ o3 =♯ o3'.
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3].
  forwards E': DObject_star_AObject_star (rm E). simpl in E'.
  forwards~ (o3'&Eo3'&Oo3'): AObject_star_commut (rm E').
  forwards~ Co3': AObject_star_correct Eo3'. exists (exist _ _ Co3'). splits~.
  apply~ AObject_star_DObject_star.
Qed.

Lemma AObject_star_assoc : forall o1 o2 o3 o12 o123,
  AObject_correct o1 →
  AObject_correct o2 →
  AObject_correct o3 →
  AObject_star o1 o2 = Some o12 →
  AObject_star o12 o3 = Some o123 →
  exists o23 o123',
    AObject_star o2 o3 = Some o23
    ∧ AObject_star o1 o23 = Some o123'
    ∧ AObject_equiv o123 o123'.
Proof.
  introv C1 C2 C3 E1 E2. unfolds AObject_star. repeat let_name. substs.
  destruct o1 as [fvs1|fvs1 fs1 vr1], o2 as [fvs2|fvs2 fs2 vr2], o3 as [fvs3|fvs3 fs3 vr3];
    let ifs := repeat let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I in
    simpl in E1; ifs; inverts E1; simpl in E2; ifs; inverts E2.
   do 2 eexists. splits*.
    unfolds let_binding. cases_if as I2.
     autos*.
     fold_bool. rew_refl in I2. apply not_not_elim in I2.
      rewrite Exists_iff_exists_mem with (CA := _) in I2. lets (a&Ma&Ea): (rm I2).
      rewrite Exists_iff_exists_mem with (CA := _) in Ea. lets (a'&Ma'&Ea'): (rm Ea).
      rewrite mem_app in Ma'. rew_refl in Ma'. inverts Ma' as M.
       false I0. rewrite Exists_iff_exists_mem with (CA := _). eexists. splits*.
        rewrite* Exists_iff_exists_mem.
       false I1. rewrite Exists_iff_exists_mem with (CA := _). eexists. splits.
        rewrite mem_app. rew_refl*.
        rewrite* Exists_iff_exists_mem.
    intro f. rew_list. destruct~ read_AObject.
   apply not_not_elim in I. rewrite Exists_iff_exists_mem with (CA := _) in I.
    lets (a&Ma&Ea): (rm I). rewrite Exists_iff_exists_mem with (CA := _) in Ea.
    lets (a'&Ma'&Ea'): (rm Ea). false I1. rewrite Exists_iff_exists_mem with (CA := _).
    eexists. splits.
     rewrite mem_app. rew_refl*.
     rewrite* Exists_iff_exists_mem.
   do 2 eexists. splits*.
    unfolds let_binding.
     repeat let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I.
      autos*.
      false I5. apply mem_incl with (H := _). introv M. applys incl_mem I3. rewrite map_app.
       rewrite mem_app. rew_refl*.
      apply not_not_elim in I4. rewrite Exists_iff_exists_mem with (CA := _) in I4.
       lets (a&Ma&Ea): (rm I4). rewrite Exists_iff_exists_mem with (CA := _) in Ea.
       lets (a'&Ma'&Ea'): (rm Ea). rewrite mem_app in Ma'. rew_refl in Ma'. inverts Ma' as M.
        false I1. rewrite Exists_iff_exists_mem. eexists. splits*.
         rewrite* Exists_iff_exists_mem.
        false I2. rewrite Exists_iff_exists_mem. eexists. splits.
         rewrite mem_app. rew_refl*.
         rewrite* Exists_iff_exists_mem.
    intro f. rew_list. destruct~ read_AObject.
   false I0. apply mem_incl with (H := _). introv M. applys incl_mem I3. rewrite map_app.
    rewrite mem_app. rew_refl*.
   apply not_not_elim in I. rewrite Exists_iff_exists_mem with (CA := _) in I.
    lets (a&Ma&Ea): (rm I). rewrite Exists_iff_exists_mem with (CA := _) in Ea.
    lets (a'&Ma'&Ea'): (rm Ea). false I1. rewrite Exists_iff_exists_mem. eexists. splits.
     rewrite mem_app. rew_refl*.
     rewrite* Exists_iff_exists_mem.
   do 2 eexists. splits*.
    unfolds let_binding.
     repeat let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I.
      autos*.
      false*.
      apply not_not_elim in I5. rewrite Exists_iff_exists_mem with (CA := _) in I5.
       lets (a&Ma&Ea): (rm I5). rewrite Exists_iff_exists_mem with (CA := _) in Ea.
       lets (a'&Ma'&Ea'): (rm Ea). rewrite mem_app in Ma'. rew_refl in Ma'. inverts Ma' as M.
        false I3. rewrite Exists_iff_exists_mem. eexists. splits*.
         rewrite Exists_iff_exists_mem. eexists. splits.
          rewrite mem_app. rew_refl*.
          autos*.
        false I1. rewrite Exists_iff_exists_mem. eexists. splits*.
         rewrite* Exists_iff_exists_mem.
    intro f. unfolds read_AObject. cases_if as I5; fold_bool; rew_refl in I5.
     repeat rewrite app_mem_assoc in *. rew_refl in I5. cases_if as I6.
      inverts I5 as I5.
       rewrite~ app_assoc_l. rewrite app_assoc_r.
        rewrite~ app_assoc_l.
        introv M. false I3. rewrite Exists_iff_exists_mem. eexists. splits.
         apply* mem_assoc_assoc.
         rewrite Exists_iff_exists_mem. eexists. splits.
          rewrite mem_app. rew_refl. left. apply* mem_assoc_assoc.
          autos*.
       inverts I5 as I5.
        rewrite app_assoc_r.
         rewrite~ app_assoc_l. rewrite~ app_assoc_l.
         introv M. false I3. rewrite Exists_iff_exists_mem. eexists. splits.
          apply* mem_assoc_assoc.
          rewrite Exists_iff_exists_mem. eexists. splits.
           rewrite mem_app. rew_refl. left. apply* mem_assoc_assoc.
           autos*.
        asserts ND1: (~ mem_assoc f fvs1).
         introv M. false I1. rewrite Exists_iff_exists_mem. eexists. splits.
          apply* mem_assoc_assoc.
          rewrite Exists_iff_exists_mem. eexists. splits.
           apply* mem_assoc_assoc.
           autos*.
         asserts ND3: (~ mem_assoc f fvs3).
          introv M. false I. rewrite Exists_iff_exists_mem. eexists. splits.
           apply* mem_assoc_assoc.
           rewrite Exists_iff_exists_mem. eexists. splits.
            apply* mem_assoc_assoc.
            autos*.
         repeat rewrite~ app_assoc_r.
      fold_bool. rew_refl* in I6.
     repeat rewrite app_mem_assoc in *. rew_refl in I5. rew_logic in I5.
      repeat cases_if; fold_bool; rew_refl in *; tryfalse*; autos*.
   false* I0.
   apply not_not_elim in I. rewrite Exists_iff_exists_mem with (CA := _) in I.
    lets (a&Ma&Ea): (rm I). rewrite Exists_iff_exists_mem with (CA := _) in Ea.
    lets (a'&Ma'&Ea'): (rm Ea). false I2. rewrite Exists_iff_exists_mem. eexists. splits*.
     rewrite* Exists_iff_exists_mem. eexists. splits.
      rewrite mem_app. rew_refl*.
      autos*.
   do 2 eexists. splits*.
    unfolds let_binding.
     repeat let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I.
      autos*.
      false I5. apply mem_incl with (H := _). introv M. rewrite map_app in M.
       rewrite mem_app in M. rew_refl in M. inverts M as M; applys* incl_mem M.
      apply not_not_elim in I4. rewrite Exists_iff_exists_mem with (CA := _) in I4.
       lets (a&Ma&Ea): (rm I4). rewrite Exists_iff_exists_mem with (CA  := _) in Ea.
       lets (a'&Ma'&Ea'): (rm Ea). rewrite mem_app in Ma. rew_refl in Ma. inverts Ma as M.
        false I0. rewrite Exists_iff_exists_mem. eexists. splits*.
         rewrite* Exists_iff_exists_mem.
        false I2. rewrite Exists_iff_exists_mem. eexists. splits*.
         rewrite* Exists_iff_exists_mem. eexists. splits.
          rewrite mem_app. rew_refl*.
          autos*.
    intro f. rew_list. unfolds read_AObject. cases_if as I4; fold_bool; rew_refl in I4.
     repeat rewrite app_mem_assoc in *. rew_refl in I4. cases_if as I5.
      inverts I4 as I4.
       rewrite~ app_assoc_l. rewrite app_assoc_r.
        rewrite~ app_assoc_l.
        introv M. false I. rewrite Exists_iff_exists_mem. eexists. splits.
         apply* mem_assoc_assoc.
         rewrite Exists_iff_exists_mem. eexists. splits.
          apply* mem_assoc_assoc.
          autos*.
       inverts I4 as I4.
        rewrite app_assoc_r.
         rewrite~ app_assoc_l. rewrite~ app_assoc_l.
         introv M. false I. rewrite Exists_iff_exists_mem. eexists. splits.
          apply* mem_assoc_assoc.
          rewrite Exists_iff_exists_mem. eexists. splits.
           apply* mem_assoc_assoc.
           autos*.
        asserts ND1: (~ mem_assoc f fvs2).
         introv M. false I0. rewrite Exists_iff_exists_mem. eexists. splits.
          apply* mem_assoc_assoc.
          rewrite Exists_iff_exists_mem. eexists. splits.
           apply* mem_assoc_assoc.
           autos*.
         asserts ND3: (~ mem_assoc f fvs3).
          introv M. false I2. rewrite Exists_iff_exists_mem. eexists. splits.
           apply* mem_assoc_assoc.
           rewrite Exists_iff_exists_mem. eexists. splits.
            rewrite mem_app. rew_refl. right. apply* mem_assoc_assoc.
            autos*.
         repeat rewrite~ app_assoc_r.
      fold_bool. rew_refl* in I5.
     repeat rewrite app_mem_assoc in *. rew_refl in I4. rew_logic in I4.
      repeat cases_if; fold_bool; rew_refl in *; tryfalse*; autos*.
   apply not_not_elim in I. rewrite Exists_iff_exists_mem with (CA := _) in I.
    lets (a&Ma&Ea): (rm I). rewrite Exists_iff_exists_mem with (CA := _) in Ea.
    lets (a'&Ma'&Ea'): (rm Ea). false I2. rewrite Exists_iff_exists_mem with (CA := _).
    eexists. splits*.
     rewrite* Exists_iff_exists_mem. eexists. splits.
      rewrite mem_app. rew_refl*.
      autos*.
Qed.

Lemma DObject_star_assoc : forall o1 o2 o3 o12 o123,
  DObject_star o1 o2 = Some o12 →
  DObject_star o12 o3 = Some o123 →
  exists o23 o123',
    DObject_star o2 o3 = Some o23
    ∧ DObject_star o1 o23 = Some o123'
    ∧ o123 =♯ o123'.
Proof.
  introv E1 E2. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3],
    o12 as [o12 OK12], o123 as [o123 OK123].
  forwards E1': DObject_star_AObject_star (rm E1). simpl in E1'.
  forwards E2': DObject_star_AObject_star (rm E2). simpl in E2'.
  forwards~ (o23'&o123'&Eo23'&Eo123'&Oo123): AObject_star_assoc (rm E1') (rm E2').
  forwards~ Co23': AObject_star_correct Eo23'. forwards~ Co123': AObject_star_correct Eo123'.
  exists (exist _ _ Co23') (exist _ _ Co123'). splits~; apply~ AObject_star_DObject_star.
Qed.

Lemma DObject_star_read_left : forall o1 o2 o3 f v,
  DObject_star o1 o2 = Some o3 →
  read_DObject o1 f = Some v →
  read_DObject o3 f = Some v.
Proof.
  introv E1 E2. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3].
  forwards E1': DObject_star_AObject_star (rm E1). unfolds in E2. unfolds. simpls.
  destruct o1 as [fvs1|fvs1 fs1 vr1]; simpl in E2; cases_if as I; inverts E2; fold_bool;
    destruct o2 as [fvs2|fvs2 fs2 vr2]; simpl in E1';
    repeat (let I := fresh "I" in cases_if as I); inverts E1'; simpl;
    let I := fresh "I" in cases_if as I; fold_bool; rew_refl in *.
   rewrite~ app_assoc_l.
   false I1. rewrite app_mem_assoc. rew_refl*.
   rewrite~ app_assoc_l.
   false I2. rewrite app_mem_assoc. rew_refl*.
   rewrite~ app_assoc_r. introv M. false I0. rewrite Exists_iff_exists_mem. eexists. splits.
    apply* mem_assoc_assoc.
    rewrite Exists_iff_exists_mem. eexists. splits.
     apply* mem_assoc_assoc.
     autos*.
   false I2. rewrite app_mem_assoc. rew_refl*.
   rewrite app_mem_assoc in I3. rew_refl in I3. inverts I3 as I3.
    false I0. applys incl_mem I2. rewrite* <- mem_assoc_map_fst.
    inverts OK1 as O1 O2 O3. false* I0.
   case_if~. false*.
Qed.

Lemma DObject_star_read_right : forall o1 o2 o3 f v,
  DObject_star o1 o2 = Some o3 →
  read_DObject o2 f = Some v →
  read_DObject o3 f = Some v.
Proof.
  introv E1 E2. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3].
  forwards E1': DObject_star_AObject_star (rm E1). unfolds in E2. unfolds. simpls.
  destruct o2 as [fvs2|fvs2 fs2 vr2]; simpl in E2; cases_if as I; inverts E2; fold_bool;
    destruct o1 as [fvs1|fvs1 fs1 vr1]; simpl in E1';
    repeat (let I := fresh "I" in cases_if as I); inverts E1'; simpl;
    let I := fresh "I" in cases_if as I; fold_bool; rew_refl in *.
   rewrite~ app_assoc_r. introv M. false I0. rewrite Exists_iff_exists_mem. eexists. splits.
    apply* mem_assoc_assoc.
    rewrite Exists_iff_exists_mem. eexists. splits.
     apply* mem_assoc_assoc.
     autos*.
   false I1. rewrite app_mem_assoc. rew_refl*.
   rewrite~ app_assoc_l.
   false I2. rewrite app_mem_assoc. rew_refl*.
   rewrite~ app_assoc_r. introv M. false I0. rewrite Exists_iff_exists_mem. eexists. splits.
    apply* mem_assoc_assoc.
    rewrite Exists_iff_exists_mem. eexists. splits.
     apply* mem_assoc_assoc.
     autos*.
   false I2. rewrite app_mem_assoc. rew_refl*.
   rewrite app_mem_assoc in I3. rew_refl in I3. inverts I3 as I3.
    false I0. applys incl_mem I2. rewrite* <- mem_assoc_map_fst.
    false*.
   case_if~. false*.
Qed.

Lemma DObject_star_undefined : forall o1 o2 f v1 v2,
  read_DObject o1 f = Some v1 →
  read_DObject o2 f = Some v2 →
  DObject_star o1 o2 = None.
Proof.
  introv E1 E2. destruct o1 as [o1 OK1], o2 as [o2 OK2].
  apply AObject_star_DObject_star_None. unfolds read_DObject.
  destruct o1 as [fvs1|fvs1 fs1 vr1]; destruct o2 as [fvs2|fvs2 fs2 vr2]; simpls~;
    repeat (let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I);
    inverts E1; inverts E2; simpls~;
    try solve [ false I;
      repeat (rewrite Exists_iff_exists_mem; eexists;
        splits; [ apply* mem_assoc_assoc | autos~ ]) ].
   false I2. applys incl_mem I0. rewrite* <- mem_assoc_map_fst.
   false I3. applys incl_mem I0. rewrite* <- mem_assoc_map_fst.
Qed.

Lemma DObject_star_defined : forall o1 o2,
  (forall f, read_DObject o1 f = None ∨ read_DObject o2 f = None) →
  DObject_star o1 o2 ≠ None.
Proof.
  introv F E. destruct o1 as [o1 OK1], o2 as [o2 OK2].
  apply DObject_star_AObject_star_None in E. unfolds read_DObject.
  destruct o1 as [fvs1|fvs1 fs1 vr1]; destruct o2 as [fvs2|fvs2 fs2 vr2]; simpls~;
    repeat (let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I);
   try solve [ apply not_not_elim in I;
     repeat (rewrite Exists_iff_exists_mem with (CA := _) in I; lets ((?&?)&?&I'): (rm I);
       rename I' into I);
     simpl in I; substs; forwards [?|?]: F; cases_if as I'; fold_bool; rew_refl in I';
     false I'; apply* mem_mem_assoc
   | rewrite incl_Forall in I0; rewrite Forall_iff_forall_mem in I0; rew_logic in I0;
     lets (f&Nf): (rm I0); rew_logic in Nf; lets (M&NM): (rm Nf);
     forwards [?|?]: F; cases_if as I'; fold_bool; rew_refl in I';
     try (cases_if; fold_bool; false*); false I'; rewrite* mem_assoc_map_fst ].
  lets (f&Nf): string_infinite (fs1 ++ fs2 ++ map fst fvs1 ++ map fst fvs2).
  repeat rewrite mem_app in Nf. rew_refl in Nf. forwards [E'|E']: F f;
    repeat (let I := fresh "I" in cases_if as I; fold_bool; rew_refl in I); false* Nf.
Qed.

Lemma AObject_star_domain : forall o1 o2 o3,
  AObject_star o1 o2 = Some o3 →
  AObject_domain_merge (AObject_dom o1) (AObject_dom o2) (AObject_dom o3).
Proof.
  introv E. destruct o1 as [fvs1|fvs1 fs1 vr1]; destruct o2 as [fvs2|fvs2 fs2 vr2];
    simpl in E; try cases_if as T1; try cases_if as T2; inverts E; simpl;
    fold_bool; rew_refl in *; constructors; simpl.
   introv. rewrite map_app. rewrite mem_app. rew_refl*.
   introv I1 I2. rewrite map_mem in *. lets ((f1&v1)&I1'&E1): (rm I1).
    lets ((f2&v2)&I2'&E2): (rm I2). simpls. substs. false T1.
    rewrite Exists_iff_exists_mem. eexists. splits*.
    rewrite Exists_iff_exists_mem. eexists. splits*.
   introv. rewrite map_app. repeat rewrite removes_mem.
    extens. rewrite mem_app. rew_refl. rew_logic*.
   introv I1 I2. rewrite map_mem with (CA := _) in I1. rewrite removes_mem in I2.
    rew_refl in I2. rew_logic in I2. lets ((f1&v1)&I1'&E1): (rm I1). substs. inverts I2 as I2.
    false I2. applys~ incl_mem T2. rewrite* map_mem.
    rewrite map_mem in I2. lets ((f2&v2)&I2'&E2): (rm I2). clear I2. simpls. substs.
     false (rm T1). rewrite Exists_iff_exists_mem. eexists. splits*.
     rewrite Exists_iff_exists_mem. eexists. simpl. splits*.
   introv. rewrite map_app. repeat rewrite removes_mem. rewrite mem_app.
    extens. rew_refl. rew_logic*.
   introv I1 I2. rewrite map_mem with (CA := _) in I2. rewrite removes_mem in I1.
    rew_refl in I1. rew_logic in I1. lets ((f2&v2)&I2'&E2): (rm I2). substs. inverts I1 as I1.
    false I1. applys~ incl_mem T2. rewrite* map_mem.
    rewrite map_mem in I1. lets ((f1&v1)&I1'&E1): (rm I1). clear I1. simpls. substs.
     false (rm T1). rewrite Exists_iff_exists_mem. eexists. splits*.
     rewrite Exists_iff_exists_mem. eexists. simpl. splits*.
Qed.

Lemma AObject_star_same_domain_defined : forall o1 o2,
  ~ AObject_domain_intersect (AObject_dom o1) (AObject_dom o2) →
  AObject_star o1 o2 ≠ None.
Proof.
  introv N D. applys (rm N). destruct o1 as [fvs1|fvs1 fs1 vr1]; destruct o2 as [fvs2|fvs2 fs2 vr2];
    simpls; try cases_if as T1; try cases_if as T2;
    fold_bool; rew_refl in *; try apply not_not_elim in T1;
    try (rewrite Exists_iff_exists_mem with (CA := _) in T1; lets ((f1&v1)&I1&E2): (rm T1));
    try (rewrite Exists_iff_exists_mem with (CA := _) in E2; lets ((f2&v2)&I2&E3): (rm E2); simpl in E3; substs);
    try (exists f2; splits; simpl; try rewrite* map_mem; try rewrite removes_mem; rew_refl; rew_logic; try rewrite* map_mem).
   lets (f&I1&I2): not_incl (rm T2). exists f. simpl. splits~. rewrite removes_mem.
    rew_refl. rew_logic*.
   lets (f&I1&I2): not_incl (rm T2). exists f. simpl. splits~. rewrite removes_mem.
    rew_refl. rew_logic*.
   forwards (f&N): string_infinite (fs1 ++ fs2). rewrite mem_app in N. rew_refl in N.
    rew_logic in N. exists f. simpl. splits~; rewrite removes_mem; rew_refl; rew_logic*.
Qed.

Lemma AObject_cofinite_star_decompose : forall fvs fs v,
  AObject_correct (AObject_cofinite fvs fs v) →
  AObject_star (AObject_finite fvs) (AObject_cofinite nil fs v)
  = Some (AObject_cofinite fvs fs v).
Proof.
  introv C. simpl. cases_if as I; fold_bool; rew_refl in I.
   cases_if as I'; fold_bool; rew_refl in I'.
    rewrite~ app_nil_r.
    false I'. apply mem_incl with (H := _). introv M. inverts C as O1 O2 O3.
     apply O3. rewrite~ mem_assoc_map_fst.
   apply not_not_elim in I. rewrite Exists_iff_exists_mem with (CA := _) in I.
    lets (?&?&E) : (rm I). inverts* E.
Qed.


(** ** Merging Objects **)

(** This operation takes two objects with the same domain and merge them. **)

Definition AObject_merge (o1 o2 : AObject) :=
  'let merge := fun fvs1 fvs2 =>
    LibList.map_partial (fun f =>
      LibOption.apply (fun v1' =>
        LibOption.map (fun v2' =>
          (f, v1' ⊔♯ v2')) (read_AObject o2 f)) (read_AObject o1 f))
      (Remove_duplicates (LibList.map fst (fvs1 ++ fvs2))) in
  match o1, o2 with
  | AObject_finite fvs1, AObject_finite fvs2 =>
    LibOption.map AObject_finite (merge fvs1 fvs2)
  | AObject_cofinite fvs1 fs1 vr1, AObject_cofinite fvs2 fs2 vr2 =>
    LibOption.map_on (merge fvs1 fvs2)
      (fun fvs =>
        let fs := Remove_duplicates (fs1 ++ fs2) in
        AObject_cofinite fvs fs (vr1 ⊔♯ vr2))
  | _, _ => None
  end.

Definition AObject_merge_check o1 o2 :=
  ifb AObject_domain_compare (AObject_dom o1) (AObject_dom o2) then
    AObject_merge o1 o2
  else None.

Lemma AObject_merge_check_correct : forall o1 o2 o3,
  AObject_correct o1 →
  AObject_correct o2 →
  AObject_merge_check o1 o2 = Some o3 →
  AObject_correct o3.
Proof.
  introv C1 C2 E. unfolds in E. cases_if as D. fold_bool. rew_refl in D.
  unfolds in E. let_name. asserts merge_correct: (forall fvs1 fvs2 fvs f v1 v2 n1 n2,
    merge fvs1 fvs2 = Some fvs →
    (forall v1 v2 n1 n2, Nth n1 fvs1 (f, v1) → Nth n2 fvs1 (f, v2) → n1 = n2) →
    (forall v1 v2 n1 n2, Nth n1 fvs2 (f, v1) → Nth n2 fvs2 (f, v2) → n1 = n2) →
    Nth n1 fvs (f, v1) → Nth n2 fvs (f, v2) → n1 = n2).
   clear E C1 C2. introv E C1 C2 N1 N2. substs.
    forwards F: map_partial_inv (rm E). apply Forall2_swap in F.
    forwards N1': Forall2_Nth_nth_def f F N1. forwards N2': Forall2_Nth_nth_def f F N2.
    forwards~ E1': Forall2_forall_Nth F N1 N1'. forwards~ E2': Forall2_forall_Nth F N2 N2'.
    forwards (v1'&E1&Ev1'): apply_on_inv (rm E1'). forwards (va&Eva&Eva'): map_inv (rm Ev1').
    forwards (v2'&E2&Ev2'): apply_on_inv (rm E2'). forwards (vb&Evb&Evb'): map_inv (rm Ev2').
    inverts Eva' as Ef Ev1. inverts Evb' as Ef' Ev2. rewrite <- Ef in *. rewrite <- Ef' in *.
    applys No_duplicates_Nth N1' N2'; try typeclass. apply* Remove_duplicates_spec.
  asserts merge_assoc: (forall fvs1 fvs2 fvs f,
    merge fvs1 fvs2 = Some fvs →
    mem_assoc f fvs →
    mem_assoc f fvs1 ∨ mem_assoc f fvs2).
   clear E. introv E I. substs. forwards F: map_partial_inv (rm E).
    rewrite mem_assoc_map_fst in I. rewrite map_mem in I. lets ((f'&v)&Mfv&Ef): (rm I).
    substs. lets (n&N): mem_Nth (rm Mfv). apply Forall2_swap in F.
    forwards N': Forall2_Nth_nth_def f' F N. forwards~ E': Forall2_forall_Nth F N N'.
    forwards (v'&E&Ev'): apply_on_inv (rm E'). forwards (v''&E'&Ev''): apply_on_inv (rm Ev').
    inverts Ev'' as Ef. rewrite Ef in *. apply Nth_mem with (CA := _) in N'.
    rewrite Remove_duplicates_mem in N'. rewrite map_mem in N'. lets ((f&v)&Mv&Ev): (rm N').
    substs. rewrite mem_app in Mv. rew_refl in Mv. do 2 rewrite mem_assoc_map_fst.
    do 2 rewrite map_mem. inverts* Mv.
  clear EQmerge. destruct o1 as [fvs1|fvs1 fs1 vr1], o2 as [fvs2|fvs2 fs2 vr2];
    simpl in E; tryfalse~.
   forwards (fvs&Efvs&Eo3): map_inv (rm E). substs. constructors. introv N1 N2.
    applys~ merge_correct Efvs N1 N2.
     inverts* C1.
     inverts* C2.
   forwards (fvs&Efvs&Eo3): map_inv (rm E). substs.
    constructors.
     introv N1 N2. applys~ merge_correct Efvs N1 N2.
      inverts* C1.
      inverts* C2.
     introv N1 N2. applys No_duplicates_Nth N1 N2; try typeclass.
      apply~ Remove_duplicates_spec.
     introv M. rewrite~ Remove_duplicates_mem. rewrite mem_app. rew_refl*.
      forwards [M'|M']: merge_assoc Efvs M.
       inverts* C1.
       inverts* C2.
Qed.

Definition DObject_merge (o1 o2 : DObject) : option DObject :=
  let r := AObject_merge_check (proj1_sig o1) (proj1_sig o2) in
  match r as r0 return r = r0 → _ with
  | Some r => fun E =>
    Some (exist _ r (AObject_merge_check_correct (proj2_sig o1) (proj2_sig o2) E))
  | None => fun _ => None
  end eq_refl.

Lemma DObject_merge_AObject : forall o1 o2 o3,
  DObject_merge o1 o2 = Some o3 →
  AObject_merge_check (proj1_sig o1) (proj1_sig o2) = Some (proj1_sig o3).
Proof.
  introv E. unfolds in E. asserts A: (forall A B (o0 : option A) o1 f,
    match o0 as o return o0 = o → option B with
    | Some o' => fun E => Some (f o' E)
    | None => fun _ => None
    end eq_refl = Some o1 →
    exists o' E, o0 = Some o' ∧ o1 = f o' E).
   clear. introv E. destruct o0; inverts* E.
  lets (o''&E'&E1&E2): (rm A) (rm E). substs*.
Qed.

Lemma AObject_merge_DObject : forall o1 o2 o3,
  AObject_merge_check (proj1_sig o1) (proj1_sig o2) = Some (proj1_sig o3) →
  DObject_merge o1 o2 = Some o3.
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3]. unfolds.
  asserts A: (forall A B (o0 : option A) o' f,
    o0 = Some o' →
    exists E', match o0 as o return o0 = o → option B with
    | Some o' => fun E => Some (f o' E)
    | None => fun _ => None
    end eq_refl = Some (f o' E')).
   clear. introv Eq. substs. eexists. fequals.
  forwards (E'&Eq): (rm A) E. rewrite Eq. repeat fequals.
Qed.

Lemma AObject_merge_check_read : forall o1 o2 o3 f v1 v2,
  AObject_correct o1 →
  AObject_correct o2 →
  AObject_merge_check o1 o2 = Some o3 →
  read_AObject o1 f = Some v1 →
  read_AObject o2 f = Some v2 →
  read_AObject o3 f = Some (v1 ⊔♯ v2).
Proof.
  introv C1 C2 E R1 R2. forwards~ C3: AObject_merge_check_correct E.
  unfolds in E. cases_if as I0. unfolds in E. let_name.
  asserts merge_read: (forall fvs1 fvs2 fvs,
    merge fvs1 fvs2 = Some fvs →
    mem_assoc f fvs →
    mem (f, v1 ⊔♯ v2) fvs).
   clear E. substs. introv E M. forwards F: map_partial_inv (rm E).
   rewrite mem_assoc_map_fst in M. rewrite map_mem in M.
   lets ((f'&v)&Mfv&E): (rm M). substs. lets (n&N): mem_Nth (rm Mfv).
   apply Forall2_swap in F. forwards N': Forall2_Nth_nth_def f' F N.
   forwards~ E': Forall2_forall_Nth F N N'. forwards (v'&E&Ev'): apply_on_inv (rm E').
   forwards (v''&E'&Ev''): apply_on_inv (rm Ev'). inverts Ev'' as Ef. rewrite Ef in *.
   simpls. rewrite E in R1. rewrite E' in R2. inverts R1. inverts R2. apply* Nth_mem.
  destruct o1 as [fvs1|fvs1 fs1 vr1], o2 as [fvs2|fvs2 fs2 vr2]; tryfalse.
   forwards (fvs&Efvs&Eo3): map_inv (rm E). substs. unfolds. cases_if as I; fold_bool.
    forwards~ M: merge_read Efvs. erewrite unique_assoc_assoc_mem; autos*.
     inverts* C3.
    rew_refl in I. false I. lets R0: R1. unfolds in R1. cases_if as I1; inverts R1.
     forwards F: map_partial_inv (rm Efvs). fold_bool. rewrite mem_assoc_map_fst in I1.
     rewrite map_mem in I1. lets ((f'&v1)&Mfv1&E1): (rm I1). substs.
     asserts Mf: (mem f' (Remove_duplicates (map fst (fvs1 ++ fvs2)))).
      rewrite Remove_duplicates_mem. rewrite map_app. rewrite mem_app. rew_refl.
      left. rewrite* map_mem.
     lets (n&N): mem_Nth (rm Mf). forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
     forwards~ E': Forall2_forall_Nth F N N'. forwards (v'&E&Ev'): apply_on_inv (rm E').
     forwards (v''&E'&Ev''): apply_on_inv (rm Ev'). inverts Ev'' as Ef. rewrite <- Ef in *.
     rewrite mem_assoc_map_fst. rewrite map_mem with (CA := _). simpl. eexists. splits.
      apply* Nth_mem.
      autos*.
   forwards (fvs&Efvs&Eo3): map_inv (rm E). substs. unfolds. cases_if as I; fold_bool.
    forwards~ M: merge_read Efvs. erewrite unique_assoc_assoc_mem; autos*.
     inverts* C3.
    rew_refl in I. cases_if as I'; fold_bool.
     false I. lets R0: R1. unfolds in R1. cases_if as I1.
      inverts R1. forwards F: map_partial_inv (rm Efvs). fold_bool.
       rewrite mem_assoc_map_fst in I1.
       rewrite map_mem in I1. lets ((f'&v1)&Mfv1&E1): (rm I1). substs.
       asserts Mf: (mem f' (Remove_duplicates (map fst (fvs1 ++ fvs2)))).
        rewrite Remove_duplicates_mem. rewrite map_app. rewrite mem_app. rew_refl.
        left. rewrite* map_mem.
       lets (n&N): mem_Nth (rm Mf). forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
       forwards~ E': Forall2_forall_Nth F N N'. forwards (v'&E&Ev'): apply_on_inv (rm E').
       forwards (v''&E'&Ev''): apply_on_inv (rm Ev'). inverts Ev'' as Ef. rewrite <- Ef in *.
       rewrite mem_assoc_map_fst. rewrite map_mem with (CA := _). simpl. eexists. splits.
        apply* Nth_mem.
        autos*.
      cases_if as I2; fold_bool. inverts R1. rew_refl in I1. rew_refl in I2.
       clear R0. lets R0: R2. unfolds in R2. cases_if as I3.
        inverts R2. forwards F: map_partial_inv (rm Efvs). fold_bool.
         rewrite mem_assoc_map_fst in I3.
         rewrite map_mem in I3. lets ((f'&v2)&Mfv2&E2): (rm I3). substs.
         asserts Mf: (mem f' (Remove_duplicates (map fst (fvs1 ++ fvs2)))).
          rewrite Remove_duplicates_mem. rewrite map_app. rewrite mem_app. rew_refl.
          right. rewrite* map_mem.
         lets (n&N): mem_Nth (rm Mf). forwards N': Forall2_Nth_nth_def (f', ⊥♯ : ASValue) F N.
         forwards~ E': Forall2_forall_Nth F N N'. forwards (v'&E&Ev'): apply_on_inv (rm E').
         forwards (v''&E'&Ev''): apply_on_inv (rm Ev'). inverts Ev'' as Ef.
         rewrite <- Ef in *. rewrite mem_assoc_map_fst. rewrite map_mem with (CA := _).
         simpl. eexists. splits.
          apply* Nth_mem.
          autos*.
       cases_if as I4. fold_bool. rew_refl in I4. rewrite Remove_duplicates_mem in I'.
        rewrite mem_app in I'. rew_refl in I'. inverts* I'.
     unfolds in R1. rewrite Remove_duplicates_mem in I'. rew_refl in I'.
      cases_if as I1; fold_bool.
       false I'. rewrite mem_app. rew_refl. left. inverts* C1.
       cases_if as I2; fold_bool. inverts R1. unfolds in R2. cases_if as I3; fold_bool.
        false I'. rewrite mem_app. rew_refl. right. inverts* C2.
        cases_if as I4; fold_bool. inverts~ R2.
Qed.

Lemma DObject_merge_read : forall o1 o2 o3 f v1 v2,
  DObject_merge o1 o2 = Some o3 →
  read_DObject o1 f = Some v1 →
  read_DObject o2 f = Some v2 →
  read_DObject o3 f = Some (v1 ⊔♯ v2).
Proof.
  introv E1 E2 E3. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3]. simpls.
  applys~ AObject_merge_check_read E2 E3. apply* DObject_merge_AObject.
Qed.

Lemma AObject_merge_check_read_None : forall o1 o2 o3 f,
  AObject_correct o1 →
  AObject_correct o2 →
  AObject_merge_check o1 o2 = Some o3 →
  read_AObject o1 f = None →
  read_AObject o2 f = None →
  read_AObject o3 f = None.
Proof.
  introv C1 C2 E R1 R2. forwards~ C3: AObject_merge_check_correct E.
  unfolds in E. cases_if as I0. unfolds in E. let_name.
  asserts merge_read: (forall fvs1 fvs2 fvs,
    merge fvs1 fvs2 = Some fvs →
    ~ mem_assoc f fvs).
   clear E. substs. introv E M. forwards F: map_partial_inv (rm E).
   rewrite mem_assoc_map_fst in M. rewrite map_mem in M. lets ((f'&v)&Mfv&Efv): (rm M).
   substs. lets (n&N): mem_Nth (rm Mfv). apply Forall2_swap in F.
   forwards N': Forall2_Nth_nth_def f' F N. forwards~ E': Forall2_forall_Nth F N N'.
   forwards (v'&E&Ev'): apply_on_inv (rm E'). forwards (v''&E'&Ev''): apply_on_inv (rm Ev').
   inverts Ev'' as Ef. rewrite Ef in *. simpl in R1. rewrite E in R1. inverts R1.
  destruct o1 as [fvs1|fvs1 fs1 vr1], o2 as [fvs2|fvs2 fs2 vr2]; tryfalse.
   forwards (fvs&Efvs&Eo3): map_inv (rm E). substs. unfolds. cases_if as I; fold_bool.
    forwards*: merge_read Efvs I.
    autos*.
   forwards (fvs&Efvs&Eo3): map_inv (rm E). substs. unfolds. cases_if as I; fold_bool.
    forwards*: merge_read Efvs I.
    cases_if as I'; fold_bool.
     autos*.
     unfolds in R1. cases_if as I1. cases_if as I2.
      fold_bool. rew_refl in I'. false I'. rewrite Remove_duplicates_mem.
      rewrite mem_app. rew_refl*.
Qed.

Lemma DObject_merge_read_None : forall o1 o2 o3 f,
  DObject_merge o1 o2 = Some o3 →
  read_DObject o1 f = None →
  read_DObject o2 f = None →
  read_DObject o3 f = None.
Proof.
  introv E1 E2 E3. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3]. simpls.
  applys~ AObject_merge_check_read_None E2 E3. apply* DObject_merge_AObject.
Qed.

Lemma AObject_merge_check_read_None_inv : forall o1 o2 o3 f,
  AObject_correct o1 →
  AObject_correct o2 →
  AObject_merge_check o1 o2 = Some o3 →
  read_AObject o3 f = None →
  read_AObject o1 f = None ∧ read_AObject o2 f = None.
Proof.
  introv C1 C2 E R. lets E0: E. unfolds in E. cases_if as I. fold_bool. rew_refl in I.
  rewrite AObject_domain_compare_correct in I.
  destruct (read_AObject o1 f) eqn: E1; destruct (read_AObject o2 f) eqn: E2.
   erewrite AObject_merge_check_read in R; try apply E0; autos*. false*.
    apply I in E2; tryfalse*. rewrite E1. discriminate.
    apply I in E1; tryfalse*. rewrite E2. discriminate.
    autos*.
Qed.

Lemma DObject_merge_read_None_inv : forall o1 o2 o3 f,
  DObject_merge o1 o2 = Some o3 →
  read_DObject o3 f = None →
  read_DObject o1 f = None ∧ read_DObject o2 f = None.
Proof.
  introv E1 E. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3]. simpls.
  applys~ AObject_merge_check_read_None_inv E. apply* DObject_merge_AObject.
Qed.

Lemma DObject_merge_same_domains : forall o1 o2 o3,
  DObject_merge o1 o2 = Some o3 →
  AObject_domain_compare (DObject_dom o1) (DObject_dom o2).
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3].
  forwards E': DObject_merge_AObject (rm E). unfolds in E'. cases_if as I.
  fold_bool. rew_refl in I. autos*.
Qed.

Lemma DObject_merge_same_domains_left : forall o1 o2 o3,
  DObject_merge o1 o2 = Some o3 →
  AObject_domain_compare (DObject_dom o1) (DObject_dom o3).
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3].
  forwards S: DObject_merge_same_domains E. forwards E': DObject_merge_AObject (rm E).
  rewrite DObject_domain_compare_correct in S. introv. lets RE: (rm S) f.
  extens. do 2 rewrite DObject_dom_correct. rewrite iff_not_cancel in *. iff R.
   applys~ AObject_merge_check_read_None E'. rewrite~ <- RE.
   applys~ AObject_merge_check_read_None_inv E'.
Qed.

Lemma DObject_merge_same_domains_right : forall o1 o2 o3,
  DObject_merge o1 o2 = Some o3 →
  AObject_domain_compare (DObject_dom o2) (DObject_dom o3).
Proof.
  introv E. destruct o1 as [o1 OK1], o2 as [o2 OK2], o3 as [o3 OK3].
  forwards S: DObject_merge_same_domains E. forwards E': DObject_merge_AObject (rm E).
  rewrite DObject_domain_compare_correct in S. introv. lets RE: (rm S) f.
  extens. do 2 rewrite DObject_dom_correct. rewrite iff_not_cancel in *. iff R.
   applys~ AObject_merge_check_read_None E'. rewrite~ RE.
   applys~ AObject_merge_check_read_None_inv E'.
Qed.


(** * Environments **)

(** ** Definition **)

Definition AEnv : Type :=
  list (var * AValue).

Definition AEnv_empty : AEnv := nil.


Global Instance AEnv_comparable : Comparable AEnv.
  typeclass.
Defined.


(** ** Reading an Writing **)

(** Reading a field of an environment. **)
Definition read_AEnv (E : AEnv) (x : var) :=
  if mem_assoc x E then Some (assoc x E)
  else None.

(** Writing a field of an environment. **)
Definition write_AEnv (E : AEnv) (x : var) v : AEnv :=
  (x, v) :: remove_assoc x E.

(** Names present in an environment. **)
Definition ENames (E : AEnv) :=
  concat (map (fun xv => AVNames (snd xv)) E).

Definition EENames (E : AEnv) :=
  concat (map (fun xv => AVENames (snd xv)) E).

(** Renaming the identifiers of an environment. **)
Definition AEnv_substs (E : AEnv) hs h : AEnv :=
  map (fun xv =>
    let '(x, v) := xv in
    (x, v [[ hs // h]])) E.

Definition AEnv_Esubsts (E : AEnv) kes ke : AEnv :=
  map (fun xv =>
    let '(x, v) := xv in
    (x, v |[[ kes // ke]])) E.


(** ** Properties **)

Lemma read_write_AEnv : forall E x v,
  read_AEnv (write_AEnv E x v) x = Some v.
Proof.
  introv. unfold read_AEnv, write_AEnv. rewrite mem_mem_assoc with (H0 := _).
   fequals. rewrite~ assoc_here.
   simpl. rew_refl*.
Qed.

Lemma read_write_AEnv_propagates : forall E x y v,
  x ≠ y →
  read_AEnv (write_AEnv E x v) y = read_AEnv E y.
Proof.
  introv D. unfolds read_AEnv, write_AEnv. cases_if as I; cases_if as I'; fold_bool.
   fequals. rewrite assoc_cons. case_if; tryfalse~. rewrite~ assoc_remove_assoc.
   rewrite mem_assoc_cons in I. rew_refl in I. inverts I as I; tryfalse~.
    rewrite~ mem_assoc_remove_assoc_neq in I. rewrite I in I'. rew_refl* in I'.
   rewrite mem_assoc_cons in I. rew_refl in I. rew_logic in I.
    lets (N1&N2): (rm I). rewrite~ mem_assoc_remove_assoc_neq in N2. false* N2.
   autos~.
Qed.

Lemma read_AEnv_empty : forall x,
  read_AEnv AEnv_empty x = None.
Proof. introv. unfolds. cases_if~. Qed.


Lemma ENames_substs : forall E h hs h0,
  mem h0 (ENames (AEnv_substs E hs h))
  = mem h (ENames E) &&& mem h0 hs
  ||| mem h0 (ENames E) &&& decide (h ≠ h0).
Proof.
  introv. unfolds AEnv_substs, ENames. extens. rew_refl. repeat rewrite concat_mem.
  iff I; [lets (L1&ML1&Mh0): (rm I)| inverts I as (I&M); lets(L1&ML1&Mh0): (rm I)];
    rewrite map_mem in ML1; lets (xv&Mxv&EL1): (rm ML1).
   rewrite map_mem in Mxv. lets ([f v]&ME&Exv): (rm Mxv). substs.
    simpl in Mh0. rewrite AVNames_substs in Mh0. rew_refl in Mh0. inverts Mh0 as (M1&M2).
     left. splits~. eexists. splits*. rewrite* map_mem.
     right. splits~. eexists. splits*. rewrite* map_mem.
   destruct xv as (f&v). substs. eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
    simpl. rewrite AVNames_substs. rew_refl*.
   destruct xv as (f&v). substs. eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
    simpl. rewrite AVNames_substs. rew_refl*.
Qed.

Lemma EENames_Esubsts : forall E ke kes ke0,
  mem ke0 (EENames (AEnv_Esubsts E kes ke))
  = mem ke (EENames E) &&& mem ke0 kes
  ||| mem ke0 (EENames E) &&& decide (ke ≠ ke0).
Proof.
  introv. unfolds AEnv_Esubsts, EENames. extens. rew_refl. repeat rewrite concat_mem.
  iff I; [lets (L1&ML1&Mh0): (rm I)| inverts I as (I&M); lets(L1&ML1&Mh0): (rm I)];
    rewrite map_mem in ML1; lets (xv&Mxv&EL1): (rm ML1).
   rewrite map_mem in Mxv. lets ([f v]&ME&Exv): (rm Mxv). substs.
    simpl in Mh0. rewrite AVENames_Esubsts in Mh0. rew_refl in Mh0. inverts Mh0 as (M1&M2).
     left. splits~. eexists. splits*. rewrite* map_mem.
     right. splits~. eexists. splits*. rewrite* map_mem.
   destruct xv as (f&v). substs. eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
    simpl. rewrite AVENames_Esubsts. rew_refl*.
   destruct xv as (f&v). substs. eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
    simpl. rewrite AVENames_Esubsts. rew_refl*.
Qed.

Lemma ENames_Esubsts : forall E ke kes h,
  mem h (ENames (AEnv_Esubsts E kes ke))
  = mem h (ENames E).
Proof.
  introv. unfolds AEnv_Esubsts, ENames. extens. rew_refl. repeat rewrite concat_mem.
  iff I; [lets (L1&ML1&Mh0): (rm I)| lets (L1&ML1&Mh0): (rm I)];
    rewrite map_mem in ML1; lets (xv&Mxv&EL1): (rm ML1).
   rewrite map_mem in Mxv. lets ([f v]&ME&Exv): (rm Mxv). substs.
    simpl in Mh0. rewrite AVNames_Esubsts in Mh0. rew_refl in Mh0.
     eexists. splits*. rewrite* map_mem.
   destruct xv as (f&v). substs. eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
    simpl. rewrite AVNames_Esubsts. rew_refl*.
Qed.

Lemma EENames_substs : forall E h hs ke,
  mem ke (EENames (AEnv_substs E hs h))
  = mem ke (EENames E).
Proof.
  introv. unfolds AEnv_substs, EENames. extens. rew_refl. repeat rewrite concat_mem.
  iff I; [lets (L1&ML1&Mh0): (rm I)| lets (L1&ML1&Mh0): (rm I)];
    rewrite map_mem in ML1; lets (xv&Mxv&EL1): (rm ML1).
   rewrite map_mem in Mxv. lets ([f v]&ME&Exv): (rm Mxv). substs.
    simpl in Mh0. rewrite AVENames_substs in Mh0. rew_refl in Mh0.
     eexists. splits*. rewrite* map_mem.
   destruct xv as (f&v). substs. eexists. splits.
    rewrite map_mem with (CA := _). eexists. splits*. rewrite map_mem. eexists. splits*.
    simpl. rewrite AVENames_substs. rew_refl*.
Qed.


(** Falsehood of an environment. **)
Definition AEnv_incoherent E :=
  exists x av, read_AEnv E x = Some av ∧ av =♯ ⊥♯.

Global Instance AEnv_incoherent_decidable : forall E,
    Decidable (AEnv_incoherent E).
  introv. applys Decidable_equiv (Exists (fun xv =>
      match read_AEnv E (fst xv) with
      | None => False
      | Some av => av =♯ ⊥♯
      end) E).
   iff I.
    rewrite Exists_iff_exists_mem with (CA := _) in I. lets ((x&xv)&Mx&Rx): (rm I).
     destruct read_AEnv eqn: Eq; tryfalse. exists* x.
    lets (x&av&B&Eq): (rm I). unfolds read_AEnv. cases_if as I; tryfalse.
     rewrite Exists_iff_exists_mem. exists (x, assoc x E). splits~.
      apply~ assoc_mem_assoc.
      simpl. rewrite I. inverts~ B.
   apply Exists_decidable. introv. destruct read_AEnv; typeclass.
Defined.


(** ** Lattice Structure **)

Definition AEnv_order E1 E2 :=
  forall x,
    match read_AEnv E1 x, read_AEnv E2 x with
    | Some av1, Some av2 => av1 ⊑♯ av2
    | None, None => True
    | _, _ => False
    end.

Local Instance AEnv_order_decidable : forall E1 E2,
    Decidable (AEnv_order E1 E2).
  introv. applys Decidable_equiv (Forall (fun x =>
      match read_AEnv E1 x, read_AEnv E2 x with
      | Some av1, Some av2 => av1 ⊑♯ av2
      | None, None => True
      | _, _ => False
      end) (map fst (E1 ++ E2))).
   rew_list. iff I.
    intro x. rewrite Forall_iff_forall_mem in I.
     tests M: (mem x (map fst (E1 ++ E2))); rew_list in M.
      applys~ I M.
      unfolds read_AEnv. cases_if as I1; fold_bool.
       false M. rewrite mem_app. rew_refl. left. rewrite~ <- mem_assoc_map_fst.
       cases_if~. cases_if as I2; fold_bool.
        false M. rewrite mem_app. rew_refl. right. rewrite~ <- mem_assoc_map_fst.
    rewrite Forall_iff_forall_mem with (CA := _). introv M. apply I.
   apply Forall_decidable. introv. do 2 destruct read_AEnv; typeclass.
Defined.

Lemma AEnv_order_refl : forall E, AEnv_order E E.
Proof. intros E x. destruct~ read_AEnv. Qed.

Lemma AEnv_order_trans : forall E1 E2 E3,
  AEnv_order E1 E2 →
  AEnv_order E2 E3 →
  AEnv_order E1 E3.
Proof.
  introv O1 O2. intro x. forwards Oa: O1 x. forwards Ob: O2 x.
  do 3 destruct~ read_AEnv.
   applys @PosetDec.trans Oa Ob.
   false*.
Qed.

Global Instance AEnv_Equiv : EquivDec.t AEnv.
  apply EquivDec.Make with (fun E1 E2 => forall x,
      match read_AEnv E2 x, read_AEnv E1 x with
      | Some av1, Some av2 => av1 =♯ av2
      | None, None => True
      | _, _ => False
      end).
  introv. destruct~ read_AEnv.
  introv Eq. intro a. forwards: Eq a. repeat destruct~ read_AEnv.
  introv E1 E2. intro a. forwards Ea: E1 a. forwards Eb: E2 a.
   do 3 destruct~ read_AEnv; tryfalse; try applys~ @EquivDec.trans Eb Ea.
  introv. apply decidable_sumbool. applys Decidable_equiv (AEnv_order x y ∧ AEnv_order y x).
   iff O.
    intro a. lets (O1&O2): (rm O). forwards Oa: O1 a. forwards Ob: O2 a.
     do 2 destruct~ read_AEnv; tryfalse.
    splits; intro a; forwards: O a; do 2 destruct~ read_AEnv.
   typeclass.
Defined.

Global Instance AEnv_Poset : PosetDec.t AEnv.
  apply PosetDec.Make with AEnv_Equiv AEnv_order.
   introv E. intro a. forwards: E a. do 2 destruct~ read_AEnv.
   introv O1 O2. intro a. forwards: O1 a. forwards: O2 a. do 2 destruct~ read_AEnv.
   apply* AEnv_order_trans.
   introv. apply decidable_sumbool. typeclass.
Defined.

Definition AEnv_join (E1 E2 : AEnv) : option AEnv :=
  let xs1 := LibList.keys E1 in
  let xs2 := LibList.keys E2 in
  ifb double_incl xs1 xs2 then
    Some (fold_left (fun x E =>
      write_AEnv E x (unsome (read_AEnv E1 x) ⊔♯ unsome (read_AEnv E2 x))) nil xs1)
  else None.

Lemma AEnv_join_read : forall E1 E2 E x v1 v2 v,
  AEnv_join E1 E2 = Some E →
  read_AEnv E1 x = Some v1 →
  read_AEnv E2 x = Some v2 →
  read_AEnv E x = Some v →
  v =♯ (v1 ⊔♯ v2).
Proof.
  introv Eq1 Eq2 Eq3 Eq4. unfolds in Eq1. cases_if as DI; inverts Eq1 as Eq1.
  rewrite fold_left_eq_fold_right in Eq4. induction (rev (keys E1)).
   simpl in Eq4. unfolds in Eq4. cases_if*.
   rew_list in Eq4. tests: (a = x).
    rewrite~ read_write_AEnv in Eq4. inverts Eq4. rewrite Eq2. rewrite~ Eq3.
    rewrite~ read_write_AEnv_propagates in Eq4.
Qed.

Lemma AEnv_join_defined_arguments : forall E1 E2 E x,
  AEnv_join E1 E2 = Some E →
  read_AEnv E1 x = None ↔ read_AEnv E2 x = None.
Proof.
  introv Eq. unfolds in Eq. cases_if as I.
  unfold read_AEnv. iff I'; repeat cases_if~; fold_bool; rew_refl in *;
    rewrite <- keys_mem_assoc in *.
   false H0. erewrite* double_incl_mem.
   false H0. erewrite* double_incl_mem. apply~ double_incl_sym.
Qed.

Lemma AEnv_join_defined : forall E1 E2 E x v1 v2,
  AEnv_join E1 E2 = Some E →
  read_AEnv E1 x = Some v1 →
  read_AEnv E2 x = Some v2 →
  read_AEnv E x ≠ None.
Proof.
  introv Eq1 Eq2 Eq3. unfolds read_AEnv, AEnv_join. cases_if as ME; try discriminate.
  cases_if as ME2. inverts Eq3. cases_if as ME1. inverts Eq2. cases_if as DI. inverts Eq1.
  fold_bool. rew_refl in *. false (rm ME).
  applys~ fold_left_ex_ind (fun x' => x' = x) (fun E : AEnv =>
    mem_assoc x E).
   introv I. unfolds write_AEnv. rewrite mem_assoc_cons. rew_refl. tests: (a = x).
    left*.
    inverts I as I; tryfalse~. right. rewrite~ mem_assoc_remove_assoc_neq.
   eexists. splits*. rewrite~ keys_mem_assoc.
Qed.

Lemma AEnv_join_undefined : forall E1 E2 E x,
  AEnv_join E1 E2 = Some E →
  read_AEnv E1 x = None →
  read_AEnv E x = None.
Proof.
  introv Eq1 Eq2. unfolds read_AEnv, AEnv_join. cases_if as ME; autos~.
  cases_if as ME1. cases_if as DI. inverts Eq1. fold_bool. rew_refl in *.
  false (rm ME1). rewrite <- keys_mem_assoc in *. lets (I1&I2): (rm DI).
  sets_eq E0: (nil : AEnv). asserts Inv: (~ mem_assoc x E0).
   substs. rewrite~ mem_assoc_nil.
  rewrite rev_mem. clear EQE0 I2. gen E0 E2. induction~ (keys E1); introv Inv ME DI.
   simpl in ME. false Inv. rewrite~ <- keys_mem_assoc.
   rewrite <- rev_mem. simpl. rew_refl. tests D: (x = a).
    left*.
    right. rew_list in ME. rewrite rev_mem. applys IHl ME.
     introv M. unfolds write_AEnv. rewrite mem_assoc_cons in M. rew_refl in M.
      inverts M as M; tryfalse~. rewrite~ mem_assoc_remove_assoc_neq in M.
     applys incl_tran DI. apply~ incl_tl. apply incl_refl.
Qed.

Lemma AEnv_join_sym : forall E1 E2 E,
  AEnv_join E1 E2 = Some E →
  exists E', AEnv_join E2 E1 = Some E' ∧ E =♯ E'.
Proof.
  introv Eq. asserts (E'&Eq'): (exists E', AEnv_join E2 E1 = Some E').
   eexists. unfolds. cases_if as DI.
    reflexivity.
    unfolds in Eq. cases_if as DI'. inverts Eq. fold_bool. rew_refl in *.
     false DI. applys double_incl_sym DI'.
   exists E'. splits~. introv.
    destruct (read_AEnv E1 x) eqn: Eq1; destruct (read_AEnv E2 x) eqn: Eq2.
     destruct (read_AEnv E x) eqn: Eq3; destruct (read_AEnv E' x) eqn: Eq4.
      forwards~ Eqa: AEnv_join_read Eq1 Eq2 Eq3.
       forwards~ Eqa': AEnv_join_read Eq2 Eq1 Eq4.
       applys @EquivDec.trans Eqa'. apply EquivDec.sym in Eqa.
        applys* @EquivDec.trans Eqa.
      forwards*: AEnv_join_defined Eq4.
      forwards*: AEnv_join_defined Eq3.
      forwards*: AEnv_join_defined Eq3.
     rewrite* AEnv_join_defined_arguments in Eq2. rewrite Eq1 in Eq2. inverts Eq2.
     rewrite* AEnv_join_defined_arguments in Eq1. rewrite Eq1 in Eq2. inverts Eq2.
     forwards* Eq3: AEnv_join_undefined Eq. forwards* Eq4: AEnv_join_undefined Eq'.
      rewrite Eq3. rewrite~ Eq4.
Qed.

Lemma AEnv_join_order_left : forall E1 E2 E,
  AEnv_join E1 E2 = Some E →
  E1 ⊑♯ E.
Proof.
  introv Eq. lets Eq0: Eq. unfolds in Eq. cases_if as DI. inverts Eq.
  introv. destruct read_AEnv eqn: Eq1.
   applys~ fold_left_ex_ind (fun x' : var => x = x') (fun E =>
     match read_AEnv E x with
     | Some av => a ⊑♯ av
     | None => False
     end).
    introv I. tests: (x = a0).
     rewrite read_write_AEnv. rewrite Eq1. apply JoinDec.bound1.
     inverts I as I; tryfalse~. rewrite~ read_write_AEnv_propagates.
    fold_bool. rew_refl in DI. eexists. splits*. rewrite keys_mem_assoc.
     unfolds in Eq1. cases_if~.
   sets_eq E: (nil : AEnv). asserts I1: (read_AEnv E x = None).
     substs. unfolds. rewrite~ mem_assoc_nil.
    asserts I2: (forall x', mem x' (keys E1) → read_AEnv E1 x' ≠ None).
      introv M Eq. unfolds in Eq. cases_if as I. rewrite keys_mem_assoc in M.
      rewrite M in I. false*.
    clear - Eq1 I1 I2. gen E. induction (keys E1); introv I1.
     simpl. rewrite~ I1.
     rewrite fold_left_cons. apply IHl.
      introv M Eq. false* I2. simpl. rew_refl*.
      rewrite~ read_write_AEnv_propagates. intro_subst. false* I2. simpl. rew_refl*.
Qed.

Lemma AEnv_join_order_right : forall E1 E2 E,
  AEnv_join E1 E2 = Some E →
  E2 ⊑♯ E.
Proof.
  introv Eq. forwards (E'&Eq'&EE'): AEnv_join_sym Eq. eapply PosetDec.trans.
   applys~ AEnv_join_order_left Eq'.
   apply~ PosetDec.refl.
Qed.

End ValueScope.

End LocationScope.

End General.

