
(** This file contains the definitions of the lattices used in the
  abstract semantics of the example language. **)

Require Export Ascii.
Require Export String.

Set Implicit Arguments.

Require Export ConcreteOWhile ConcretisationsSep.

Notation " [ ] " := nil : list_scope.
Notation " [ x ] " := (cons x nil) : list_scope.
Notation " [ x ; .. ; y ] " := (cons x .. (cons y nil) ..) : list_scope.


Require Export SignProps.

Definition L := nat.
Definition K := nat.
Definition KE := nat.
Definition ABasic := Sign.t.

Definition AValueSign : Type :=
  AValue L K KE ABasic.

Definition AValueSign_gamma : valuation _ _ _ → AValueSign → ℘ val :=
  AValue_gamma _ _ _ Sign.gamma.

Definition ASValueSign : Type :=
  ASValue L K KE ABasic.

Definition ASValueSign_gamma : valuation _ _ _ → ASValueSign → ℘ (option val) :=
  ASValue_gamma _ _ _ Sign.gamma.

Global Instance AValueSign_Inhab : Inhab AValueSign.
  apply~ prove_Inhab. repeat constructors.
Defined.

Global Instance ASValueSign_Inhab : Inhab ASValueSign.
  applys prove_Inhab (AValue_ASValue (arbitrary : AValueSign)).
Defined.

Global Instance AValueSign_Lat : LatticeDec.t AValueSign.
  apply AValue_Lat; typeclass.
Defined.

Global Instance ASValueSign_Lat : LatticeDec.t ASValueSign.
  apply ASValue_Lat; typeclass.
Defined.

Global Instance AValueSign_Top : TopDec.t AValueSign.
  typeclass.
Defined.

Global Instance ASValueSign_Top : TopDec.t ASValueSign.
  typeclass.
Defined.


(** * Formulae Extensions **)

(** ** Definition **)

(** Formula extensions ([Ext] types) carry a lot of properties which
  we do not want to precise at each use.  This type packages the
  needed properties. **)
Record FormulaExtension (Ext : Type) := {
    FormulaExtension_Poset : PosetDec.t Ext ;

    FormulaExtension_substs : Ext → list (ALoc L K) → ALoc L K → Ext ;
    FormulaExtension_Esubsts : Ext → list KE → KE → Ext ;
    FormulaExtension_substs_congruence : forall hs h ext1 ext2,
      ext1 ⊑♯ ext2 →
      FormulaExtension_substs ext1 hs h ⊑♯ FormulaExtension_substs ext2 hs h ;
    FormulaExtension_Esubsts_congruence : forall kes ke ext1 ext2,
      ext1 ⊑♯ ext2 →
      FormulaExtension_Esubsts ext1 kes ke ⊑♯ FormulaExtension_Esubsts ext2 kes ke ;

    FormulaExtension_Names : Ext → list (ALoc L K) ;
    FormulaExtension_ENames : Ext → list KE ;
    FormulaExtension_Names_substs : forall ext h hs h0,
      mem h0 (FormulaExtension_Names (FormulaExtension_substs ext hs h))
        = mem h (FormulaExtension_Names ext) &&& mem h0 hs
        ||| mem h0 (FormulaExtension_Names ext) &&& decide (h ≠ h0) ;
    FormulaExtension_ENames_Esubsts : forall ext ke kes ke0,
      mem ke0 (FormulaExtension_ENames (FormulaExtension_Esubsts ext kes ke))
        = mem ke (FormulaExtension_ENames ext) &&& mem ke0 kes
        ||| mem ke0 (FormulaExtension_ENames ext) &&& decide (ke ≠ ke0) ;
    FormulaExtension_Names_Esubsts : forall ext ke kes h,
      mem h (FormulaExtension_Names (FormulaExtension_Esubsts ext kes ke))
      = mem h (FormulaExtension_Names ext) ;
    FormulaExtension_ENames_substs : forall ext h hs ke,
      mem ke (FormulaExtension_ENames (FormulaExtension_substs ext hs h))
      = mem ke (FormulaExtension_ENames ext) ;

    FormulaExtension_C : Type ;
    FormulaExtension_gamma : valuation L K KE → Ext → ℘ FormulaExtension_C ;
    FormulaExtension_gamma_monotone : forall rho ext1 ext2,
      valid_valuation rho →
      ext1 ⊑♯ ext2 →
      FormulaExtension_gamma rho ext1 ⊑ FormulaExtension_gamma rho ext2
  }.

Global Instance : forall Ext,
    FormulaExtension Ext →
    PosetDec.t Ext.
  introv FExt. apply FExt.
Defined.

(** It is sometimes needed to have some additional properties. **)
Record FormulaExtensionBot (Ext : Type) := {
    FormulaExtensionNormal :> FormulaExtension Ext ;

    FormulaExtension_Bot : @BotDec.t _ (FormulaExtension_Poset FormulaExtensionNormal) ;
    FormulaExtension_bot_gamma : forall rho c,
      valid_valuation rho →
      ~ FormulaExtension_gamma FormulaExtensionNormal rho
          (BotDec.elem (t := FormulaExtension_Bot)) c ;
    FormulaExtension_bot_Names :
      FormulaExtension_Names FormulaExtensionNormal
        (BotDec.elem (t := FormulaExtension_Bot)) = nil ;
    FormulaExtension_bot_ENames :
      FormulaExtension_ENames FormulaExtensionNormal
        (BotDec.elem (t := FormulaExtension_Bot)) = nil
  }.

Global Instance : forall Ext,
    FormulaExtensionBot Ext →
    PosetDec.t Ext.
  introv FExt. apply FExt.
Defined.


(** ** Generic Instantion of Extensions **)

(** *** Product **)

Definition FormulaExtension_prod : forall A B,
    FormulaExtension A →
    FormulaExtension B →
    FormulaExtension (A * B).
  introv FExtA FExtB. refine {|
      FormulaExtension_Poset := ProductPosetDec _ _ ;
      FormulaExtension_substs := fun ab hs h =>
        (FormulaExtension_substs FExtA (fst ab) hs h,
         FormulaExtension_substs FExtB (snd ab) hs h) ;
      FormulaExtension_Esubsts := fun ab kes ke =>
        (FormulaExtension_Esubsts FExtA (fst ab) kes ke,
         FormulaExtension_Esubsts FExtB (snd ab) kes ke) ;
      FormulaExtension_Names := fun ab =>
        Remove_duplicates (FormulaExtension_Names FExtA (fst ab)
          ++ FormulaExtension_Names FExtB (snd ab)) ;
      FormulaExtension_ENames := fun ab =>
        Remove_duplicates (FormulaExtension_ENames FExtA (fst ab)
          ++ FormulaExtension_ENames FExtB (snd ab)) ;
      FormulaExtension_C := FormulaExtension_C FExtA * FormulaExtension_C FExtB ;
      FormulaExtension_gamma := fun rho =>
        productProdGamma (FormulaExtension_gamma FExtA rho)
          (FormulaExtension_gamma FExtB rho)
    |}.
   introv (O1&O2). split.
    apply~ FExtA.
    apply~ FExtB.
   introv (O1&O2). split.
    apply~ FExtA.
    apply~ FExtB.
   introv. repeat rewrite Remove_duplicates_mem. repeat rewrite mem_app.
    simpl. repeat rewrite FormulaExtension_Names_substs. extens. rew_refl*.
   introv. repeat rewrite Remove_duplicates_mem. repeat rewrite mem_app.
    simpl. repeat rewrite FormulaExtension_ENames_Esubsts. extens. rew_refl*.
   introv. repeat rewrite Remove_duplicates_mem. repeat rewrite mem_app.
    simpl. repeat rewrite FormulaExtension_Names_Esubsts. extens. rew_refl*.
   introv. repeat rewrite Remove_duplicates_mem. repeat rewrite mem_app.
    simpl. repeat rewrite FormulaExtension_ENames_substs. extens. rew_refl*.
   introv V (O1&O2) (G1&G2). split.
    applys~ FormulaExtension_gamma_monotone O1.
    applys~ FormulaExtension_gamma_monotone O2.
Defined.

Definition FormulaExtensionBot_prod : forall A B,
    FormulaExtensionBot A →
    FormulaExtensionBot B →
    FormulaExtensionBot (A * B).
  introv FExtA FExtB. refine {|
      FormulaExtensionNormal := FormulaExtension_prod FExtA FExtB ;
      FormulaExtension_Bot :=
        ProductBotDec (FormulaExtension_Bot FExtA) (FormulaExtension_Bot FExtB)
    |}.
   introv V (G1&G2). applys* FormulaExtension_bot_gamma G1.
   simpl. repeat rewrite~ FormulaExtension_bot_Names.
   simpl. repeat rewrite~ FormulaExtension_bot_ENames.
Defined.


(** *** [unit] **)

Definition FormulaExtension_unit : FormulaExtension (LiftBot.t UnitLattice.t).
  refine {|
      FormulaExtension_Poset := LiftBot.PosetDec _ ;
      FormulaExtension_substs := fun u _ _ => u ;
      FormulaExtension_Esubsts := fun u _ _ => u ;
      FormulaExtension_Names := fun u => nil ;
      FormulaExtension_ENames := fun u => nil ;
      FormulaExtension_C := unit ;
      FormulaExtension_gamma := fun rho => LiftBot.gamma UnitLattice.gamma
    |}; autos~.
  introv V O G. destruct ext1; destruct ext2; inverts~ O. false*.
Defined.

Definition FormulaExtensionBot_unit : FormulaExtensionBot (LiftBot.t UnitLattice.t).
  refine {|
      FormulaExtensionNormal := FormulaExtension_unit ;
      FormulaExtension_Bot := LatticeDec.bot (t := LiftBot.LatDec _ _ _ _)
    |}; autos*.
Defined.


(** ** Instantiations of Extensions for the Domains **)

(** *** [KE] **)

Definition FormulaExtension_KE : FormulaExtension (FinitePowerSet.t KE).
  refine (let _ := Comparable_EquivDec _ : EquivDec.t KE in {|
      FormulaExtension_substs := fun ake hs h => ake ;
      FormulaExtension_Esubsts := fun ake kes ke =>
        FinitePowerSet.substs _ _ ake ke (FinitePowerSet.from_list _ _ kes) ;
      FormulaExtension_Names := fun ake => nil ;
      FormulaExtension_ENames := fun ake =>
        LibOption.unsome_default nil (FinitePowerSet.to_list ake) ;
      FormulaExtension_C := floc ;
      FormulaExtension_gamma := fun rho =>
        immerse_gamma_prop (FinitePowerSet.gamma _) (fevalrho rho)
    |}).
   autos*.
   introv O. apply~ FinitePowerSet.substs_congruence.
   reflexivity.
   introv. extens. rew_refl. iff I.
    forwards (?&C): @FinitePowerSet.from_list_spec (@Comparable_EquivDec KE _).
     forwards O: @id C.
      eexists. splits; [ apply I |]. autos*.
     clear - O. destruct FinitePowerSet.substs eqn: E.
      destruct ext; tryfalse.
       forwards [(O'&[N|E'])|(O'&E')]: FinitePowerSet.substs_inv O.
        eapply EquivDec.trans; [| apply EquivDec.sym; apply FinitePowerSet.from_list_simpl ].
         apply* eq_EquivDec.
        right. splits~. forwards O'': @PosetDec.trans O'.
          apply PosetDec.refl. apply EquivDec.sym. apply FinitePowerSet.from_list_simpl.
         apply FinitePowerSet.from_list_spec in O''. lets (e'&Me'&Ee'): (rm O'').
         simpl in Ee'. substs~.
        false. applys* FinitePowerSet.some_not_top E'. introv (ke'&Mke'&Eke').
         simpl in Eke'. substs. false* pick_spec. apply nat_infinite.
        left. splits.
         simpl. forwards (C&?): FinitePowerSet.from_list_spec (@Comparable_EquivDec KE _).
          forwards (e'&Me'&Ee'): @id C.
           eapply @PosetDec.trans.
            apply E'.
            apply PosetDec.refl. apply EquivDec.sym. apply FinitePowerSet.from_list_simpl.
          simpl in Ee'. substs~.
         forwards (C&?): FinitePowerSet.from_list_spec (@Comparable_EquivDec KE _).
          forwards (e'&Me'&Ee'): C O'. substs~.
       false. applys* FinitePowerSet.const_bot O.
    inverts I as (I1&I2).
     skip. (* Admitted *)
     skip. (* Admitted *)
   reflexivity.
   reflexivity.
   introv V O (c&G&E). exists c. splits~. applys FinitePowerSet.Gamma; eassumption.
Defined.

Definition FormulaExtensionBot_KE : FormulaExtensionBot (FinitePowerSet.t KE).
  refine {|
      FormulaExtensionNormal := FormulaExtension_KE
    |}; autos*.
   introv (ke&G&E). eapply FinitePowerSet.gamma_bot in G. autos*.
Defined.


(** *** [AValue] **)

Definition FormulaExtension_AValue : FormulaExtension AValueSign.
  refine ({|
      FormulaExtension_Poset := LatticeDec.porder (t := AValue_Lat _ _ _ _) ;
      FormulaExtension_substs := AValue_substs _ _ (ABasic := _) ;
      FormulaExtension_Esubsts := AValue_Esubsts _ (ABasic := _) ;
      FormulaExtension_Names := AVNames (ABasic := _) ;
      FormulaExtension_ENames := AVENames (ABasic := _) ;
      FormulaExtension_C := val ;
      FormulaExtension_gamma := AValueSign_gamma
    |}).
   apply~ AValue_substs_congruence.
   apply~ AValue_Esubsts_congruence. apply~ nat_infinite.
   apply~ AVNames_substs; try typeclass. apply~ nat_infinite.
   apply~ AVENames_Esubsts. apply~ nat_infinite.
   apply~ AVNames_Esubsts.
   apply~ AVENames_substs.
   introv V. apply~ AValue_Gamma. typeclass.
Defined.

Definition FormulaExtensionBot_AValue : FormulaExtensionBot AValueSign.
  refine {|
      FormulaExtensionNormal := FormulaExtension_AValue
    |}; autos*.
   introv V. apply~ AValue_gamma_bot. apply~ Sign.bottom_empty.
Defined.


(** *** [ASValue] **)

Definition FormulaExtension_ASValue : FormulaExtension ASValueSign.
  refine ({|
      FormulaExtension_Poset := LatticeDec.porder (t := ASValue_Lat _ _ _ _) ;
      FormulaExtension_substs := ASValue_substs _ _ (ABasic := _) ;
      FormulaExtension_Esubsts := ASValue_Esubsts _ (ABasic := _) ;
      FormulaExtension_Names := ASVNames (ABasic := _) ;
      FormulaExtension_ENames := ASVENames (ABasic := _) ;
      FormulaExtension_C := option val ;
      FormulaExtension_gamma := ASValueSign_gamma
    |}).
   apply~ ASValue_substs_congruence.
   apply~ ASValue_Esubsts_congruence. apply~ nat_infinite.
   apply~ ASVNames_substs; try typeclass. apply~ nat_infinite.
   apply~ ASVENames_Esubsts. apply~ nat_infinite.
   apply~ ASVNames_Esubsts.
   apply~ ASVENames_substs.
   introv V. apply~ ASValue_Gamma. typeclass.
Defined.

Definition FormulaExtensionBot_ASValue : FormulaExtensionBot ASValueSign.
  refine {|
      FormulaExtensionNormal := FormulaExtension_ASValue
    |}; autos*.
   introv V. apply~ ASValue_gamma_bot. apply~ Sign.bottom_empty.
Defined.


(** *** [ALocS] **)

Definition FormulaExtension_ALocS : FormulaExtension (ALocS L K).
  refine ({|
      FormulaExtension_substs := fun l hs h =>
        ALocS_substs _ _ l h (FinitePowerSet.from_list _ _ hs) ;
      FormulaExtension_Esubsts := fun l _ _ => l ;
      FormulaExtension_Names := fun l =>
        unsome_default nil (FinitePowerSet.to_list l) ;
      FormulaExtension_ENames := fun l => nil ;
      FormulaExtension_C := loc ;
      FormulaExtension_gamma := ALocS_gamma _ _
    |}).
   introv O. apply~ FinitePowerSet.substs_congruence.
   autos*.
   introv. skip. (* Admitted *)
   reflexivity.
   reflexivity.
   reflexivity.
   introv V O (ke&G&E). exists ke. splits~. eapply FinitePowerSet.Gamma; eassumption.
Defined.

Definition FormulaExtensionBot_ALocS : FormulaExtensionBot (ALocS L K).
  refine {|
      FormulaExtensionNormal := FormulaExtension_ALocS
    |}; autos*.
   introv V (ke&G&E). eapply FinitePowerSet.gamma_bot in G. autos*.
Defined.


(** * Implementations **)

(** ** Instantiation For The Sign Abstraction **)

Definition IFormulaSign :=
  IFormula L K KE ABasic.

Definition DMembraneSign :=
  @DMembrane L K KE _ _ _.

Definition EFormulaSign (Ext : Type) : Type :=
  EFormula L K KE ABasic Ext.

Definition EFormulaSign_gamma Ext (FExt : FormulaExtension Ext) :=
  EFormulaGamma _ _ _ _ Sign.gamma (FormulaExtension_gamma FExt).

Definition DFormulaSign (Ext : Type) (FExt : FormulaExtension Ext) : Type :=
  @DFormula L K KE _ _ _ ABasic Ext
    (FormulaExtension_Names FExt) (FormulaExtension_ENames FExt).

Global Instance DFormulaSign_Poset : forall Ext (FExt : FormulaExtension Ext),
    Poset.t (DFormulaSign FExt).
  introv. eapply DFormula_Poset.
   typeclass.
   exact Sign.gamma.
   applys FormulaExtension_gamma FExt.
Defined.

Definition DFormulaSign_gamma Ext (FExt : FormulaExtension Ext) : DFormulaSign FExt → _ :=
  DFormulaGamma _ Sign.gamma
    (ExtNames := FormulaExtension_Names FExt) (ExtENames := FormulaExtension_ENames FExt)
    (FormulaExtension_gamma FExt).

Definition DFrameMembraneSign (Ext : Type) (FExt : FormulaExtension Ext)
    : DMembraneSign → DFormulaSign FExt → option (DFormulaSign FExt) :=
  DFrameMembrane nat_infinite nat_infinite nat_infinite
    (ExtENames := FormulaExtension_ENames FExt).

Definition DFrameFormulaSign (Ext : Type) (FExt : FormulaExtension Ext)
    : IFormulaSign → DFormulaSign FExt → option (DFormulaSign FExt) :=
  DFrameFormula nat_infinite nat_infinite nat_infinite
    nat_infinite_pick nat_infinite_pick nat_infinite_pick
    _ _ Sign.Gamma Sign.bottom_empty Sign.gamma_top
    (FormulaExtension_substs FExt) (FormulaExtension_Esubsts FExt)
    (ExtENames := _).

Notation "M 'D°' Phi" := (DFrameMembraneSign M Phi)
  (at level 10) : sep_scope.

Notation "phi 'D*' Phi" := (DFrameFormulaSign phi Phi)
  (at level 10) : sep_scope.

Definition DFormulaExtends Ext (FExt : FormulaExtension Ext) Phi1 C Phi2 :=
  formulaExtends (DFrameMembraneSign (FExt := _)) (DFrameFormulaSign (FExt := FExt))
    Phi1 C Phi2.

Definition proj_formulaContext (C : formulaContext DMembraneSign IFormulaSign) :=
  match C with
  | formulaContextMembrane M => formulaContextMembrane _ (proj1_sig M)
  | formulaContextFormula phi => formulaContextFormula _ phi
  end.

Lemma DFormulaExtends_EFormulaExtens : forall Ext (FExt : FormulaExtension Ext) Phi1 C Phi2,
  DFormulaExtends (FExt := FExt) Phi1 C Phi2 →
  @EFormulaExtends L K KE _ _ _ nat_infinite nat_infinite nat_infinite
    nat_infinite_pick nat_infinite_pick nat_infinite_pick
    _ _ _ _ (FormulaExtension_substs FExt) (FormulaExtension_Esubsts FExt)
    (FormulaExtension_Names FExt) (FormulaExtension_ENames FExt)
    (proj1_sig Phi1) (LibList.map proj_formulaContext C) (proj1_sig Phi2).
Proof.
  asserts Deriv: (forall A B (o : option A) v (f : forall v', o = Some v' → B),
    match o as o' return o = o' → option B with
    | Some v => fun E => Some (f v E)
    | None => fun _ => None
    end eq_refl = Some v →
    exists vo E, v = f vo E).
   introv Eq. destruct o; inverts* Eq.
  introv Ex. induction Ex.
   constructors~.
   destruct M. destruct Phi1. forwards (Phi&Eq1&Eq2): Deriv H. substs. simpl.
    constructors*.
   destruct Phi1. forwards (Phi&Eq1&Eq2): Deriv H. substs. simpl.
    constructors*.
Qed.


(** * Abstract Domains *)

Notation "'emp'" := (IFormula_emp L K KE ABasic) : sep_scope.
Notation "phi1 '*' phi2" := (IFormula_star phi1 phi2) : sep_scope.
Notation "h '-->' o" := (IFormula_loc h o)
  (at level 39, no associativity) : sep_scope.
Notation "ke '~~>' E" := (IFormula_env ke E)
  (at level 39, no associativity) : sep_scope.
Notation "h '--->' hs" := (Membrane_rewriting KE h hs)
  (at level 56) : sep_scope.
Notation "'ν' h" := (Membrane_new KE h)
  (at level 56) : sep_scope.
Notation "ke '~~~>' kes" := (Membrane_Erewriting L K ke kes)
  (at level 56) : sep_scope.
Notation "'ν~' ke" := (Membrane_Enew L K ke)
  (at level 56) : sep_scope.

Open Scope sep_scope.


Section Domains.

(** ** Errors **)

Definition aError := FinitePowerSet.t error.

Instance error_Comparable : Comparable error.
  prove_comparable_simple_inductive.
Defined.

Global Instance aErrorLat : LatticeDec.t aError.
  apply FinitePowerSet.Lat.
   typeclass.
   apply~ Comparable_EquivDec. typeclass.
Defined.

Global Instance aErrorTop : TopDec.t aError.
  typeclass.
Defined.

Definition aError_gamma : aError → ℘ error :=
  FinitePowerSet.gamma (Comparable_EquivDec _).

Instance aError_Gamma : Gamma.t _ _ aError_gamma.
Proof. typeclass. Qed.


(** ** Initial State **)

Definition initialEFormula : EFormulaSign (FinitePowerSet.t KE) :=
  let lg := singleton_loc _ (@pick _ _ (nat_infinite_pick nil)) in
  let E := write_AEnv (AEnv_empty _ _ _ _) "global"%string (ALoc_AValue _ _ lg) in
  let ke := @pick _ _ (nat_infinite_pick nil) in
  EFormula_make (ke ~~> E * lg --> DObject_empty _ _ _ _)
    (FinitePowerSet.const ke)
    ([lg ---> [lg] ; ke ~~~> [ke]]).

Definition initialDFormula : DFormulaSign FormulaExtensionBot_KE.
  refine (exist _ initialEFormula _).
  constructors.
  (** Membrane **)
  simpl. constructors.
   (** Unique **)
   introv N1 N2. repeat inverts N1 as N1. repeat inverts N2 as N2. autos*.
   (** Unique Env **)
   introv N1 N2. repeat inverts N1 as N1. repeat inverts N2 as N2. autos*.
   (** Rewriting Unique **)
   introv M. simpl in M. rew_refl in M. repeat inverts M as M.
    constructors~. introv I. inverts I.
   (** Rewriting Unique Env **)
   introv M. repeat rewrite mem_cons in M. rew_refl in M. repeat inverts M as M.
    constructors~. introv M. inverts M.
  (** Naming **)
  simpl. introv M. rew_refl in M. repeat inverts M as M;
    (eapply MInInnerInterface_rewriting; [| simpl; rew_refl* ]); simpl; rew_refl*.
  (** Naming Env **)
  simpl. introv M. rew_refl in M. repeat inverts M as M.
   eapply MEInInnerInterface_rewriting; [| repeat rewrite mem_cons; rew_refl* ].
   simpl. rew_refl*.
  (** Naming Ext **)
  simpl. introv M. false*.
  (** Naming Ext Env **)
  simpl. introv M. rew_refl in M. repeat inverts M as M.
   eapply MEInInnerInterface_rewriting; [| repeat rewrite mem_cons; rew_refl* ].
   simpl. rew_refl*.
  (** Rewriting Complete **)
  simpl. introv E M1 M2 (o&S). cases_if~. inverts E. simpls. fold_bool. rew_refl in *.
  repeat inverts M1 as M1. repeat inverts M2 as M2. substs. eexists. autos*.
  (** Rewriting Complete Env **)
  simpl. introv E M1 M2 (o&S). cases_if~. inverts E. simpls. fold_bool. rew_refl in *.
  repeat inverts M1 as M1. repeat inverts M2 as M2. substs. eexists. autos*.
Defined.

Lemma initialEFormula_initialState : DFormulaSign_gamma initialDFormula
    (initialState, (snd (alloc_state emptyState))).
Proof.
  evar (l0 : L). evar (ke0 : KE). evar (lc0 : loc). evar (lf0 : floc).
  set (rho :=
    make_valuation (fun l => ifb l = l0 then Some lc0 else None) (fun _ : K => None)
      (fun ke => ifb ke = ke0 then Some (fun lf => lf = lf0) else None)).
  eexists. do 2 exists rho. asserts V: (valid_valuation rho).
    constructors.
     introv E1 E2. destruct h1; inverts E1 as E1.
      destruct h2; inverts E2 as E2.
       repeat cases_if. fold_bool. rew_refl in *. substs~.
       do 2 inverts E2 as E2.
      do 2 inverts E1 as E1.
     introv E1 E2. inverts E1 as (E1&E1'). inverts E2 as (E2&E2').
      simpls. repeat cases_if. fold_bool. rew_refl in *. substs~.
  splits~.
   constructors~.
    introv. iff I.
     inverts I as I; simpl in I; tryfalse. cases_if as I'. fold_bool. rew_refl in I'.
      eexists. simpl. substs. unfolds l0. rew_refl*.
     lets (hs&Mhs): (rm I). simpl in Mhs. rew_refl in Mhs. repeat inverts Mhs as Mhs.
      asserts WTF: (h = singleton_loc K 0).
       clear - Mhs. skip. (* Tactic [injection] unexplainably fails. *)
      substs. constructors~. simpl. unfolds l0. cases_if as I; autos~. fold_bool. rew_refl* in I.
    introv. iff I.
     inverts I as I; simpl in I. cases_if as I'. fold_bool. rew_refl in I'.
      eexists. simpl. cases_if as I0; autos~. fold_bool. rew_refl in I0. false I0.
      substs. unfolds~ ke0.
     lets (kes&Mkes): (rm I). simpl in Mkes. cases_if as I; tryfalse. fold_bool.
      rew_refl in I. repeat inverts I as I.
      asserts WTF: (ke = 0).
       clear - I. skip. (* Tactic [injection] unexplainably fails. *)
      substs. constructors~. simpl. unfolds ke0. cases_if as I'; autos~. fold_bool.
      rew_refl* in I'.
    introv. iff I.
     inverts I as I; simpl in I; tryfalse. cases_if as I'. fold_bool. rew_refl in I'.
      substs. eapply MInInnerInterface_rewriting.
       rewrite mem_cons. rew_refl*.
       simpl. rew_refl*.
     inverts I as M1 M2.
      simpl in M2. rew_refl in M2. repeat inverts M2 as M2.
       simpl in M1. rew_refl in M1. repeat inverts M1 as M1.
       constructors~. simpl. cases_if as I; autos~. fold_bool. rew_refl* in I.
      simpl in M1. rew_refl in M1. repeat inverts M1 as M1.
    introv. iff I.
     inverts I as I; simpl in I; tryfalse. cases_if as I'. fold_bool. rew_refl in I'.
      substs. eapply MEInInnerInterface_rewriting.
       rewrite mem_cons. rew_refl. left. reflexivity.
       simpl. cases_if as I'; autos~. fold_bool. rew_refl in I'. false~ I'.
     inverts I as M1 M2.
      simpl in M2. cases_if as I; tryfalse. fold_bool. rew_refl in I.
       repeat inverts I as I. simpl in M1. rew_refl in M1. repeat inverts M1 as M1.
       constructors~. simpl. cases_if as I'; autos~. fold_bool. rew_refl* in I'.
      simpl in M1. false*.
    introv M E. simpl in M. rew_refl in M. repeat inverts M as M.
      asserts WTF: (h = singleton_loc K 0).
       clear - M. skip. (* Tactic [injection] unexplainably fails. *)
     substs. eexists. simpl. rew_refl. splits*.
    introv M E. simpl in M. cases_if as I; tryfalse. fold_bool. rew_refl in I.
     repeat inverts I as I. asserts WTF: (ke = 0).
       clear - I. skip. (* Tactic [injection] unexplainably fails. *)
     substs. eexists. simpl. rew_refl. splits*.
    introv E. right*.
    introv E. right*.
   constructors.
    eapply (IFormulaEntails_env {| state_fresh_loc := S lc0 ; state_fresh_floc := S lf0 |}).
     introv A. apply not_indom_empty in A. autos~.
     introv NE A. simpl in A. false NE. eexists. splits.
      simpl. cases_if as I; autos~. fold_bool. rew_refl* in I.
      tests~: (k = lf0). forwards* A': @indom_write_inv A.
       apply not_indom_empty in A'. false*.
     introv (s&Eq&P). evar (l : loc). evar (x : var). exists (write empty x (val_loc l)). splits.
      intro y. tests: (y = "global"%string).
       rewrite read_write_AEnv. unfolds x. rewrite read_option_write_eq.
        do 3 constructors. splits.
         simpl. rewrite <- Exists_exists_st. apply~ Exists_here. case_if~.
         simpl. unfolds lc0. cases_if as I; autos~. fold_bool. rew_refl* in I.
        typeclass.
        typeclass.
       rewrite~ read_write_AEnv_propagates. rewrite read_AEnv_empty.
        rewrite~ read_option_write_neq.
         rewrite~ @not_indom_read_option. apply not_indom_empty.
        typeclass.
      simpl in Eq. cases_if as I. inverts Eq. fold_bool.  rew_refl in I. substs.
       apply Heap.binds_write_eq.
    eapply (IFormulaEntails_loc {| state_fresh_loc := S lc0 ; state_fresh_floc := S lf0 |}).
     introv NE A. simpl in A. false NE. simpl. cases_if as I.
      tests~: (k = lc0). forwards* A': @indom_write_inv A.
       apply not_indom_empty in A'. false*.
      fold_bool. rew_refl* in I.
     introv Eq. simpl in Eq. case_if. inverts Eq. exists (empty : obj). splits.
      intro y. rewrite not_indom_read_option.
       simpl. rew_refl*.
       apply not_indom_empty.
      simpl. apply Heap.binds_write_eq.
     introv A. apply not_indom_empty in A. false~.
    simpl. constructors.
     introv I. apply not_indom_empty in I. false*.
     introv B. apply not_binds_empty in B. false*.
     introv B. forwards [(E1&E2)|(D&B')]: @binds_write_inv B.
      substs. unfolds lc0. apply binds_write_eq.
      apply not_binds_empty in B'. false*.
    simpl. constructors.
     introv I1 I2. apply not_indom_empty in I2. false*.
     introv B. forwards [(E1&E2)|(D&B')]: @binds_write_inv B.
      substs. unfolds lf0. apply binds_write_eq.
      apply not_binds_empty in B'. false*.
     introv B. apply not_binds_empty in B. autos*.
    constructors~.
     introv A. false*.
     introv I. exact I.
     autos~.
   exists ke0. splits.
    simpl. rewrite <- Exists_exists_st. constructors. case_if*.
    exists (fun lf => lf = lf0). splits.
     simpl. cases_if as I; autos~. fold_bool. rew_refl* in I.
     unfolds lf0. reflexivity.
Qed.


(** ** Results **)

Section Param.

(** The type [owhile_ares] is a membraned type: it will be embedded in
  the type [owhile_ast] with additional information.  However, these
  additional data will never be carrying membranes. The idea is thus
  to put this additional information inside the [owhile_ares] type,
  which is parameterised. **)

(** For instance, the constructor [owhile_st_op_1] is composed of an
  environment location (of type [floc]) and an output (of type
  [owhile_res]. The output will be abstracted by a membrane formula
  [Phi]. However, the carried information of type [floc] will never be
  abstracted by a membranes formula, just as a normal abstract value,
  which has to be inside the membrane. **)

(** The type [owhile_ares] is thus parameterised by some additional
  abstract values of type [Ext] which should be considered in the
  context of the membrane of output (of type [owhile_ares]). **)

(** The type corresponding to [owhile_res] is thus [owhile_ares unit],
  the abstract result with no extension. **)

Variable Ext : Type.
Variable FExt : FormulaExtension Ext.

(** *** Outputs **)

(** [oute] is either a normal output or an error. A normal output
  carries a value, as well as the potential extension [Ext]. **)

Definition aOute :=
  (DFormulaSign (FormulaExtension_prod FormulaExtension_AValue FExt) * aError)%type.

(** [outs] is more complex: it can either be a normal output, but can
  also be a return output (and it then accompagnied by a value), or an
  error. To avoid having to deal with two different membranes, there
  is only one membrane to frame the state. In this membrane is put a
  potentially undefined value; if this value is undefined, this
  [aOuts] represents a normal output; if this value is not, then this
  [aOuts] represents a return output accompagnied with this value; if
  we don’t know whether the value is undefined or not (it has its own
  lattice which allows this state), then we don’t know whether the
  output is normal or return, but we still know some of its
  structure. In addition to all of this, [aOuts] of course also tracks
  its potential extension [Ext]. **)

Definition aOuts :=
  (DFormulaSign (FormulaExtension_prod FormulaExtension_ASValue FExt) * aError)%type.

Definition aOute_gamma : aOute → ℘ (oute * _) :=
  immerse_gamma (productSumGamma
      (DFormulaSign_gamma (FExt := FormulaExtension_prod FormulaExtension_AValue FExt))
      aError_gamma)
    (fun o =>
       match o return State * (val * _) + error with
       | (oute_val St v, c) => inl (St, (v, c))
       | (oute_error err, _) => inr err
       end).

Definition aOuts_gamma : aOuts → ℘ (outs * _) :=
  immerse_gamma (productSumGamma
      (DFormulaSign_gamma (FExt := FormulaExtension_prod FormulaExtension_ASValue FExt))
      aError_gamma)
    (fun o =>
       match o return State * (option val * _) + error with
       | (outs_normal St, c) => inl (St, (None, c))
       | (outs_return St v, c) => inl (St, (Some v, c))
       | (outs_error err, _) => inr err
       end).

Local Instance AOute_Equiv : Equiv.t aOute.
  apply ProductEquiv.
   eapply DFormula_Equiv.
    typeclass.
    exact Sign.gamma.
    apply (@FormulaExtension_gamma _ (FormulaExtension_prod FormulaExtension_AValue FExt)).
   typeclass.
Defined.

Local Instance aOuts_Equiv : Equiv.t aOuts.
  apply ProductEquiv.
   eapply DFormula_Equiv.
    typeclass.
    exact Sign.gamma.
    apply (@FormulaExtension_gamma _ (FormulaExtension_prod FormulaExtension_ASValue FExt)).
   typeclass.
Defined.

Local Instance AOute_Poset : Poset.t aOute.
  apply ProductPoset.
   eapply DFormula_Poset.
    typeclass.
    exact Sign.gamma.
    apply (@FormulaExtension_gamma _ (FormulaExtension_prod FormulaExtension_AValue FExt)).
   typeclass.
Defined.

Local Instance aOuts_Poset : Poset.t aOuts.
  apply ProductPoset.
   eapply DFormula_Poset.
    typeclass.
    exact Sign.gamma.
    apply (@FormulaExtension_gamma _ (FormulaExtension_prod FormulaExtension_ASValue FExt)).
   typeclass.
Defined.

Definition AOuteFrameMembrane M (o : aOute) : option aOute :=
  let (Phi, e) := o in
  LibOption.map_on (M D° Phi) (fun Phi => (Phi, e)).

Definition AOuteFrameFormula phi (o : aOute) : option aOute :=
  let (Phi, e) := o in
  LibOption.map_on (phi D* Phi) (fun Phi => (Phi, e)).

Definition AOutsFrameMembrane M (o : aOuts) : option aOuts :=
  let (Phi, e) := o in
  LibOption.map_on (M D° Phi) (fun Phi => (Phi, e)).

Definition AOutsFrameFormula phi (o : aOuts) : option aOuts :=
  let (Phi, e) := o in
  LibOption.map_on (phi D* Phi) (fun Phi => (Phi, e)).


(** *** Results **)

Inductive owhile_ares : Type :=
  | owhile_ares_expr : aOute -> owhile_ares
  | owhile_ares_stat : aOuts -> owhile_ares
  .

Definition owhile_ares_eq ar1 ar2 :=
  match ar1, ar2 with
  | owhile_ares_expr r1, owhile_ares_expr r2 => Equiv.eq r1 r2
  | owhile_ares_stat r1, owhile_ares_stat r2 => Equiv.eq r1 r2
  | _, _ => False
  end.

Global Instance owhile_ares_Equiv : Equiv.t owhile_ares.
  apply Equiv.Make with (eq := owhile_ares_eq).
   introv. destruct x; try apply Equiv.refl; exact I.
   introv E. destruct x; destruct y; try (apply Equiv.sym; apply E);
     try exact I; inverts E.
   introv E1 E2. destruct x; destruct y; destruct z; tryfalse;
     try apply (Equiv.trans _ _ _ E1 E2); exact I.
Defined.

Definition owhile_ares_order ar1 ar2 :=
  match ar1, ar2 with
  | owhile_ares_expr r1, owhile_ares_expr r2 => r1 ⊑ r2
  | owhile_ares_stat r1, owhile_ares_stat r2 => r1 ⊑ r2
  | _, _ => False
  end.

Global Instance owhile_ares_Poset : Poset.t owhile_ares.
  apply Poset.Make with (eq := owhile_ares_Equiv) (order := owhile_ares_order).
   introv E. destruct x; destruct y; try solve [ inverts E ]; try exact I;
     apply Poset.refl; apply E.
   introv R1 R2. destruct x; destruct y; try solve [ try inverts R1; inverts R2 ]; try exact I.
     applys @Poset.antisym R1 R2.
     applys @Poset.antisym R1 R2.
   introv R1 R2. destruct x; destruct y; destruct z;
     try solve [ try inverts R1; inverts R2 ]; try exact I;
     (eapply @Poset.trans; [ apply R1 | apply R2 ]).
Defined.

Definition owhile_ares_gamma : owhile_ares → ℘ (owhile_res * _) := fun ar r =>
  match ar, r with
  | owhile_ares_expr ao, (owhile_res_expr o, c) => aOute_gamma ao (o, c)
  | owhile_ares_stat ao, (owhile_res_stat o, c) => aOuts_gamma ao (o, c)
  | _, _ => False
  end.


Definition owhile_ares_frameMembrane M ar :=
  match ar with
  | owhile_ares_expr o =>
    LibOption.map_on (AOuteFrameMembrane M o) owhile_ares_expr
  | owhile_ares_stat o =>
    LibOption.map_on (AOutsFrameMembrane M o) owhile_ares_stat
  end.

Definition owhile_ares_frameFormula phi ar :=
  match ar with
  | owhile_ares_expr o =>
    LibOption.map_on (AOuteFrameFormula phi o) owhile_ares_expr
  | owhile_ares_stat o =>
    LibOption.map_on (AOutsFrameFormula phi o) owhile_ares_stat
  end.

Definition owhile_ares_extends :=
  formulaExtends owhile_ares_frameMembrane owhile_ares_frameFormula.

Lemma owhile_ares_extends_expr : forall o C r,
  owhile_ares_extends (owhile_ares_expr o) C r →
  exists o', r = owhile_ares_expr o'.
Proof.
  introv Ex. gen o r. induction C; introv Ex; inverts Ex as Eq Ex.
   autos*.
   simpl in Eq. destruct AOuteFrameMembrane; inverts* Eq.
   simpl in Eq. destruct AOuteFrameFormula; inverts* Eq.
Qed.

Lemma owhile_ares_extends_stat : forall o C r,
  owhile_ares_extends (owhile_ares_stat o) C r →
  exists o', r = owhile_ares_stat o'.
Proof.
  introv Ex. gen o r. induction C; introv Ex; inverts Ex as Eq Ex.
   autos*.
   simpl in Eq. destruct AOutsFrameMembrane; inverts* Eq.
   simpl in Eq. destruct AOutsFrameFormula; inverts* Eq.
Qed.

Lemma owhile_ares_extends_expr_extends : forall Phi1 Phi2 ext1 ext2 C,
  owhile_ares_extends (owhile_ares_expr (Phi1, ext1)) C (owhile_ares_expr (Phi2, ext2)) →
  DFormulaExtends Phi1 C Phi2 ∧ ext1 = ext2.
Proof.
  introv Ex. gen Phi1 ext1 Phi2 ext2. induction C; introv Ex; inverts Ex as E Ex.
   splits~. constructors.
   simpl in E. destruct (M D° Phi1) as [Phi1'|] eqn: E'; inverts E.
    forwards (Ex'&Eq): IHC Ex. splits~. constructors*.
   simpl in E. destruct (phi D* Phi1) as [Phi1'|] eqn: E'; inverts E.
    forwards (Ex'&Eq): IHC Ex. splits~. constructors*.
Qed.

Lemma owhile_ares_extends_stat_extends : forall Phi1 Phi2 ext1 ext2 C,
  owhile_ares_extends (owhile_ares_stat (Phi1, ext1)) C (owhile_ares_stat (Phi2, ext2)) →
  DFormulaExtends Phi1 C Phi2 ∧ ext1 = ext2.
Proof.
  introv Ex. gen Phi1 ext1 Phi2 ext2. induction C; introv Ex; inverts Ex as E Ex.
   splits~. constructors.
   simpl in E. destruct (M D° Phi1) as [Phi1'|] eqn: E'; inverts E.
    forwards (Ex'&Eq): IHC Ex. splits~. constructors*.
   simpl in E. destruct (phi D* Phi1) as [Phi1'|] eqn: E'; inverts E.
    forwards (Ex'&Eq): IHC Ex. splits~. constructors*.
Qed.

Lemma owhile_ares_gamma_correct : forall o1 o2 : owhile_ares,
  o1 ⊑ o2 →
  forall o1' o2' C,
    owhile_ares_extends o1 C o1' →
    owhile_ares_extends o2 C o2' →
    forall StExt,
      owhile_ares_gamma o1' StExt →
      owhile_ares_gamma o2' StExt.
Proof.
  introv O Ex1 Ex2 G. destruct o2.
   forwards ((Phi'&e')&Eo'): owhile_ares_extends_expr Ex2. substs. destruct o1; tryfalse.
    forwards ((Phi''&e'')&Eo''): owhile_ares_extends_expr Ex1. substs.
     destruct a0 as [Phi1 ext1]. destruct a as [Phi2 ext2]. lets (O1&O2): (rm O).
     forwards (Ex1'&Eq1): owhile_ares_extends_expr_extends Ex1.
     forwards (Ex2'&Eq2): owhile_ares_extends_expr_extends Ex2.
     simpl. destruct StExt as [r xt]. destruct r; tryfalse. destruct o; inverts G as G.
      constructors. refine (let Concr := DFormulaOrder_dom_concr O1 in _). applys~ Concr G.
       applys~ DFormulaExtends_EFormulaExtens Ex1'.
       applys~ DFormulaExtends_EFormulaExtens Ex2'.
      constructors. substs. applys~ @Gamma.monotone aError_Gamma O2.
   forwards ((Phi'&e')&Eo'): owhile_ares_extends_stat Ex2. substs. destruct o1; tryfalse.
    forwards ((Phi''&e'')&Eo''): owhile_ares_extends_stat Ex1. substs.
     destruct a0 as [Phi1 ext1]. destruct a as [Phi2 ext2]. lets (O1&O2): (rm O).
     forwards (Ex1'&Eq1): owhile_ares_extends_stat_extends Ex1.
     forwards (Ex2'&Eq2): owhile_ares_extends_stat_extends Ex2.
     simpl. destruct StExt as [r xt]. destruct r; tryfalse. destruct o; inverts G as G.
      constructors. refine (let Concr := DFormulaOrder_dom_concr O1 in _). applys~ Concr G.
       applys~ DFormulaExtends_EFormulaExtens Ex1'.
       applys~ DFormulaExtends_EFormulaExtens Ex2'.
      constructors. refine (let Concr := DFormulaOrder_dom_concr O1 in _). applys~ Concr G.
       applys~ DFormulaExtends_EFormulaExtens Ex1'.
       applys~ DFormulaExtends_EFormulaExtens Ex2'.
      constructors. substs. applys~ @Gamma.monotone aError_Gamma O2.
Qed.

End Param.


(** ** Semantic Contexts **)

Inductive owhile_ast : Type :=
  | owhile_ast_expr : DFormulaSign FormulaExtensionBot_KE -> owhile_ast
  | owhile_ast_stat : DFormulaSign FormulaExtensionBot_KE -> owhile_ast
  | owhile_ast_prog : DFormulaSign FormulaExtensionBot_KE -> owhile_ast

  | owhile_ast_in_1 : owhile_ares FormulaExtensionBot_unit -> owhile_ast

  | owhile_ast_property_1 : owhile_ares FormulaExtensionBot_unit -> owhile_ast
  | owhile_ast_op_1 : owhile_ares FormulaExtensionBot_KE -> owhile_ast
  | owhile_ast_op_2 : ABasic -> owhile_ares FormulaExtensionBot_unit -> owhile_ast
  | owhile_ast_not_1 : owhile_ares FormulaExtensionBot_unit -> owhile_ast
  | owhile_ast_eq_1 : owhile_ares FormulaExtensionBot_KE -> owhile_ast
  | owhile_ast_eq_2 : owhile_ares FormulaExtensionBot_AValue -> owhile_ast

  | owhile_ast_app_1 : owhile_ares FormulaExtensionBot_KE -> owhile_ast
  (*LATER: do it relationnaly.
  | owhile_ast_app_2 : FinitePowerSet.t (var * prog) -> owhile_ares FormulaExtensionBot_KE_KE -> owhile_ast*)
  | owhile_ast_app_2 :
      FinitePowerSet.t (var * prog) ->
      owhile_ares FormulaExtensionBot_KE ->
      owhile_ast
  | owhile_ast_app_3 : owhile_ares FormulaExtensionBot_unit -> owhile_ast

  | owhile_ast_seq_1 : owhile_ares FormulaExtensionBot_KE -> owhile_ast

  | owhile_ast_if_1 : owhile_ares FormulaExtensionBot_KE -> owhile_ast
  | owhile_ast_while_1 : owhile_ares FormulaExtensionBot_KE -> owhile_ast
  | owhile_ast_while_2 : owhile_ares FormulaExtensionBot_KE -> owhile_ast

  | owhile_ast_property_asgn_1 : owhile_ares FormulaExtensionBot_KE -> owhile_ast
  | owhile_ast_property_asgn_2 : owhile_ares FormulaExtensionBot_ALocS -> owhile_ast
  | owhile_ast_delete_1 : owhile_ares FormulaExtensionBot_unit -> owhile_ast

  | owhile_ast_return_1 : owhile_ares FormulaExtensionBot_unit -> owhile_ast
  | owhile_ast_run_1 : owhile_ares FormulaExtensionBot_unit -> owhile_ast
  .


Local Instance : EquivDec.t (var * prog).
  apply Comparable_EquivDec. typeclass.
Defined.

Local Instance : Equiv.t (DFormulaSign FormulaExtensionBot_KE).
  eapply DFormula_Equiv.
   typeclass.
   exact Sign.gamma.
   apply (@FormulaExtension_gamma _ FormulaExtensionBot_KE).
Defined.

Definition owhile_ast_eq asigma1 asigma2 :=
  match asigma1, asigma2 with
  | owhile_ast_expr Phi1, owhile_ast_expr Phi2 => Equiv.eq Phi1 Phi2
  | owhile_ast_stat Phi1, owhile_ast_stat Phi2 => Equiv.eq Phi1 Phi2
  | owhile_ast_prog Phi1, owhile_ast_prog Phi2 => Equiv.eq Phi1 Phi2

  | owhile_ast_in_1 ar1, owhile_ast_in_1 ar2 => Equiv.eq ar1 ar2

  | owhile_ast_property_1 ar1, owhile_ast_property_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_op_1 ar1, owhile_ast_op_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_op_2 av1 ar1, owhile_ast_op_2 av2 ar2 => Equiv.eq av1 av2 ∧ Equiv.eq ar1 ar2
  | owhile_ast_not_1 ar1, owhile_ast_not_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_eq_1 ar1, owhile_ast_eq_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_eq_2 ar1, owhile_ast_eq_2 ar2 => Equiv.eq ar1 ar2

  | owhile_ast_app_1 ar1, owhile_ast_app_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_app_2 ps1 ar1, owhile_ast_app_2 ps2 ar2 => Equiv.eq ps1 ps2 ∧ Equiv.eq ar1 ar2
  | owhile_ast_app_3 ar1, owhile_ast_app_3 ar2 => Equiv.eq ar1 ar2

  | owhile_ast_seq_1 ar1, owhile_ast_seq_1 ar2 => Equiv.eq ar1 ar2

  | owhile_ast_if_1 ar1, owhile_ast_if_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_while_1 ar1, owhile_ast_while_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_while_2 ar1, owhile_ast_while_2 ar2 => Equiv.eq ar1 ar2

  | owhile_ast_property_asgn_1 ar1, owhile_ast_property_asgn_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_property_asgn_2 ar1, owhile_ast_property_asgn_2 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_delete_1 ar1, owhile_ast_delete_1 ar2 => Equiv.eq ar1 ar2

  | owhile_ast_return_1 ar1, owhile_ast_return_1 ar2 => Equiv.eq ar1 ar2
  | owhile_ast_run_1 ar1, owhile_ast_run_1 ar2 => Equiv.eq ar1 ar2

  | _, _ => False
  end.

Global Instance owhile_ast_Equiv : Equiv.t owhile_ast.
  apply Equiv.Make with (eq := owhile_ast_eq).
   introv. destruct x; try apply Equiv.refl; try exact I; splits; apply Equiv.refl.
   introv E. destruct x; destruct y; try (apply Equiv.sym; apply E);
     try exact I; try solve [ inverts E ]; splits; apply Equiv.sym; apply E.
   introv E1 E2. destruct x; destruct y; try solve [ inverts E1 ]; destruct z;
     try exact I; try solve [ inverts E2 ]; try apply (Equiv.trans _ _ _ E1 E2).
    inverts E1. inverts E2. split*.
    inverts E1. inverts E2. split*.
Defined.

Local Instance : Poset.t (FinitePowerSet.t (var * prog)).
  apply PosetDec_Poset. apply FinitePowerSet.Lat.
   typeclass.
   apply Comparable_EquivDec. typeclass.
Defined.

Local Instance : Poset.t (DFormulaSign FormulaExtensionBot_KE).
  eapply DFormula_Poset.
   typeclass.
   exact Sign.gamma.
   apply (@FormulaExtension_gamma _ FormulaExtensionBot_KE).
Defined.

Definition owhile_ast_order asigma1 asigma2 :=
  match asigma1, asigma2 with
  | owhile_ast_expr Phi1, owhile_ast_expr Phi2 => Phi1 ⊑ Phi2
  | owhile_ast_stat Phi1, owhile_ast_stat Phi2 => Phi1 ⊑ Phi2
  | owhile_ast_prog Phi1, owhile_ast_prog Phi2 => Phi1 ⊑ Phi2

  | owhile_ast_in_1 ar1, owhile_ast_in_1 ar2 => ar1 ⊑ ar2

  | owhile_ast_property_1 ar1, owhile_ast_property_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_op_1 ar1, owhile_ast_op_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_op_2 av1 ar1, owhile_ast_op_2 av2 ar2 => av1 ⊑ av2 ∧ ar1 ⊑ ar2
  | owhile_ast_not_1 ar1, owhile_ast_not_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_eq_1 ar1, owhile_ast_eq_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_eq_2 ar1, owhile_ast_eq_2 ar2 => ar1 ⊑ ar2

  | owhile_ast_app_1 ar1, owhile_ast_app_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_app_2 ps1 ar1, owhile_ast_app_2 ps2 ar2 => ps1 ⊑ ps2 ∧ ar1 ⊑ ar2
  | owhile_ast_app_3 ar1, owhile_ast_app_3 ar2 => ar1 ⊑ ar2

  | owhile_ast_seq_1 ar1, owhile_ast_seq_1 ar2 => ar1 ⊑ ar2

  | owhile_ast_if_1 ar1, owhile_ast_if_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_while_1 ar1, owhile_ast_while_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_while_2 ar1, owhile_ast_while_2 ar2 => ar1 ⊑ ar2

  | owhile_ast_property_asgn_1 ar1, owhile_ast_property_asgn_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_property_asgn_2 ar1, owhile_ast_property_asgn_2 ar2 => ar1 ⊑ ar2
  | owhile_ast_delete_1 ar1, owhile_ast_delete_1 ar2 => ar1 ⊑ ar2

  | owhile_ast_return_1 ar1, owhile_ast_return_1 ar2 => ar1 ⊑ ar2
  | owhile_ast_run_1 ar1, owhile_ast_run_1 ar2 => ar1 ⊑ ar2

  | _, _ => False
  end.

Global Instance owhile_ast_Poset : Poset.t owhile_ast.
  apply Poset.Make with (eq := owhile_ast_Equiv) (order := owhile_ast_order).
   introv E. destruct x; destruct y; try solve [ inverts E ]; try exact I;
    try (apply Poset.refl; apply E); inverts E; splits; apply~ @Poset.refl.
   introv R1 R2. destruct x; destruct y; try solve [ try inverts R1; inverts R2 ]; try exact I;
     unfolds; unfolds owhile_ast_Equiv; unfolds; try apply~ @Poset.antisym;
     try apply @Poset.antisym with (t := owhile_ares_Poset _); autos~; try solve [ split* ].
    inverts R1. inverts R2. split.
     applys~ (Poset.antisym a a0).
     apply @Poset.antisym with (t := owhile_ares_Poset _); autos~.
    inverts R1. inverts R2. split.
     applys~ (Poset.antisym t t0).
     apply @Poset.antisym with (t := owhile_ares_Poset _); autos~.
   introv R1 R2. destruct x; destruct y; try solve [ inverts R1 ]; destruct z;
     try solve [ inverts R2 ]; try exact I;
     try (eapply @Poset.trans; [ apply R1 | apply R2 ]).
    inverts R1. inverts R2. split; eapply @Poset.trans; eassumption.
    inverts R1. inverts R2. split; eapply @Poset.trans; eassumption.
Defined.


Definition owhile_ast_gamma : owhile_ast → ℘ owhile_st := fun asigma sigma =>
  match asigma, sigma with
  | owhile_ast_expr Phi, owhile_st_expr le St =>
    DFormulaSign_gamma Phi (St, le)
  | owhile_ast_stat Phi, owhile_st_stat le St =>
    DFormulaSign_gamma Phi (St, le)
  | owhile_ast_prog Phi, owhile_st_prog le St =>
    DFormulaSign_gamma Phi (St, le)

  | owhile_ast_in_1 ar, owhile_st_in_1 r =>
    owhile_ares_gamma ar (r, tt)

  | owhile_ast_property_1 ar, owhile_st_property_1 r =>
    owhile_ares_gamma ar (r, tt)
  | owhile_ast_op_1 ar, owhile_st_op_1 le r =>
    owhile_ares_gamma ar (r, le)
  | owhile_ast_op_2 av ar, owhile_st_op_2 v r =>
    Sign.gamma av v
    ∧ owhile_ares_gamma ar (r, tt)
  | owhile_ast_not_1 ar, owhile_st_not_1 r =>
    owhile_ares_gamma ar (r, tt)
  | owhile_ast_eq_1 ar, owhile_st_eq_1 le r =>
    owhile_ares_gamma ar (r, le)
  | owhile_ast_eq_2 ar, owhile_st_eq_2 v r =>
    owhile_ares_gamma ar (r, v)

  | owhile_ast_app_1 ar, owhile_st_app_1 le r =>
    owhile_ares_gamma ar (r, le)
  | owhile_ast_app_2 ps ar, owhile_st_app_2 x p le r =>
    FinitePowerSet.gamma _ ps (x, p)
    ∧ owhile_ares_gamma ar (r, le)
  | owhile_ast_app_3 ar, owhile_st_app_3 r =>
    owhile_ares_gamma ar (r, tt)

  | owhile_ast_seq_1 ar, owhile_st_seq_1 le r =>
    owhile_ares_gamma ar (r, le)

  | owhile_ast_if_1 ar, owhile_st_if_1 le r =>
    owhile_ares_gamma ar (r, le)
  | owhile_ast_while_1 ar, owhile_st_while_1 le r =>
    owhile_ares_gamma ar (r, le)
  | owhile_ast_while_2 ar, owhile_st_while_2 le r =>
    owhile_ares_gamma ar (r, le)

  | owhile_ast_property_asgn_1 ar, owhile_st_property_asgn_1 le r =>
    owhile_ares_gamma ar (r, le)
  | owhile_ast_property_asgn_2 ar, owhile_st_property_asgn_2 l r =>
    owhile_ares_gamma ar (r, l)
  | owhile_ast_delete_1 ar, owhile_st_delete_1 r =>
    owhile_ares_gamma ar (r, tt)

  | owhile_ast_return_1 ar, owhile_st_return_1 r =>
    owhile_ares_gamma ar (r, tt)
  | owhile_ast_run_1 ar, owhile_st_run_1 r =>
    owhile_ares_gamma ar (r, tt)

  | _, _ => False
  end.


Definition owhile_ast_frameMembrane M ar :=
  match ar with
  | owhile_ast_expr Phi =>
      LibOption.map_on (M D° Phi) owhile_ast_expr
  | owhile_ast_stat Phi =>
      LibOption.map_on (M D° Phi) owhile_ast_stat
  | owhile_ast_prog Phi =>
      LibOption.map_on (M D° Phi) owhile_ast_prog

  | owhile_ast_in_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_in_1

  | owhile_ast_property_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_property_1
  | owhile_ast_op_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_op_1
  | owhile_ast_op_2 av ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) (owhile_ast_op_2 av)
  | owhile_ast_not_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_not_1
  | owhile_ast_eq_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_eq_1
  | owhile_ast_eq_2 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_eq_2

  | owhile_ast_app_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_app_1
  | owhile_ast_app_2 ps ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) (owhile_ast_app_2 ps)
  | owhile_ast_app_3 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_app_3

  | owhile_ast_seq_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_seq_1

  | owhile_ast_if_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_if_1
  | owhile_ast_while_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_while_1
  | owhile_ast_while_2 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_while_2

  | owhile_ast_property_asgn_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_property_asgn_1
  | owhile_ast_property_asgn_2 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_property_asgn_2
  | owhile_ast_delete_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_delete_1

  | owhile_ast_return_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_return_1
  | owhile_ast_run_1 ar =>
      LibOption.map_on (owhile_ares_frameMembrane M ar) owhile_ast_run_1
  end.

Definition owhile_ast_frameFormula phi ar :=
  match ar with
  | owhile_ast_expr Phi =>
      LibOption.map_on (phi D* Phi) owhile_ast_expr
  | owhile_ast_stat Phi =>
      LibOption.map_on (phi D* Phi) owhile_ast_stat
  | owhile_ast_prog Phi =>
      LibOption.map_on (phi D* Phi) owhile_ast_prog

  | owhile_ast_in_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_in_1

  | owhile_ast_property_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_property_1
  | owhile_ast_op_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_op_1
  | owhile_ast_op_2 av ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) (owhile_ast_op_2 av)
  | owhile_ast_not_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_not_1
  | owhile_ast_eq_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_eq_1
  | owhile_ast_eq_2 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_eq_2

  | owhile_ast_app_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_app_1
  | owhile_ast_app_2 ps ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) (owhile_ast_app_2 ps)
  | owhile_ast_app_3 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_app_3

  | owhile_ast_seq_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_seq_1

  | owhile_ast_if_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_if_1
  | owhile_ast_while_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_while_1
  | owhile_ast_while_2 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_while_2

  | owhile_ast_property_asgn_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_property_asgn_1
  | owhile_ast_property_asgn_2 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_property_asgn_2
  | owhile_ast_delete_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_delete_1

  | owhile_ast_return_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_return_1
  | owhile_ast_run_1 ar =>
      LibOption.map_on (owhile_ares_frameFormula phi ar) owhile_ast_run_1
  end.

Definition owhile_ast_extends :=
  formulaExtends owhile_ast_frameMembrane owhile_ast_frameFormula.

Lemma DFormulaOrder_dom_extends : forall Phi1 Phi2 Phi1' Phi2' C,
  Phi1 ⊑ Phi2 →
  DFormulaExtends Phi1 C Phi1' →
  DFormulaExtends Phi2 C Phi2' →
  Phi1' ⊑ Phi2'.
Proof.
  introv O Ex1 Ex2. destruct Phi1 as [Phi1 OKPhi1]. destruct Phi2 as [Phi2 OKPhi2].
  destruct Phi1' as [Phi1' OKPhi1']. destruct Phi2' as [Phi2' OKPhi2'].
  apply DFormulaExtends_EFormulaExtens in Ex1. apply DFormulaExtends_EFormulaExtens in Ex2.
  eapply EFormulaOrder_dom_extends.
   apply O.
   apply Ex1.
   apply Ex2.
Qed.

Lemma owhile_ast_gamma_correct : forall o1 o2 : owhile_ast,
  o1 ⊑ o2 →
  forall o1' o2' C,
    owhile_ast_extends o1 C o1' →
    owhile_ast_extends o2 C o2' →
    forall StExt,
      owhile_ast_gamma o1' StExt →
      owhile_ast_gamma o2' StExt.
Proof.
  introv O Ex1 Ex2 G. destruct o1; destruct o2; tryfalse;
   try solve [
     repeat match goal with
     | Ex : owhile_ast_extends (?f ?o) C ?o' |- _ =>
       let on := fresh "o" in let En := fresh "E" in let Exn := fresh "Ex" in
       asserts (on&En&Exn): (exists on, o' = f on ∧ owhile_ares_extends o C on);
       [ let E := fresh "E" in let E' := fresh "E'" in
         let prove_branch frame :=
           simpl in E; destruct frame eqn: E'; inverts E;
           let on := fresh "o" in let En := fresh "E" in let Exn := fresh "Ex" in
           forwards~ (on&En&Exn): IHC Ex; substs; eexists; splits*; constructors*
         in
         clear - Ex; gen o o'; induction C; introv Ex; inverts Ex as E Ex;
          [ eexists; splits*; constructors*
          | prove_branch owhile_ares_frameMembrane
          | prove_branch owhile_ares_frameFormula ]
       | clear Ex ]
     end; substs; destruct StExt; tryfalse; simpls;
     first [
       applys owhile_ares_gamma_correct G; try eassumption
     | lets (G1&G2): (rm G); lets (O1&O2): (rm O); splits;
       [ first [
           applys~ Sign.gamma_monotone G1
         | eapply FinitePowerSet.Gamma; [ apply O1 | apply G1 ] ]
       | applys owhile_ares_gamma_correct G2; try eassumption ] ]
   | repeat match goal with
     | Ex : owhile_ast_extends (?f ?o) C ?o' |- _ =>
       let on := fresh "o" in let En := fresh "E" in let Exn := fresh "Ex" in
       asserts (on&En&Exn): (exists on, o' = f on ∧ DFormulaExtends o C on);
       [ let E := fresh "E" in let E' := fresh "E'" in
         let prove_branch frame :=
           simpl in E; destruct frame eqn: E'; inverts E;
           let on := fresh "o" in let En := fresh "E" in let Exn := fresh "Ex" in
           forwards~ (on&En&Exn): IHC Ex; substs; eexists; splits*; constructors*
         in
         clear - Ex; gen o o'; induction C; introv Ex; inverts Ex as E Ex;
          [ eexists; splits*; constructors*
          | prove_branch DFrameMembraneSign
          | prove_branch DFrameFormulaSign ]
       | clear Ex ]
     end; substs; destruct StExt; tryfalse;
     applys EFormulaOrder_dom_gamma G;
     applys* DFormulaOrder_dom_extends O ].
Qed.

End Domains.

