
(** This file contains defines functions to manipulate membranes. **)

Set Implicit Arguments.

Require Export FormulaeSep.

Section General.

(** [L] represents singleton locations, [K] set locations and [KE]
  environments, which behave like set locations. **)
Variable L K KE : Type. (** Typically, set of integers **)
Hypothesis L_comparable : Comparable L.
Hypothesis K_comparable : Comparable K.
Hypothesis KE_comparable : Comparable KE.

(** These sets have to be infinite. **)
Hypothesis L_infinite : forall ls,
  exists (l : L), ~ mem l ls.
Hypothesis K_infinite : forall ks,
  exists (k : K), ~ mem k ks.
Hypothesis KE_infinite : forall ks,
  exists (k : KE), ~ mem k ks.
Hypothesis L_infinite_pick : forall ls,
  Pickable (fun l : L => ~ mem l ls).
Hypothesis K_infinite_pick : forall ks,
  Pickable (fun k : K => ~ mem k ks).
Hypothesis KE_infinite_pick : forall ks,
  Pickable (fun k : KE => ~ mem k ks).

Variable ABasic : Type. (** Abstract basic values **)
Hypothesis ABasic_comparable : Comparable ABasic.
Hypothesis ABasicLat : LatticeDec.t ABasic.
Hypothesis ABasicTop : TopDec.t ABasic.
Variable ABasic_gamma : ABasic → ℘ basic_val.
Hypothesis ABasic_Gamma : Gamma.t _ _ ABasic_gamma.
Hypothesis ABasic_gamma_bot : forall b,
  ~ ABasic_gamma (⊥♯) b.
Hypothesis ABasic_gamma_top : forall b,
  ABasic_gamma (⊤♯) b.

Let ALoc := ALoc L K.
Let d_singleton_loc := singleton_loc _ : L → ALoc.
Let d_set_loc := set_loc _ (K := K) : K → ALoc.
Coercion d_singleton_loc : L >-> ALoc.
Coercion d_set_loc : K >-> ALoc.

Local Instance Local_Inhab_ALoc : Inhab ALoc.
  apply~ ALoc_inhab.
Defined.

Local Instance Local_Comparable_ALoc : Comparable ALoc.
  apply~ ALoc_comparable.
Defined.

Let AValue := AValue L K KE ABasic.
Let ASValue := ASValue L K KE ABasic.
Let DObject := DObject L K KE ABasic.
Let AEnv := AEnv L K KE ABasic.

Variable Ext : Type.
Variable Ext_substs : Ext → list ALoc → ALoc → Ext.
Variable Ext_Esubsts : Ext → list KE → KE → Ext.
Variable ExtNames : Ext → list ALoc.
Variable ExtENames : Ext → list KE.

Let IFormula := IFormula L K KE ABasic.
Let EFormula := EFormula L K KE ABasic Ext.
Let IMembrane := IMembrane L K KE.
Let Membrane := Membrane L K KE.
Let DFormula := DFormula _ _ _ ABasic ExtNames ExtENames.

Notation "'emp'" := (IFormula_emp _ _ _ _) : sep_scope.
Notation "phi1 '*' phi2" := (IFormula_star phi1 phi2) : sep_scope.
Notation "h '-->' o" := (IFormula_loc h o)
  (at level 39, no associativity) : sep_scope.
Notation "ke '~~>' E" := (IFormula_env ke E)
  (at level 39, no associativity) : sep_scope.

Notation "h '--->' hs" := (Membrane_rewriting _ h hs)
  (at level 56) : sep_scope.
Notation "'ν' h" := (Membrane_new _ h)
  (at level 56) : sep_scope.
Notation "ke '~~~>' kes" := (Membrane_Erewriting _ _ ke kes)
  (at level 56) : sep_scope.
Notation "'ν~' ke" := (Membrane_Enew _ _ ke)
  (at level 56) : sep_scope.

Open Scope sep_scope.

Instance IFormula_Inhab : Inhab IFormula.
  applys prove_Inhab (emp : IFormula).
Defined.

Let Mread := Mread _ _ : Membrane → ALoc → option (list ALoc).
Let Mread_env := Mread_env _ : Membrane → KE → option (list KE).

Let MInnerInterface_correct :=
  MInnerInterface_correct _ _ _ : forall (h : ALoc) (M : Membrane), _.
Let MInInnerInterfaceDecidable :=
  MInInnerInterfaceDecidable _ _ _ : forall (h : ALoc) (M : Membrane), _.
Let MEInnerInterface_correct :=
  MEInnerInterface_correct _ _ _ : forall (ke : KE) (M : Membrane), _.
Let MEInInnerInterfaceDecidable :=
  MEInInnerInterfaceDecidable _ _ _ : forall (ke : KE) (M : Membrane), _.

Let Membrane_correct : Membrane → Prop := Membrane_correct _ _ _.

Local Instance Local_AEnv_Inhab : Inhab AEnv.
  applys prove_Inhab. constructors.
Defined.

Local Instance Local_DObject_Inhab : Inhab DObject.
  applys prove_Inhab DObject_empty.
Defined.

Local Instance Local_locInDom_decidable : forall (phi : IFormula) (h : ALoc) f,
    Decidable (locInDom phi h f).
  apply locInDom_decidable; typeclass.
Defined.

Local Instance Local_envInDom_decidable : forall (phi : IFormula) ke,
    Decidable (envInDom phi ke).
  apply envInDom_decidable; typeclass.
Defined.


(** * Manipulating Abstract State Formulae **)

(** ** Passing Values Through Membranes **)

(** A general definition to apply a membrane, to be instantiated for
  [AValue] and [ASValue].  It can fail in case there is a name
  conflict: the names of the argument are supposed to be different
  from the inner interface of the formula. **)
Fixpoint applyMembrane (A : Type) (rew : A → list ALoc → ALoc → A) (names : A → list ALoc)
    (Erew : A → list KE → KE → A) (Enames : A → list KE) (M : Membrane) (av : A) :=
  match M with
  | nil => Some av
  | ν h :: M' =>
    if mem h (names av) then None (** Name conflict **)
    else applyMembrane rew names Erew Enames M' av
  | h ---> hs :: M' =>
    applyMembrane rew names Erew Enames M' (rew av hs h)
  | ν~ ke :: M' =>
    if mem ke (Enames av) then None (** Name conflict **)
    else applyMembrane rew names Erew Enames M' av
  | ke ~~~> kes :: M' =>
    applyMembrane rew names Erew Enames M' (Erew av kes ke)
  end.

(** No “ν” appear in the membrane.  This is enough for [applyMembrane]
  to be defined. **)
Definition noNewMembrane : Membrane → _ :=
  Forall (fun IM =>
    match IM with
    | ν h => False
    | ν~ ke => False
    | _ => True
    end).

(** No more than one location [h] is mapped to a given [h'] in the
  membrane: there are no two [l] and [l'] such that [l ---> k; l' --->
  k] is in the membrane for the same [k]. This is typical of renaming
  membranes, but not specific to. **)
Record noMergingMembrane (M : Membrane) := {
  noMergingMembrane_obj : forall h, length (MgetDependencies _ _ M h) <= 1 ;
  noMergingMembrane_env : forall ke, length (MgetDependencies_env _ M ke) <= 1 }.


Lemma noMergingMembrane_unique_obj : forall M n1 n2 h1 h2 hs,
  Membrane_correct M →
  noMergingMembrane M →
  Nth n1 M (h1 ---> hs) →
  Nth n2 M (h2 ---> hs) →
  hs ≠ nil →
  n1 = n2.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C NM N1 N2 D. asserts (h&I): (exists h, mem h hs).
    destruct hs; tryfalse~. eexists. simpl. rew_refl*.
  asserts: (h1 = h2).
    forwards GD: noMergingMembrane_obj NM h. apply Nth_mem with (CA := _) in N1.
    apply Nth_mem with (CA := _) in N2. rewrite Mread_mem in N1, N2; autos*.
    forwards* I1: MgetDependencies_Mread N1 I. forwards* I2: MgetDependencies_Mread N2 I.
    destruct MgetDependencies as [|? [|]]; tryfalse.
     simpl in I1, I2. rew_refl in I1. rew_refl in I2. inverts I1; tryfalse. inverts* I2.
     repeat rewrite length_cons in GD. math.
  substs. applys* Membrane_unique N1 N2.
Qed.

Lemma noMergingMembrane_unique_env : forall M n1 n2 ke1 ke2 kes,
  Membrane_correct M →
  noMergingMembrane M →
  Nth n1 M (ke1 ~~~> kes) →
  Nth n2 M (ke2 ~~~> kes) →
  kes ≠ nil →
  n1 = n2.
Proof.
  (** This lemma does not need these hypotheses, and I clear them to avoid unwanted
    dependencies. **)
  clear ABasic_Gamma ABasic_gamma_bot ABasic_gamma_top.
  introv C NM N1 N2 D. asserts (ke&I): (exists ke, mem ke kes).
    destruct kes; tryfalse~. eexists. simpl. rew_refl*.
  asserts: (ke1 = ke2).
    forwards GD: noMergingMembrane_env NM ke. apply Nth_mem with (CA := _) in N1.
    apply Nth_mem with (CA := _) in N2. rewrite Mread_env_mem in N1, N2; autos*.
    forwards* I1: MgetDependencies_Mread_env N1 I. forwards* I2: MgetDependencies_Mread_env N2 I.
    destruct MgetDependencies_env as [|? [|]]; tryfalse.
     simpl in I1, I2. rew_refl in I1. rew_refl in I2. inverts I1; tryfalse. inverts* I2.
     repeat rewrite length_cons in GD. math.
  substs. applys* Membrane_unique_env N1 N2.
Qed.


Definition applyMembraneAValue :=
  applyMembrane (AValue_substs _ _ (ABasic := _)) (AVNames (ABasic := _))
    (AValue_Esubsts _ (ABasic := _)) (AVENames (ABasic := ABasic)).

Definition applyMembraneASValue :=
  applyMembrane (ASValue_substs _ _ (ABasic := _)) (ASVNames (ABasic := _))
    (ASValue_Esubsts _ (ABasic := _)) (ASVENames (ABasic := ABasic)).

Lemma applyMembrane_Names_none : forall A rew Names Erew ENames (av av' : A) h M,
  (forall av h hs h',
     mem h' (Names (rew av hs h)) =
       mem h (Names av) &&& mem h' hs ||| mem h' (Names av) &&& decide (h ≠ h')) →
  (forall av ke kes h,
     mem h (Names (Erew av kes ke)) = mem h (Names av)) →
  applyMembrane rew Names Erew ENames M av = Some av' →
  mem h (Names av) →
  Mread M h = None →
  mem h (Names av').
Proof.
  introv NS NSE Ea Mh Er. gen av. induction M as [|IM M]; introv Ea Mh.
   inverts~ Ea.
   simpls. destruct IM as [h' hs|h'|ke' kes|ke'].
    cases_if as I; tryfalse. applys~ IHM Er Ea. rewrite NS.
     rew_refl. right. splits~. fold_bool. rew_refl~ in I.
    cases_if; tryfalse. applys~ IHM Ea.
    applys~ IHM Ea. rewrite~ NSE.
    cases_if; tryfalse. applys~ IHM Ea.
Qed.

Lemma applyMembrane_ENames_none : forall A rew Names Erew ENames (av av' : A) ke M,
  (forall av ke kes ke',
     mem ke' (ENames (Erew av kes ke)) =
       mem ke (ENames av) &&& mem ke' kes ||| mem ke' (ENames av) &&& decide (ke ≠ ke')) →
  (forall av h hs ke,
     mem ke (ENames (rew av hs h)) = mem ke (ENames av)) →
  applyMembrane rew Names Erew ENames M av = Some av' →
  mem ke (ENames av) →
  Mread_env M ke = None →
  mem ke (ENames av').
Proof.
  introv NESE NES Ea Mh Er. gen av. induction M as [|IM M]; introv Ea Mh.
   inverts~ Ea.
   simpls. destruct IM as [h' hs|h'|ke' kes|ke'].
    applys~ IHM Ea. rewrite~ NES.
    cases_if; tryfalse. applys~ IHM Ea.
    cases_if as I; tryfalse. applys~ IHM Er Ea. rewrite NESE.
     rew_refl. right. splits~. fold_bool. rew_refl~ in I.
    cases_if; tryfalse. applys~ IHM Ea.
Qed.

Lemma applyMembrane_Names_some : forall A rew Names Erew ENames (av av' : A) h h' hs M,
  (forall av h hs h',
     mem h' (Names (rew av hs h)) =
       mem h (Names av) &&& mem h' hs ||| mem h' (Names av) &&& decide (h ≠ h')) →
  (forall av ke kes h,
     mem h (Names (Erew av kes ke)) = mem h (Names av)) →
  Membrane_correct M →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  (forall h, MInInnerInterface _ _ _ h M → ~ MInOuterInterface _ _ _ h M) →
  applyMembrane rew Names Erew ENames M av = Some av' →
  mem h (Names av) →
  Mread M h = Some hs →
  mem h' hs →
  mem h' (Names av').
Proof.
  introv NS NSE C D Ea Mh Er Mh'. gen av. induction M as [|IM M]; introv Ea Mh.
   inverts Er.
   asserts CM: (Membrane_correct M).
     applys~ Membrane_correct_sub C; try rewrite~ app_nil_l; typeclass.
    asserts D': (forall h, MInInnerInterface _ _ _ h M → ~ MInOuterInterface _ _ _ h M).
      introv I1 I2. false D.
       applys~ MInInnerInterface_monotone I1.
       applys~ MInOuterInterface_monotone I2.
    destruct IM as [hr hrs|hr|kers ker|ker]; simpls.
     cases_if as I; fold_bool; rew_refl in I.
      substs. inverts Er. destruct (Mread M hr) eqn: Er.
       destruct C as [C]. unfolds in Er. forwards (hs'&Mhs'): Mread_defined_inv Er.
        forwards (n&N): mem_Nth Mhs'. false C.
         apply~ Nth_here.
         apply* Nth_next.
         false*.
       applys~ applyMembrane_Names_none Ea.
        rewrite NS. rew_refl. left*.
        destruct (Mread M h') eqn: Eh'; autos~. false D.
         applys~ MInInnerInterface_rewriting.
          apply Mh'.
          simpl. rew_refl*.
         apply~ MInOuterInterface_monotone. unfolds in Eh'. applys* Mread_defined_inv Eh'.
      applys~ IHM Ea. rewrite NS. repeat rewrite <- and_andb.
       rew_refl. right~.
     cases_if*.
     applys~ IHM Ea. rewrite~ NSE.
     cases_if*.
Qed.

Lemma applyMembrane_ENames_some : forall A rew Names Erew ENames (av av' : A) ke ke' kes M,
  (forall av ke kes ke',
     mem ke' (ENames (Erew av kes ke)) =
       mem ke (ENames av) &&& mem ke' kes ||| mem ke' (ENames av) &&& decide (ke ≠ ke')) →
  (forall av h hs ke,
     mem ke (ENames (rew av hs h)) = mem ke (ENames av)) →
  Membrane_correct M →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  (forall ke, MEInInnerInterface _ _ _ ke M → ~ MEInOuterInterface _ _ _ ke M) →
  applyMembrane rew Names Erew ENames M av = Some av' →
  mem ke (ENames av) →
  Mread_env M ke = Some kes →
  mem ke' kes →
  mem ke' (ENames av').
Proof.
  introv NESE NES C D Ea Mh Er Mh'. gen av. induction M as [|IM M]; introv Ea Mh.
   inverts Er.
   asserts CM: (Membrane_correct M).
     applys~ Membrane_correct_sub C; try rewrite~ app_nil_l; typeclass.
    asserts D': (forall ke, MEInInnerInterface _ _ _ ke M → ~ MEInOuterInterface _ _ _ ke M).
      introv I1 I2. false D.
       applys~ MEInInnerInterface_monotone I1.
       applys~ MEInOuterInterface_monotone I2.
    destruct IM as [hr hrs|hr|kers ker|ker]; simpls.
     applys~ IHM Ea. rewrite~ NES.
     cases_if*.
     cases_if as I; fold_bool; rew_refl in I.
      substs. inverts Er. destruct (Mread_env M kers) eqn: Er.
       destruct C as [? C]. unfolds in Er. forwards (hs'&Mhs'): Mread_env_defined_inv Er.
        forwards (n&N): mem_Nth Mhs'. false C.
         apply~ Nth_here.
         apply* Nth_next.
         false*.
       applys~ applyMembrane_ENames_none Ea.
        rewrite NESE. rew_refl. left*.
        destruct (Mread_env M ke') eqn: Eh'; autos~. false D.
         applys~ MEInInnerInterface_rewriting.
          apply Mh'.
          simpl. rew_refl*.
         apply~ MEInOuterInterface_monotone. unfolds in Eh'. applys* Mread_env_defined_inv Eh'.
      applys~ IHM Ea. rewrite NESE. repeat rewrite <- and_andb.
       rew_refl. right~.
     cases_if*.
Qed.

Lemma noNewMembrane_applyMembrane : forall A M rew Erew names Enames (av : A),
  noNewMembrane M → applyMembrane rew names Erew Enames M av ≠ None.
Proof.
  introv OKM. gen av; induction OKM; introv.
   discriminate.
   destruct x as [h hs|?| |]; tryfalse.
    simpl. apply~ IHOKM.
    simpl. apply~ IHOKM.
Qed.

Lemma applyMembraneAValue_AVNames : forall av av' h h' hs M,
  Membrane_correct M →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  (forall h, MInInnerInterface _ _ _ h M → ~ MInOuterInterface _ _ _ h M) →
  applyMembraneAValue M av = Some av' →
  mem h (AVNames av) →
  Mread M h = Some hs →
  mem h' hs →
  mem h' (AVNames av').
Proof.
  introv C D Ea Mh Er Mh'. applys* applyMembrane_Names_some Ea Er.
   apply AVNames_substs; typeclass.
   apply AVNames_Esubsts.
Qed.

Lemma applyMembraneAValue_AVENames : forall av av' ke ke' kes M,
  Membrane_correct M →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  (forall ke, MEInInnerInterface _ _ _ ke M → ~ MEInOuterInterface _ _ _ ke M) →
  applyMembraneAValue M av = Some av' →
  mem ke (AVENames av) →
  Mread_env M ke = Some kes →
  mem ke' kes →
  mem ke' (AVENames av').
Proof.
  introv C D Ea Mh Er Mh'. applys* applyMembrane_ENames_some Ea Er.
   apply AVENames_Esubsts; typeclass.
   apply AVENames_substs.
Qed.

Lemma applyMembraneASValue_ASVNames : forall av av' h h' hs M,
  Membrane_correct M →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  (forall h, MInInnerInterface _ _ _ h M → ~ MInOuterInterface _ _ _ h M) →
  applyMembraneASValue M av = Some av' →
  mem h (ASVNames av) →
  Mread M h = Some hs →
  mem h' hs →
  mem h' (ASVNames av').
Proof.
  introv C D Ea Mh Er Mh'. applys* applyMembrane_Names_some Ea Er.
   apply ASVNames_substs; typeclass.
   apply ASVNames_Esubsts.
Qed.

Lemma applyMembraneASValue_ASVENames : forall av av' ke ke' kes M,
  Membrane_correct M →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  (forall ke, MEInInnerInterface _ _ _ ke M → ~ MEInOuterInterface _ _ _ ke M) →
  applyMembraneASValue M av = Some av' →
  mem ke (ASVENames av) →
  Mread_env M ke = Some kes →
  mem ke' kes →
  mem ke' (ASVENames av').
Proof.
  introv C D Ea Mh Er Mh'. applys* applyMembrane_ENames_some Ea Er.
   apply ASVENames_Esubsts; typeclass.
   apply ASVENames_substs.
Qed.

Lemma applyMembraneAValue_AValue_substs : forall hs h v v',
  applyMembraneAValue ([h ---> hs]) v = Some v' →
  v' = AValue_substs _ _ v hs h.
Proof. introv E. inverts~ E. Qed.

Lemma applyMembraneASValue_ASValue_substs : forall hs h v v',
  applyMembraneASValue ([h ---> hs]) v = Some v' →
  v' = ASValue_substs _ _ v hs h.
Proof. introv E. inverts~ E. Qed.

Lemma applyMembraneAValue_AValue_Esubsts : forall kes ke v v',
  applyMembraneAValue ([ke ~~~> kes]) v = Some v' →
  v' = AValue_Esubsts _ v kes ke.
Proof. introv E. inverts~ E. Qed.

Lemma applyMembraneASValue_ASValue_Esubsts : forall kes ke v v',
  applyMembraneASValue ([ke ~~~> kes]) v = Some v' →
  v' = ASValue_Esubsts _ v kes ke.
Proof. introv E. inverts~ E. Qed.


(** Applying a membrane to an object.  Maybe [applyMembrane] could
  have been used to define this function, but iterating through the
  object appears natural. **)
Definition applyMembraneDObject M : DObject → option DObject :=
  DObject_map_partial _ _ _ _ _ (applyMembraneASValue M).

Local Instance : Inhab (property * ASValue). (* Coq is much slower without this hint. *)
  typeclass.
Defined.

Lemma applyMembraneDObject_ONames : forall o o' h h' hs M,
  Membrane_correct M →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  (forall h, MInInnerInterface _ _ _ h M → ~ MInOuterInterface _ _ _ h M) →
  applyMembraneDObject M o = Some o' →
  mem h (DNames o) →
  Mread M h = Some hs →
  mem h' hs →
  mem h' (DNames o').
Proof.
  introv C D E Ih Er Ih'. eapply ONames_param_inv in Ih; try apply (proj2_sig o); autos*.
  lets* (f&v&Efv&Mh): Ih. forwards~ (v'&E1&E2): read_DObject_DObject_map_partial E Efv.
  forwards* M': applyMembraneASValue_ASVNames E1.
  applys ONames_param_spec M'; try apply (proj2_sig o'); autos*.
Qed.

Lemma applyMembraneDObject_OENames : forall o o' ke ke' kes M,
  Membrane_correct M →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  (forall ke, MEInInnerInterface _ _ _ ke M → ~ MEInOuterInterface _ _ _ ke M) →
  applyMembraneDObject M o = Some o' →
  mem ke (DENames o) →
  Mread_env M ke = Some kes →
  mem ke' kes →
  mem ke' (DENames o').
Proof.
  introv C D E Ih Er Ih'. eapply ONames_param_inv in Ih; try apply (proj2_sig o); autos*.
  lets* (f&v&Efv&Mh): Ih. forwards~ (v'&E1&E2): read_DObject_DObject_map_partial E Efv.
  forwards* M': applyMembraneASValue_ASVENames E1.
  applys ONames_param_spec M'; try apply (proj2_sig o'); autos*.
Qed.

Lemma applyMembraneDObject_dom : forall o o' M,
  applyMembraneDObject M o = Some o' →
  AObject_domain_compare (DObject_dom o) (DObject_dom o').
Proof. introv E. applys~ DObject_map_partial_dom E. typeclass. Qed.

Lemma noNewMembrane_applyMembraneDObject : forall M o,
  noNewMembrane M →
  applyMembraneDObject M o ≠ None.
Proof.
  introv OKM E. forwards (f&v&R&E'): DObject_map_partial_None (rm E).
  applys~ noNewMembrane_applyMembrane E'.
Qed.

(** Applying a membrane to an environment. **)
Definition applyMembraneAEnv M (E : AEnv) : option AEnv :=
  map_partial (fun fv =>
   let (f, v) := fv : _ * _ in
   LibOption.map (fun v => (f, v)) (applyMembraneAValue M v)) E.

Local Instance : Inhab (var * AValue). (* Coq is much slower without this hint. *)
  typeclass.
Defined.

Lemma applyMembraneAEnv_ONames : forall E E' h h' hs M,
  Membrane_correct M →
  (forall h, MInInnerInterface _ _ _ h M → ~ MInOuterInterface _ _ _ h M) →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  applyMembraneAEnv M E = Some E' →
  mem h (ENames E) →
  Mread M h = Some hs →
  mem h' hs →
  mem h' (ENames E').
Proof.
  introv C D Eq Ih Er Ih'. unfolds in Eq. forwards F: map_partial_inv (rm Eq).
  unfolds ENames. rewrite concat_mem in *. lets (hs'&Mhs'&Mh): (rm Ih).
  rewrite map_mem in Mhs'. lets ([f av]&Mfav&Efav): (rm Mhs').
  substs. simpl in Mh. lets (n&N): mem_Nth (rm Mfav).
  forwards~ N': Forall2_Nth_nth_def (arbitrary : var * AValue) F N.
  forwards~ Eq: Forall2_forall_Nth (rm F) N N'.
  forwards (vr&Eq'&ET): LibOption.map_inv (rm Eq). rewrite ET in N'.
  apply concat_mem. exists (AVNames vr). splits~.
   rewrite map_mem. exists (f, vr). splits~. applys~ Nth_mem N'.
   applys* applyMembraneAValue_AVNames Eq'.
Qed.

Lemma applyMembraneAEnv_OENames : forall E E' ke ke' kes M,
  Membrane_correct M →
  (forall ke, MEInInnerInterface _ _ _ ke M → ~ MEInOuterInterface _ _ _ ke M) →
  (** The inner and outer interface have already been renamed in order to not intersect. **)
  applyMembraneAEnv M E = Some E' →
  mem ke (EENames E) →
  Mread_env M ke = Some kes →
  mem ke' kes →
  mem ke' (EENames E').
Proof.
  introv C D Eq Ih Er Ih'. unfolds in Eq. forwards F: map_partial_inv (rm Eq).
  unfolds EENames. rewrite concat_mem in *. lets (hs'&Mhs'&Mh): (rm Ih).
  rewrite map_mem in Mhs'. lets ([f av]&Mfav&Efav): (rm Mhs').
  substs. simpl in Mh. lets (n&N): mem_Nth (rm Mfav).
  forwards~ N': Forall2_Nth_nth_def (arbitrary : var * AValue) F N.
  forwards~ Eq: Forall2_forall_Nth (rm F) N N'.
  forwards (vr&Eq'&ET): LibOption.map_inv (rm Eq). rewrite ET in N'.
  exists (AVENames vr). splits~.
   rewrite map_mem. exists (f, vr). splits~. applys~ Nth_mem N'.
   applys* applyMembraneAValue_AVENames Eq'.
Qed.

Lemma applyMembraneAEnv_AEnv_substs : forall hs h E E',
  applyMembraneAEnv ([h ---> hs]) E = Some E' →
  E' = AEnv_substs _ _ E hs h.
Proof.
  introv Eq. unfolds in Eq. apply map_partial_inv in Eq. induction Eq.
   autos~.
   unfolds. rewrite map_cons. fequals~. destruct x1 as [f' v'].
    lets (v2&Ev1&Ev2): LibOption.map_inv (rm H). inverts Ev2. inverts Ev1. fequals.
Qed.

Lemma applyMembraneAEnv_AEnv_Esubsts : forall kes ke E E',
  applyMembraneAEnv ([ke ~~~> kes]) E = Some E' →
  E' = AEnv_Esubsts _ E kes ke.
Proof.
  introv Eq. unfolds in Eq. apply map_partial_inv in Eq. induction Eq.
   autos~.
   unfolds. rewrite map_cons. fequals~. destruct x1 as [f' v'].
    lets (v2&Ev1&Ev2): LibOption.map_inv (rm H). inverts Ev2. inverts Ev1. fequals.
Qed.

Lemma noNewMembrane_applyMembraneAEnv : forall M E,
  noNewMembrane M → applyMembraneAEnv M E ≠ None.
Proof.
  introv OKM Eq. unfolds in Eq. apply map_partial_inv_none in Eq.
  induction Eq as [hs (f&v) Eq|hs (f&v) Eq].
   apply option_map_none_inv in Eq. forwards~: noNewMembrane_applyMembrane Eq.
   autos~.
Qed.



(** ** Context Operators  **)

(** *** Composing Membranes **)

(** Composing two membranes. **)
Definition MembraneCompose (Minner Mouter : Membrane) : Membrane :=
  let mread h :=
    option_case ([h]) id (Mread Minner h) in
  let newify := map (fun h => ν h) in
  let mread_env ke :=
    option_case ([ke]) id (Mread_env Minner ke) in
  let newify_env := map (fun ke => ν~ ke) in
  concat (map (fun MI =>
    match MI with
    | h ---> hs =>
      [h ---> Remove_duplicates (concat (map mread hs))]
    | ν h =>
      newify (mread h)
    | ke ~~~> kes =>
      [ke ~~~> Remove_duplicates (concat (map mread_env kes))]
    | ν~ ke =>
      newify_env (mread_env ke)
    end) Mouter)
  ++ newify (Mnew Minner)
  ++ newify_env (Mnew_env Minner).

(** Checking whether the outer membrane fits the interface of the inner one. **)
Definition MembraneCheckCompose (Minner Mouter : Membrane) : option Membrane :=
  ifb incl (MOuterInterface Minner) (MInnerInterface Mouter)
   ∧ ~ Exists (fun h => mem h (MInnerInterface Mouter)
         ∧ ~ mem h (MOuterInterface Minner)) (MInnerInterface Minner)
   ∧ incl (MEOuterInterface Minner) (MEInnerInterface Mouter)
   ∧ ~ Exists (fun ke => mem ke (MEInnerInterface Mouter)
         ∧ ~ mem ke (MEOuterInterface Minner)) (MEInnerInterface Minner) then
    Some (MembraneCompose Minner Mouter)
  else None.

Lemma MembraneCheckCompose_MembraneCompose : forall Minner Mouter M,
  MembraneCheckCompose Minner Mouter = Some M →
  MembraneCompose Minner Mouter = M.
Proof. introv E. unfolds in E. cases_if; inverts~ E. Qed.

Lemma MembraneCheckCompose_cond : forall Minner Mouter M,
  MembraneCheckCompose Minner Mouter = Some M →
  incl (MOuterInterface Minner) (MInnerInterface Mouter)
  ∧ ~ Exists (fun h => mem h (MInnerInterface Mouter)
        ∧ ~ mem h (MOuterInterface Minner)) (MInnerInterface Minner)
  ∧ incl (MEOuterInterface Minner) (MEInnerInterface Mouter)
  ∧ ~ Exists (fun ke => mem ke (MEInnerInterface Mouter)
        ∧ ~ mem ke (MEOuterInterface Minner)) (MEInnerInterface Minner).
Proof. introv E. unfolds in E. cases_if as I; fold_bool; rew_refl~ in I. Qed.

Lemma MembraneCompose_correct : forall Minner Mouter,
  Membrane_correct Minner →
  Membrane_correct Mouter →
  Membrane_correct (MembraneCompose Minner Mouter).
Proof.
  introv Ci Co. unfolds MembraneCompose. constructors.
  (** Uniqueness **)
  introv I1 I2. forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1);
    [| forwards [I1|(m'&Em'&I1)]: Nth_app_inv (rm I1');
       forwards I1': Nth_mem I1; rewrite map_mem with (CA := _) in I1';
       lets (?&?&A): (rm I1'); inverts A ].
  forwards [I2'|(m&Em&I2')]: Nth_app_inv (rm I2);
    [| forwards [I2|(m'&Em'&I2)]: Nth_app_inv (rm I2');
       forwards I2': Nth_mem I2; rewrite map_mem with (CA := _) in I2';
       lets (?&?&A): (rm I2'); inverts A ].
  gen n1 n2. induction Mouter as [|IM Mouter]; introv I1 I2.
   inverts I1.
   rewrite map_cons in *. rewrite concat_cons in *. destruct IM as [h' hs'|h'|ke' kes'|ke'].
    forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1); [repeat inverts I1' as I1'|];
      (forwards [I2'|(m'&Em'&I2')]: Nth_app_inv (rm I2); [repeat inverts I2' as I2'|]).
     autos*.
     clear Em' IHMouter. asserts I: (mem h (MOuterInterface Mouter)).
       lets I: Nth_mem (rm I2'). rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
       rewrite map_mem in Mhs. lets (IM&MIM&EIM): (rm Mhs).
       unfolds Local_Comparable_ALoc. erewrite MOuterInterface_correct. rew_refl.
       destruct IM as [h0 hs0|h0|ke0 kes0|ke0]; substs.
        unfolds in Mh. fold mem in Mh. rew_refl in Mh. repeat inverts Mh as Mh. exists~ hs0.
        rewrite map_mem with (CA := _) in Mh. lets (h'&Mh'&Eh'): Mh. inverts Eh'.
        simpl in Mh. rew_refl in Mh. repeat inverts Mh as Mh.
        rewrite map_mem with (CA := _) in Mh. lets (h'&Mh'&Eh'): Mh. inverts Eh'.
      lets (hs&Mhs): MOuterInterface_mem_inv (rm I). lets (n&N): mem_Nth (rm Mhs).
      destruct Co as [Co]. forwards*: Co Nth_here Nth_next. false*.
     clear Em IHMouter. asserts I: (mem h (MOuterInterface Mouter)).
       lets I: Nth_mem (rm I1'). rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
       rewrite map_mem with (CA := _) in Mhs. lets (IM&MIM&EIM): (rm Mhs).
       unfolds Local_Comparable_ALoc. erewrite MOuterInterface_correct. rew_refl.
       destruct IM as [h0 hs0|h0|ke0 kes0|ke0]; substs.
        unfolds in Mh. fold mem in Mh. rew_refl in Mh. repeat inverts Mh as Mh. exists~ hs0.
        rewrite map_mem with (CA := _) in Mh. lets (h'&Mh'&Eh'): Mh. inverts Eh'.
        simpl in Mh. rew_refl in Mh. repeat inverts Mh as Mh.
        rewrite map_mem with (CA := _) in Mh. lets (h'&Mh'&Eh'): Mh. inverts Eh'.
      lets (hs&Mhs): MOuterInterface_mem_inv (rm I). lets (n&N): mem_Nth (rm Mhs).
      destruct Co as [Co]. forwards*: Co Nth_here Nth_next. false*.
     forwards: (rm IHMouter) (rm I1') (rm I2'); [| substs* ].
      applys~ Membrane_correct_sub Co; try rewrite~ app_nil_l; typeclass.
    forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1);
      [forwards I1: Nth_mem I1'; rewrite map_mem with (CA := _) in I1;
       lets (?&?&A): (rm I1); inverts A|].
     forwards [I2'|(m'&Em'&I2')]: Nth_app_inv (rm I2);
       [forwards I2: Nth_mem I2'; rewrite map_mem with (CA := _) in I2;
        lets (?&?&A): (rm I2); inverts A|].
     forwards: (rm IHMouter) (rm I1') (rm I2'); [| substs* ].
     applys~ Membrane_correct_sub Co; try rewrite~ app_nil_l; typeclass.
    forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1).
     forwards I1: Nth_mem I1'. simpl in I1. rew_refl in I1. repeat inverts I1 as I1.
     forwards [I2'|(m'&Em'&I2')]: Nth_app_inv (rm I2).
      forwards I2: Nth_mem I2'. simpl in I2. rew_refl in I2. repeat inverts I2 as I2.
      forwards~: IHMouter I1' I2'.
       applys~ Membrane_correct_sub Co; try rewrite~ app_nil_l; typeclass.
       substs. reflexivity.
    forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1).
     forwards I1: Nth_mem I1'. rewrite map_mem with (CA := _) in I1.
      lets (?&?&A): (rm I1). inverts A.
     forwards [I2'|(m'&Em'&I2')]: Nth_app_inv (rm I2).
      forwards I2: Nth_mem I2'. rewrite map_mem with (CA := _) in I2.
      lets (?&?&A): (rm I2). inverts A.
      forwards~: IHMouter I1' I2'.
       applys~ Membrane_correct_sub Co; try rewrite~ app_nil_l; typeclass.
       substs. reflexivity.
  (** Uniqueness env **)
  introv I1 I2. forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1);
    [| forwards [I1|(m'&Em'&I1)]: Nth_app_inv (rm I1');
       forwards I1': Nth_mem I1; rewrite map_mem with (CA := _) in I1';
       lets (?&?&A): (rm I1'); inverts A ].
  forwards [I2'|(m&Em&I2')]: Nth_app_inv (rm I2);
    [| forwards [I2|(m'&Em'&I2)]: Nth_app_inv (rm I2');
       forwards I2': Nth_mem I2; rewrite map_mem with (CA := _) in I2';
       lets (?&?&A): (rm I2'); inverts A ].
  gen n1 n2. induction Mouter as [|IM Mouter]; introv I1 I2.
   inverts I1.
   rewrite map_cons in *. rewrite concat_cons in *. destruct IM as [h' hs'|h'|ke' kes'|ke'].
    forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1).
     forwards I1: Nth_mem I1'. simpl in I1. rew_refl in I1. repeat inverts I1 as I1.
     forwards [I2'|(m'&Em'&I2')]: Nth_app_inv (rm I2).
      forwards I2: Nth_mem I2'. simpl in I2. rew_refl in I2. repeat inverts I2 as I2.
      forwards~: IHMouter I1' I2'.
       applys~ Membrane_correct_sub Co; try rewrite~ app_nil_l; typeclass.
       substs. reflexivity.
    forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1).
     forwards I1: Nth_mem I1'. rewrite map_mem with (CA := _) in I1.
      lets (?&?&A): (rm I1). inverts A.
     forwards [I2'|(m'&Em'&I2')]: Nth_app_inv (rm I2).
      forwards I2: Nth_mem I2'. rewrite map_mem with (CA := _) in I2.
       lets (?&?&A): (rm I2). inverts A.
      forwards~: IHMouter I1' I2'.
       applys~ Membrane_correct_sub Co; try rewrite~ app_nil_l; typeclass.
       substs. reflexivity.
    forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1); [repeat inverts I1' as I1'|];
      (forwards [I2'|(m'&Em'&I2')]: Nth_app_inv (rm I2); [repeat inverts I2' as I2'|]).
     autos*.
     clear Em' IHMouter. asserts I: (mem ke (MEOuterInterface Mouter)).
       lets I: Nth_mem (rm I2'). rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
       rewrite map_mem with (CA := _) in Mhs. lets (IM&MIM&EIM): (rm Mhs).
       erewrite MEOuterInterface_correct. rew_refl.
       destruct IM as [h0 hs0|h0|ke0 kes0|ke0]; substs.
        simpl in Mh. rew_refl in Mh. repeat inverts Mh as Mh.
        rewrite map_mem with (CA := _) in Mh. lets (h'&Mh'&Eh'): Mh. inverts Eh'.
        unfolds in Mh. fold mem in Mh. rew_refl in Mh. repeat inverts Mh as Mh. exists~ kes0.
        rewrite map_mem with (CA := _) in Mh. lets (h'&Mh'&Eh'): Mh. inverts Eh'.
      lets (hs&Mhs): MEOuterInterface_mem_inv (rm I). lets (n&N): mem_Nth (rm Mhs).
      destruct Co as [? Co]. forwards*: Co Nth_here Nth_next. false*.
     clear Em IHMouter. asserts I: (mem ke (MEOuterInterface Mouter)).
       lets I: Nth_mem (rm I1'). rewrite concat_mem in I. lets (hs&Mhs&Mh): (rm I).
       rewrite map_mem with (CA := _) in Mhs. lets (IM&MIM&EIM): (rm Mhs).
       erewrite MEOuterInterface_correct. rew_refl.
       destruct IM as [h0 hs0|h0|ke0 kes0|ke0]; substs.
        simpl in Mh. rew_refl in Mh. repeat inverts Mh as Mh.
        rewrite map_mem with (CA := _) in Mh. lets (h'&Mh'&Eh'): Mh. inverts Eh'.
        unfolds in Mh. fold mem in Mh. rew_refl in Mh. repeat inverts Mh as Mh. exists~ kes0.
        rewrite map_mem with (CA := _) in Mh. lets (h'&Mh'&Eh'): Mh. inverts Eh'.
      lets (hs&Mhs): MEOuterInterface_mem_inv (rm I). lets (n&N): mem_Nth (rm Mhs).
      destruct Co as [? Co]. forwards*: Co Nth_here Nth_next. false*.
     forwards: (rm IHMouter) (rm I1') (rm I2'); [| substs* ].
      applys~ Membrane_correct_sub Co; try rewrite~ app_nil_l; typeclass.
    forwards [I1'|(m&Em&I1')]: Nth_app_inv (rm I1);
      [forwards I1: Nth_mem I1'; rewrite map_mem with (CA := _) in I1;
       lets (?&?&A): (rm I1); inverts A|].
     forwards [I2'|(m'&Em'&I2')]: Nth_app_inv (rm I2);
       [forwards I2: Nth_mem I2'; rewrite map_mem with (CA := _) in I2;
        lets (?&?&A): (rm I2); inverts A|].
     forwards: (rm IHMouter) (rm I1') (rm I2'); [| substs* ].
     applys~ Membrane_correct_sub Co; try rewrite~ app_nil_l; typeclass.
  (** Rewriting uniqueness **)
  introv I. rewrite mem_app in I. rew_refl in I. inverts I as I.
   rewrite concat_mem in I. lets (M&MI&MM): (rm I). rewrite map_mem with (CA := _) in MI.
    lets (IM&MIM&EIM): (rm MI). destruct IM as [h1 hs1|h1|ke1 kes1|ke1]; substs.
     simpl in MM. rew_refl in MM. repeat inverts MM as MM. apply~ Remove_duplicates_spec.
     rewrite map_mem with (CA := _) in MM. lets (h2&M&E): (rm MM). inverts E.
     simpl in MM. rew_refl in MM. repeat inverts MM as MM.
     rewrite map_mem with (CA := _) in MM. lets (h2&M&E): (rm MM). inverts E.
   rewrite mem_app in I. rew_refl in I. inverts I as I;
     rewrite map_mem with (CA := _) in I; lets (h''&M&E): (rm I); inverts E.
  (** Rewriting uniqueness env **)
  introv I. rewrite mem_app in I. rew_refl in I. inverts I as I.
   rewrite concat_mem in I. lets (M&MI&MM): (rm I). rewrite map_mem with (CA := _) in MI.
    lets (IM&MIM&EIM): (rm MI). destruct IM as [h1 hs1|h1|ke1 kes1|ke1]; substs.
     simpl in MM. rew_refl in MM. repeat inverts MM as MM.
     rewrite map_mem with (CA := _) in MM. lets (h2&M&E): (rm MM). inverts E.
     simpl in MM. rew_refl in MM. repeat inverts MM as MM. apply~ Remove_duplicates_spec.
     rewrite map_mem with (CA := _) in MM. lets (h2&M&E): (rm MM). inverts E.
   rewrite mem_app in I. rew_refl in I. inverts I as I;
     rewrite map_mem with (CA := _) in I; lets (h''&M&E): (rm I); inverts E.
Qed.

Lemma MembraneCheckCompose_correct : forall Minner Mouter M,
  Membrane_correct Minner →
  Membrane_correct Mouter →
  MembraneCheckCompose Minner Mouter = Some M →
  Membrane_correct M.
Proof.
  introv C1 C2 E. forwards E': MembraneCheckCompose_MembraneCompose E.
  substs. applys* MembraneCompose_correct C1 C2.
Qed.

Lemma MembraneCompose_Mread_inv : forall Minner Mouter h hs h',
  Membrane_correct Minner →
  Membrane_correct Mouter →
  Mread (MembraneCompose Minner Mouter) h = Some hs →
  mem h' hs
  ↔ (exists hs0 h0 hs1,
      Mread Mouter h = Some hs0 ∧ mem h0 hs0 ∧ Mread Minner h0 = Some hs1 ∧ mem h' hs1)
  ∨ (exists hs0,
      Mread Mouter h = Some hs0 ∧ mem h' hs0 ∧ ~ MInOuterInterface _ _ _ h' Minner).
Proof.
  introv C1 C2 E. forwards C: MembraneCompose_correct C1 C2.
  unfolds Mread, ALoc. rewrite* <- Mread_mem in E. unfold MembraneCompose in E.
  rewrite mem_app in E. rew_refl in E. inverts E as E.
   rewrite concat_mem in E. lets (IMs&MIM&IIM): (rm E). rewrite map_mem in MIM.
    lets (IM&IMO&IME): (rm MIM). clear MIM. destruct IM as [h2 hs2|h2|ke2 kes2|ke2]; substs.
     unfolds in IIM. rew_refl in IIM. do 2 inverts IIM as IIM.
      lets (n1&N1): mem_Nth IMO. rewrite Remove_duplicates_mem. iff I.
       rewrite concat_mem in I. lets (hs&Mhs&Ihs): (rm I).
        rewrite map_mem in Mhs. lets (h&Mh&Ehs): (rm Mhs). clear Mhs.
        destruct (Mread Minner h) as [hs'|] eqn: Eread; substs; simpl in Ihs.
         left. exists hs2 h hs'. splits*. rewrite* <- Mread_mem.
         right. rew_refl in Ihs. repeat inverts Ihs as Ihs. exists hs2. splits~.
          rewrite* <- Mread_mem.
          rewrite~ <- Mread_None.
       inverts I as I.
        lets (hs0&h0&hs1&E1&I1&E2&I2): (rm I).
         rewrite <- Mread_mem in E1, E2; autos~; try typeclass.
         lets (n2&N2): mem_Nth E1. forwards~: Membrane_unique C2 N1 N2. substs.
         forwards E: Nth_func N1 N2. inverts E.
         rewrite concat_mem. exists hs1. splits~. rewrite map_mem with (CA := _).
         exists h0. splits*. rewrite~ Mread_mem in E2; try typeclass.
         unfolds Mread. rewrite~ E2.
        lets (hs0&E1&I1&NI): (rm I). rewrite~ <- Mread_mem in E1; try typeclass.
         lets (n2&N2): mem_Nth E1. forwards~: Membrane_unique C2 N1 N2. substs.
         forwards E: Nth_func N1 N2. inverts E.
         rewrite concat_mem. exists [h']. splits~.
          rewrite map_mem with (CA := _). exists h'. splits*. rewrite~ <- Mread_None in NI.
           unfolds Mread. rewrite~ NI.
          simpl. rew_refl*.
     rewrite map_mem with (CA := _) in IIM. lets (?&?&A): (rm IIM). inverts A.
     simpl in IIM. rew_refl in IIM. repeat inverts IIM as IIM.
     rewrite map_mem with (CA := _) in IIM. lets (?&?&A): (rm IIM). inverts A.
   rewrite mem_app with (CA := _) in E. rew_refl in E. inverts E as E;
     rewrite map_mem with (CA := _) in E; lets (h''&M&A): (rm E); inverts A.
Qed.

Lemma MembraneCompose_Mread_env_inv : forall Minner Mouter ke kes ke',
  Membrane_correct Minner →
  Membrane_correct Mouter →
  Mread_env (MembraneCompose Minner Mouter) ke = Some kes →
  mem ke' kes
  ↔ (exists kes0 ke0 kes1,
      Mread_env Mouter ke = Some kes0 ∧ mem ke0 kes0
      ∧ Mread_env Minner ke0 = Some kes1 ∧ mem ke' kes1)
  ∨ (exists kes0,
      Mread_env Mouter ke = Some kes0 ∧ mem ke' kes0 ∧ ~ MEInOuterInterface _ _ _ ke' Minner).
Proof.
  introv C1 C2 E. forwards C: MembraneCompose_correct C1 C2.
  unfolds Mread_env. rewrite* <- Mread_env_mem in E. unfold MembraneCompose in E.
  rewrite mem_app in E. rew_refl in E. inverts E as E.
   rewrite concat_mem in E. lets (IMs&MIM&IIM): (rm E). rewrite map_mem with (CA := _) in MIM.
    lets (IM&IMO&IME): (rm MIM). destruct IM as [h2 hs2|h2|ke2 kes2|ke2]; substs.
     simpl in IIM. rew_refl in IIM. repeat inverts IIM as IIM.
     rewrite map_mem with (CA := _) in IIM. lets (?&?&A): (rm IIM). inverts A.
     unfolds in IIM. rew_refl in IIM. do 2 inverts IIM as IIM.
      lets (n1&N1): mem_Nth IMO. rewrite Remove_duplicates_mem. iff I.
       rewrite concat_mem in I. lets (hs&Mhs&Ihs): (rm I).
        rewrite map_mem with (CA := _) in Mhs. lets (h&Mh&Ehs): (rm Mhs).
        destruct (Mread_env Minner h) as [hs'|] eqn: Eread; substs; simpl in Ihs.
         left. exists kes2 h hs'. splits*. rewrite* <- Mread_env_mem.
         right. rew_refl in Ihs. repeat inverts Ihs as Ihs. exists kes2. splits~.
          rewrite* <- Mread_env_mem.
          rewrite~ <- Mread_env_None.
       inverts I as I.
        lets (hs0&h0&hs1&E1&I1&E2&I2): (rm I).
         rewrite <- Mread_env_mem in E1, E2; autos~; try typeclass.
         lets (n2&N2): mem_Nth E1. forwards~: Membrane_unique_env C2 N1 N2. substs.
         forwards E: Nth_func N1 N2. inverts E.
         rewrite concat_mem. exists hs1. splits~. rewrite map_mem with (CA := _). exists h0. splits*.
         rewrite~ Mread_env_mem in E2; try typeclass. unfolds Mread_env. rewrite~ E2.
        lets (hs0&E1&I1&NI): (rm I). rewrite~ <- Mread_env_mem in E1; try typeclass.
         lets (n2&N2): mem_Nth E1. forwards~: Membrane_unique_env C2 N1 N2. substs.
         forwards E: Nth_func N1 N2. inverts E.
         rewrite concat_mem. exists [ke']. splits~.
          rewrite map_mem with (CA := _). exists ke'. splits*. rewrite~ <- Mread_env_None in NI.
           unfolds Mread_env. rewrite~ NI.
          simpl. rew_refl*.
     rewrite map_mem with (CA := _) in IIM. lets (?&?&A): (rm IIM). inverts A.
   rewrite mem_app with (CA := _) in E. rew_refl in E. inverts E as E;
     rewrite map_mem with (CA := _) in E; lets (h''&M&A): (rm E); inverts A.
Qed.

Lemma MembraneCompose_Mread_cancel_outer : forall Minner Mouter h hs h' hs',
  Membrane_correct Minner →
  Membrane_correct Mouter →
  Mread (MembraneCompose Minner Mouter) h = Some hs →
  Mread Mouter h = Some hs' →
  mem h' hs
  ↔ (exists h0 hs1,
      mem h0 hs' ∧ Mread Minner h0 = Some hs1 ∧ mem h' hs1)
  ∨ (mem h' hs' ∧ ~ MInOuterInterface _ _ _ h' Minner).
Proof.
  introv C1 C2 E1 E2. forwards~ MR: MembraneCompose_Mread_inv E1. rewrites (rm MR). iff I.
   repeat rewrite E2 in I. lets [(hs0&h0&hs1&Ehs0&Mhs0&MRh0&Mh1)|(hs0&Ehs0&Mh1&NI)]: (rm I).
    inverts* Ehs0.
    inverts* Ehs0.
   lets [(h0&hs1&Mhs0&MRh0&Mh1)|(Mh1&NI)]: (rm I).
    left. exists* hs' h0.
    right*.
Qed.

Lemma MembraneCompose_Mread_env_cancel_outer : forall Minner Mouter ke kes ke' kes',
  Membrane_correct Minner →
  Membrane_correct Mouter →
  Mread_env (MembraneCompose Minner Mouter) ke = Some kes →
  Mread_env Mouter ke = Some kes' →
  mem ke' kes
  ↔ (exists ke0 kes1,
      mem ke0 kes' ∧ Mread_env Minner ke0 = Some kes1 ∧ mem ke' kes1)
  ∨ (mem ke' kes' ∧ ~ MEInOuterInterface _ _ _ ke' Minner).
Proof.
  introv C1 C2 E1 E2. forwards~ MR: MembraneCompose_Mread_env_inv E1. rewrites (rm MR). iff I.
   repeat rewrite E2 in I. lets [(hs0&h0&hs1&Ehs0&Mhs0&MRh0&Mh1)|(hs0&Ehs0&Mh1&NI)]: (rm I).
    inverts* Ehs0.
    inverts* Ehs0.
   lets [(h0&hs1&Mhs0&MRh0&Mh1)|(Mh1&NI)]: (rm I).
    left. exists* kes' h0.
    right*.
Qed.

Lemma MembraneCompose_same_interface : forall Minner Mouter h,
  MInOuterInterface _ _ _ h Mouter
  ↔ MInOuterInterface _ _ _ h (MembraneCompose Minner Mouter).
Proof.
  introv. unfolds MembraneCompose. iff I; lets (hs&Ihs): (rm I).
   eexists. rewrite mem_app with (CA := _). rew_refl. left.
    rewrite concat_mem. eexists. rewrite map_mem with (CA := _). splits.
     eexists. splits*.
     simpl. rew_refl*.
   rewrite mem_app with (CA := _) in Ihs. rew_refl in Ihs. inverts Ihs as I.
    rewrite concat_mem in I. lets (IMs&MIMs&IIMs): (rm I). rewrite map_mem in MIMs.
     lets (IM&MIM&EIM): (rm MIMs). clear MIMs. substs.
     destruct IM as [h' hs'|h'|ke' kes'|ke'].
      unfolds in IIMs. rew_refl in IIMs. repeat inverts IIMs as IIMs. exists* hs'.
      rewrite map_mem with (CA := _) in IIMs. lets (ha&_&A): (rm IIMs). inverts A.
      simpl in IIMs. rew_refl in IIMs. repeat inverts IIMs as IIMs.
      rewrite map_mem with (CA := _) in IIMs. lets (ha&_&A): (rm IIMs). inverts A.
    rewrite mem_app with (CA := _) in I. rew_refl in I. inverts I as I;
      rewrite map_mem with (CA := _) in I; lets (h''&M&E): (rm I); inverts E.
Qed.

Lemma MembraneCompose_same_interface_env : forall Minner Mouter ke,
  MEInOuterInterface _ _ _ ke Mouter
  ↔ MEInOuterInterface _ _ _ ke (MembraneCompose Minner Mouter).
Proof.
  introv. unfolds MembraneCompose. iff I; lets (hs&Ihs): (rm I).
   eexists. rewrite mem_app. rew_refl. left.
    rewrite concat_mem. eexists. rewrite map_mem. splits.
     eexists. splits*.
     simpl. rew_refl*.
   rewrite mem_app in Ihs. rew_refl in Ihs. inverts Ihs as I.
    rewrite concat_mem in I. lets (IMs&MIMs&IIMs): (rm I). rewrite map_mem in MIMs.
     lets (IM&MIM&EIM): (rm MIMs). substs. destruct IM as [h' hs'|h'|ke' kes'|ke'].
      simpl in IIMs. rew_refl in IIMs. repeat inverts IIMs as IIMs.
      rewrite map_mem with (CA := _) in IIMs. lets (ha&_&A): (rm IIMs). inverts A.
      unfolds in IIMs. rew_refl in IIMs. repeat inverts IIMs as IIMs. exists~ kes'.
      rewrite map_mem with (CA := _) in IIMs. lets (ha&_&A): (rm IIMs). inverts A.
    rewrite mem_app with (CA := _) in I. rew_refl in I. inverts I as I;
      rewrite map_mem with (CA := _) in I; lets (h''&M&E): (rm I); inverts E.
Qed.

Lemma MembraneCompose_inner_interface : forall Minner Mouter h,
  Membrane_correct Minner →
  Membrane_correct Mouter →
  MInInnerInterface _ _ _ h (MembraneCompose Minner Mouter) →
    MInInnerInterface _ _ _ h Minner
    ∨ (MInInnerInterface _ _ _ h Mouter ∧ ~ MInOuterInterface _ _ _ h Minner).
Proof.
  introv C1 C2. forwards C': MembraneCompose_correct C1 C2. introv I. inverts I as I1 I2.
   rewrite~ Mread_mem in I2; try typeclass. forwards~ E: MembraneCompose_Mread_inv I2.
    unfolds ALoc, Local_Comparable_ALoc, Local_ALoc_Comparable. rewrite E in I1.
    inverts I1 as I.
     lets (hs'&h'&hs''&Ehs'&Ihs'&Ehs''&Ihs''): (rm I). unfolds Mread.
      rewrite~ <- Mread_mem in Ehs''; try typeclass.
      left. applys~ MInInnerInterface_rewriting Ehs''.
     lets (hs'&Ehs'&Ihs'&NInner): (rm I). unfolds Mread.
      rewrite~ <- Mread_mem in Ehs'; try typeclass.
      right. splits~. applys~ MInInnerInterface_rewriting Ehs'.
   unfold MembraneCompose in I1. rewrite mem_app in I1. rew_refl in I1. inverts I1 as I.
    rewrite concat_mem in I. lets (hs&Mhs&Ihs): (rm I).
     rewrite map_mem with (CA := _) in Mhs. lets (IM&MIM&EIM): (rm Mhs). substs.
     destruct IM as [h' hs'|h'|ke' kes'|ke'].
      simpl in Ihs. rew_refl in Ihs. repeat inverts Ihs as Ihs.
      rewrite map_mem in Ihs. lets (h''&Mh''&Eh''): (rm Ihs). inverts Eh''.
       destruct (Mread Minner h') eqn: EM; simpl in Mh''.
        left. unfolds Mread, ALoc. rewrite~ <- Mread_mem in EM; try typeclass.
         applys~ MInInnerInterface_rewriting EM.
        rew_refl in Mh''. repeat inverts Mh'' as Mh''. unfolds Mread, ALoc.
         rewrite Mread_None in EM. right. splits*. constructors*.
      simpl in Ihs. rew_refl in Ihs. repeat inverts Ihs as Ihs.
      rewrite map_mem with (CA := _) in Ihs. lets (h''&Mh''&Eh''): (rm Ihs). inverts Eh''.
    rewrite mem_app with (CA := _) in I. rew_refl in I. inverts I as I;
      rewrite map_mem with (CA := _) in I; lets (h''&M&E): (rm I); inverts E.
     unfolds Local_Comparable_ALoc. erewrite Mnew_correct in M. left. constructors*.
Qed.

Lemma MembraneCheckCompose_inner_interface_inner : forall Minner Mouter M h,
  Membrane_correct Minner →
  Membrane_correct Mouter →
  MembraneCheckCompose Minner Mouter = Some M →
  MInInnerInterface _ _ _ h Minner →
  MInInnerInterface _ _ _ h M.
Proof.
  introv C1 C2 E I. unfolds in E. cases_if as II; fold_bool; rew_refl in II.
  lets (Incl&D&?&?): (rm II). inverts E. inverts I as I1 I2.
   forwards I: incl_mem Incl.
     applys* MOuterInterface_mem I2.
    rewrite MInnerInterface_correct in I. rew_refl in I. inverts I as I.
     set (MID := MInInnerInterfaceDecidable).
      rewrite <- istrue_decide. rewrite <- MInnerInterface_correct.
      unfolds MInnerInterface, MembraneCompose. rewrite concat_mem. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits.
        rewrite mem_app. rew_refl. left. rewrite concat_mem. eexists. splits.
         rewrite map_mem with (CA := _). eexists. splits*.
         simpl. rew_refl*.
        reflexivity.
       rewrite Remove_duplicates_mem. rewrite concat_mem. eexists. splits.
        rewrite map_mem with (CA := _). eexists. splits*.
        rewrite~ Mread_mem in I2; try typeclass. unfolds Mread. rewrite* I2.
     apply MInInnerInterface_new. unfold MembraneCompose. rewrite mem_app. rew_refl.
      left. rewrite concat_mem. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits*.
       rewrite~ Mread_mem in I2; try typeclass. unfolds Mread. rewrite I2. rewrite* map_mem.
   apply MInInnerInterface_new. unfolds MembraneCompose. rewrite mem_app. rew_refl. right.
    rewrite <- Mnew_correct in I1. rewrite mem_app with (CA := _). rew_refl.
    left. rewrite map_mem with (CA := _). eexists. splits*.
Qed.

Lemma MembraneCheckCompose_inner_interface_inner_env : forall Minner Mouter M ke,
  Membrane_correct Minner →
  Membrane_correct Mouter →
  MembraneCheckCompose Minner Mouter = Some M →
  MEInInnerInterface _ _ _ ke Minner →
  MEInInnerInterface _ _ _ ke M.
Proof.
  introv C1 C2 E I. unfolds in E. cases_if as II; fold_bool; rew_refl in II.
  lets (?&?&Incl&D): (rm II). inverts E. inverts I as I1 I2.
   forwards I: incl_mem Incl.
     applys* MEOuterInterface_mem I2.
    rewrite MEInnerInterface_correct in I. rew_refl in I. inverts I as I.
     set (MID := MEInInnerInterfaceDecidable).
      rewrite <- istrue_decide. rewrite <- MEInnerInterface_correct.
      unfolds MEInnerInterface, MembraneCompose. rewrite concat_mem. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits.
        rewrite mem_app. rew_refl. left. rewrite concat_mem. eexists. splits.
         rewrite map_mem with (CA := _). eexists. splits*.
         simpl. rew_refl*.
        reflexivity.
       rewrite Remove_duplicates_mem. rewrite concat_mem. eexists. splits.
        rewrite map_mem with (CA := _). eexists. splits*.
        rewrite~ Mread_env_mem in I2; try typeclass. unfolds Mread_env. rewrite* I2.
     apply MEInInnerInterface_new. unfold MembraneCompose. rewrite mem_app. rew_refl.
      left. rewrite concat_mem. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits*.
       rewrite~ Mread_env_mem in I2; try typeclass. unfolds Mread_env. rewrite I2.
        rewrite* map_mem.
   apply MEInInnerInterface_new. unfolds MembraneCompose. rewrite mem_app. rew_refl. right.
    rewrite <- Mnew_env_correct in I1. rewrite mem_app. rew_refl.
    right. rewrite map_mem with (CA := _). eexists. splits*.
Qed.

Lemma MembraneCompose_inner_interface_outer : forall Minner Mouter h,
  Membrane_correct Minner →
  Membrane_correct Mouter →
  MInInnerInterface _ _ _ h Mouter →
  ~ MInOuterInterface _ _ _ h Minner →
  MInInnerInterface _ _ _ h (MembraneCompose Minner Mouter).
Proof.
  introv C1 C2 I NI. rewrite <- Mread_None in NI. inverts I as I1 I2.
   set (MID := MInInnerInterfaceDecidable).
    rewrite <- istrue_decide. rewrite <- MInnerInterface_correct.
    unfolds MInnerInterface, MembraneCompose. rewrite concat_mem. eexists. splits.
     rewrite map_mem with (CA := _). eexists. splits.
      rewrite mem_app. rew_refl. left. rewrite concat_mem. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits*.
       simpl. rew_refl*.
      reflexivity.
     rewrite Remove_duplicates_mem. rewrite concat_mem. eexists. splits.
      rewrite map_mem with (CA := _). eexists. splits*.
      unfolds Mread. rewrite NI. simpl. rew_refl*.
   apply~ MInInnerInterface_new. unfolds MembraneCompose.
    rewrite mem_app. rew_refl. left. rewrite concat_mem. eexists. splits.
     rewrite map_mem with (CA := _). eexists. splits*.
     rewrite map_mem with (CA := _). eexists. splits*. unfolds Mread. rewrite NI. simpl. rew_refl*.
Qed.

Lemma MembraneCompose_inner_interface_outer_env : forall Minner Mouter ke,
  Membrane_correct Minner →
  Membrane_correct Mouter →
  MEInInnerInterface _ _ _ ke Mouter →
  ~ MEInOuterInterface _ _ _ ke Minner →
  MEInInnerInterface _ _ _ ke (MembraneCompose Minner Mouter).
Proof.
  introv C1 C2 I NI. rewrite <- Mread_env_None in NI. inverts I as I1 I2.
   set (MID := MEInInnerInterfaceDecidable).
    rewrite <- istrue_decide. rewrite <- MEInnerInterface_correct.
    unfolds MEInnerInterface, MembraneCompose. rewrite concat_mem. eexists. splits.
     rewrite map_mem with (CA := _). eexists. splits.
      rewrite mem_app. rew_refl. left. rewrite concat_mem. eexists. splits.
       rewrite map_mem with (CA := _). eexists. splits*.
       simpl. rew_refl*.
      reflexivity.
     rewrite Remove_duplicates_mem. rewrite concat_mem. eexists. splits.
      rewrite map_mem with (CA := _). eexists. splits*.
      unfolds Mread_env. rewrite NI. simpl. rew_refl*.
   apply~ MEInInnerInterface_new. unfolds MembraneCompose.
    rewrite mem_app. rew_refl. left. rewrite concat_mem. eexists. splits.
     rewrite map_mem with (CA := _). eexists. splits*.
     rewrite map_mem with (CA := _). eexists. splits*. unfolds Mread_env. rewrite NI. simpl. rew_refl*.
Qed.

Lemma MembraneCompose_eq_incl : forall Minner Minner' Mouter Mouter',
  Membrane_correct Minner →
  Membrane_correct Mouter →
  Membrane_correct Minner' →
  Membrane_correct Mouter' →
  Membrane_eq _ _ _ Minner Minner' →
  Membrane_eq _ _ _ Mouter Mouter' →
  Membrane_incl _ _ _ (MembraneCompose Minner Mouter) (MembraneCompose Minner' Mouter').
Proof.
  introv C1 C2 C1' C2' E1 E2. constructors~.
  (** News **)
  apply~ mem_double_incl. intro h. extens. unfolds Local_Comparable_ALoc.
  repeat erewrite Mnew_correct. unfolds MembraneCompose.
  repeat rewrite mem_app. rew_refl. repeat rewrite concat_mem.
  iff [(IMs&MIMs&IMsM)|M].
   rewrite map_mem with (CA := _) in MIMs. lets (IM&MIM&EIM): (rm MIMs).
    destruct IM as [h' hs'|h'|ke' kes'|ke']; substs.
     simpl in IMsM. rew_refl in IMsM. repeat inverts IMsM as IMsM.
     left. rewrite map_mem with (CA := _) in IMsM. lets (h''&Mh&E): (rm IMsM). inverts E.
      destruct (Mread Minner' h') as [hs'|] eqn: E'; simpl in Mh.
       destruct (Mread Minner h') as [hs|] eqn: E.
        eexists. splits.
         rewrite map_mem with (CA := _). exists (ν h' : IMembrane). splits.
          rewrite <- Mnew_correct in *. rewrite (double_incl_mem _ (Membrane_eq_new E2)).
           autos~.
          rewrite~ E.
         rewrite map_mem with (CA := _). exists h''. splits~. simpl.
          forwards* (_&I): Membrane_eq_read_defined E1. rewrite <- In_mem in *. autos~.
        unfolds Mread, ALoc. rewrite* Membrane_eq_read_undefined in E.
         rewrite E' in E. inverts E.
       rew_refl in Mh. inverts Mh as Mh; tryfalse. exists [ν h' : IMembrane].
        splits.
         rewrite map_mem with (CA := _). exists (ν h' : IMembrane). splits.
          rewrite <- Mnew_correct in *. rewrite (double_incl_mem _ (Membrane_eq_new E2)).
           autos~.
          unfolds Mread, ALoc. rewrite Membrane_eq_read_undefined in E'.
           rewrite~ E'.
           applys~ Membrane_eq_sym E1.
         simpl. rew_refl*.
     simpl in IMsM. rew_refl in IMsM. repeat inverts IMsM as IMsM.
     rewrite map_mem with (CA := _) in IMsM. lets (h''&Mh&E): (rm IMsM). inverts E.
   right. repeat rewrite map_mem with (CA := _) in *. lets [(h'&Mh&E)|(h'&Mh&E)]: (rm M); inverts E.
    rewrite <- (double_incl_mem _ (Membrane_eq_new E1)) in Mh. left*.
   rewrite map_mem with (CA := _) in MIMs. lets (IM&MIM&EIM): (rm MIMs).
    destruct IM as [h' hs'|h'|ke' kes'|ke']; substs.
     simpl in IMsM. rew_refl in IMsM. repeat inverts IMsM as IMsM.
     left. rewrite map_mem with (CA := _) in IMsM. lets (h''&Mh&E): (rm IMsM). inverts E.
      destruct (Mread Minner h') as [hs|] eqn: E; simpl in Mh.
       destruct (Mread Minner' h') as [hs'|] eqn: E'.
        eexists. splits.
         rewrite map_mem with (CA := _). exists (ν h' : IMembrane). splits.
          rewrite <- Mnew_correct in *. rewrite <- (double_incl_mem _ (Membrane_eq_new E2)).
           autos~.
          rewrite~ E'.
         rewrite map_mem with (CA := _). exists h''. splits~. simpl.
          forwards* (I&_): Membrane_eq_read_defined E1. rewrite <- In_mem in *. autos~.
        unfolds Mread, ALoc. rewrite Membrane_eq_read_undefined in E'.
         rewrite E' in E. inverts E.
         applys~ Membrane_eq_sym E1.
       rew_refl in Mh. inverts Mh as Mh; tryfalse. exists [ν h' : IMembrane].
        splits.
         rewrite map_mem with (CA := _). exists (ν h' : IMembrane). splits.
          rewrite <- Mnew_correct in *. rewrite <- (double_incl_mem _ (Membrane_eq_new E2)).
           autos~.
          unfolds Mread, ALoc. rewrite* Membrane_eq_read_undefined in E. rewrite~ E.
         simpl. rew_refl*.
     simpl in IMsM. rew_refl in IMsM. repeat inverts IMsM as IMsM.
     rewrite map_mem with (CA := _) in IMsM. lets (h''&Mh&E): (rm IMsM). inverts E.
   right. repeat rewrite map_mem with (CA := _) in *. lets [(h'&Mh&E)|(h'&Mh&E)]: (rm M); inverts E.
    rewrite (double_incl_mem _ (Membrane_eq_new E1)) in Mh. left*.
  (** News Env **)
  apply~ mem_double_incl. intro ke. extens. repeat erewrite Mnew_env_correct.
  unfolds MembraneCompose. repeat rewrite mem_app. rew_refl. repeat rewrite concat_mem.
  iff [(IMs&MIMs&IMsM)|M].
   rewrite map_mem with (CA := _) in MIMs. lets (IM&MIM&EIM): (rm MIMs).
    destruct IM as [h' hs'|h'|ke' kes'|ke']; substs.
     simpl in IMsM. rew_refl in IMsM. repeat inverts IMsM as IMsM.
     rewrite map_mem with (CA := _) in IMsM. lets (h''&Mh&E): (rm IMsM). inverts E.
     simpl in IMsM. rew_refl in IMsM. repeat inverts IMsM as IMsM.
     left. rewrite map_mem with (CA := _) in IMsM. lets (h''&Mh&E): (rm IMsM). inverts E.
      destruct (Mread_env Minner' ke') as [kes'|] eqn: E'; simpl in Mh.
       destruct (Mread_env Minner ke') as [kes|] eqn: E.
        eexists. splits.
         rewrite map_mem with (CA := _). exists (ν~ ke' : IMembrane). splits.
          rewrite <- Mnew_env_correct in *.
           rewrite (double_incl_mem _ (Membrane_eq_new_env E2)). autos~.
          rewrite~ E.
         rewrite map_mem with (CA := _). exists h''. splits~. simpl.
          forwards* (_&I): Membrane_eq_read_env_defined E1. rewrite <- In_mem in *. autos~.
        unfolds Mread_env. rewrite* Membrane_eq_read_env_undefined in E. rewrite E' in E. inverts E.
       rew_refl in Mh. inverts Mh as Mh; tryfalse. exists [ν~ ke' : IMembrane].
        splits.
         rewrite map_mem with (CA := _). exists (ν~ ke' : IMembrane). splits.
          rewrite <- Mnew_env_correct in *.
           rewrite (double_incl_mem _ (Membrane_eq_new_env E2)). autos~.
          unfolds Mread_env. rewrite Membrane_eq_read_env_undefined in E'.
           rewrite~ E'.
           applys~ Membrane_eq_sym E1.
         simpl. rew_refl*.
   right. repeat rewrite map_mem with (CA := _) in *. lets [(h'&Mh&E)|(h'&Mh&E)]: (rm M); inverts E.
    rewrite <- (double_incl_mem _ (Membrane_eq_new_env E1)) in Mh. right*.
   rewrite map_mem with (CA := _) in MIMs. lets (IM&MIM&EIM): (rm MIMs).
    destruct IM as [h' hs'|h'|ke' kes'|ke']; substs.
     simpl in IMsM. rew_refl in IMsM. repeat inverts IMsM as IMsM.
     rewrite map_mem with (CA := _) in IMsM. lets (h''&Mh&E): (rm IMsM). inverts E.
     simpl in IMsM. rew_refl in IMsM. repeat inverts IMsM as IMsM.
     left. rewrite map_mem with (CA := _) in IMsM. lets (h''&Mh&E): (rm IMsM). inverts E.
      destruct (Mread_env Minner ke') as [hs|] eqn: E; simpl in Mh.
       destruct (Mread_env Minner' ke') as [hs'|] eqn: E'.
        eexists. splits.
         rewrite map_mem with (CA := _). exists (ν~ ke' : IMembrane). splits.
          rewrite <- Mnew_env_correct in *.
           rewrite <- (double_incl_mem _ (Membrane_eq_new_env E2)). autos~.
          rewrite~ E'.
         rewrite map_mem with (CA := _). exists h''. splits~. simpl.
          forwards* (I&_): Membrane_eq_read_env_defined E1. rewrite <- In_mem in *. autos~.
        unfolds Mread_env. rewrite Membrane_eq_read_env_undefined in E'.
         rewrite E' in E. inverts E.
         applys~ Membrane_eq_sym E1.
       rew_refl in Mh. inverts Mh as Mh; tryfalse. exists [ν~ ke' : IMembrane].
        splits.
         rewrite map_mem with (CA := _). exists (ν~ ke' : IMembrane). splits.
          rewrite <- Mnew_env_correct in *.
           rewrite <- (double_incl_mem _ (Membrane_eq_new_env E2)). autos~.
          unfolds Mread_env. rewrite* Membrane_eq_read_env_undefined in E. rewrite~ E.
         simpl. rew_refl*.
   right. repeat rewrite map_mem with (CA := _) in *. lets [(h'&Mh&E)|(h'&Mh&E)]: (rm M); inverts E.
    rewrite (double_incl_mem _ (Membrane_eq_new_env E1)) in Mh. right*.
  (** Read Undefined **)
  introv. repeat rewrite Mread_None with (KE_comparable := _). apply contrapose_intro. introv I.
  forwards (hs&E): Mread_outer_interface (rm I). rewrite~ <- Mread_mem in E; try typeclass.
   unfolds MembraneCompose. rewrite mem_app in E. rew_refl in E. inverts E as I.
    rewrite concat_mem in I. lets (IMs&MIMs&IIMs): (rm I). rewrite map_mem with (CA := _) in MIMs.
     lets (IM&MIM&EIM): (rm MIMs). substs. destruct IM as [h' hs'|h'|ke' kes'|ke'].
      simpl in IIMs. rew_refl in IIMs. repeat inverts IIMs as IIMs.
       destruct (Mread Mouter' h') as [hs|] eqn: E.
        eapply Mread_defined_inv. rewrite <- Mread_mem; try typeclass.
         rewrite mem_app. rew_refl. left. rewrite concat_mem. eexists. splits.
          rewrite map_mem with (CA := _). unfolds Mread, ALoc. rewrite~ <- Mread_mem in E; try typeclass.
          simpl. rew_refl. left~.
         apply~ MembraneCompose_correct.
        unfolds Mread, ALoc. rewrite* <- Membrane_eq_read_undefined in E.
         rewrite Mread_None in E. unfold MInOuterInterface in E. rew_logic in E. false* E.
      rewrite map_mem with (CA := _) in IIMs. lets (?&_&A): (rm IIMs). inverts A.
      simpl in IIMs. rew_refl in IIMs. repeat inverts IIMs as IIMs.
      rewrite map_mem with (CA := _) in IIMs. lets (?&_&A): (rm IIMs). inverts A.
    rewrite mem_app in I. rew_refl in I.
     inverts I as I; rewrite map_mem with (CA := _) in I; lets (h'&_&A): (rm I); inverts A.
   apply~ MembraneCompose_correct.
  (** Read Undefined Env **)
  introv. repeat rewrite Mread_env_None with (L_comparable := _) (K_comparable := _).
  apply contrapose_intro. introv I.
  forwards (hs&E): Mread_env_outer_interface (rm I). rewrite~ <- Mread_env_mem in E; try typeclass.
   unfolds MembraneCompose. rewrite mem_app in E. rew_refl in E. inverts E as I.
    rewrite concat_mem in I. lets (IMs&MIMs&IIMs): (rm I). rewrite map_mem with (CA := _) in MIMs.
     lets (IM&MIM&EIM): (rm MIMs). substs. destruct IM as [h' hs'|h'|ke' kes'|ke'].
      simpl in IIMs. rew_refl in IIMs. repeat inverts IIMs as IIMs.
      rewrite map_mem with (CA := _) in IIMs. lets (?&_&A): (rm IIMs). inverts A.
      simpl in IIMs. rew_refl in IIMs. repeat inverts IIMs as IIMs.
       destruct (Mread_env Mouter' ke') as [hs|] eqn: E.
        eapply Mread_env_defined_inv. rewrite <- Mread_env_mem; try typeclass.
         rewrite mem_app. rew_refl. left. rewrite concat_mem. eexists. splits.
          rewrite map_mem with (CA := _). unfolds Mread_env. rewrite~ <- Mread_env_mem in E; try typeclass.
          simpl. rew_refl. left~.
         apply~ MembraneCompose_correct.
        unfolds Mread_env. rewrite* <- Membrane_eq_read_env_undefined in E.
         rewrite Mread_env_None in E. unfold MEInOuterInterface in E. rew_logic in E. false* E.
      rewrite map_mem with (CA := _) in IIMs. lets (?&_&A): (rm IIMs). inverts A.
    rewrite mem_app in I. rew_refl in I.
     inverts I as I; rewrite map_mem with (CA := _) in I; lets (h'&_&A): (rm I); inverts A.
   apply~ MembraneCompose_correct.
  (** Read Defined **)
  introv I1 I2. apply mem_incl with (H := _). introv I.
  forwards~ R: MembraneCompose_Mread_inv I2. unfolds ALoc. rewrites (rm R).
  forwards~ R: MembraneCompose_Mread_inv I1. unfolds ALoc. rewrites (rm R) in I.
  lets [(hs3&h3&hs4&E3&I3&E4&I4)|(hs3&E3&I3&NI)]: (rm I).
   left. destruct (Mread Mouter' h) as [hs5|] eqn: E5.
    destruct (Mread Minner' h3) as [hs6|] eqn: E6.
     exists hs5 h3 hs6. splits~.
      unfolds Mread. forwards~ D: Membrane_eq_read_defined E3 E5.
       rewrite <- (double_incl_mem _ D); autos~.
      unfolds Mread. forwards~ D: Membrane_eq_read_defined E4 E6.
       rewrite <- (double_incl_mem _ D); autos~.
     unfolds Mread. rewrite* <- Membrane_eq_read_undefined in E6. rewrite E6 in E4. inverts E4.
    unfolds Mread. rewrite* <- Membrane_eq_read_undefined in E5. rewrite E5 in E3. inverts E3.
   right. destruct (Mread Mouter' h) as [hs4|] eqn: E4.
    exists hs4. splits~.
     unfolds Mread. forwards~ D: Membrane_eq_read_defined E3 E4.
      rewrite <- (double_incl_mem _ D); autos~.
     applys contrapose_intro NI. applys~ Membrane_eq_outer_interface.
      apply~ Membrane_eq_sym.
    unfolds Mread. rewrite* <- Membrane_eq_read_undefined in E4. rewrite E4 in E3. inverts E3.
  (** Read Defined Env **)
  introv I1 I2. apply mem_incl with (H := _). introv I.
  rewrite MembraneCompose_Mread_env_inv; try apply I2; autos~.
  rewrite MembraneCompose_Mread_env_inv in I; try apply I1; autos~.
  lets [(hs3&h3&hs4&E3&I3&E4&I4)|(hs3&E3&I3&NI)]: (rm I).
   left. destruct (Mread_env Mouter' ke) as [hs5|] eqn: E5.
    destruct (Mread_env Minner' h3) as [hs6|] eqn: E6.
     exists hs5 h3 hs6. splits~.
      unfolds Mread_env. forwards~ D: Membrane_eq_read_env_defined E3 E5.
       rewrite <- (double_incl_mem _ D); autos~.
      unfolds Mread_env. forwards~ D: Membrane_eq_read_env_defined E4 E6.
       rewrite <- (double_incl_mem _ D); autos~.
     unfolds Mread_env. rewrite* <- Membrane_eq_read_env_undefined in E6. rewrite E6 in E4. inverts E4.
    unfolds Mread_env. rewrite* <- Membrane_eq_read_env_undefined in E5. rewrite E5 in E3. inverts E3.
   right. destruct (Mread_env Mouter' ke) as [hs4|] eqn: E4.
    exists hs4. splits~.
     unfolds Mread_env. forwards~ D: Membrane_eq_read_env_defined E3 E4.
      rewrite <- (double_incl_mem _ D); autos~.
     applys contrapose_intro NI. applys~ Membrane_eq_outer_interface_env.
      apply~ Membrane_eq_sym.
    unfolds Mread_env. rewrite* <- Membrane_eq_read_env_undefined in E4. rewrite E4 in E3. inverts E3.
Qed.

Lemma MembraneCompose_eq : forall Minner Minner' Mouter Mouter',
  Membrane_correct Minner →
  Membrane_correct Mouter →
  Membrane_correct Minner' →
  Membrane_correct Mouter' →
  Membrane_eq _ _ _ Minner Minner' →
  Membrane_eq _ _ _ Mouter Mouter' →
  Membrane_eq _ _ _ (MembraneCompose Minner Mouter) (MembraneCompose Minner' Mouter').
Proof.
  introv C1 C2 C1' C2' E1 E2. apply~ Membrane_incl_Membrane_eq;
    apply~ MembraneCompose_eq_incl; apply~ Membrane_eq_sym.
Qed.


(** *** Passing Through Membranes **)

(** A datatype used in the auxiliary function of
  [passThroughMembrane_aux]. It can be read as follows: when passing a
  membrane of the form [l ---> k, l' ---> k], then the subformulae [l
  --> o] and [l' --> o'] has to be merged into a single [k --> o'']
  (and not [k --> o * k --> o'], which would mean something else). It
  is supposed for simplicifation purposes reasons that each location
  [l] only appears once.  This is not a problem as each formula
  passing through a membrane are generated by oracles: it is the
  oracle’s job to split them nicely.  We provide a function
  [mergeSameObjects], on which is not proven anything, which perform
  this operation beforehand. The structure accumulates the current
  object, as well as the list of already seen locations [l].  For
  practicability, the dependencies [d] of the first location is
  conserved. **)
Definition passThroughMembrane_acc :=
  (ALoc * (list ALoc (** Already encountered locations **)
           * list ALoc (** Dependencies **)
           * DObject (** Resulting object **)))%type.

Local Instance passThroughMembrane_acc_comparable : Comparable passThroughMembrane_acc.
  repeat apply prod_comparable; typeclass.
Defined.

Local Instance passThroughMembrane_acc_snd_Inhab : Inhab (list ALoc * list ALoc * DObject).
  (* Coq is much slower without this hint. *)
  repeat apply @prod_inhab; typeclass.
Defined.

Local Instance IFormula_passThroughMembrane_acc_Inhab :
    Inhab (IFormula * list passThroughMembrane_acc).
  (* Coq is much slower without this hint. *)
   apply @prod_inhab; typeclass.
Defined.

(** Auxiliary function for [passThroughMembrane_aux]. **)
Definition passThroughMembrane_fold_star_aux (Dph : passThroughMembrane_acc) :=
  let '(h, (hs, d, o)) := Dph in
  (** Invariant: [d = MgetDependencies M h]. **)
  (** Invariant: [incl hs d]. **)
  LibOption.apply (fun phiDp : IFormula * list passThroughMembrane_acc =>
    let (phi, Dp) := phiDp in
    if mem_assoc h Dp then
      let '(hs', d', o') :=
        assoc (h : ALoc) (Dp : list passThroughMembrane_acc) in
      (** Invariant: [d = d' = MgetDependencies M h]. **)
      (** Invariant: [incl (map fst hso') d']. **)
      let Dp' := filter (fun Dph => decide (fst Dph <> h)) Dp in
      (** We now have new information about [h]: let’s merge what we
        know about [h]. **)
      let hs'' :=
        Remove_duplicates (hs ++ hs') in
      match DObject_merge _ _ _ _ _ o o' with
      | Some o'' =>
        (** The merge was successful: these two objects had the
          same domain. **)
        ifb incl d hs'' then
          (** We now have all the locations of the depency of [h],
            which we can now migrate to the current formula. **)
            Some (phi * h --> o'', Dp')
        else
          (** There still lacks some informations about [h]. **)
          Some (phi, (h, (hs'', d, o'')) :: Dp')
      | None =>
        (** The domains of all these objects are not identical:
          either the given formula is wrong, either it contains more
          than one subformula with a fixed location [h0] in the
          dependency of [h], which we forbig here (see note above
          about the simplification made). **)
        None
      end
    else
      (** No new information about [h]. **)
      Some (phi, Dph :: Dp)).

Definition passThroughMembrane_fold_star :=
  fold_left passThroughMembrane_fold_star_aux.

(** A type similar to [passThroughMembrane_acc], but for environments. **)
Definition passThroughMembrane_acc_env :=
  (KE * (list KE * list KE * AEnv))%type.

Local Instance passThroughMembrane_acc_env_snd_comparable :
    Comparable (list KE * list KE * AEnv).
  repeat apply prod_comparable; typeclass.
Defined.

Local Instance passThroughMembrane_acc_env_comparable : Comparable passThroughMembrane_acc_env.
  apply prod_comparable; typeclass.
Defined.

Local Instance passThroughMembrane_acc_env_snd_Inhab : Inhab (list KE * list KE * AEnv).
  (* Coq is much slower without this hint. *)
  repeat apply @prod_inhab; typeclass.
Defined.

Local Instance IFormula_passThroughMembrane_acc_env_Inhab :
    Inhab (IFormula * list passThroughMembrane_acc_env).
  (* Coq is much slower without this hint. *)
   apply @prod_inhab; typeclass.
Defined.

(** Auxiliary function for [passThroughMembrane_aux] **)
Definition passThroughMembrane_env_fold_star :=
  fold_left (fun DphE : passThroughMembrane_acc_env =>
    let '(ke, (kes, d, E)) := DphE in
    (** Invariant: [d = MgetDependencies_env M ke]. **)
    (** Invariant: [incl kes d]. **)
    LibOption.apply (fun phiDpE : IFormula * list passThroughMembrane_acc_env =>
      let (phi, DpE) := phiDpE in
      if mem_assoc ke DpE then
        let '(kes', d', E') := assoc ke DpE in
        (** Invariant: [ke = ke' = MgetDependencies_env M ke]. **)
        (** Invariant: [incl kes' d']. **)
        let DpE' := filter (fun DphE => decide (fst DphE <> ke)) DpE in
        (** We now have a new information about [h]: let’s merge them. **)
        LibOption.map_on (AEnv_join _ _ _ _ E E') (fun E'' =>
          let kes'' := Remove_duplicates (kes ++ kes') in
          (** There might be duplicates in [kes ++ kes'].  If this is
          the case, it means that some of the replaced abstract
          locations [ke] were incoherent (on the form [ke ~~> E * ke
          ~~> E']).  But having incoherent abstract location does not
          mean that the formula is incoherent, only that the valuation
          of these abstract locations have to be empty in every
          concretisation.  However, for well chosen membranes (or
          should I say ill-chosen?), passing such a formula through a
          membrane makes this incoherence disappears.  This is still
          correct: from something absurd, we are allowed to infer
          anything. **)
          (** Invariant: [incl kes'' d]. **)
          ifb incl d kes'' then
            (** We now have the full information about [ke]. **)
            (phi * ke ~~> E'', DpE')
          else
            (** There still lacks some informations about [ke]. **)
            (phi, (ke, (kes'', d, E'')) :: DpE'))
      else
        (** No new information about [ke]. **)
        Some (phi, DphE :: DpE))).

(** Auxiliary function for [passThroughMembrane_aux] **)
Definition passThroughMembrane_fold_object (M : Membrane) (h : ALoc) o' :=
  fold_left (fun (h' : ALoc) (phiDp : IFormula * list passThroughMembrane_acc) =>
    let '(phi, Dp) := phiDp in
    match MgetDependencies _ _ M h' with
    | [] =>
      (** Never happens **)
      arbitrary
    | [_] =>
      (** The rewriting is on the form [h ---> h']: [h'] is
       alone at right hand side and we can rewrite it
       directly. **)
      (h' --> o' * phi, Dp)
    | d =>
      (** This list is not empty. **)
      (phi, (h', ([h], d, o')) :: Dp)
    end).

(** Auxiliary function for [passThroughMembrane_aux] **)
Definition passThroughMembrane_fold_env (M : Membrane) (ke : KE) E' :=
  fold_left (fun ke' (phiDpe : IFormula * list passThroughMembrane_acc_env) =>
    let (phi, Dpe) := phiDpe in
    match MgetDependencies_env _ M ke' with
    | [] =>
      (** Never happens **)
      arbitrary
    | [_] =>
      (** The rewriting is on the form [ke ~~~> ke']: [ke'] is
       alone at right hand side and we can rewrite it
       directly. **)
      (ke' ~~> E' * phi, Dpe)
    | d =>
      (** This list is not empty. **)
      (phi, (ke', ([ke], d, E')) :: Dpe)
    end).

(** Applying a renaming membrane to an inner formula.  This function
  is supposed to also merge the different memory cells represented by
  the same abstract location.  For instance, if the membrane is of the
  form [l ---> k, l' ---> k], then the transformation of [l --> o * l'
  --> o'] should be of the form [k --> o''] and not [k --> o * k -->
  o'].  It thus also returns a list of memory cells/objects waiting
  for the whole information to be added to the formula: the first
  abstract location is the abstract location [k] needing additionnal
  information, the second is the list of (outside) abstract locations
  whose information we already have, the third is the list which we
  need to complete to build the final object, and the object is the
  current built object. **)
Fixpoint passThroughMembrane_aux (phi : IFormula) (M : Membrane)
    : option (IFormula
        * list passThroughMembrane_acc
        * list passThroughMembrane_acc_env) :=
  match phi with
  | emp => Some (emp, nil, nil)
  | phi1 * phi2 =>
    match passThroughMembrane_aux phi1 M, passThroughMembrane_aux phi2 M with
    | Some (phi1', Dp1, Dpe1), Some (phi2', Dp2, Dpe2) =>
      match passThroughMembrane_fold_star (Some (phi1' * phi2', Dp1)) Dp2 with
      | Some (phi, Dp) =>
        match passThroughMembrane_env_fold_star (Some (phi, Dpe1)) Dpe2 with
        | Some (phi, Dpe) =>
          Some (phi, Dp, Dpe)
        | None => None
        end
      | None => None
      end
    | _, _ => None
    end
  | h --> o =>
    LibOption.map_on (applyMembraneDObject M o) (fun o' =>
      match Mread M h with
      | None =>
        (** No special replacement: the membrane does not change the identifier [h]. **)
        (h --> o', nil, nil)
      | Some hs =>
        (** The identifier [h] has been split into [hs], thus splitting the memory. **)
        let (phi, Dp) :=
          passThroughMembrane_fold_object M h o' (emp, nil) hs
        in (phi, Dp, nil)
      end)
  | ke ~~> E =>
    LibOption.map_on (applyMembraneAEnv M E) (fun E' =>
      match Mread_env M ke with
      | None =>
        (** No special replacement: the membrane does not change the identifier [ke]. **)
        (ke ~~> E', nil, nil)
      | Some kes =>
        (** The identifier [ke] has been split into [kes], thus splitting the memory. **)
        let (phi, Dpe) :=
          passThroughMembrane_fold_env M ke E' (emp, nil) kes
        in (phi, nil, Dpe)
      end)
  end.


Lemma passThroughMembrane_fold_star_none : forall Dp,
  passThroughMembrane_fold_star None Dp = None.
Proof. introv. unfolds. induction* Dp. simpl. destruct a as (h'&((hs'&d')&o')). autos*. Qed.


(** An invariant conserved by each instance of
  [passThroughMembrane_acc] in these computations. **)
Record passThroughMembrane_acc_invariant (Dp : list passThroughMembrane_acc) := {
    passThroughMembrane_acc_unique : forall h h' a a',
      mem (h, a) Dp →
      mem (h', a') Dp →
      h = h' →
      a = a'
  }.

Lemma passThroughMembrane_fold_star_aux_invariant : forall phi phi' Dpa Dph Dpr,
  passThroughMembrane_fold_star_aux Dph (Some (phi, Dpa)) = Some (phi', Dpr) →
  passThroughMembrane_acc_invariant Dpa →
  passThroughMembrane_acc_invariant Dpr.
Proof.
  introv E I. constructors.
  introv I1 I2 ?. subst h'. destruct Dph as [h' [[hs' d'] o']]. simpl in E. cases_if as MA.
   destruct (assoc h' Dpa) as ((hs&d)&o) eqn: EA. destruct DObject_merge eqn: EO.
    cases_if as DI.
     inverts E. rewrite filter_mem_eq in I1, I2. rew_refl in I1. rew_refl in I2.
      applys~ passThroughMembrane_acc_unique I1 I2.
     inverts E. simpl in I1, I2. rewrite filter_mem_eq in I1, I2. rew_refl in *.
      inverts I1 as I1; inverts I2 as I2.
       inverts I1. inverts~ I2.
       inverts I1. false*.
       inverts I2. false*.
       applys~ passThroughMembrane_acc_unique I1 I2.
    false*.
   fold_bool. rew_refl in MA. inverts E. simpl in I1. rew_refl in I1. inverts I1 as I1.
    inverts I1. simpl in I2. rew_refl in I2. inverts I2 as I2.
     inverts~ I2.
     false MA. applys~ mem_mem_assoc I2.
    simpl in I2. rew_refl in I2. inverts I2 as I2.
     inverts I2. false MA. applys~ mem_mem_assoc I1.
     applys~ passThroughMembrane_acc_unique I1 I2.
Qed.

Lemma passThroughMembrane_fold_star_invariant : forall phi phi' Dpa Dpb Dpr,
  passThroughMembrane_fold_star (Some (phi, Dpa)) Dpb = Some (phi', Dpr) →
  passThroughMembrane_acc_invariant Dpa →
  passThroughMembrane_acc_invariant Dpr.
Proof.
  introv E I. unfolds in E. rewrite fold_left_eq_fold_right in E.
  gen phi' Dpr. induction (rev Dpb); introv E.
   inverts~ E.
   simpl in E. destruct fold_right as [[? ?]|].
    applys* passThroughMembrane_fold_star_aux_invariant E.
    destruct a as [? [[? ?] ?]]. inverts E.
Qed.

(** A global property which is preserved: the [d] does not change
  (they are conserved just to avoid recomputing them at each step). **)
Record passThroughMembrane_acc_corresponds
  (Dph : passThroughMembrane_acc) (Dp : list passThroughMembrane_acc) := {
    passThroughMembrane_acc_corresponds_dependencies : forall h hs hs' d d' o o',
      Dph = (h, (hs, d, o)) →
      mem (h, (hs', d', o')) Dp →
      d = d'
  }.

Lemma passThroughMembrane_fold_star_aux_corresponds : forall phi phi' Dpa Dph Dpr Dph',
  passThroughMembrane_fold_star_aux Dph (Some (phi, Dpa)) = Some (phi', Dpr) →
  passThroughMembrane_acc_corresponds Dph Dpa →
  passThroughMembrane_acc_corresponds Dph' Dpa →
  passThroughMembrane_acc_corresponds Dph' (Dph :: nil) →
  passThroughMembrane_acc_corresponds Dph' Dpr.
Proof.
  introv E C1 C2 CT. constructors.
  introv E' I. subst Dph'. destruct Dph as [h' [[hs2 d2] o2]]. simpl in E. cases_if as MA.
   destruct (assoc h' Dpa) as ((hs3&d3)&o3) eqn: EA. destruct DObject_merge eqn: EO.
    cases_if as DI.
     inverts E. rewrite filter_mem_eq in I. rew_refl in I.
      applys~ passThroughMembrane_acc_corresponds_dependencies C2 I.
     inverts E. simpl in I. rewrite filter_mem_eq in I. rew_refl in I. inverts I as I.
       inverts I. lets I: mem_assoc_assoc EA MA. transitivity d3.
        applys~ passThroughMembrane_acc_corresponds_dependencies C2 I.
        symmetry. applys~ passThroughMembrane_acc_corresponds_dependencies C1 I.
       applys~ passThroughMembrane_acc_corresponds_dependencies C2 I.
    false*.
   fold_bool. rew_refl in MA. inverts E. simpl in I. rew_refl in I. inverts I as I.
    inverts I. applys~ passThroughMembrane_acc_corresponds_dependencies CT.
     simpl. rew_refl*.
    applys~ passThroughMembrane_acc_corresponds_dependencies C2 I.
Qed.

Lemma passThroughMembrane_fold_star_corresponds : forall phi phi' Dpa Dpb Dpr,
  passThroughMembrane_fold_star (Some (phi, Dpa)) Dpb = Some (phi', Dpr) →
  passThroughMembrane_acc_invariant Dpa →
  passThroughMembrane_acc_invariant Dpb →
  Forall (fun Dph => passThroughMembrane_acc_corresponds Dph Dpa) Dpb →
  Forall (fun Dph => passThroughMembrane_acc_corresponds Dph Dpr) Dpb.
Proof.
  introv E C1 C2. do 2 rewrite Forall_iff_forall_mem with (CA := _).
  gen Dpa phi phi' Dpr. induction Dpb; introv C1 E F I.
   false*.
   unfolds in E. simpl in E. simpl in I. rew_refl in I. inverts I as I.
    forwards Ca: F.
      simpl. rew_refl*.
     destruct passThroughMembrane_fold_star_aux as [[? ?]|] eqn: EO.
      forwards~ Ca': passThroughMembrane_fold_star_aux_corresponds EO.
       apply~ Ca.
       constructors. introv ? I. simpl in I. rew_refl in I. substs. repeat inverts I as I.
        autos~.
       forwards~ Cl: passThroughMembrane_fold_star_aux_invariant EO.
       asserts Inv: (forall x, mem x Dpb → passThroughMembrane_acc_corresponds x l).
        introv I. applys~ passThroughMembrane_fold_star_aux_corresponds EO.
         apply F. simpl. rew_refl*.
         constructors. introv E' I'. simpl in I'. rew_refl in I'.
          repeat inverts I' as I'. forwards~ Eq: passThroughMembrane_acc_unique C2.
           simpl. rew_refl. left*.
           simpl. rew_refl. right. substs*.
          inverts~ Eq.
       clear Ca EO C1 IHDpb. gen i l phi' Dpr. induction Dpb; introv Cl Il Inv E.
        inverts~ E.
        simpl in E. destruct passThroughMembrane_fold_star_aux as [[? ?]|] eqn: EO.
         applys~ IHDpb E.
          constructors. introv I1 I2 Eq. simpl in I1, I2. rew_refl in I1.
           rew_refl in I2. applys~ passThroughMembrane_acc_unique C2; simpl; rew_refl.
            inverts* I1.
            substs. inverts* I2.
          introv I. apply F. simpl in I. rew_refl in I. simpl. rew_refl*.
          tests D: (fst a = fst a0).
           forwards Ea: passThroughMembrane_acc_unique C2 (snd a) (snd a0) D.
             simpl. rew_refl. destruct* a.
             simpl. rew_refl. destruct* a0.
            asserts: (a = a0).
             destruct a. destruct a0. simpls. substs*.
            substs. applys~ passThroughMembrane_fold_star_aux_corresponds EO.
            clear. constructors. introv E I. simpl in I. rew_refl in I. substs.
            repeat inverts I as I. autos~.
           applys~ passThroughMembrane_fold_star_aux_corresponds EO.
            apply~ Inv. simpl. rew_refl*.
            constructors. introv Eq I. simpl in I. rew_refl in I.
             repeat inverts I as I. subst a. false~ D.
          applys~ passThroughMembrane_fold_star_aux_invariant EO.
          introv I. applys~ passThroughMembrane_fold_star_aux_corresponds EO.
           apply~ Inv. simpl. rew_refl*.
           apply~ Inv. simpl. rew_refl*.
           constructors. introv E' I'. subst x. simpl in I'. rew_refl in I'.
            repeat inverts I' as I'. forwards~ E': passThroughMembrane_acc_unique C2.
             simpl. rew_refl*.
             simpl. rew_refl. right. right*.
             inverts~ E'.
         rewrite passThroughMembrane_fold_star_none in E. inverts E.
      rewrite passThroughMembrane_fold_star_none in E. inverts E.
    destruct passThroughMembrane_fold_star_aux as [[? ?]|] eqn: E0.
     applys~ IHDpb E.
      constructors. introv I1 I2 Eq. applys~ passThroughMembrane_acc_unique C2 Eq;
        simpl; rew_refl*.
      applys~ passThroughMembrane_fold_star_aux_invariant E0.
      introv I'. applys~ passThroughMembrane_fold_star_aux_corresponds E0.
       apply~ F. simpl. rew_refl*.
       apply~ F. simpl. rew_refl*.
       constructors. introv E' I0. subst x0. simpl in I0. rew_refl in I0.
        repeat inverts I0 as I0. forwards~ Ea: passThroughMembrane_acc_unique C2.
         simpl. rew_refl*.
         simpl. rew_refl. right*.
        inverts~ Ea.
     rewrite passThroughMembrane_fold_star_none in E. inverts E.
Qed.


Lemma passThroughMembrane_fold_star_sub : forall phi phi' Dpa Dpb Dpr,
  passThroughMembrane_fold_star (Some (phi, Dpa)) Dpb = Some (phi', Dpr) →
  IFormula_sub phi' phi.
Proof.
  introv Eq. gen phi phi' Dpa. induction Dpb; introv Eq.
   inverts Eq. apply rtclosure_refl.
   destruct a as (h'&((hs'&d')&o')). unfolds in Eq. simpl in Eq. cases_if as I.
    destruct (assoc h' Dpa) as ((hs&d)&o) eqn: EA.
     destruct DObject_merge eqn: EO.
      cases_if as DI.
       forwards S': IHDpb (rm Eq). applys~ rtclosure_trans S'.
        apply rtclosure_once. constructors~.
      applys IHDpb Eq.
     rewrite passThroughMembrane_fold_star_none in Eq. inverts Eq.
    applys IHDpb Eq.
Qed.

Lemma passThroughMembrane_env_fold_star_none : forall Dp,
  passThroughMembrane_env_fold_star None Dp = None.
Proof. introv. unfolds. induction* Dp. simpl. destruct a as (h'&((hs'&d')&o')). autos*. Qed.

Lemma passThroughMembrane_env_fold_star_sub : forall phi phi' Dpa Dpb Dpr,
  passThroughMembrane_env_fold_star (Some (phi, Dpa)) Dpb = Some (phi', Dpr) →
  IFormula_sub phi' phi.
Proof.
  introv Eq. gen phi phi' Dpa. induction Dpb; introv Eq.
   inverts Eq. apply rtclosure_refl.
   destruct a as (h'&((hs'&d')&o')). unfolds in Eq. simpl in Eq. cases_if.
    destruct (assoc h' Dpa) as ((hs&d)&o) eqn: EA. destruct AEnv_join.
     cases_if as DI.
      forwards S': IHDpb (rm Eq). applys~ rtclosure_trans S'.
       apply rtclosure_once. constructors~.
      applys IHDpb Eq.
     rewrite passThroughMembrane_env_fold_star_none in Eq. inverts Eq.
    applys IHDpb Eq.
Qed.

Lemma passThroughMembrane_fold_star_aux_mem_assoc_left : forall phi phi' Dpa Dph Dpr h o f hs d,
  passThroughMembrane_acc_invariant Dpa →
  passThroughMembrane_acc_corresponds Dph Dpa →
  passThroughMembrane_fold_star_aux Dph (Some (phi, Dpa)) = Some (phi', Dpr) →
  mem (h, (hs, d, o)) Dpa →
  in_AObject_domain f (DObject_dom o) →
  locInDom phi' h f ∨ exists o' hs',
    mem (h, (hs', d, o')) Dpr ∧ in_AObject_domain f (DObject_dom o').
Proof.
  introv CDpa Cor E MA I. lets E0: E. unfolds in E. destruct Dph as [h' [[hs' d'] o']].
  simpl in E. tests T: (h = h').
   subst. cases_if as I1.
    destruct assoc as [[hs'' d''] o''] eqn: Ea.
     asserts Eo: ((hs'', d'', o'') = (hs, d, o)).
       applys~ passThroughMembrane_acc_unique MA. apply~ mem_assoc_assoc.
      forwards~: passThroughMembrane_acc_corresponds_dependencies Cor MA. substs.
      inverts Eo. destruct DObject_merge eqn: EO.
      forwards~ Eq: DObject_merge_same_domains_right EO.
       typeclass.
       cases_if as I2; inverts E.
        left. eexists. splits.
         applys rtclosure_once. constructors*.
          rewrite~ <- (Eq f).
        right. do 2 eexists. splits.
         simpl. rew_refl*.
         rewrite~ <- (Eq f).
     false*.
    rewrite mem_assoc_map_fst in I1. fold_bool. rew_refl in I1. false I1. rewrite* map_mem.
   cases_if as I1.
    destruct assoc as [[hs'' d''] o'']. destruct DObject_merge eqn: EO.
     cases_if as I2.
      inverts E. right. do 2 eexists. rewrite filter_mem_eq. rew_refl. splits*.
      inverts E. right. do 2 eexists. simpl. rewrite filter_mem_eq. rew_refl. splits*.
     false*.
    inverts E. right. do 2 eexists. simpl. rew_refl*.
Qed.

Lemma passThroughMembrane_fold_star_mem_assoc_left : forall phi phi' Dpa Dpb Dpr h o f hs d,
  passThroughMembrane_acc_invariant Dpa →
  passThroughMembrane_acc_invariant Dpb →
  Forall (fun Dph => passThroughMembrane_acc_corresponds Dph Dpa) Dpb →
  passThroughMembrane_fold_star (Some (phi, Dpa)) Dpb = Some (phi', Dpr) →
  mem (h, (hs, d, o)) Dpa →
  in_AObject_domain f (DObject_dom o) →
  locInDom phi' h f ∨ exists o' hs',
    mem (h, (hs', d, o')) Dpr ∧ in_AObject_domain f (DObject_dom o').
Proof.
  introv CDpa CDpb F E MA I. gen phi phi' Dpa Dpr hs d o. induction Dpb; introv CDpa F E MA I.
   inverts* E.
   unfolds in E. rewrite fold_left_cons in E.
    destruct passThroughMembrane_fold_star_aux as [[phi'' Dpr']|] eqn: E'.
     forwards~ [(o'&S&I')|(o'&hs'&Mh&I')]: passThroughMembrane_fold_star_aux_mem_assoc_left E' MA I.
      applys* Forall_inv_head F.
      left. exists o'. splits~. applys rtclosure_trans S.
       applys* passThroughMembrane_fold_star_sub E.
      applys* IHDpb E.
       constructors. introv I1 I2 Eq.
        applys passThroughMembrane_acc_unique CDpb Eq; simpl; rew_refl*.
       apply* passThroughMembrane_fold_star_aux_invariant.
       rewrite Forall_iff_forall_mem in *. introv II. forwards C: F.
        simpl. rew_refl. right*.
        applys~ passThroughMembrane_fold_star_aux_corresponds E'.
         apply F. simpl. rew_refl*.
         constructors. introv Eq I2. simpl in I2. rew_refl in I2. repeat inverts I2 as I2.
          forwards~ Ea: passThroughMembrane_acc_unique CDpb.
           simpl. rew_refl. left*.
           simpl. rew_refl. subst x. right*.
           inverts~ Ea.
     rewrite passThroughMembrane_fold_star_none in E. false*.
Qed.

Lemma passThroughMembrane_env_fold_star_mem_assoc_left : forall phi phi' Dpa Dpb Dpr ke,
  passThroughMembrane_env_fold_star (Some (phi, Dpa)) Dpb = Some (phi', Dpr) →
  mem_assoc ke Dpa →
  envInDom phi' ke ∨ mem_assoc ke Dpr.
Proof.
  introv Eq MA. unfolds in Eq. rewrite mem_assoc_map_fst in MA.
  rewrite mem_assoc_map_fst. gen phi Dpa. induction Dpb; introv E I.
   inverts~ E.
   simpl in E. destruct a as [ke' hsdo]. destruct hsdo as [hsd o]. destruct hsd as [kes d].
    simpl in E. rewrite map_mem with (CA := _) in I. lets (hd&Mhd&Ehd): (rm I). tests: (ke = ke').
     cases_if as I1.
      destruct assoc as [hsd' o']. destruct hsd' as [hs' d']. destruct AEnv_join eqn: E'.
       cases_if as I2.
        left. forwards S: passThroughMembrane_env_fold_star_sub E. eexists.
         applys rtclosure_trans S. substs. apply rtclosure_once. constructors*.
        applys IHDpb E. rewrite map_cons. simpl. rew_refl*.
       rewrite passThroughMembrane_env_fold_star_none in E. inverts E.
      rewrite mem_assoc_map_fst in I1. fold_bool. rew_refl in I1. false I1. rewrite* map_mem.
     cases_if as I1.
      destruct assoc as [hsd' o']. destruct hsd' as [hs' d']. destruct AEnv_join eqn: E'.
       simpl in E. cases_if as I2.
        applys IHDpb E. rewrite map_mem with (CA := _). eexists. splits*.
         rewrite filter_mem_eq. rew_refl. substs. splits*.
        applys IHDpb E. rewrite map_cons. simpl. rew_refl. right.
         rewrite map_mem with (CA := _). eexists. splits*.
         rewrite filter_mem_eq. rew_refl. substs. splits*.
       rewrite passThroughMembrane_env_fold_star_none in E. inverts E.
      applys IHDpb E. rewrite map_cons. simpl. rew_refl. right. rewrite* map_mem.
Qed.

Lemma passThroughMembrane_fold_star_aux_mem_assoc_right : forall phi phi' Dpa Dpr h o f hs d,
  passThroughMembrane_acc_invariant Dpa →
  passThroughMembrane_acc_corresponds (h, (hs, d, o)) Dpa →
  passThroughMembrane_fold_star_aux (h, (hs, d, o)) (Some (phi, Dpa)) = Some (phi', Dpr) →
  in_AObject_domain f (DObject_dom o) →
  locInDom phi' h f ∨ exists o' hs',
    mem (h, (hs', d, o')) Dpr ∧ in_AObject_domain f (DObject_dom o').
Proof.
  introv CDpa Cor E I. lets E0: E. unfolds in E. simpl in E. cases_if as I1.
   destruct assoc as [[hs'' d''] o''] eqn: Ea. destruct DObject_merge eqn: EO.
    forwards~ Eq: DObject_merge_same_domains_left EO.
     typeclass.
     cases_if as I2; inverts E.
      left. eexists. splits.
       applys rtclosure_once. constructors*.
        rewrite~ <- (Eq f).
      right. do 2 eexists. splits.
       simpl. rew_refl*.
       rewrite~ <- (Eq f).
    false*.
   inverts E. right. do 2 eexists. splits*. simpl. rew_refl*.
Qed.

Lemma passThroughMembrane_fold_star_mem_assoc_right : forall phi phi' Dpa Dpb Dpr h o f hs d,
  passThroughMembrane_acc_invariant Dpa →
  passThroughMembrane_acc_invariant Dpb →
  Forall (fun Dph => passThroughMembrane_acc_corresponds Dph Dpa) Dpb →
  passThroughMembrane_fold_star (Some (phi, Dpa)) Dpb = Some (phi', Dpr) →
  mem (h, (hs, d, o)) Dpb →
  in_AObject_domain f (DObject_dom o) →
  locInDom phi' h f ∨ exists o' hs',
    mem (h, (hs', d, o')) Dpr ∧ in_AObject_domain f (DObject_dom o').
Proof.
  introv CDpa CDpb F Eq MA I. unfolds in Eq. gen phi Dpa. induction Dpb; introv CDpa F E.
   inverts~ E. false*.
   simpl in E. destruct a as [h' [[hs' d'] o']].
    destruct passThroughMembrane_fold_star_aux as [[? ?]|] eqn: E0.
     tests T: ((h, (hs, d, o)) = (h', (hs', d', o'))).
      introv ? ?. subst.
       forwards* [(o&S&I')|(o&hs&Mh&I')]: passThroughMembrane_fold_star_aux_mem_assoc_right E0.
        applys~ Forall_inv_head F.
        left. exists o. splits~. applys rtclosure_trans S.
         applys~ passThroughMembrane_fold_star_sub E.
        applys* passThroughMembrane_fold_star_mem_assoc_left E.
         applys~ passThroughMembrane_fold_star_aux_invariant E0.
         constructors. introv I1 I2 Eq.
          applys passThroughMembrane_acc_unique CDpb Eq; simpl; rew_refl*.
         rewrite Forall_iff_forall_mem in *. introv Ix.
          applys~ passThroughMembrane_fold_star_aux_corresponds E0.
           apply~ F. simpl. rew_refl*.
           apply~ F. simpl. rew_refl. right*.
           constructors. introv Eq I2. simpl in I2. rew_refl in I2. repeat inverts I2 as I2.
            forwards~ Ea: passThroughMembrane_acc_unique CDpb.
             simpl. rew_refl. left*.
             simpl. rew_refl. subst x. right*.
             inverts~ Ea.
      simpl in MA. rew_refl in MA. inverts MA as MA; tryfalse~. applys~ IHDpb MA E.
       constructors. introv I1 I2 Eq.
        applys passThroughMembrane_acc_unique CDpb Eq; simpl; rew_refl*.
       applys~ passThroughMembrane_fold_star_aux_invariant E0.
       rewrite Forall_iff_forall_mem in *. introv Ix.
        applys~ passThroughMembrane_fold_star_aux_corresponds E0.
         apply~ F. simpl. rew_refl*.
         apply~ F. simpl. rew_refl. right*.
         constructors. introv Eq I2. simpl in I2. rew_refl in I2. repeat inverts I2 as I2.
          forwards~ Ea: passThroughMembrane_acc_unique CDpb.
           simpl. rew_refl. left*.
           simpl. rew_refl. subst x. right*.
           inverts~ Ea.
     rewrite passThroughMembrane_fold_star_none in E. false*.
Qed.

Lemma passThroughMembrane_env_fold_star_mem_assoc_right : forall phi phi' Dpa Dpb Dpr ke,
  passThroughMembrane_env_fold_star (Some (phi, Dpa)) Dpb = Some (phi', Dpr) →
  mem_assoc ke Dpb →
  envInDom phi' ke ∨ mem_assoc ke Dpr.
Proof.
  introv Eq MA. unfolds in Eq. rewrite mem_assoc_map_fst in MA.
  rewrite mem_assoc_map_fst. gen phi Dpa. induction Dpb; introv E.
   inverts~ E. false*.
   simpl in E. destruct a as [h' hsdo]. destruct hsdo as [hsd o]. destruct hsd as [hs d].
    simpl in E. rewrite map_mem with (CA := _) in MA. lets (hd&Mhd&Ehd): (rm MA).
    tests: (hd = (h', (hs, d, o))).
     simpl in Ehd. substs. cases_if as I1.
      destruct assoc as [hsd' o']. destruct hsd' as [hs' d']. destruct AEnv_join eqn: E'.
       cases_if as I2.
        left. forwards S: passThroughMembrane_env_fold_star_sub E. eexists.
         applys rtclosure_trans S. substs. apply rtclosure_once. constructors*.
        forwards I: passThroughMembrane_env_fold_star_mem_assoc_left E;
          [| rewrite* mem_assoc_map_fst in I ].
         rewrite mem_assoc_map_fst. rewrite map_cons. simpl. rew_refl*.
       rewrite passThroughMembrane_env_fold_star_none in E. inverts E.
      forwards I: passThroughMembrane_env_fold_star_mem_assoc_left E.
       rewrite mem_assoc_cons. rew_refl*.
       rewrite* mem_assoc_map_fst in I.
     simpl in Mhd. rew_refl in Mhd. inverts* Mhd. cases_if as I1.
      destruct assoc as [hsd' o']. destruct hsd' as [hs' d']. destruct AEnv_join eqn: E'.
       simpl in E. cases_if as I2.
        applys IHDpb E. rewrite map_mem with (CA := _). eexists. splits*.
        applys IHDpb E. rewrite* map_mem.
       rewrite passThroughMembrane_env_fold_star_none in E. inverts E.
      applys IHDpb E. rewrite* map_mem.
Qed.

Lemma passThroughMembrane_aux_invariant_obj : forall phi M phi' Dp Dpe,
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  Forall (fun Dph =>
    let '(h, (hs, d, o)) := Dph in
    d = MgetDependencies _ _ M h ∧ incl hs d) Dp.
Proof.
  introv C E. gen phi' Dp Dpe. induction phi; introv E.
   inverts E. constructors~.
   simpl in E. destruct (passThroughMembrane_aux phi1 M); tryfalse.
    destruct p as [Dpphi1' Dpe1]. destruct Dpphi1' as [phi1' Dp1].
    destruct (passThroughMembrane_aux phi2 M); tryfalse.
    destruct p as [Dpphi2' Dpe2]. destruct Dpphi2' as [phi2' Dp2].
    forwards~ F1: (rm IHphi1). forwards~ F2: (rm IHphi2).
    gen phi1' phi2' Dp1 Dpe1. induction F2; introv F1 E.
     simpl in E. destruct passThroughMembrane_env_fold_star; tryfalse.
      destruct p. inverts~ E.
     destruct x as (h'&((hs'&d')&o')). unfold passThroughMembrane_fold_star in E.
      simpl in E. cases_if as IA.
       destruct (assoc h' Dp1) as ((hs&d)&o) eqn: EA. unfolds ALoc. rewrite EA in E.
        destruct DObject_merge eqn: EO.
         cases_if as DI; fold_bool; rew_refl in DI.
          applys~ IHF2 (rm E). rewrite Forall_iff_forall_mem in *. introv Mx.
           apply~ F1. rewrite filter_mem_eq in Mx. rew_refl* in Mx.
          applys~ IHF2 (rm E). lets (Ed'&Id'): (rm H). constructors~.
           splits~.
            apply mem_incl with (H := _). introv I. rewrite Remove_duplicates_mem in I.
            rewrite mem_app in I. rew_refl in I. inverts I as I.
             apply* incl_mem.
             forwards~ ID: mem_assoc_assoc EA. rewrite Forall_iff_forall_mem in F1.
              forwards (?&I'): F1 ID. substs. applys~ incl_mem I'.
           repeat rewrite Forall_iff_forall_mem with (CA := _) in *. introv Mx. apply~ F1.
            rewrite filter_mem_eq in Mx. rew_refl* in Mx.
        rewrite passThroughMembrane_fold_star_none in E. inverts E.
       forwards~: IHF2 (rm E). constructors~.
   forwards (v'&Ev'&E'): apply_on_inv (rm E). destruct (Mread M a) as [hs|] eqn: Er.
    asserts ID: (forall h, mem h (rev hs) → mem a (MgetDependencies _ _ M h)).
      introv I. rewrite <- Mem_mem in I. rewrite <- Mem_rev_iff in I.
      rewrite Mem_mem with (H := _) in I.
      unfolds Mread. applys~ MgetDependencies_Mread Er I.
     unfold passThroughMembrane_fold_object in E'. rewrite fold_left_eq_fold_right in E'.
     gen phi' Dp. unfolds ALoc. induction (rev hs); introv E.
      inverts E. constructors~.
      simpl in E. match type of E with context [fold_right ?f ?i ?l] =>
        destruct (fold_right f i l) as [phi Dp'] eqn: E' end. forwards~ F: IHl.
         introv I. apply~ ID. simpl. rew_refl*.
         repeat fequals. destruct MgetDependencies as [|? ()]; try solve [ inverts~ E ].
         destruct arbitrary in E. inverts~ E.
       simpl in E. destruct (MgetDependencies _ _ M a0) as [|? l'] eqn: ED; inverts E as E.
        forwards A: ID.
         simpl. rew_refl. left~.
         rewrite ED in A. false*.
        destruct l' as [|? l']; inverts~ E. constructors~. splits~.
         rewrite <- ED. apply mem_incl with (H := _). introv I. simpl in I. rew_refl in I.
          inverts I; tryfalse. apply ID. simpl. rew_refl*.
         simpl. rew_refl*.
    inverts E'. constructors~.
   simpl in E. destruct applyMembraneAEnv; tryfalse. simpl in E. destruct Mread_env.
    destruct passThroughMembrane_fold_env; tryfalse. inverts E. constructors~.
    inverts E. constructors~.
Qed.

Lemma passThroughMembrane_aux_invariant_invariant : forall phi M phi' Dp Dpe,
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  passThroughMembrane_acc_invariant Dp.
Proof.
  induction phi; introv C E.
   inverts E. constructors. introv I. false*.
   simpl in E. destruct passThroughMembrane_aux as [[[phi1' Dp1] Dpe1]|] eqn: E1; tryfalse~.
    destruct passThroughMembrane_aux as [[[phi2' Dp2] Dpe2]|] eqn: E2 in E; tryfalse~.
    destruct passThroughMembrane_fold_star as [[phi3 Dp3]|] eqn: E3; tryfalse~.
    destruct passThroughMembrane_env_fold_star as [[phi4 Dpe4]|] eqn: E4; tryfalse~.
    inverts E. applys~ passThroughMembrane_fold_star_invariant E3. applys~ IHphi1 E1.
   forwards (o&E1&E2): map_on_inv E. destruct Mread as [hs|] eqn: E3.
    destruct passThroughMembrane_fold_object as [phi2 Dp2] eqn: E4. inverts E2.
     unfolds in E4. rewrite fold_left_eq_fold_right in E4.
     asserts Inv: (forall h, mem h (rev hs) → mem a (MgetDependencies _ _ M h)).
      unfolds in E3. introv I. rewrite <- rev_mem in I.
       forwards~: MgetDependencies_Mread E3 I.
     asserts ND: (No_duplicates (rev hs)).
      applys~ No_duplicates_rev.
       typeclass.
       unfolds in E3. applys~ Mread_No_Duplicates E3.
     apply proj2 with (forall h, mem h (map fst Dp2) → mem h (rev hs)).
     clear E3 E1 E. gen phi2 Dp2. induction (rev hs); introv E1.
      inverts E1. splits*. constructors. introv I. false*.
      simpl in E1. destruct fold_right as [phi Dp] eqn: E3. forwards~ (IH&IDp): IHl.
       introv I. apply~ Inv. simpl. rew_refl*.
       inverts~ ND.
       destruct MgetDependencies as [|? [|]] eqn: E2; inverts~ E1.
        forwards A: Inv.
         simpl. rew_refl*.
         rewrite E2 in A. false*.
        splits~. introv I. simpl. rew_refl*.
        splits.
         introv I. rewrite map_mem in I. lets ((h'&a')&Ih&Eh): (rm I).
          simpl in Ih. rew_refl in Ih. subst h. inverts Ih as Ih.
           inverts Ih. simpl. rew_refl*.
           simpl. rew_refl. right. apply IH. rewrite* map_mem.
         constructors. introv I1 I2 E. simpl in I1, I2. rew_refl in I1. rew_refl in I2.
          inverts I1 as I1; inverts I2 as I2.
           inverts I1. inverts~ I2.
           inverts I1. substs. inverts ND as N1 N2. false N1. rewrite Mem_mem with (H := _).
            apply IH. rewrite* map_mem.
           inverts I2. substs. inverts ND as N1 N2. false N1. rewrite Mem_mem with (H := _).
            apply IH. rewrite* map_mem.
           applys~ passThroughMembrane_acc_unique I1 I2.
    inverts E2. constructors. introv I. false*.
   forwards (o&E1&E2): map_on_inv E. destruct Mread_env as [hs|] eqn: E3.
    destruct passThroughMembrane_fold_env as [phi2 Dpe2] eqn: E4. inverts E2.
     constructors. introv I. false*.
    inverts E2. constructors. introv I. false*.
Qed.

Lemma passThroughMembrane_aux_invariant_env : forall phi M phi' Dp Dpe,
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  Forall (fun Dph =>
    let '(h, (hs, d, o)) := Dph in
    d = MgetDependencies_env _ M h ∧ incl hs d) Dpe.
Proof.
  introv C E. gen phi' Dp Dpe. induction phi; introv E.
   inverts E. constructors~.
   simpl in E. destruct (passThroughMembrane_aux phi1 M); tryfalse.
    destruct p as [Dpphi1' Dpe1]. destruct Dpphi1' as [phi1' Dp1].
    destruct (passThroughMembrane_aux phi2 M); tryfalse.
    destruct p as [Dpphi2' Dpe2]. destruct Dpphi2' as [phi2' Dp2].
    destruct passThroughMembrane_fold_star as [[phi Dp']|]; tryfalse.
    destruct passThroughMembrane_env_fold_star eqn: E'; tryfalse. destruct p as [phi'' Dpe'].
    inverts E. forwards~ F1: (rm IHphi1). forwards~ F2: (rm IHphi2).
    gen phi Dpe1. induction F2; introv E F1.
     simpl in E. inverts~ E.
     destruct x as (h'&((hs'&d')&o')). unfolds passThroughMembrane_env_fold_star.
      simpl in E. cases_if as IA.
       destruct (assoc h' Dpe1) as ((hs&d)&o) eqn: EA.
        destruct AEnv_join.
         cases_if as DI; fold_bool; rew_refl in DI.
          applys~ IHF2 (rm E). rewrite Forall_iff_forall_mem in *. introv Mx.
           apply~ F1. rewrite filter_mem_eq in Mx. rew_refl* in Mx.
          applys~ IHF2 (rm E). lets (Ed'&Id'): (rm H). constructors~.
           splits~.
            apply mem_incl with (H := _). introv I. rewrite Remove_duplicates_mem in I.
            rewrite mem_app in I. rew_refl in I. inverts I as I.
             apply* incl_mem.
             forwards~ ID: mem_assoc_assoc EA. rewrite Forall_iff_forall_mem in F1.
              forwards (?&I'): F1 ID. substs. applys* incl_mem I'.
           repeat rewrite Forall_iff_forall_mem with (CA := _) in *. introv Mx. apply F1.
            rewrite filter_mem_eq in Mx. rew_refl* in Mx.
         simpl in E. clear - E. false. induction l.
          inverts E.
          destruct a as [? a]. do 2 destruct a as [a ?]. applys~ IHl E.
       forwards~: IHF2 (rm E). constructors~.
   simpl in E. destruct applyMembraneDObject; tryfalse. simpl in E. destruct Mread.
    destruct passThroughMembrane_fold_object. inverts E. constructors~.
    inverts E. constructors~.
   forwards (v'&Ev'&E'): apply_on_inv (rm E). destruct (Mread_env M k) as [hs|] eqn: Er.
    asserts ID: (forall h, mem h (rev hs) → mem k (MgetDependencies_env _ M h)).
      introv I. rewrite <- Mem_mem in I. rewrite <- Mem_rev_iff in I. rewrite Mem_mem with (H := _) in I.
      unfolds Mread_env. applys~ MgetDependencies_Mread_env Er I.
     unfold passThroughMembrane_fold_env in E'. rewrite fold_left_eq_fold_right in E'.
     gen phi' Dpe. induction (rev hs); introv E.
      inverts E. constructors~.
      simpl in E. match type of E with context [fold_right ?f ?i ?l] =>
        destruct (fold_right f i l) as [phi Dp'] eqn: E' end. forwards~ F: IHl.
         introv I. apply~ ID. simpl. rew_refl*.
         repeat fequals. destruct MgetDependencies_env as [|? ()]; try solve [ inverts~ E ].
         destruct arbitrary in E. inverts~ E.
       simpl in E. destruct (MgetDependencies_env _ M a0) as [|? l'] eqn: ED; inverts E as E.
        forwards A: ID.
         simpl. rew_refl. left~.
         rewrite ED in A. false*.
        destruct l' as [|? l']; inverts~ E. constructors~. splits~.
         rewrite <- ED. apply mem_incl with (H := _). introv I. simpl in I. rew_refl in I.
          inverts I; tryfalse. apply ID. simpl. rew_refl*.
         simpl. rew_refl*.
    inverts E'. constructors~.
Qed.

Lemma passThroughMembrane_aux_defined : forall phi membrane,
  noNewMembrane membrane →
  (** This property is added for proof simplification purposes (see
    comment in [passThroughMembrane_aux]). *)
  noMergingMembrane membrane →
  Membrane_correct membrane →
  passThroughMembrane_aux phi membrane ≠ None.
Proof.
  introv NN NM C. asserts (phi'&Ephi'): (exists phi',
    passThroughMembrane_aux phi membrane = Some (phi', nil, nil)).
   induction phi; simpl.
    eexists. autos*.
    lets (phi1'&Ephi1'): (rm IHphi1). lets (phi2'&Ephi2'): (rm IHphi2).
     rewrite Ephi1'. rewrite Ephi2'. simpl. autos*.
    destruct (applyMembraneDObject _ _) eqn: E'.
     simpl. destruct (Mread membrane a) as [l|] eqn: I.
      unfolds passThroughMembrane_fold_object.
       asserts F: (forall h',
         mem h' l → MgetDependencies _ _ membrane h' = [a]).
        introv M. forwards LI: (noMergingMembrane_obj NM) h'.
        destruct MgetDependencies as [|h l'] eqn: E.
         unfolds in I. forwards* A: MgetDependencies_Mread I. rewrite E in A. false*.
         rewrite length_cons in LI. destruct l'.
          unfolds Mread. eapply MgetDependencies_Mread in I;
            try solve [ rewrite E in I; simpl in I; rew_refl in I; inverts* I ]; autos*.
          rewrite length_cons in LI. simpl in LI. repeat inverts LI as LI.
       clear I. generalize (emp : IFormula). induction l; introv.
        simpl. autos*.
        forwards (phi'&Ephi'): (rm IHl).
          introv M. apply F. simpl. rew_refl*.
         inverts Ephi' as E. simpl. rewrite F.
          rewrite* E.
          simpl. rew_refl*.
      autos*.
     forwards*: noNewMembrane_applyMembraneDObject E'.
    destruct (applyMembraneAEnv _ _) eqn: E'.
     simpl. destruct (Mread_env membrane k) as [kes|] eqn: I.
      unfolds passThroughMembrane_fold_env.
       asserts F: (forall ke,
         mem ke kes → MgetDependencies_env _ membrane ke = [k]).
        introv M. forwards LI: (noMergingMembrane_env NM) ke.
         destruct MgetDependencies_env as [|h l'] eqn: E.
          unfolds in I. forwards* A: MgetDependencies_Mread_env I. rewrite E in A. false*.
          destruct l'.
           unfolds Mread_env. eapply MgetDependencies_Mread_env in I;
             try solve [ rewrite E in I; simpl in I; rew_refl in I; inverts* I ]; autos*.
           rewrite length_cons in LI. simpl in LI. repeat inverts LI as LI.
       clear I. generalize (emp : IFormula). induction kes; introv.
        simpl. autos*.
        forwards (phi'&Ephi'): (rm IHkes).
          introv M. apply F. simpl. rew_refl*.
         inverts Ephi' as E. simpl. rewrite F.
          rewrite* E.
          simpl. rew_refl*.
      autos*.
     forwards*: noNewMembrane_applyMembraneAEnv E'.
   rewrite Ephi'. discriminate.
Qed.

(** [passThroughMembrane] forget about the additional memory which
  [passThroughMembrane_aux] provides. **)
Definition passThroughMembrane phi M : option IFormula :=
  LibOption.map (compose fst fst) (passThroughMembrane_aux phi M).


(** [passThroughMembrane] has some prerequisites for being
  defined. The function [mergeSameObjects] transforms a formula [phi]
  into another formula [phi'] such that [phi'] has these
  prerequisites, if at all possible. **)
Fixpoint mergeSameObjects_aux (phi : IFormula) (acc : list (ALoc * DObject))
    : option (IFormula * list (ALoc * DObject)) :=
  match phi with
  | emp => Some (emp, nil)
  | phi1 * phi2 =>
    match mergeSameObjects_aux phi1 acc with
    | Some (phi1', acc) =>
      match mergeSameObjects_aux phi2 acc with
      | Some (phi2', acc) => Some (phi1' * phi2', acc)
      | None => None
      end
    | None => None
    end
  | h --> o =>
    if mem_assoc h acc then
      let o' := assoc h acc in
      match DObject_star _ _ _ _ o o' with
      | Some o'' => Some (emp, (h, o'') :: filter (fun ho => decide (h <> fst ho)) acc)
      | None => None
      end
    else Some (emp, (h, o) :: acc)
  | ke ~~> E => Some (ke ~~> E, acc)
  end.

Definition mergeSameObjects (phi : IFormula) : option IFormula :=
  LibOption.map_on (mergeSameObjects_aux phi nil) (fun phiacc =>
    let (phi, acc) := phiacc in
    fold_left (fun ho phi =>
      let (h, o) := (ho : _ * _) in h --> o * phi) phi acc).


Lemma passThroughMembrane_defined : forall phi membrane,
  noNewMembrane membrane →
  noMergingMembrane membrane →
  Membrane_correct membrane →
  passThroughMembrane phi membrane ≠ None.
Proof.
  introv NN NM C E. apply option_map_none_inv in E. apply* passThroughMembrane_aux_defined.
Qed.

Lemma passThroughMembrane_aux_singleton_membrane : forall hs h phi phi' Dp Dpe,
  passThroughMembrane_aux phi ([h ---> hs]) = Some (phi', Dp, Dpe) →
  Dp = nil ∧ Dpe = nil.
Proof.
  introv E. gen Dp Dpe phi'. induction phi; introv E.
   inverts~ E.
   simpl in E. destruct (passThroughMembrane_aux phi1 ([h ---> hs])) as [p|]; tryfalse.
    destruct p as [phi1Dp' Dpe1]. destruct phi1Dp' as [phi1' Dp1]. forwards~ (?&?): IHphi1.
    destruct (passThroughMembrane_aux phi2 ([h ---> hs])) as [p|]; tryfalse.
    destruct p as [phi2Dp' Dpe2]. destruct phi2Dp' as [phi2' Dp2]. forwards~ (?&?): IHphi2.
    substs. inverts~ E.
   simpl in E. destruct applyMembraneDObject as [o|]; tryfalse. simpl in E.
    cases_if as I; try solve [ inverts~ E ]. fold_bool. rew_refl in I. substs.
    sets_eq phi: (emp : IFormula). unfold passThroughMembrane_fold_object in E.
    match type of E with context [ fold_left ?f0 ] =>
      sets_eq f: f0 end. destruct fold_left as [phi'' Dp''] eqn: E'.
    inverts E. sets_eq hs': hs. rewrite EQhs' in EQf. subst f.
    asserts I: (forall h, mem h hs' → mem h hs). substs~.
    clear EQphi EQhs'. gen Dp phi phi'. induction hs'; introv E.
     inverts~ E.
     simpl in E. cases_if as I'.
      applys~ IHhs' E. introv I0. apply I. simpl. rew_refl*.
      fold_bool. rew_refl in I'. false I'. splits~. apply I. simpl. rew_refl*.
   simpl in E. destruct applyMembraneAEnv; tryfalse. inverts~ E.
Qed.

Lemma passThroughMembrane_aux_singleton_membrane_env : forall kes ke phi phi' Dp Dpe,
  passThroughMembrane_aux phi ([ke ~~~> kes]) = Some (phi', Dp, Dpe) →
  Dp = nil ∧ Dpe = nil.
Proof.
  introv E. gen Dp Dpe phi'. induction phi; introv E.
   inverts~ E.
   simpl in E. destruct (passThroughMembrane_aux phi1 ([ke ~~~> kes])) as [p|]; tryfalse.
    destruct p as [phi1Dp' Dpe1]. destruct phi1Dp' as [phi1' Dp1]. forwards~ (?&?): IHphi1.
    destruct (passThroughMembrane_aux phi2 ([ke ~~~> kes])) as [p|]; tryfalse.
    destruct p as [phi2Dp' Dpe2]. destruct phi2Dp' as [phi2' Dp2]. forwards~ (?&?): IHphi2.
    substs. inverts~ E.
   simpl in E. destruct applyMembraneDObject; tryfalse. inverts~ E.
   simpl in E. destruct applyMembraneAEnv; tryfalse. simpl in E.
    cases_if as I; try solve [ inverts~ E ]. fold_bool. rew_refl in I. substs.
    sets_eq phi: (emp : IFormula). unfold passThroughMembrane_fold_env in E.
    match type of E with context [ fold_left ?f0 ] =>
      sets_eq f: f0 end. destruct fold_left as [phi'' Dp''] eqn: E'.
    inverts E. sets_eq kes': kes. rewrite EQkes' in EQf. subst f.
    asserts I: (forall ke, mem ke kes' → mem ke kes). substs~.
    clear EQphi EQkes'. gen Dpe phi phi'. induction kes'; introv E.
     inverts~ E.
     simpl in E. cases_if as I'.
      applys~ IHkes' E. introv I0. apply I. simpl. rew_refl*.
      fold_bool. rew_refl in I'. false I'. splits~. apply I. simpl. rew_refl*.
Qed.

Lemma passThroughMembrane_fold_object_sub : forall M h o hs phi Dp phi' Dp',
  (forall h, mem h hs → MgetDependencies _ _ M h ≠ nil) →
  passThroughMembrane_fold_object M h o (phi, Dp) hs = (phi', Dp') →
  IFormula_sub phi' phi.
Proof.
  introv A E. unfolds in E. gen phi Dp. induction hs; introv E.
   inverts E. apply~ IFormula_sub_refl.
   simpl in E. destruct MgetDependencies as [|a' l] eqn: E'.
    false* A. simpl. rew_refl*.
    destruct l; forwards~ S: IHhs E.
     introv I. apply A. simpl. rew_refl*.
     applys rtclosure_trans S. apply rtclosure_once. constructors*.
     introv I. apply A. simpl. rew_refl*.
Qed.

Lemma passThroughMembrane_fold_object_sub_Dp : forall M h o hs phi Dp phi' Dp',
  (forall h, mem h hs → MgetDependencies _ _ M h ≠ nil) →
  passThroughMembrane_fold_object M h o (phi, Dp) hs = (phi', Dp') →
  incl Dp Dp'.
Proof.
  introv A E. unfolds in E. gen phi Dp. induction hs; introv E.
   inverts E. apply~ incl_refl.
   simpl in E. destruct MgetDependencies as [|a' l] eqn: E'.
    false* A. simpl. rew_refl*.
    destruct l; forwards~ S: IHhs E.
     introv I. apply A. simpl. rew_refl*.
     introv I. apply A. simpl. rew_refl*.
     applys incl_tran S. apply incl_tl. apply incl_refl.
Qed.

Lemma passThroughMembrane_fold_env_sub : forall M ke o kes phi Dp phi' Dp',
  (forall ke, mem ke kes → MgetDependencies_env _ M ke ≠ nil) →
  passThroughMembrane_fold_env M ke o (phi, Dp) kes = (phi', Dp') →
  IFormula_sub phi' phi.
Proof.
  introv A E. unfolds in E. gen phi Dp. induction kes; introv E.
   inverts E. apply~ IFormula_sub_refl.
   simpl in E. destruct MgetDependencies_env as [|a' l] eqn: E'.
    false* A. simpl. rew_refl*.
    destruct l; forwards~ S: IHkes E.
     introv I. apply A. simpl. rew_refl*.
     applys rtclosure_trans S. apply rtclosure_once. constructors*.
     introv I. apply A. simpl. rew_refl*.
Qed.

Lemma passThroughMembrane_fold_object_mem_assoc_left : forall M h0 o0 hs phi h Dp phi' Dp' f,
  (forall h, mem h hs → MgetDependencies _ _ M h ≠ nil) →
  mem_assoc h Dp →
  in_AObject_domain f (DObject_dom o0) →
  passThroughMembrane_fold_object M h0 o0 (phi, Dp) hs = (phi', Dp') →
  locInDom phi' h f ∨ mem_assoc h Dp'.
Proof.
  introv A MA I E. unfolds in E. gen phi Dp. induction hs; introv MA E.
   inverts* E.
   simpl in E. destruct MgetDependencies as [|a' l] eqn: E'.
    false* A. simpl. rew_refl*.
    destruct l; forwards~ S: IHhs E.
     introv I'. apply A. simpl. rew_refl*.
     introv I'. apply A. simpl. rew_refl*.
     rewrite mem_assoc_cons. rew_refl*.
Qed.

Lemma passThroughMembrane_fold_env_mem_assoc_left : forall M h0 o0 kes phi ke Dp phi' Dp',
  (forall ke, mem ke kes → MgetDependencies_env _ M ke ≠ nil) →
  mem_assoc ke Dp →
  passThroughMembrane_fold_env M h0 o0 (phi, Dp) kes = (phi', Dp') →
  envInDom phi' ke ∨ mem_assoc ke Dp'.
Proof.
  introv A MA E. unfolds in E. gen phi Dp. induction kes; introv MA E.
   inverts* E.
   simpl in E. destruct MgetDependencies_env as [|a' l] eqn: E'.
    false* A. simpl. rew_refl*.
    destruct l; forwards~ S: IHkes E.
     introv I. apply A. simpl. rew_refl*.
     introv I. apply A. simpl. rew_refl*.
     rewrite mem_assoc_cons. rew_refl*.
Qed.

Lemma passThroughMembrane_fold_object_mem_assoc_right_precise : forall M h0 o0 hs phi h' Dp phi' Dp' f,
  (forall h, mem h hs → MgetDependencies _ _ M h ≠ nil) →
  mem h' hs →
  in_AObject_domain f (DObject_dom o0) →
  passThroughMembrane_fold_object M h0 o0 (phi, Dp) hs = (phi', Dp') →
  locInDom phi' h' f ∨ exists o' hs' d',
    mem (h', (hs', d', o')) Dp' ∧ in_AObject_domain f (DObject_dom o').
Proof.
  introv A Ih' I E. unfolds in E. gen phi Dp. induction hs; introv E.
   false*.
   simpl in E. destruct MgetDependencies as [|a' l] eqn: E'.
    false A.
     simpl. rew_refl*.
     autos*.
    simpl in Ih'. rew_refl in Ih'. inverts Ih' as Ih'.
     destruct l.
      left. exists o0. splits~. applys rtclosure_trans.
       applys~ passThroughMembrane_fold_object_sub E.
        introv I'. apply~ A. simpl. rew_refl*.
       applys rtclosure_once. constructors~.
      forwards I': passThroughMembrane_fold_object_sub_Dp E.
       introv I'. apply A. simpl. rew_refl*.
       right. do 3 eexists. splits.
        applys incl_mem I'. simpl. rew_refl*.
        autos*.
     destruct l; forwards~ S: IHhs E.
      introv I'. apply A. simpl. rew_refl*.
      introv I'. apply A. simpl. rew_refl*.
Qed.

Lemma passThroughMembrane_fold_object_mem_assoc_right : forall M h0 o0 hs phi h Dp phi' Dp' f,
  (forall h, mem h hs → MgetDependencies _ _ M h ≠ nil) →
  mem h hs →
  in_AObject_domain f (DObject_dom o0) →
  passThroughMembrane_fold_object M h0 o0 (phi, Dp) hs = (phi', Dp') →
  locInDom phi' h f ∨ mem_assoc h Dp'.
Proof.
  introv A I fI E. forwards [I'|(o'&hs'&d'&I1&I2)]: passThroughMembrane_fold_object_mem_assoc_right_precise A I fI E.
   left~.
   right. rewrite mem_assoc_map_fst. rewrite* map_mem.
Qed.

Lemma passThroughMembrane_fold_env_mem_assoc_right : forall M h0 o0 kes phi ke Dp phi' Dp',
  (forall ke, mem ke kes → MgetDependencies_env _ M ke ≠ nil) →
  mem ke kes →
  passThroughMembrane_fold_env M h0 o0 (phi, Dp) kes = (phi', Dp') →
  envInDom phi' ke ∨ mem_assoc ke Dp'.
Proof.
  introv A I E. unfolds in E. gen phi Dp. induction kes; introv E.
   inverts E. false*.
   simpl in E. destruct MgetDependencies_env as [|a' l] eqn: E'.
    false A; [| apply E' ]. simpl. rew_refl*.
    tests: (ke = a).
     destruct l.
      forwards~ S: passThroughMembrane_fold_env_sub E.
        introv I'. apply A. simpl. rew_refl*.
       left. eexists. applys rtclosure_trans S. apply rtclosure_once. constructors*.
      applys~ passThroughMembrane_fold_env_mem_assoc_left E.
        introv I'. apply A. simpl. rew_refl*.
       rewrite mem_assoc_map_fst. rewrite map_cons. simpl. rew_refl*.
     simpl in I. rew_refl in I. inverts I as I; tryfalse~.
      destruct l; applys~ IHkes E; introv I' Eq'; false* A; simpl; rew_refl*.
Qed.

Lemma passThroughMembrane_aux_dom : forall M phi phi' Dp Dpe h f hs h',
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  locInDom phi h f →
  Mread M h = Some hs →
  mem h' hs →
  locInDom phi' h' f ∨ exists o' hs' d',
    mem (h', (hs', d', o')) Dp ∧ in_AObject_domain f (DObject_dom o').
Proof.
  introv C Ep Mh Er Mh'. gen phi' Dp Dpe. induction phi; introv Ep; simpls.
   apply locInDom_emp in Mh; try typeclass. false*.
   apply locInDom_star in Mh. destruct (passThroughMembrane_aux phi1 M) eqn: E1; tryfalse.
    destruct p as [phiDp1' Dpe1]. destruct phiDp1' as [phi1' Dp1'].
    destruct (passThroughMembrane_aux phi2 M) eqn: E2; tryfalse.
    destruct p as [phiDp2' Dpe2]. destruct phiDp2' as [phi2' Dp2'].
    destruct passThroughMembrane_fold_star as [[phi Dp']|] eqn: E; tryfalse.
    destruct passThroughMembrane_env_fold_star eqn: E'; tryfalse. destruct p as [phi'' Dpe'].
    inverts Ep. asserts FS: (IFormula_sub phi' (phi1' * phi2')).
      forwards S1: passThroughMembrane_fold_star_sub E.
      forwards S2: passThroughMembrane_env_fold_star_sub E'.
      applys~ rtclosure_trans S2 S1.
    inverts Mh as Mh.
     forwards~ [LD|(o'&hs'&d'&I1'&I2')]: IHphi1.
      left. apply* locInDom_sub. applys~ rtclosure_trans FS.
       apply~ rtclosure_once. constructors*.
       forwards~ [(o2&S2&I2)|(o2&hs2&MA2&I2)]: passThroughMembrane_fold_star_mem_assoc_left E I1' I2'.
        applys~ passThroughMembrane_aux_invariant_invariant E1.
        applys~ passThroughMembrane_aux_invariant_invariant E2.
        rewrite Forall_iff_forall_mem. introv I. constructors. introv Eq I'.
         subst x. forwards~ F1: passThroughMembrane_aux_invariant_obj E1.
         forwards~ F2: passThroughMembrane_aux_invariant_obj E2.
         rewrite Forall_iff_forall_mem in F1, F2. lets (?&?): F1 I'. lets (?&?): F2 I.
         substs~.
        left. eexists. splits*. applys rtclosure_trans S2.
         apply* passThroughMembrane_env_fold_star_sub.
        right*.
     forwards~ [LD|(o'&hs'&d'&I1'&I2')]: IHphi2.
      left. apply* locInDom_sub. applys~ rtclosure_trans FS.
       apply~ rtclosure_once. constructors*.
       forwards~ [(o2&S2&I2)|(o2&hs2&MA2&I2)]: passThroughMembrane_fold_star_mem_assoc_right E I1' I2'.
        applys~ passThroughMembrane_aux_invariant_invariant E1.
        applys~ passThroughMembrane_aux_invariant_invariant E2.
        rewrite Forall_iff_forall_mem. introv I. constructors. introv Eq I'.
         subst x. forwards~ F1: passThroughMembrane_aux_invariant_obj E1.
         forwards~ F2: passThroughMembrane_aux_invariant_obj E2.
         rewrite Forall_iff_forall_mem in F1, F2. lets (?&?): F1 I'. lets (?&?): F2 I.
         substs~.
        left. eexists. splits*. applys rtclosure_trans S2.
         apply* passThroughMembrane_env_fold_star_sub.
        right*.
   lets (o&S&I): (rm Mh). repeat inverts S as S.
    destruct (applyMembraneDObject M _) eqn: E1; try solve [ inverts Ep ]. simpl in Ep.
    rewrite Er in Ep. destruct passThroughMembrane_fold_object as [phi Dp'] eqn: EF.
    inverts Ep. forwards~ [I'|MA]: passThroughMembrane_fold_object_mem_assoc_right_precise Mh' EF.
     introv I' Eq. unfolds Mread. forwards* I2: MgetDependencies_Mread Er. rewrite* Eq in I2.
     erewrite <- applyMembraneDObject_dom; try apply E1; autos*.
     left~.
     right~.
   lets (o&S): (rm Mh). repeat inverts S as S.
Qed.

Lemma passThroughMembrane_aux_dom_Some : forall phi M phi' h h' hs Dp Dpe f,
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  locInDom phi h f →
  Mread M h = Some hs →
  mem h' hs →
  locInDom phi' h' f ∨ mem_assoc h' Dp.
Proof.
  introv C Ep Mh Er Mh'.
  forwards~ [I|(o'&hs'&d'&I&I')]: passThroughMembrane_aux_dom C Ep Mh Er Mh'.
  right. rewrite mem_assoc_map_fst. rewrite* map_mem.
Qed.

Lemma passThroughMembrane_aux_dom_env_Some : forall phi M phi' ke ke' kes Dp Dpe,
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  envInDom phi ke →
  Mread_env M ke = Some kes →
  mem ke' kes →
  envInDom phi' ke' ∨ mem_assoc ke' Dpe.
Proof.
  introv C Ep Mh Er Mh'. gen phi' Dp Dpe. induction phi; introv Ep; simpls.
   apply envInDom_emp in Mh; try typeclass. false*.
   apply envInDom_star in Mh. destruct (passThroughMembrane_aux phi1 M) eqn: E1; tryfalse.
    destruct p as [phiDp1' Dpe1]. destruct phiDp1' as [phi1' Dp1'].
    destruct (passThroughMembrane_aux phi2 M) eqn: E2; tryfalse.
    destruct p as [phiDp2' Dpe2]. destruct phiDp2' as [phi2' Dp2'].
    destruct passThroughMembrane_fold_star as [[phi Dp']|] eqn: E; tryfalse.
    destruct passThroughMembrane_env_fold_star eqn: E'; tryfalse. destruct p as [phi'' Dpe'].
    inverts Ep. asserts FS: (IFormula_sub phi' (phi1' * phi2')).
      forwards S1: passThroughMembrane_fold_star_sub E.
      forwards S2: passThroughMembrane_env_fold_star_sub E'.
      applys~ rtclosure_trans S2 S1.
    inverts Mh as Mh.
     forwards~ [LD|MA]: IHphi1.
      left. apply* envInDom_sub. applys~ rtclosure_trans FS.
       apply~ rtclosure_once. constructors*.
      forwards* [(o&S)|MA']: passThroughMembrane_env_fold_star_mem_assoc_left E'.
       left. constructors*.
     forwards~ [LD|MA]: IHphi2.
      left. apply* envInDom_sub. applys~ rtclosure_trans FS.
       apply~ rtclosure_once. constructors*.
      forwards* [(o&S)|MA']: passThroughMembrane_env_fold_star_mem_assoc_right E'.
       left. constructors*.
   lets (o&S): (rm Mh). repeat inverts S as S.
   lets (o&S): (rm Mh). repeat inverts S as S.
    destruct (applyMembraneAEnv M _) eqn: E1; try solve [ inverts Ep ]. simpl in Ep.
    rewrite Er in Ep. destruct passThroughMembrane_fold_env as [phi Dp'] eqn: EF.
    inverts Ep. applys~ passThroughMembrane_fold_env_mem_assoc_right EF.
    introv I Eq. unfolds Mread_env. forwards* I': MgetDependencies_Mread_env Er. rewrite* Eq in I'.
Qed.

Lemma passThroughMembrane_aux_IFNames : forall phi M phi' h' Dp Dpe,
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  mem h' (IFNames phi') →
  ~ MInOuterInterface _ _ _ h' M ∧ mem h' (IFNames phi)
  ∨ exists h hs,
    Mread M h = Some hs ∧ mem h' hs ∧ mem h (IFNames phi).
Proof.
  introv C E I. gen phi' Dp Dpe. induction phi; introv I Ep; simpls.
   inverts Ep. false*.
   skip. (* Admitted *)
   destruct (applyMembraneDObject M _) eqn: E1; try solve [ inverts Ep ]. simpl in Ep.
    destruct Mread eqn: E.
     right. destruct passThroughMembrane_fold_object as [phi Dp'] eqn: EF.
      inversion Ep. substs.
       unfolds in EF. rewrite fold_left_eq_fold_right in EF.
       asserts Inv: (forall h, mem h (rev l) → mem a (MgetDependencies _ _ M h)).
        introv I'. rewrite <- rev_mem in I'. unfolds in E. applys~ MgetDependencies_Mread E.
       clear Ep E E1. gen phi' Dp. induction (rev l); introv I E.
        inversion E. substs. false*.
        simpl in E. destruct fold_right eqn: E'. destruct MgetDependencies eqn: Ed.
         forwards A: Inv.
          simpl. rew_refl*.
          rewrite Ed in A. false*.
         destruct l2 eqn: El; inversion E; substs.
          simpl in I. rew_refl in I. destruct I as [I|I].
           forwards Ia: Inv.
            simpl. rew_refl*.
            rewrite Ed in Ia. simpl in Ia. rew_refl in Ia. destruct Ia as [Ia|()].
             exists a. forwards~ (hs&Ea&I'): MgetDependencies_correct C.
              rewrite Ed. simpl. rew_refl*.
              exists hs. rew_refl. substs~.
           skip.
          skip.
     skip.
   skip.
Qed.

Lemma passThroughMembrane_aux_IFENames : forall phi M phi' ke' Dp Dpe,
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  mem ke' (IFENames phi') →
  ~ MEInOuterInterface _ _ _ ke' M ∧ mem ke' (IFENames phi)
  ∨ exists ke kes,
    Mread_env M ke = Some kes ∧ mem ke' kes ∧ mem ke (IFENames phi).
Proof.
  skip.
Qed.

Lemma passThroughMembrane_IFNames : forall phi M phi' h',
  Membrane_correct M →
  passThroughMembrane phi M = Some phi' →
  mem h' (IFNames phi') →
  ~ MInOuterInterface _ _ _ h' M ∧ mem h' (IFNames phi)
  ∨ exists h hs,
    Mread M h = Some hs ∧ mem h' hs ∧ mem h (IFNames phi).
Proof.
  introv C E I. unfolds in E.
  destruct passThroughMembrane_aux as [[[? ?] ?]|] eqn: E'; inverts E.
  applys~ passThroughMembrane_aux_IFNames E'.
Qed.

Lemma passThroughMembrane_IFENames : forall phi M phi' ke' Dp Dpe,
  Membrane_correct M →
  passThroughMembrane_aux phi M = Some (phi', Dp, Dpe) →
  mem ke' (IFENames phi') →
  ~ MEInOuterInterface _ _ _ ke' M ∧ mem ke' (IFENames phi)
  ∨ exists ke kes,
    Mread_env M ke = Some kes ∧ mem ke' kes ∧ mem ke (IFENames phi).
Proof.
  introv C E I. unfolds in E.
  destruct passThroughMembrane_aux as [[[? ?] ?]|] eqn: E'; inverts E.
  applys~ passThroughMembrane_aux_IFENames C E' I.
Qed.

End General.

